/*
 * $RCSfile: TIFFDirectory.java,v $
 *
 *
 * Copyright (c) 2006 Sun Microsystems, Inc. All  Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistribution of source code must retain the above copyright
 *   notice, this  list of conditions and the following disclaimer.
 *
 * - Redistribution in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * This software is provided "AS IS," without a warranty of any
 * kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
 * EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL
 * NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF
 * USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR
 * ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
 * CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND
 * REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * You acknowledge that this software is not designed or intended for
 * use in the design, construction, operation or maintenance of any
 * nuclear facility.
 *
 * $Revision: 1.4 $
 * $Date: 2006/08/25 00:16:49 $
 * $State: Exp $
 */
package com.sun.media.imageio.plugins.tiff;

import java.util.*;

import javax.imageio.metadata.IIOMetadata;

import com.sun.media.imageioimpl.plugins.tiff.TIFFIFD;

/**
 * A convenience class for simplifying interaction with TIFF native
 * image metadata. A TIFF image metadata tree represents an Image File
 * Directory (IFD) from a TIFF 6.0 stream. An IFD consists of a number of
 * IFD Entries each of which associates an identifying tag number with
 * a compatible value. A <code>TIFFDirectory</code> instance corresponds
 * to an IFD and contains a set of {@link TIFFField}s each of which
 * corresponds to an IFD Entry in the IFD.
 *
 * <p>
 * A <code>TIFFDirectory</code> is aware of the tag numbers in the
 * group of {@link TIFFTagSet}s associated with it. When
 * a <code>TIFFDirectory</code> is created from a native image metadata
 * object, these tag sets are derived from the <tt>tagSets</tt> attribute
 * of the <tt>TIFFIFD</tt> node.</p>
 *
 * <p>
 * A <code>TIFFDirectory</code> might also have a parent {@link TIFFTag}.
 * This will occur if the directory represents an IFD other than the root
 * IFD of the image. The parent tag is the tag of the IFD Entry which is a
 * pointer to the IFD represented by this <code>TIFFDirectory</code>. The
 * {@link TIFFTag#isIFDPointer} method of this parent <code>TIFFTag</code>
 * must return <code>true</code>. When a <code>TIFFDirectory</code> is
 * created from a native image metadata object, the parent tag set is set
 * from the <tt>parentTagName</tt> attribute of the corresponding
 * <tt>TIFFIFD</tt> node. Note that a <code>TIFFDirectory</code> instance
 * which has a non-<code>null</code> parent tag will be contained in the
 * data field of a <code>TIFFField</code> instance which has a tag field
 * equal to the contained directory's parent tag.</p>
 *
 * <p>
 * As an example consider an EXIF image. The <code>TIFFDirectory</code>
 * instance corresponding to the EXIF IFD in the EXIF stream would have parent
 * tag {@link EXIFParentTIFFTagSet#TAG_EXIF_IFD_POINTER TAG_EXIF_IFD_POINTER}
 * and would include {@link EXIFTIFFTagSet} in its group of known tag sets.
 * The <code>TIFFDirectory</code> corresponding to this EXIF IFD will be
 * contained in the data field of a <code>TIFFField</code> which will in turn
 * be contained in the <code>TIFFDirectory</code> corresponding to the primary
 * IFD of the EXIF image which will itself have a <code>null</code>-valued
 * parent tag.</p>
 *
 * <p>
 * <b>Note that this implementation is not synchronized.</b> If multiple
 * threads use a <code>TIFFDirectory</code> instance concurrently, and at
 * least one of the threads modifies the directory, for example, by adding
 * or removing <code>TIFFField</code>s or <code>TIFFTagSet</code>s, it
 * <i>must</i> be synchronized externally.</p>
 *
 * @see IIOMetadata
 * @see TIFFField
 * @see TIFFTag
 * @see TIFFTagSet
 *
 * @since 1.1-beta
 */
public class TIFFDirectory implements Cloneable {

   /**
    * The largest low-valued tag number in the TIFF 6.0 specification.
    */
   private static final int MAX_LOW_FIELD_TAG_NUM
       = BaselineTIFFTagSet.TAG_REFERENCE_BLACK_WHITE;

   /**
    * The <code>TIFFTagSets</code> associated with this directory.
    */
   private final List<TIFFTagSet> _tagSets;

   /**
    * The parent <code>TIFFTag</code> of this directory.
    */
   private final TIFFTag _parentTag;

   /**
    * The fields in this directory which have a low tag number. These are
    * managed as an array for efficiency as they are the most common fields.
    */
   private final TIFFField[] _lowFields = new TIFFField[MAX_LOW_FIELD_TAG_NUM + 1];

   /**
    * The number of low tag numbered fields in the directory.
    */
   private int _numLowFields;

   /**
    * A mapping of <code>Integer</code> tag numbers to <code>TIFFField</code>s
    * for fields which are not low tag numbered.
    */
   private final Map<Integer, TIFFField> _highFields = new TreeMap<>();

   /**
    * Constructs a <code>TIFFDirectory</code> which is aware of a given
    * group of {@link TIFFTagSet}s. An optional parent {@link TIFFTag}
    * may also be specified.
    *
    * @param tagSets The <code>TIFFTagSets</code> associated with this
    * directory.
    * @param parentTag The parent <code>TIFFTag</code> of this directory;
    * may be <code>null</code>.
    * @throws IllegalArgumentException if <code>tagSets</code> is
    * <code>null</code>.
    */
   protected TIFFDirectory( TIFFTagSet[] tagSets, TIFFTag parentTag ) {
      if (tagSets == null) {
         throw new IllegalArgumentException("tagSets == null!");
      }
      this._tagSets = new ArrayList<>(tagSets.length);
      Collections.addAll(this._tagSets, tagSets);
      this._parentTag = parentTag;
   }

   /**
    * Returns the {@link TIFFTagSet}s of which this directory is aware.
    *
    * @return The <code>TIFFTagSet</code>s associated with this
    * <code>TIFFDirectory</code>.
    */
   public final TIFFTagSet[] getTagSets() {
      return _tagSets.toArray(new TIFFTagSet[_tagSets.size()]);
   }

   /**
    * Adds an element to the group of {@link TIFFTagSet}s of which this
    * directory is aware.
    *
    * @param tagSet The <code>TIFFTagSet</code> to add.
    * @throws IllegalArgumentException if <code>tagSet</code> is
    * <code>null</code>.
    */
   public final void addTagSet( TIFFTagSet tagSet ) {
      if (tagSet == null) {
         throw new IllegalArgumentException("tagSet == null");
      }

      if (!_tagSets.contains(tagSet)) {
         _tagSets.add(tagSet);
      }
   }

   /**
    * Returns the parent {@link TIFFTag} of this directory if one
    * has been defined or <code>null</code> otherwise.
    *
    * @return The parent <code>TIFFTag</code> of this
    * <code>TIFFDirectory</code> or <code>null</code>.
    */
   private TIFFTag getParentTag() {
      return _parentTag;
   }

   /**
    * Returns the {@link TIFFTag} which has tag number equal to
    * <code>tagNumber</code> or <code>null</code> if no such tag
    * exists in the {@link TIFFTagSet}s associated with this
    * directory.
    *
    * @param tagNumber The tag number of interest.
    * @return The corresponding <code>TIFFTag</code> or <code>null</code>.
    */
   public final TIFFTag getTag( int tagNumber ) {
      return TIFFIFD.getTag(tagNumber, _tagSets);
   }

   /**
    * Adds a TIFF field to the directory.
    *
    * @param f The field to add.
    * @throws IllegalArgumentException if <code>f</code> is <code>null</code>.
    */
   public final void addTIFFField( TIFFField f ) {
      if (f == null) {
         throw new IllegalArgumentException("f == null");
      }
      int tagNumber = f.getTagNumber();
      if (tagNumber >= 0 && tagNumber <= MAX_LOW_FIELD_TAG_NUM) {
         if (_lowFields[tagNumber] == null) {
            _numLowFields++;
         }
         _lowFields[tagNumber] = f;
      } else {
         _highFields.put(tagNumber, f);
      }
   }

   /**
    * Retrieves a TIFF field from the directory.
    *
    * @param tagNumber The tag number of the tag associated with the field.
    * @return A <code>TIFFField</code> with the requested tag number of
    * <code>null</code> if no such field is present.
    */
   public final TIFFField getTIFFField( int tagNumber ) {
      TIFFField field;
      if (tagNumber >= 0 && tagNumber <= MAX_LOW_FIELD_TAG_NUM) {
         field = _lowFields[tagNumber];
      } else {
         field = _highFields.get(tagNumber);
      }
      return field;
   }

   /**
    * Retrieves all TIFF fields from the directory.
    *
    * @return An array of all TIFF fields in order of numerically increasing tag number.
    */
   public final TIFFField[] getTIFFFields() {
      // Allocate return value.
      TIFFField[] fields = new TIFFField[_numLowFields + _highFields.size()];

      // Copy any low-index fields.
      int nextIndex = 0;
      for (int i = 0; i <= MAX_LOW_FIELD_TAG_NUM; i++) {
         if (_lowFields[i] != null) {
            fields[nextIndex++] = _lowFields[i];
            if (nextIndex == _numLowFields) {
               break;
            }
         }
      }

      // Copy any high-index fields.
      if (!_highFields.isEmpty()) {
         for (Integer integer : _highFields.keySet()) {
            fields[nextIndex++] = _highFields.get(integer);
         }
      }

      return fields;
   }

   /**
    * Removes all TIFF fields from the directory.
    */
   protected final void removeTIFFFields() {
      Arrays.fill(_lowFields, null);
      _numLowFields = 0;
      _highFields.clear();
   }

   /**
    * Clones the directory and all the fields contained therein.
    *
    * @return A clone of this <code>TIFFDirectory</code>.
    */
   @Override
   public final Object clone() throws CloneNotSupportedException {
      super.clone();
      TIFFDirectory dir = new TIFFDirectory(getTagSets(), getParentTag());
      TIFFField[] fields = getTIFFFields();
      for (TIFFField field : fields) {
         dir.addTIFFField(field);
      }

      return dir;
   }
}
