/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */

package com.sun.media.imageio.plugins.tiff;

import java.awt.*;
import java.awt.image.*;

class TIFFImageUtil {

    /**
     * For the case of binary data (<code>isBinary()</code> returns
     * <code>true</code>), return the binary data as a packed byte array.
     * The data will be packed as eight bits per byte with no bit offset,
     * i.e., the first bit in each image line will be the left-most of the
     * first byte of the line.  The line stride in bytes will be
     * <code>(int)((getWidth()+7)/8)</code>.  The length of the returned
     * array will be the line stride multiplied by <code>getHeight()</code>
     *
     * @return the binary data as a packed array of bytes with zero offset
     * of <code>null</code> if the data are not binary.
     * @throws IllegalArgumentException if <code>isBinary()</code> returns
     *                                  <code>false</code> with the <code>SampleModel</code> of the
     *                                  supplied <code>Raster</code> as argument.
     */
    public static byte[] getPackedBinaryData( Raster raster, Rectangle rect ) {
        SampleModel sm = raster.getSampleModel();
        if (!isBinary(sm)) {
            throw new IllegalArgumentException("Not a binary file.");
        }

        int rectX = rect.x;
        int rectY = rect.y;
        int rectWidth = rect.width;
        int rectHeight = rect.height;

        DataBuffer dataBuffer = raster.getDataBuffer();

        int dx = rectX - raster.getSampleModelTranslateX();
        int dy = rectY - raster.getSampleModelTranslateY();

        MultiPixelPackedSampleModel mpp = (MultiPixelPackedSampleModel) sm;
        int lineStride = mpp.getScanlineStride();
        int eltOffset = dataBuffer.getOffset() + mpp.getOffset(dx, dy);
        int bitOffset = mpp.getBitOffset(dx);

        int numBytesPerRow = (rectWidth + 7) / 8;
        if (dataBuffer instanceof DataBufferByte &&
                eltOffset == 0 && bitOffset == 0 &&
                numBytesPerRow == lineStride &&
                ((DataBufferByte) dataBuffer).getData().length ==
                        numBytesPerRow * rectHeight) {
            return ((DataBufferByte) dataBuffer).getData();
        }

        byte[] binaryDataArray = new byte[numBytesPerRow * rectHeight];

        int b = 0;

        if (bitOffset == 0) {
            if (dataBuffer instanceof DataBufferByte) {
                byte[] data = ((DataBufferByte) dataBuffer).getData();
                int offset = 0;
                for (int y = 0; y < rectHeight; y++) {
                    System.arraycopy(data, eltOffset,
                            binaryDataArray, offset,
                            numBytesPerRow);
                    offset += numBytesPerRow;
                    eltOffset += lineStride;
                }
            } else if (dataBuffer instanceof DataBufferShort ||
                    dataBuffer instanceof DataBufferUShort) {
                short[] data = dataBuffer instanceof DataBufferShort ?
                        ((DataBufferShort) dataBuffer).getData() :
                        ((DataBufferUShort) dataBuffer).getData();

                for (int y = 0; y < rectHeight; y++) {
                    int xRemaining = rectWidth;
                    int i = eltOffset;
                    while (xRemaining > 8) {
                        short datum = data[i++];
                        binaryDataArray[b++] = (byte) ((datum >>> 8) & 0xFF);
                        binaryDataArray[b++] = (byte) (datum & 0xFF);
                        xRemaining -= 16;
                    }
                    if (xRemaining > 0) {
                        binaryDataArray[b++] = (byte) ((data[i] >>> 8) & 0XFF);
                    }
                    eltOffset += lineStride;
                }
            } else if (dataBuffer instanceof DataBufferInt) {
                int[] data = ((DataBufferInt) dataBuffer).getData();

                for (int y = 0; y < rectHeight; y++) {
                    int xRemaining = rectWidth;
                    int i = eltOffset;
                    while (xRemaining > 24) {
                        int datum = data[i++];
                        binaryDataArray[b++] = (byte) ((datum >>> 24) & 0xFF);
                        binaryDataArray[b++] = (byte) ((datum >>> 16) & 0xFF);
                        binaryDataArray[b++] = (byte) ((datum >>> 8) & 0xFF);
                        binaryDataArray[b++] = (byte) (datum & 0xFF);
                        xRemaining -= 32;
                    }
                    int shift = 24;
                    while (xRemaining > 0) {
                        binaryDataArray[b++] =
                                (byte) ((data[i] >>> shift) & 0xFF);
                        shift -= 8;
                        xRemaining -= 8;
                    }
                    eltOffset += lineStride;
                }
            }
        } else { // bitOffset != 0
            if (dataBuffer instanceof DataBufferByte) {
                byte[] data = ((DataBufferByte) dataBuffer).getData();

                if ((bitOffset & 7) == 0) {
                    int offset = 0;
                    for (int y = 0; y < rectHeight; y++) {
                        System.arraycopy(data, eltOffset,
                                binaryDataArray, offset,
                                numBytesPerRow);
                        offset += numBytesPerRow;
                        eltOffset += lineStride;
                    }
                } else { // bitOffset % 8 != 0
                    int leftShift = bitOffset & 7;
                    int rightShift = 8 - leftShift;
                    for (int y = 0; y < rectHeight; y++) {
                        int i = eltOffset;
                        int xRemaining = rectWidth;
                        while (xRemaining > 0) {
                            if (xRemaining > rightShift) {
                                binaryDataArray[b++] =
                                        (byte) (((data[i++] & 0xFF) << leftShift) |
                                                ((data[i] & 0xFF) >>> rightShift));
                            } else {
                                binaryDataArray[b++] =
                                        (byte) ((data[i] & 0xFF) << leftShift);
                            }
                            xRemaining -= 8;
                        }
                        eltOffset += lineStride;
                    }
                }
            } else if (dataBuffer instanceof DataBufferShort ||
                    dataBuffer instanceof DataBufferUShort) {
                short[] data = dataBuffer instanceof DataBufferShort ?
                        ((DataBufferShort) dataBuffer).getData() :
                        ((DataBufferUShort) dataBuffer).getData();

                for (int y = 0; y < rectHeight; y++) {
                    int bOffset = bitOffset;
                    for (int x = 0; x < rectWidth; x += 8, bOffset += 8) {
                        int i = eltOffset + bOffset / 16;
                        int mod = bOffset % 16;
                        int left = data[i] & 0xFFFF;
                        if (mod <= 8) {
                            binaryDataArray[b++] = (byte) (left >>> (8 - mod));
                        } else {
                            int delta = mod - 8;
                            int right = data[i + 1] & 0xFFFF;
                            binaryDataArray[b++] =
                                    (byte) ((left << delta) |
                                            (right >>> (16 - delta)));
                        }
                    }
                    eltOffset += lineStride;
                }
            } else if (dataBuffer instanceof DataBufferInt) {
                int[] data = ((DataBufferInt) dataBuffer).getData();

                for (int y = 0; y < rectHeight; y++) {
                    int bOffset = bitOffset;
                    for (int x = 0; x < rectWidth; x += 8, bOffset += 8) {
                        int i = eltOffset + bOffset / 32;
                        int mod = bOffset % 32;
                        int left = data[i];
                        if (mod <= 24) {
                            binaryDataArray[b++] =
                                    (byte) (left >>> (24 - mod));
                        } else {
                            int delta = mod - 24;
                            int right = data[i + 1];
                            binaryDataArray[b++] =
                                    (byte) ((left << delta) |
                                            (right >>> (32 - delta)));
                        }
                    }
                    eltOffset += lineStride;
                }
            }
        }

        return binaryDataArray;
    }

    /**
     * Sets the supplied <code>Raster</code>'s data from an array
     * of packed binary data of the form returned by
     * <code>getPackedBinaryData()</code>.
     *
     * @throws IllegalArgumentException if <code>isBinary()</code> returns
     *                                  <code>false</code> with the <code>SampleModel</code> of the
     *                                  supplied <code>Raster</code> as argument.
     */
    public static void setPackedBinaryData( byte[] binaryDataArray, WritableRaster raster, Rectangle rect ) {
        SampleModel sm = raster.getSampleModel();
        if (!isBinary(sm)) {
            throw new IllegalArgumentException("Not a binary file.");
        }

        int rectX = rect.x;
        int rectY = rect.y;
        int rectWidth = rect.width;
        int rectHeight = rect.height;

        DataBuffer dataBuffer = raster.getDataBuffer();

        int dx = rectX - raster.getSampleModelTranslateX();
        int dy = rectY - raster.getSampleModelTranslateY();

        MultiPixelPackedSampleModel mpp = (MultiPixelPackedSampleModel) sm;
        int lineStride = mpp.getScanlineStride();
        int eltOffset = dataBuffer.getOffset() + mpp.getOffset(dx, dy);
        int bitOffset = mpp.getBitOffset(dx);

        int b = 0;

        if (bitOffset == 0) {
            if (dataBuffer instanceof DataBufferByte) {
                byte[] data = ((DataBufferByte) dataBuffer).getData();
                if (data == binaryDataArray) {
                    // Optimal case: simply return.
                    return;
                }
                int stride = (rectWidth + 7) / 8;
                int offset = 0;
                for (int y = 0; y < rectHeight; y++) {
                    System.arraycopy(binaryDataArray, offset,
                            data, eltOffset,
                            stride);
                    offset += stride;
                    eltOffset += lineStride;
                }
            } else if (dataBuffer instanceof DataBufferShort ||
                    dataBuffer instanceof DataBufferUShort) {
                short[] data = dataBuffer instanceof DataBufferShort ?
                        ((DataBufferShort) dataBuffer).getData() :
                        ((DataBufferUShort) dataBuffer).getData();

                for (int y = 0; y < rectHeight; y++) {
                    int xRemaining = rectWidth;
                    int i = eltOffset;
                    while (xRemaining > 8) {
                        data[i++] =
                                (short) (((binaryDataArray[b++] & 0xFF) << 8) |
                                        (binaryDataArray[b++] & 0xFF));
                        xRemaining -= 16;
                    }
                    if (xRemaining > 0) {
                        data[i] = (short) ((binaryDataArray[b++] & 0xFF) << 8);
                    }
                    eltOffset += lineStride;
                }
            } else if (dataBuffer instanceof DataBufferInt) {
                int[] data = ((DataBufferInt) dataBuffer).getData();

                for (int y = 0; y < rectHeight; y++) {
                    int xRemaining = rectWidth;
                    int i = eltOffset;
                    while (xRemaining > 24) {
                        data[i++] = ((binaryDataArray[b++] & 0xFF) << 24) |
                                        ((binaryDataArray[b++] & 0xFF) << 16) |
                                        ((binaryDataArray[b++] & 0xFF) << 8) |
                                        (binaryDataArray[b++] & 0xFF);
                        xRemaining -= 32;
                    }
                    int shift = 24;
                    while (xRemaining > 0) {
                        data[i] |= (binaryDataArray[b++] & 0xFF) << shift;
                        shift -= 8;
                        xRemaining -= 8;
                    }
                    eltOffset += lineStride;
                }
            }
        } else { // bitOffset != 0
            int stride = (rectWidth + 7) / 8;
            int offset = 0;
            if (dataBuffer instanceof DataBufferByte) {
                byte[] data = ((DataBufferByte) dataBuffer).getData();

                if ((bitOffset & 7) == 0) {
                    for (int y = 0; y < rectHeight; y++) {
                        System.arraycopy(binaryDataArray, offset,
                                data, eltOffset,
                                stride);
                        offset += stride;
                        eltOffset += lineStride;
                    }
                } else { // bitOffset % 8 != 0
                    int rightShift = bitOffset & 7;
                    int leftShift = 8 - rightShift;
                    int leftShift8 = 8 + leftShift;
                    int mask = (byte) (255 << leftShift);
                    int mask1 = (byte) ~mask;

                    for (int y = 0; y < rectHeight; y++) {
                        int i = eltOffset;
                        int xRemaining = rectWidth;
                        while (xRemaining > 0) {
                            byte datum = binaryDataArray[b++];

                            if (xRemaining > leftShift8) {
                                // when all the bits in this BYTE will be set
                                // into the data buffer.
                                data[i] = (byte) ((data[i] & mask) |
                                        ((datum & 0xFF) >>> rightShift));
                                data[++i] = (byte) ((datum & 0xFF) << leftShift);
                            } else if (xRemaining > leftShift) {
                                // All the "leftShift" high bits will be set
                                // into the data buffer.  But not all the
                                // "rightShift" low bits will be set.
                                data[i] = (byte) ((data[i] & mask) |
                                        ((datum & 0xFF) >>> rightShift));
                                i++;
                                data[i] =
                                        (byte) ((data[i] & mask1) | ((datum & 0xFF) << leftShift));
                            } else {
                                // Less than "leftShift" high bits will be set.
                                int remainMask = (1 << leftShift - xRemaining) - 1;
                                data[i] =
                                        (byte) ((data[i] & (mask | remainMask)) |
                                                (datum & 0xFF) >>> rightShift & ~remainMask);
                            }
                            xRemaining -= 8;
                        }
                        eltOffset += lineStride;
                    }
                }
            } else if (dataBuffer instanceof DataBufferShort ||
                    dataBuffer instanceof DataBufferUShort) {
                short[] data = dataBuffer instanceof DataBufferShort ?
                        ((DataBufferShort) dataBuffer).getData() :
                        ((DataBufferUShort) dataBuffer).getData();

                int rightShift = bitOffset & 7;
                int leftShift = 8 - rightShift;
                int leftShift16 = 16 + leftShift;
                int mask = (short) (~(255 << leftShift));
                int mask1 = (short) (65535 << leftShift);
                int mask2 = (short) ~mask1;

                for (int y = 0; y < rectHeight; y++) {
                    int bOffset = bitOffset;
                    int xRemaining = rectWidth;
                    for (int x = 0; x < rectWidth;
                         x += 8, bOffset += 8, xRemaining -= 8) {
                        int i = eltOffset + (bOffset >> 4);
                        int mod = bOffset & 15;
                        int datum = binaryDataArray[b++] & 0xFF;
                        if (mod <= 8) {
                            // This BYTE is set into one SHORT
                            if (xRemaining < 8) {
                                // Mask the bits to be set.
                                datum &= 255 << 8 - xRemaining;
                            }
                            data[i] = (short) ((data[i] & mask) | (datum << leftShift));
                        } else if (xRemaining > leftShift16) {
                            // This BYTE will be set into two SHORTs
                            data[i] = (short) ((data[i] & mask1) | ((datum >>> rightShift) & 0xFFFF));
                            data[++i] =
                                    (short) ((datum << leftShift) & 0xFFFF);
                        } else if (xRemaining > leftShift) {
                            // This BYTE will be set into two SHORTs;
                            // But not all the low bits will be set into SHORT
                            data[i] = (short) ((data[i] & mask1) | ((datum >>> rightShift) & 0xFFFF));
                            i++;
                            data[i] =
                                    (short) ((data[i] & mask2) | ((datum << leftShift) & 0xFFFF));
                        } else {
                            // Only some of the high bits will be set into
                            // SHORTs
                            int remainMask = (1 << leftShift - xRemaining) - 1;
                            data[i] = (short) ((data[i] & (mask1 | remainMask)) |
                                    ((datum >>> rightShift) & 0xFFFF & ~remainMask));
                        }
                    }
                    eltOffset += lineStride;
                }
            } else if (dataBuffer instanceof DataBufferInt) {
                int[] data = ((DataBufferInt) dataBuffer).getData();
                int rightShift = bitOffset & 7;
                int leftShift = 8 - rightShift;
                int leftShift32 = 32 + leftShift;
                int mask = 0xFFFFFFFF << leftShift;
                int mask1 = ~mask;

                for (int y = 0; y < rectHeight; y++) {
                    int bOffset = bitOffset;
                    int xRemaining = rectWidth;
                    for (int x = 0; x < rectWidth;
                         x += 8, bOffset += 8, xRemaining -= 8) {
                        int i = eltOffset + (bOffset >> 5);
                        int mod = bOffset & 31;
                        int datum = binaryDataArray[b++] & 0xFF;
                        if (mod <= 24) {
                            // This BYTE is set into one INT
                            int shift = 24 - mod;
                            if (xRemaining < 8) {
                                // Mask the bits to be set.
                                datum &= 255 << 8 - xRemaining;
                            }
                            data[i] = (data[i] & (~(255 << shift))) | (datum << shift);
                        } else if (xRemaining > leftShift32) {
                            // All the bits of this BYTE will be set into two INTs
                            data[i] = (data[i] & mask) | (datum >>> rightShift);
                            data[++i] = datum << leftShift;
                        } else if (xRemaining > leftShift) {
                            // This BYTE will be set into two INTs;
                            // But not all the low bits will be set into INT
                            data[i] = (data[i] & mask) | (datum >>> rightShift);
                            i++;
                            data[i] = (data[i] & mask1) | (datum << leftShift);
                        } else {
                            // Only some of the high bits will be set into INT
                            int remainMask = (1 << leftShift - xRemaining) - 1;
                            data[i] = (data[i] & (mask | remainMask)) |
                                    (datum >>> rightShift & ~remainMask);
                        }
                    }
                    eltOffset += lineStride;
                }
            }
        }
    }

    public static boolean isBinary( SampleModel sm ) {
        return sm instanceof MultiPixelPackedSampleModel &&
                ((MultiPixelPackedSampleModel) sm).getPixelBitStride() == 1 &&
                sm.getNumBands() == 1;
    }

    /**
     * Returns whether the image has contiguous data across rows.
     */
    public static boolean imageIsContiguous( RenderedImage image ) {
        SampleModel sm;
        if (image instanceof BufferedImage) {
            WritableRaster ras = ((BufferedImage) image).getRaster();
            sm = ras.getSampleModel();
        } else {
            sm = image.getSampleModel();
        }

        if (sm instanceof ComponentSampleModel) {
            // Ensure image rows samples are stored contiguously
            // in a single bank.
            ComponentSampleModel csm = (ComponentSampleModel) sm;

            if (csm.getPixelStride() != csm.getNumBands()) {
                return false;
            }

            int[] bandOffsets = csm.getBandOffsets();
            for (int i = 0; i < bandOffsets.length; i++) {
                if (bandOffsets[i] != i) {
                    return false;
                }
            }

            int[] bankIndices = csm.getBankIndices();
            for (int i = 0; i < bandOffsets.length; i++) {
                if (bankIndices[i] != 0) {
                    return false;
                }
            }

            return true;
        }

        // Otherwise true if and only if it's a bilevel image with
        // a MultiPixelPackedSampleModel, 1 bit per pixel, and 1 bit
        // pixel stride.
        return TIFFImageUtil.isBinary(sm);
    }
}
