/*
 * $RCSfile: TIFFTagSet.java,v $
 *
 * 
 * Copyright (c) 2005 Sun Microsystems, Inc. All  Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met: 
 * 
 * - Redistribution of source code must retain the above copyright 
 *   notice, this  list of conditions and the following disclaimer.
 * 
 * - Redistribution in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in 
 *   the documentation and/or other materials provided with the
 *   distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of 
 * contributors may be used to endorse or promote products derived 
 * from this software without specific prior written permission.
 * 
 * This software is provided "AS IS," without a warranty of any 
 * kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND 
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
 * EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL
 * NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF 
 * USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR 
 * ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
 * CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND
 * REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES. 
 * 
 * You acknowledge that this software is not designed or intended for 
 * use in the design, construction, operation or maintenance of any 
 * nuclear facility. 
 *
 * $Revision: 1.1 $
 * $Date: 2005/02/11 05:01:19 $
 * $State: Exp $
 */
package com.sun.media.imageio.plugins.tiff;

// Should implement Set?

import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * A class representing a set of TIFF tags.  Each tag in the set must
 * have a unique number (this is a limitation of the TIFF
 * specification itself).
 *
 * <p> This class and its subclasses are responsible for mapping
 * between raw tag numbers and <code>TIFFTag</code> objects, which
 * contain additional information about each tag, such as the tag's
 * name, legal data types, and mnemonic names for some or all of ts
 * data values.
 *
 * @see TIFFTag
 */
public class TIFFTagSet {

	private final SortedMap<Integer, TIFFTag> _allowedTagsByNumber = new TreeMap<>();

	private final SortedMap<String, TIFFTag> _allowedTagsByName = new TreeMap<>();

	/**
	 * Constructs a <code>TIFFTagSet</code>, given a <code>List</code>
	 * of <code>TIFFTag</code> objects.
	 *
	 * @param tags a <code>List</code> object containing
	 * <code>TIFFTag</code> objects to be added to this tag set.
	 *
	 * @throws IllegalArgumentException if <code>tags</code> is
	 * <code>null</code>, or contains objects that are not instances
	 *     of the <code>TIFFTag</code> class.
	 */
    TIFFTagSet( List<TIFFTag> tags ) {
		if (tags == null) {
			throw new IllegalArgumentException("tags == null!");
		}
        for (TIFFTag tag : tags) {
            _allowedTagsByNumber.put(tag.getNumber(), tag);
            _allowedTagsByName.put(tag.getName(), tag);
        }
	}

	/**
	 * Returns the <code>TIFFTag</code> from this set that is
	 * associated with the given tag number, or <code>null</code> if
	 * no tag exists for that number.
	 *
	 * @param tagNumber the number of the tag to be retrieved.
	 *
	 * @return the numbered <code>TIFFTag</code>, or <code>null</code>.
	 */
	public final TIFFTag getTag( int tagNumber ) {
		return _allowedTagsByNumber.get(tagNumber);
	}

	/**
	 * Returns the <code>TIFFTag</code> having the given tag name, or
	 * <code>null</code> if the named tag does not belong to this tag set.
	 *
	 * @param tagName the name of the tag to be retrieved, as a
	 * <code>String</code>.
	 *
	 * @return the named <code>TIFFTag</code>, or <code>null</code>.
	 *
	 * @throws IllegalArgumentException if <code>tagName</code> is
	 * <code>null</code>.
	 */
	public final TIFFTag getTag( String tagName ) {
		if (tagName == null) {
			throw new IllegalArgumentException("tagName == null!");
		}
		return _allowedTagsByName.get(tagName);
	}

}
