/*
 * $RCSfile: SimpleCMYKColorSpace.java,v $
 *
 *
 * Copyright (c) 2005 Sun Microsystems, Inc. All  Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistribution of source code must retain the above copyright
 *   notice, this  list of conditions and the following disclaimer.
 *
 * - Redistribution in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * This software is provided "AS IS," without a warranty of any
 * kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
 * EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL
 * NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF
 * USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR
 * ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
 * CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND
 * REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * You acknowledge that this software is not designed or intended for
 * use in the design, construction, operation or maintenance of any
 * nuclear facility.
 *
 * $Revision: 1.5 $
 * $Date: 2006/02/17 19:08:24 $
 * $State: Exp $
 */
package com.sun.media.imageioimpl.common;

import java.awt.color.ColorSpace;

/**
 * Singleton class representing a simple, mathematically defined CMYK
 * color space.
 */
public final class SimpleCMYKColorSpace extends ColorSpace {
    /**
     * The exponent for gamma correction.
     */
    private static final double POWER1 = 1.0 / 2.4;
    private static ColorSpace _theInstance;
    private final ColorSpace _csRGB;

    private SimpleCMYKColorSpace() {
        super(TYPE_CMYK, 4);
        _csRGB = ColorSpace.getInstance(ColorSpace.CS_LINEAR_RGB);
    }

    public static synchronized ColorSpace getInstance() {
        if (_theInstance == null) {
            _theInstance = new SimpleCMYKColorSpace();
        }
        return _theInstance;
    }

    @Override
    public boolean equals( Object object ) {
        return object instanceof SimpleCMYKColorSpace;
    }

    @Override
    public float[] toRGB( float[] colorvalue ) {
        float cC = colorvalue[0];
        float mM = colorvalue[1];
        float yY = colorvalue[2];
        float kK = colorvalue[3];

        float kK1 = 1.0F - kK;

        // Convert from CMYK to linear RGB.
        float[] rgbvalue = new float[]{kK1 * (1.0F - cC),
                kK1 * (1.0F - mM),
                kK1 * (1.0F - yY)};

        // Convert from linear RGB to sRGB.
        for (int i = 0; i < 3; i++) {
            float value = rgbvalue[i];

            if (value < 0.0F) {
                value = 0.0F;
            }

            if (value < 0.0031308F) {
                rgbvalue[i] = 12.92F * value;
            } else {
                if (value > 1.0F) {
                    value = 1.0F;
                }
                rgbvalue[i] = (float) (1.055 * Math.pow(value, POWER1) - 0.055);
            }
        }

        return rgbvalue;
    }

    @Override
    public float[] fromRGB( float[] rgbvalue ) {
        // Convert from sRGB to linear RGB.
        for (int i = 0; i < 3; i++) {
            if (rgbvalue[i] < 0.040449936F) {
                rgbvalue[i] /= 12.92F;
            } else {
                rgbvalue[i] = (float) (Math.pow((rgbvalue[i] + 0.055) / 1.055, 2.4));
            }
        }

        // Convert from linear RGB to CMYK.
        float cC = 1.0F - rgbvalue[0];
        float mM = 1.0F - rgbvalue[1];
        float yY = 1.0F - rgbvalue[2];
        float kK = Math.min(cC, Math.min(mM, yY));

        // If K == 1.0F, then C = M = Y = 1.0F.
        if (kK != 1.0F) {
            float kK1 = 1.0F - kK;
            cC = (cC - kK) / kK1;
            mM = (mM - kK) / kK1;
            yY = (yY - kK) / kK1;
        } else {
            cC = 0;
            mM = 0;
            yY = 0;
        }

        return new float[]{cC, mM, yY, kK};
    }

    @Override
    public float[] toCIEXYZ( float[] colorvalue ) {
        return _csRGB.toCIEXYZ(toRGB(colorvalue));
    }

    @Override
    public float[] fromCIEXYZ( float[] xyzvalue ) {
        return fromRGB(_csRGB.fromCIEXYZ(xyzvalue));
    }
}
