/*
 * $RCSfile: TIFFIFD.java,v $
 *
 *
 * Copyright (c) 2005 Sun Microsystems, Inc. All  Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistribution of source code must retain the above copyright
 *   notice, this  list of conditions and the following disclaimer.
 *
 * - Redistribution in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * This software is provided "AS IS," without a warranty of any
 * kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
 * EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL
 * NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF
 * USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR
 * ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
 * CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND
 * REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * You acknowledge that this software is not designed or intended for
 * use in the design, construction, operation or maintenance of any
 * nuclear facility.
 *
 * $Revision: 1.7 $
 * $Date: 2006/09/27 23:56:30 $
 * $State: Exp $
 */
package com.sun.media.imageioimpl.plugins.tiff;

import java.io.EOFException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import javax.imageio.stream.ImageInputStream;

import com.sun.media.imageio.plugins.tiff.BaselineTIFFTagSet;
import com.sun.media.imageio.plugins.tiff.TIFFDirectory;
import com.sun.media.imageio.plugins.tiff.TIFFField;
import com.sun.media.imageio.plugins.tiff.TIFFTag;
import com.sun.media.imageio.plugins.tiff.TIFFTagSet;

public final class TIFFIFD extends TIFFDirectory {

    private TIFFIFD( List<TIFFTagSet> tagSets, TIFFTag parentTag ) {
        super(tagSets.toArray(new TIFFTagSet[tagSets.size()]),
                parentTag);
    }

    TIFFIFD( List<TIFFTagSet> tagSets ) {
        this(tagSets, null);
    }

    public static TIFFTag getTag( int tagNumber, List<TIFFTagSet> tagSets ) {
        for (TIFFTagSet tagSet : tagSets) {
            TIFFTag tag = tagSet.getTag(tagNumber);
            if (tag != null) {
                return tag;
            }
        }

        return null;
    }

    public static TIFFTag getTag( String tagName, List<TIFFTagSet> tagSets ) {
        for (TIFFTagSet tagSet : tagSets) {
            TIFFTag tag = tagSet.getTag(tagName);
            if (tag != null) {
                return tag;
            }
        }

        return null;
    }

    public List<TIFFTagSet> getTagSetList() {
        return Arrays.asList(getTagSets());
    }

    /**
     * Returns an <code>Iterator</code> over the TIFF fields. The
     * traversal is in the order of increasing tag number.
     */
    // Note: the sort is guaranteed for low fields by the use of an
    // array wherein the index corresponds to the tag number and for
    // the high fields by the use of a TreeMap with tag number keys.
    public Iterator<TIFFField> iterator() {
        return Arrays.asList(getTIFFFields()).iterator();
    }

    // Stream position initially at beginning, left at end
    // if ignoreUnknownFields is true, do not load fields for which
    // a tag cannot be found in an allowed TagSet.
    public void initialize( ImageInputStream stream,
                            boolean ignoreUnknownFields ) throws IOException {
        removeTIFFFields();

        List<TIFFTagSet> tagSetList = getTagSetList();

        int numEntries = stream.readUnsignedShort();
        for (int i = 0; i < numEntries; i++) {
            // Read tag number, value type, and value count.
            int tag = stream.readUnsignedShort();
            int type = stream.readUnsignedShort();
            int count = (int) stream.readUnsignedInt();

            // Get the associated TIFFTag.
            TIFFTag tiffTag = getTag(tag, tagSetList);

            // Ignore unknown fields.
            if (ignoreUnknownFields && tiffTag == null) {
                // Skip the value/offset so as to leave the stream
                // position at the start of the next IFD entry.
                stream.skipBytes(4);
                // Continue with the next IFD entry.
                continue;
            }

            long nextTagOffset = stream.getStreamPosition() + 4;

            int sizeOfType = TIFFTag.getSizeOfType(type);
            if (count * sizeOfType > 4) {
                long value = stream.readUnsignedInt();
                stream.seek(value);
            }

            if (tag == BaselineTIFFTagSet.TAG_STRIP_BYTE_COUNTS
                    || tag == BaselineTIFFTagSet.TAG_TILE_BYTE_COUNTS
                    || tag == BaselineTIFFTagSet.TAG_JPEG_INTERCHANGE_FORMAT_LENGTH) {
                long stripOrTileByteCountsPosition = stream.getStreamPosition();
            } else if (tag == BaselineTIFFTagSet.TAG_STRIP_OFFSETS
                    || tag == BaselineTIFFTagSet.TAG_TILE_OFFSETS
                    || tag == BaselineTIFFTagSet.TAG_JPEG_INTERCHANGE_FORMAT) {
                long stripOrTileOffsetsPosition = stream.getStreamPosition();
            }

            Object obj = null;

            try {
                switch (type) {
                    case TIFFTag.TIFF_BYTE:
                    case TIFFTag.TIFF_SBYTE:
                    case TIFFTag.TIFF_UNDEFINED:
                    case TIFFTag.TIFF_ASCII:
                        byte[] bvalues = new byte[count];
                        stream.readFully(bvalues, 0, count);

                        if (type == TIFFTag.TIFF_ASCII) {
                            // Can be multiple strings
                            List<String> vector = new ArrayList<>();
                            boolean inString = false;
                            int prevIndex = 0;
                            for (int index = 0; index <= count; index++) {
                                if (index < count && bvalues[index] != 0) {
                                    if (!inString) {
                                        // start of string
                                        prevIndex = index;
                                        inString = true;
                                    }
                                } else // null or special case at end of string
                                    if (inString) {
                                        // end of string
                                        String string = new String(bvalues, prevIndex,
                                                index - prevIndex);
                                        vector.add(string);
                                        inString = false;
                                    }
                            }

                            count = vector.size();
                            String[] strings;
                            if (count != 0) {
                                strings = new String[count];
                                for (int c = 0; c < count; c++) {
                                    strings[c] = vector.get(c);
                                }
                            } else {
                                // This case has been observed when the value of
                                // 'count' recorded in the field is non-zero but
                                // the value portion contains all nulls.
                                count = 1;
                                strings = new String[]{""};
                            }

                            obj = strings;
                        } else {
                            obj = bvalues;
                        }
                        break;

                    case TIFFTag.TIFF_SHORT:
                        char[] cvalues = new char[count];
                        for (int j = 0; j < count; j++) {
                            cvalues[j] = (char) (stream.readUnsignedShort());
                        }
                        obj = cvalues;
                        break;

                    case TIFFTag.TIFF_LONG:
                    case TIFFTag.TIFF_IFD_POINTER:
                        long[] lvalues = new long[count];
                        for (int j = 0; j < count; j++) {
                            lvalues[j] = stream.readUnsignedInt();
                        }
                        obj = lvalues;
                        break;

                    case TIFFTag.TIFF_RATIONAL:
                        long[][] llvalues = new long[count][2];
                        for (int j = 0; j < count; j++) {
                            llvalues[j][0] = stream.readUnsignedInt();
                            llvalues[j][1] = stream.readUnsignedInt();
                        }
                        obj = llvalues;
                        break;

                    case TIFFTag.TIFF_SSHORT:
                        short[] svalues = new short[count];
                        for (int j = 0; j < count; j++) {
                            svalues[j] = stream.readShort();
                        }
                        obj = svalues;
                        break;

                    case TIFFTag.TIFF_SLONG:
                        int[] ivalues = new int[count];
                        for (int j = 0; j < count; j++) {
                            ivalues[j] = stream.readInt();
                        }
                        obj = ivalues;
                        break;

                    case TIFFTag.TIFF_SRATIONAL:
                        int[][] iivalues = new int[count][2];
                        for (int j = 0; j < count; j++) {
                            iivalues[j][0] = stream.readInt();
                            iivalues[j][1] = stream.readInt();
                        }
                        obj = iivalues;
                        break;

                    case TIFFTag.TIFF_FLOAT:
                        float[] fvalues = new float[count];
                        for (int j = 0; j < count; j++) {
                            fvalues[j] = stream.readFloat();
                        }
                        obj = fvalues;
                        break;

                    case TIFFTag.TIFF_DOUBLE:
                        double[] dvalues = new double[count];
                        for (int j = 0; j < count; j++) {
                            dvalues[j] = stream.readDouble();
                        }
                        obj = dvalues;
                        break;

                    default:
                        break;
                }
            } catch (EOFException eofe) {
                // The TIFF 6.0 fields have tag numbers less than or equal
                // to 532 (ReferenceBlackWhite) or equal to 33432 (Copyright).
                // If there is an error reading a baseline tag, then re-throw
                // the exception and fail; otherwise continue with the next
                // field.
                if (BaselineTIFFTagSet.getInstance().getTag(tag) == null) {
                    throw eofe;
                }
            }

            if (tiffTag == null) {
            } else if (!tiffTag.isDataTypeOK(type)) {
            } else if (tiffTag.isIFDPointer() && obj != null) {
                stream.mark();
                stream.seek(((long[]) obj)[0]);
                List<TIFFTagSet> tagSets = new ArrayList<>(1);
                tagSets.add(tiffTag.getTagSet());
                TIFFIFD subIFD = new TIFFIFD(tagSets);
                subIFD.initialize(stream, ignoreUnknownFields);
                obj = subIFD;
                stream.reset();
            }

            if (tiffTag == null) {
                tiffTag = new TIFFTag(null, tag, 1 << type, null);
            }

            // Add the field if its contents have been initialized which
            // will not be the case if an EOF was ignored above.
            if (obj != null) {
                TIFFField field = new TIFFField(tiffTag, type, count, obj);
                addTIFFField(field);
            }

            stream.seek(nextTagOffset);
        }

        long lastPosition = stream.getStreamPosition();
    }
}
