/*
 * $RCSfile: TIFFImageReader.java,v $
 *
 *
 * Copyright (c) 2005 Sun Microsystems, Inc. All  Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistribution of source code must retain the above copyright
 *   notice, this  list of conditions and the following disclaimer.
 *
 * - Redistribution in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * This software is provided "AS IS," without a warranty of any
 * kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
 * EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL
 * NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF
 * USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR
 * ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
 * CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND
 * REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * You acknowledge that this software is not designed or intended for
 * use in the design, construction, operation or maintenance of any
 * nuclear facility.
 *
 * $Revision: 1.10 $
 * $Date: 2006/06/28 21:29:03 $
 * $State: Exp $
 */
package com.sun.media.imageioimpl.plugins.tiff;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.color.ColorSpace;
import java.awt.color.ICC_ColorSpace;
import java.awt.color.ICC_Profile;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.Raster;
import java.awt.image.RenderedImage;
import java.awt.image.SampleModel;
import java.io.IOException;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.imageio.IIOException;
import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.ImageReadParam;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.spi.ImageReaderSpi;
import javax.imageio.stream.ImageInputStream;

import org.w3c.dom.Node;

import com.sun.media.imageio.plugins.tiff.BaselineTIFFTagSet;
import com.sun.media.imageio.plugins.tiff.TIFFColorConverter;
import com.sun.media.imageio.plugins.tiff.TIFFDecompressor;
import com.sun.media.imageio.plugins.tiff.TIFFFaxDecompressor;
import com.sun.media.imageio.plugins.tiff.TIFFField;
import com.sun.media.imageio.plugins.tiff.TIFFImageReadParam;
import com.sun.media.imageio.plugins.tiff.TIFFTagSet;

final class TIFFImageReader extends ImageReader {
    // The current ImageInputStream source.
    private ImageInputStream _stream;

    // True if the file header has been read.
    private boolean _gotHeader;

    private ImageReadParam _imageReadParam = getDefaultReadParam();

    // Stream metadata, or null.
    private TIFFStreamMetadata _streamMetadata;

    // The current image index.
    private int _currIndex = -1;

    // Metadata for image at 'currIndex', or null.
    private TIFFImageMetadata _imageMetadata;

    // A <code>List</code> of <code>Long</code>s indicating the stream
    // positions of the start of the IFD for each image.  Entries
    // are added as needed.
    private List<Long> _imageStartPosition = new ArrayList<>();

    // The number of images in the stream, if known, otherwise -1.
    private int _numImages = -1;

    // The ImageTypeSpecifiers of the images in the stream.
    // Contains a map of Integers to Lists.
    private HashMap<Integer, List<ImageTypeSpecifier>> _imageTypeMap = new HashMap<>();

    private BufferedImage _theImage;

    private int _width = -1;
    private int _height = -1;
    private int _numBands = -1;
    private int _tileOrStripWidth = -1;
    private int _tileOrStripHeight = -1;

    private int _planarConfiguration = BaselineTIFFTagSet.PLANAR_CONFIGURATION_CHUNKY;

    private int _compression;
    private int _photometricInterpretation;
    private int _samplesPerPixel;
    private int[] _sampleFormat;
    private int[] _bitsPerSample;
    private int[] _extraSamples;
    private char[] _colorMap;

    private int _sourceXOffset;
    private int _sourceYOffset;
    private int _srcXSubsampling;
    private int _srcYSubsampling;

    private int _dstWidth;
    private int _dstHeight;
    private int _dstMinX;
    private int _dstMinY;
    private int _dstXOffset;
    private int _dstYOffset;

    private int _tilesAcross;
    private int _tilesDown;

    private int _pixelsRead;
    private int _pixelsToRead;

    private int[] _sourceBands;
    private int[] _destinationBands;
    private TIFFDecompressor _decompressor;

    TIFFImageReader( ImageReaderSpi originatingProvider ) {
        super(originatingProvider);
    }

    // floor(num/den)
    private static int ifloor( int num, int den ) {
        if (num < 0) {
            num -= den - 1;
        }
        return num / den;
    }

    // ceil(num/den)
    private static int iceil( int num, int den ) {
        if (num > 0) {
            num += den - 1;
        }
        return num / den;
    }

    /**
     * Gets the destination image type.
     */
    // NOTE: Code shamelessly copied from ImageReader.getDestination().
    private static ImageTypeSpecifier getDestinationType( ImageReadParam param, Iterator<ImageTypeSpecifier> imageTypes ) throws IIOException {
        if (imageTypes == null || !imageTypes.hasNext()) {
            throw new IllegalArgumentException("imageTypes null or empty!");
        }

        ImageTypeSpecifier imageType = null;

        // If param is non-null, use it
        if (param != null) {
            imageType = param.getDestinationType();
        }

        // No info from param, use fallback image type
        if (imageType == null) {
            ImageTypeSpecifier object = imageTypes.next();
            if (object == null) {
                throw new IllegalArgumentException("Non-ImageTypeSpecifier retrieved from imageTypes!");
            }
            imageType = object;
        } else {
            boolean foundIt = false;
            while (imageTypes.hasNext()) {
                ImageTypeSpecifier type = imageTypes.next();
                if (type.equals(imageType)) {
                    foundIt = true;
                    break;
                }
            }

            if (!foundIt) {
                throw new IIOException("Destination type from ImageReadParam does not match!");
            }
        }

        return imageType;
    }

    /**
     * Converts a pixel's X coordinate into a horizontal tile index
     * relative to a given tile grid layout specified by its X offset
     * and tile width.
     * <p>
     * <p> If <code>tileWidth < 0</code>, the results of this method
     * are undefined.  If <code>tileWidth == 0</code>, an
     * <code>ArithmeticException</code> will be thrown.
     *
     * @throws ArithmeticException If <code>tileWidth == 0</code>.
     */
    private static int xToTileX( int x, int tileGridXOffset, int tileWidth ) {
        x -= tileGridXOffset;
        if (x < 0) {
            x += 1 - tileWidth;        // force round to -infinity (ceiling)
        }
        return x / tileWidth;
    }

    /**
     * Converts a pixel's Y coordinate into a vertical tile index
     * relative to a given tile grid layout specified by its Y offset
     * and tile height.
     * <p>
     * <p> If <code>tileHeight < 0</code>, the results of this method
     * are undefined.  If <code>tileHeight == 0</code>, an
     * <code>ArithmeticException</code> will be thrown.
     *
     * @throws ArithmeticException If <code>tileHeight == 0</code>.
     */
    private static int yToTileY( int y, int tileGridYOffset, int tileHeight ) {
        y -= tileGridYOffset;
        if (y < 0) {
            y += 1 - tileHeight;     // force round to -infinity (ceiling)
        }
        return y / tileHeight;
    }

    @Override
    public void setInput( Object input, boolean seekForwardOnly, boolean ignoreMetadata ) {
        super.setInput(input, seekForwardOnly, ignoreMetadata);

        // Clear all local values based on the previous stream contents.
        resetLocal();

        if (input != null) {
            if (!(input instanceof ImageInputStream)) {
                throw new IllegalArgumentException("input not an ImageInputStream!");
            }
            this._stream = (ImageInputStream) input;
        } else {
            this._stream = null;
        }
    }

    // Do not seek to the beginning of the stream so as to allow users to
    // point us at an IFD within some other file format
    private void readHeader() throws IIOException {
        if (_gotHeader) {
            return;
        }
        if (_stream == null) {
            throw new IllegalStateException("Input not set!");
        }

        // Create an object to store the stream metadata
        this._streamMetadata = new TIFFStreamMetadata();

        try {
            int byteOrder = _stream.readUnsignedShort();
            if (byteOrder == 0x4d4d) {
                _streamMetadata.byteOrder = ByteOrder.BIG_ENDIAN;
                _stream.setByteOrder(ByteOrder.BIG_ENDIAN);
            } else if (byteOrder == 0x4949) {
                _streamMetadata.byteOrder = ByteOrder.LITTLE_ENDIAN;
                _stream.setByteOrder(ByteOrder.LITTLE_ENDIAN);
            } else {
                processWarningOccurred("Bad byte order in header, assuming little-endian");
                _streamMetadata.byteOrder = ByteOrder.LITTLE_ENDIAN;
                _stream.setByteOrder(ByteOrder.LITTLE_ENDIAN);
            }

            int magic = _stream.readUnsignedShort();
            if (magic != 42) {
                processWarningOccurred("Bad magic number in header, continuing");
            }

            // Seek to start of first IFD
            long offset = _stream.readUnsignedInt();
            _imageStartPosition.add(offset);
            _stream.seek(offset);
        } catch (IOException e) {
            throw new IIOException("I/O error reading header!", e);
        }

        _gotHeader = true;
    }

    private int locateImage( int imageIndex ) throws IIOException {
        readHeader();

        try {
            // Find closest known index
            int index = Math.min(imageIndex, _imageStartPosition.size() - 1);

            // Seek to that position
            Long position = _imageStartPosition.get(index);
            _stream.seek(position);

            // Skip IFDs until at desired index or last image found
            while (index < imageIndex) {
                int count = _stream.readUnsignedShort();
                _stream.skipBytes(12 * count);

                long offset = _stream.readUnsignedInt();
                if (offset == 0) {
                    return index;
                }

                _imageStartPosition.add(offset);
                _stream.seek(offset);
                ++index;
            }
        } catch (IOException e) {
            throw new IIOException("Couldn't seek!", e);
        }

        if (_currIndex != imageIndex) {
            _imageMetadata = null;
        }
        _currIndex = imageIndex;
        return imageIndex;
    }

    @Override
    public int getNumImages( boolean allowSearch ) throws IOException {
        if (_stream == null) {
            throw new IllegalStateException("Input not set!");
        }
        if (seekForwardOnly && allowSearch) {
            throw new IllegalStateException("seekForwardOnly and allowSearch can't both be true!");
        }

        if (_numImages > 0) {
            return _numImages;
        }
        if (allowSearch) {
            this._numImages = locateImage(Integer.MAX_VALUE) + 1;
        }
        return _numImages;
    }

    @Override
    public IIOMetadata getStreamMetadata() throws IIOException {
        readHeader();
        return _streamMetadata;
    }

    // Throw an IndexOutOfBoundsException if index < minIndex,
    // and bump minIndex if required.
    private void checkIndex( int imageIndex ) {
        if (imageIndex < minIndex) {
            throw new IndexOutOfBoundsException("imageIndex < minIndex!");
        }
        if (seekForwardOnly) {
            minIndex = imageIndex;
        }
    }

    // Verify that imageIndex is in bounds, find the image IFD, read the
    // image metadata, initialize instance variables from the metadata.
    private void seekToImage( int imageIndex ) throws IIOException {
        checkIndex(imageIndex);

        int index = locateImage(imageIndex);
        if (index != imageIndex) {
            throw new IndexOutOfBoundsException("imageIndex out of bounds!");
        }

        readMetadata();

        initializeFromMetadata();
    }

    // Stream must be positioned at start of IFD for 'currIndex'
    private void readMetadata() throws IIOException {
        if (_stream == null) {
            throw new IllegalStateException("Input not set!");
        }

        if (_imageMetadata != null) {
            return;
        }
        try {
            // Create an object to store the image metadata
            List<TIFFTagSet> tagSets;
            if (_imageReadParam instanceof TIFFImageReadParam) {
                tagSets = ((TIFFImageReadParam) _imageReadParam).getAllowedTagSets();
            } else {
                tagSets = new ArrayList<>(1);
                tagSets.add(BaselineTIFFTagSet.getInstance());
            }

            this._imageMetadata = new TIFFImageMetadata(tagSets);
            _imageMetadata.initializeFromStream(_stream, ignoreMetadata);
        } catch (IIOException iioe) {
            throw iioe;
        } catch (IOException ioe) {
            throw new IIOException("I/O error reading image metadata!", ioe);
        }
    }

    private int getWidth() {
        return this._width;
    }

    private int getHeight() {
        return this._height;
    }

    // Returns tile width if image is tiled, else image width
    private int getTileOrStripWidth() {
        TIFFField field =
                _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_TILE_WIDTH);
        return (field == null) ? getWidth() : field.getAsInt(0);
    }

    // Returns tile height if image is tiled, else strip height
    private int getTileOrStripHeight() {
        TIFFField field =
                _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_TILE_LENGTH);
        if (field != null) {
            return field.getAsInt(0);
        }

        field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_ROWS_PER_STRIP);
        // Default for ROWS_PER_STRIP is 2^32 - 1, i.e., infinity
        int height = (field == null) ? -1 : field.getAsInt(0);
        return (height == -1) ? getHeight() : height;
    }

    private int getPlanarConfiguration() {
        TIFFField field =
                _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_PLANAR_CONFIGURATION);
        if (field != null) {
            int planarConfigurationValue = field.getAsInt(0);
            if (planarConfigurationValue == BaselineTIFFTagSet.PLANAR_CONFIGURATION_PLANAR) {
                // Some writers (e.g. Kofax standard Multi-Page TIFF
                // Storage Filter v2.01.000; cf. bug 4929147) do not
                // correctly set the value of this field. Attempt to
                // ascertain whether the value is correctly Planar.
                if (getCompression() == BaselineTIFFTagSet.COMPRESSION_OLD_JPEG
                        && _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_JPEG_INTERCHANGE_FORMAT) != null) {
                    // JPEG interchange format cannot have
                    // PlanarConfiguration value Chunky so reset.
                    processWarningOccurred("PlanarConfiguration \"Planar\" value inconsistent with JPEGInterchangeFormat; resetting to \"Chunky\".");
                    planarConfigurationValue =
                            BaselineTIFFTagSet.PLANAR_CONFIGURATION_CHUNKY;
                } else {
                    TIFFField offsetField =
                            _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_TILE_OFFSETS);
                    if (offsetField == null) {
                        // Tiles
                        offsetField =
                                _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_STRIP_OFFSETS);
                        int tw = getTileOrStripWidth();
                        int th = getTileOrStripHeight();
                        int tAcross = (getWidth() + tw - 1) / tw;
                        int tDown = (getHeight() + th - 1) / th;
                        int tilesPerImage = tAcross * tDown;
                        long[] offsetArray = offsetField.getAsLongs();
                        if (offsetArray != null
                                && offsetArray.length == tilesPerImage) {
                            // Length of offsets array is
                            // TilesPerImage for Chunky and
                            // SamplesPerPixel*TilesPerImage for Planar.
                            processWarningOccurred("PlanarConfiguration \"Planar\" value inconsistent with TileOffsets field value count; resetting to \"Chunky\".");
                            planarConfigurationValue =
                                    BaselineTIFFTagSet.PLANAR_CONFIGURATION_CHUNKY;
                        }
                    } else {
                        // Strips
                        int rowsPerStrip = getTileOrStripHeight();
                        int stripsPerImage = (getHeight() + rowsPerStrip - 1) / rowsPerStrip;
                        long[] offsetArray = offsetField.getAsLongs();
                        if (offsetArray != null
                                && offsetArray.length == stripsPerImage) {
                            // Length of offsets array is
                            // StripsPerImage for Chunky and
                            // SamplesPerPixel*StripsPerImage for Planar.
                            processWarningOccurred("PlanarConfiguration \"Planar\" value inconsistent with StripOffsets field value count; resetting to \"Chunky\".");
                            planarConfigurationValue =
                                    BaselineTIFFTagSet.PLANAR_CONFIGURATION_CHUNKY;
                        }
                    }
                }
            }
            return planarConfigurationValue;
        }

        return BaselineTIFFTagSet.PLANAR_CONFIGURATION_CHUNKY;
    }

    private long getTileOrStripOffset( int tileIndex ) throws IIOException {
        TIFFField field =
                _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_TILE_OFFSETS);
        if (field == null) {
            field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_STRIP_OFFSETS);
        }
        if (field == null) {
            field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_JPEG_INTERCHANGE_FORMAT);
        }

        if (field == null) {
            throw new IIOException("Missing required strip or tile offsets field.");
        }

        return field.getAsLong(tileIndex);
    }

    private long getTileOrStripByteCount( int tileIndex ) throws IOException {
        TIFFField field =
                _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_TILE_BYTE_COUNTS);
        if (field == null) {
            field =
                    _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_STRIP_BYTE_COUNTS);
        }
        if (field == null) {
            field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_JPEG_INTERCHANGE_FORMAT_LENGTH);
        }

        long tileOrStripByteCount;
        if (field != null) {
            tileOrStripByteCount = field.getAsLong(tileIndex);
        } else {
            processWarningOccurred("TIFF directory contains neither StripByteCounts nor TileByteCounts field: attempting to calculate from strip or tile width and height.");

            // Initialize to number of bytes per strip or tile assuming
            // no compression.
            int bitsPerPixel = _bitsPerSample[0];
            for (int i = 1; i < _samplesPerPixel; i++) {
                bitsPerPixel += _bitsPerSample[i];
            }
            int bytesPerRow = (getTileOrStripWidth() * bitsPerPixel + 7) / 8;
            tileOrStripByteCount = bytesPerRow * getTileOrStripHeight();

            // Clamp to end of stream if possible.
            long streamLength = _stream.length();
            if (streamLength != -1) {
                tileOrStripByteCount =
                        Math.min(tileOrStripByteCount,
                                streamLength - getTileOrStripOffset(tileIndex));
            } else {
                processWarningOccurred("Stream length is unknown: cannot clamp estimated strip or tile byte count to EOF.");
            }
        }

        return tileOrStripByteCount;
    }

    private int getCompression() {
        TIFFField field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_COMPRESSION);
        if (field == null) {
            return BaselineTIFFTagSet.COMPRESSION_NONE;
        } else {
            return field.getAsInt(0);
        }
    }

    @Override
    public int getWidth( int imageIndex ) throws IOException {
        seekToImage(imageIndex);
        return getWidth();
    }

    // Thumbnails

    @Override
    public int getHeight( int imageIndex ) throws IOException {
        seekToImage(imageIndex);
        return getHeight();
    }

    /**
     * Initializes these instance variables from the image metadata:
     * <pre>
     * compression
     * width
     * height
     * samplesPerPixel
     * numBands
     * colorMap
     * photometricInterpretation
     * sampleFormat
     * bitsPerSample
     * extraSamples
     * tileOrStripWidth
     * tileOrStripHeight
     * </pre>
     */
    private void initializeFromMetadata() {
        TIFFField field;

        // Compression
        field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_COMPRESSION);
        if (field == null) {
            processWarningOccurred("Compression field is missing; assuming no compression");
            _compression = BaselineTIFFTagSet.COMPRESSION_NONE;
        } else {
            _compression = field.getAsInt(0);
        }

        // Whether key dimensional information is absent.
        boolean isMissingDimension = false;

        // ImageWidth -> width
        field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_IMAGE_WIDTH);
        if (field != null) {
            this._width = field.getAsInt(0);
        } else {
            processWarningOccurred("ImageWidth field is missing.");
            isMissingDimension = true;
        }

        // ImageLength -> height
        field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_IMAGE_LENGTH);
        if (field != null) {
            this._height = field.getAsInt(0);
        } else {
            processWarningOccurred("ImageLength field is missing.");
            isMissingDimension = true;
        }

        // SamplesPerPixel
        field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_SAMPLES_PER_PIXEL);
        if (field != null) {
            _samplesPerPixel = field.getAsInt(0);
        } else {
            _samplesPerPixel = 1;
            isMissingDimension = true;
        }

        // If any dimension is missing and there is a JPEG stream available
        // get the information from it.
        int defaultBitDepth = 1;
        field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_JPEG_INTERCHANGE_FORMAT);
        if (isMissingDimension && field != null) {
            Iterator<ImageReader> iter = ImageIO.getImageReadersByFormatName("JPEG");
            if (iter != null && iter.hasNext()) {
                ImageReader jreader = iter.next();
                try {
                    _stream.mark();
                    _stream.seek(field.getAsLong(0));
                    jreader.setInput(_stream);
                    if (_imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_IMAGE_WIDTH) == null) {
                        this._width = jreader.getWidth(0);
                    }
                    if (_imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_IMAGE_LENGTH) == null) {
                        this._height = jreader.getHeight(0);
                    }
                    ImageTypeSpecifier imageType = jreader.getRawImageType(0);
                    if (_imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_SAMPLES_PER_PIXEL) == null) {
                        this._samplesPerPixel = imageType.getSampleModel().getNumBands();
                    }
                    _stream.reset();
                    defaultBitDepth = imageType.getColorModel().getComponentSize(0);
                } catch (IOException e) {
                    // Ignore it and proceed: an error will occur later.
                }
                jreader.dispose();
            }
        }

        if (_samplesPerPixel < 1) {
            processWarningOccurred("Samples per pixel < 1!");
        }

        // SamplesPerPixel -> numBands
        _numBands = _samplesPerPixel;

        // ColorMap
        this._colorMap = null;
        field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_COLOR_MAP);
        if (field != null) {
            // Grab color map
            _colorMap = field.getAsChars();
        }

        // PhotometricInterpretation
        field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_PHOTOMETRIC_INTERPRETATION);
        if (field == null) {
            if (_compression == BaselineTIFFTagSet.COMPRESSION_CCITT_RLE
                    || _compression == BaselineTIFFTagSet.COMPRESSION_CCITT_T_4
                    || _compression == BaselineTIFFTagSet.COMPRESSION_CCITT_T_6) {
                processWarningOccurred("PhotometricInterpretation field is missing; " + "assuming WhiteIsZero");
                _photometricInterpretation = BaselineTIFFTagSet.PHOTOMETRIC_INTERPRETATION_WHITE_IS_ZERO;
            } else if (this._colorMap != null) {
                _photometricInterpretation = BaselineTIFFTagSet.PHOTOMETRIC_INTERPRETATION_PALETTE_COLOR;
            } else if (_samplesPerPixel == 3 || _samplesPerPixel == 4) {
                _photometricInterpretation = BaselineTIFFTagSet.PHOTOMETRIC_INTERPRETATION_RGB;
            } else {
                processWarningOccurred("PhotometricInterpretation field is missing; " + "assuming BlackIsZero");
                _photometricInterpretation = BaselineTIFFTagSet.PHOTOMETRIC_INTERPRETATION_BLACK_IS_ZERO;
            }
        } else {
            _photometricInterpretation = field.getAsInt(0);
        }

        // SampleFormat
        boolean replicateFirst;
        int first = -1;

        field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_SAMPLE_FORMAT);
        _sampleFormat = new int[_samplesPerPixel];
        replicateFirst = false;
        if (field == null) {
            replicateFirst = true;
            first = BaselineTIFFTagSet.SAMPLE_FORMAT_UNDEFINED;
        } else if (field.getCount() != _samplesPerPixel) {
            replicateFirst = true;
            first = field.getAsInt(0);
        }

        for (int i = 0; i < _samplesPerPixel; i++) {
            _sampleFormat[i] = replicateFirst ? first : field.getAsInt(i);
            if (_sampleFormat[i] != BaselineTIFFTagSet.SAMPLE_FORMAT_UNSIGNED_INTEGER
                    && _sampleFormat[i] != BaselineTIFFTagSet.SAMPLE_FORMAT_SIGNED_INTEGER
                    && _sampleFormat[i] != BaselineTIFFTagSet.SAMPLE_FORMAT_FLOATING_POINT
                    && _sampleFormat[i] != BaselineTIFFTagSet.SAMPLE_FORMAT_UNDEFINED) {
                processWarningOccurred("Illegal value for SAMPLE_FORMAT, assuming SAMPLE_FORMAT_UNDEFINED");
                _sampleFormat[i] = BaselineTIFFTagSet.SAMPLE_FORMAT_UNDEFINED;
            }
        }

        // BitsPerSample
        field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_BITS_PER_SAMPLE);
        this._bitsPerSample = new int[_samplesPerPixel];
        replicateFirst = false;
        if (field == null) {
            replicateFirst = true;
            first = defaultBitDepth;
        } else if (field.getCount() != _samplesPerPixel) {
            replicateFirst = true;
            first = field.getAsInt(0);
        }

        for (int i = 0; i < _samplesPerPixel; i++) {
            // Replicate initial value if not enough values provided
            _bitsPerSample[i] = replicateFirst ? first : field.getAsInt(i);
        }

        // ExtraSamples
        this._extraSamples = null;
        field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_EXTRA_SAMPLES);
        if (field != null) {
            _extraSamples = field.getAsInts();
        }
    }

    @Override
    public Iterator<ImageTypeSpecifier> getImageTypes( int imageIndex ) throws IIOException {
        List<ImageTypeSpecifier> lImageTypeSpecifiers; // List of ImageTypeSpecifiers

        Integer imageIndexInteger = imageIndex;
        if (_imageTypeMap.containsKey(imageIndexInteger)) {
            // Return the cached ITS List.
            lImageTypeSpecifiers = _imageTypeMap.get(imageIndexInteger);
        } else {
            // Create a new ITS List.
            lImageTypeSpecifiers = new ArrayList<>(1);

            // Create the ITS and cache if for later use so that this method
            // always returns an Iterator containing the same ITS objects.
            seekToImage(imageIndex);
            ImageTypeSpecifier itsRaw = TIFFDecompressor.getRawImageTypeSpecifier(_photometricInterpretation,
                    _compression,
                    _samplesPerPixel,
                    _bitsPerSample,
                    _sampleFormat,
                    _extraSamples,
                    _colorMap);
            // Check for an ICCProfile field.
            TIFFField iccProfileField = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_ICC_PROFILE);

            // If an ICCProfile field is present change the ImageTypeSpecifier
            // to use it if the data layout is component type.
            if (iccProfileField != null && itsRaw.getColorModel() instanceof ComponentColorModel) {
                // Create a ColorSpace from the profile.
                byte[] iccProfileValue = iccProfileField.getAsBytes();
                ICC_Profile iccProfile = ICC_Profile.getInstance(iccProfileValue);
                ICC_ColorSpace iccColorSpace = new ICC_ColorSpace(iccProfile);

                // Get the raw sample and color information.
                ColorModel cmRaw = itsRaw.getColorModel();
                ColorSpace csRaw = cmRaw.getColorSpace();
                SampleModel smRaw = itsRaw.getSampleModel();

                // Get the number of samples per pixel and the number
                // of color components.
                int numBands = smRaw.getNumBands();
                int numComponents = iccColorSpace.getNumComponents();

                // Replace the ColorModel with the ICC ColorModel if the
                // numbers of samples and color components are amenable.
                if (numBands == numComponents || numBands == numComponents + 1) {
                    // Set alpha flags.
                    boolean hasAlpha = numComponents != numBands;
                    boolean isAlphaPre = hasAlpha && cmRaw.isAlphaPremultiplied();

                    // Create a ColorModel of the same class and with
                    // the same transfer type.
                    ColorModel iccColorModel = new ComponentColorModel(iccColorSpace,
                            cmRaw.getComponentSize(),
                            hasAlpha,
                            isAlphaPre,
                            cmRaw.getTransparency(),
                            cmRaw.getTransferType());

                    // Prepend the ICC profile-based ITS to the List. The
                    // ColorModel and SampleModel are guaranteed to be
                    // compatible as the old and new ColorModels are both
                    // ComponentColorModels with the same transfer type
                    // and the same number of components.
                    lImageTypeSpecifiers.add(new ImageTypeSpecifier(iccColorModel, smRaw));

                    // Append the raw ITS to the List if and only if its
                    // ColorSpace has the same type and number of components
                    // as the ICC ColorSpace.
                    if (csRaw.getType() == iccColorSpace.getType()
                            && csRaw.getNumComponents() == iccColorSpace.getNumComponents()) {
                        lImageTypeSpecifiers.add(itsRaw);
                    }
                } else { // ICCProfile not compatible with SampleModel.
                    // Append the raw ITS to the List.
                    lImageTypeSpecifiers.add(itsRaw);
                }
            } else { // No ICCProfile field or raw ColorModel not component.
                // Append the raw ITS to the List.
                lImageTypeSpecifiers.add(itsRaw);
            }

            // Cache the ITS List.
            _imageTypeMap.put(imageIndexInteger, lImageTypeSpecifiers);
        }

        return lImageTypeSpecifiers.iterator();
    }

    @Override
    public IIOMetadata getImageMetadata( int imageIndex ) throws IIOException {
        seekToImage(imageIndex);
        TIFFImageMetadata im =
                new TIFFImageMetadata(_imageMetadata.getRootIFD().getTagSetList());
        Node root =
                _imageMetadata.getAsTree(TIFFImageMetadata.NATIVE_METADATA_FORMAT_NAME);
        im.setFromTree(TIFFImageMetadata.NATIVE_METADATA_FORMAT_NAME, root);
        return im;
    }

    @Override
    public boolean isRandomAccessEasy( int imageIndex ) throws IOException {
        if (_currIndex != -1) {
            seekToImage(_currIndex);
            return getCompression() == BaselineTIFFTagSet.COMPRESSION_NONE;
        } else {
            return false;
        }
    }

    @Override
    public boolean hasThumbnails( int imageIndex ) {
        return false;
    }

    @Override
    public int getNumThumbnails( int imageIndex ) {
        return 0;
    }

    @Override
    public ImageReadParam getDefaultReadParam() {
        return new TIFFImageReadParam();
    }

    @Override
    public boolean isImageTiled( int imageIndex ) throws IOException {
        seekToImage(imageIndex);

        TIFFField field = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_TILE_WIDTH);
        return field != null;
    }

    @Override
    public int getTileWidth( int imageIndex ) throws IOException {
        seekToImage(imageIndex);
        return getTileOrStripWidth();
    }

    @Override
    public int getTileHeight( int imageIndex ) throws IOException {
        seekToImage(imageIndex);
        return getTileOrStripHeight();
    }

    @Override
    public BufferedImage readTile( int imageIndex, int tileX, int tileY ) throws IOException {
        int width = getWidth(imageIndex);
        int height = getHeight(imageIndex);
        int tw = getTileWidth(imageIndex);
        int th = getTileHeight(imageIndex);

        int x = tw * tileX;
        int y = th * tileY;

        if (tileX < 0 || tileY < 0 || x >= width || y >= height) {
            throw new IllegalArgumentException("Tile indices are out of bounds!");
        }

        if (x + tw > width) {
            tw = width - x;
        }

        if (y + th > height) {
            th = height - y;
        }

        ImageReadParam param = getDefaultReadParam();
        Rectangle tileRect = new Rectangle(x, y, tw, th);
        param.setSourceRegion(tileRect);

        return read(imageIndex, param);
    }

    @Override
    public boolean canReadRaster() {
        // Enable this?
        return false;
    }

    @Override
    public Raster readRaster( int imageIndex, ImageReadParam param ) {
        // Enable this?
        throw new UnsupportedOperationException();
    }

    private void prepareRead( int imageIndex, ImageReadParam param ) throws IOException {
        if (_stream == null) {
            throw new IllegalStateException("Input not set!");
        }

        // A null ImageReadParam means we use the default
        if (param == null) {
            param = getDefaultReadParam();
        }

        _imageReadParam = param;

        seekToImage(imageIndex);

        _tileOrStripWidth = getTileOrStripWidth();
        _tileOrStripHeight = getTileOrStripHeight();
        _planarConfiguration = getPlanarConfiguration();

        _sourceBands = param.getSourceBands();
        if (_sourceBands == null) {
            _sourceBands = new int[_numBands];
            for (int i = 0; i < _numBands; i++) {
                _sourceBands[i] = i;
            }
        }

        // Initialize the destination image
        Iterator<ImageTypeSpecifier> imageTypes = getImageTypes(imageIndex);
        ImageTypeSpecifier theImageType = getDestinationType(param, imageTypes);

        int destNumBands = theImageType.getSampleModel().getNumBands();

        this._destinationBands = param.getDestinationBands();
        if (_destinationBands == null) {
            _destinationBands = new int[destNumBands];
            for (int i = 0; i < destNumBands; i++) {
                _destinationBands[i] = i;
            }
        }

        if (_sourceBands.length != _destinationBands.length) {
            throw new IllegalArgumentException("sourceBands.length != destinationBands.length");
        }

        for (int i = 0; i < _sourceBands.length; i++) {
            int sb = _sourceBands[i];
            if (sb < 0 || sb >= _numBands) {
                throw new IllegalArgumentException("Source band out of range!");
            }
            int db = _destinationBands[i];
            if (db < 0 || db >= destNumBands) {
                throw new IllegalArgumentException("Destination band out of range!");
            }
        }
    }

    @Override
    public RenderedImage readAsRenderedImage( int imageIndex, ImageReadParam param ) throws IOException {
        prepareRead(imageIndex, param);
        return new TIFFRenderedImage(this, imageIndex, _imageReadParam,
                _width, _height);
    }

    private void decodeTile( int ti, int tj, int band ) throws IOException {
        // Compute the region covered by the strip or tile
        Rectangle tileRect = new Rectangle(ti * _tileOrStripWidth,
                tj * _tileOrStripHeight,
                _tileOrStripWidth,
                _tileOrStripHeight);

        // Clip against the image bounds if the image is not tiled. If it
        // is tiled, the tile may legally extend beyond the image bounds.
        if (!isImageTiled(_currIndex)) {
            tileRect = tileRect.intersection(new Rectangle(0, 0, _width, _height));
        }

        // Return if the intersection is empty.
        if (tileRect.width <= 0 || tileRect.height <= 0) {
            return;
        }

        int srcMinX = tileRect.x;
        int srcMinY = tileRect.y;
        int srcWidth = tileRect.width;
        int srcHeight = tileRect.height;

        // Determine dest region that can be derived from the
        // source region

        _dstMinX = iceil(srcMinX - _sourceXOffset, _srcXSubsampling);
        int dstMaxX = ifloor(srcMinX + srcWidth - 1 - _sourceXOffset, _srcXSubsampling);

        _dstMinY = iceil(srcMinY - _sourceYOffset, _srcYSubsampling);
        int dstMaxY = ifloor(srcMinY + srcHeight - 1 - _sourceYOffset, _srcYSubsampling);

        _dstWidth = dstMaxX - _dstMinX + 1;
        _dstHeight = dstMaxY - _dstMinY + 1;

        _dstMinX += _dstXOffset;
        _dstMinY += _dstYOffset;

        // Clip against image bounds

        Rectangle dstRect = new Rectangle(_dstMinX, _dstMinY, _dstWidth, _dstHeight);
        dstRect = dstRect.intersection(_theImage.getRaster().getBounds());

        _dstMinX = dstRect.x;
        _dstMinY = dstRect.y;
        _dstWidth = dstRect.width;
        _dstHeight = dstRect.height;

        if (_dstWidth <= 0 || _dstHeight <= 0) {
            return;
        }

        // Backwards map dest region to source to determine
        // active source region

        int activeSrcMinX = (_dstMinX - _dstXOffset) * _srcXSubsampling + _sourceXOffset;
        int sxmax = (_dstMinX + _dstWidth - 1 - _dstXOffset) * _srcXSubsampling + _sourceXOffset;
        int activeSrcWidth = sxmax - activeSrcMinX + 1;

        int activeSrcMinY = (_dstMinY - _dstYOffset) * _srcYSubsampling + _sourceYOffset;
        int symax = (_dstMinY + _dstHeight - 1 - _dstYOffset) * _srcYSubsampling + _sourceYOffset;
        int activeSrcHeight = symax - activeSrcMinY + 1;

        _decompressor.setSrcMinX(srcMinX);
        _decompressor.setSrcMinY(srcMinY);
        _decompressor.setSrcWidth(srcWidth);
        _decompressor.setSrcHeight(srcHeight);

        _decompressor.setDstMinX(_dstMinX);
        _decompressor.setDstMinY(_dstMinY);
        _decompressor.setDstWidth(_dstWidth);
        _decompressor.setDstHeight(_dstHeight);

        _decompressor.setActiveSrcMinX(activeSrcMinX);
        _decompressor.setActiveSrcMinY(activeSrcMinY);
        _decompressor.setActiveSrcWidth(activeSrcWidth);
        _decompressor.setActiveSrcHeight(activeSrcHeight);

        int tileIndex = tj * _tilesAcross + ti;

        if (_planarConfiguration == BaselineTIFFTagSet.PLANAR_CONFIGURATION_PLANAR) {
            tileIndex += band * _tilesAcross * _tilesDown;
        }

        long offset = getTileOrStripOffset(tileIndex);
        long byteCount = getTileOrStripByteCount(tileIndex);

        _decompressor.setStream(_stream);
        _decompressor.setOffset(offset);
        _decompressor.setByteCount((int) byteCount);

        _decompressor.beginDecoding();

        _stream.mark();
        _decompressor.decode();
        _stream.reset();
    }

    private void reportProgress() {
        // Report image progress/update to listeners after each tile
        _pixelsRead += _dstWidth * _dstHeight;
        processImageProgress(100.0f * _pixelsRead / _pixelsToRead);
        processImageUpdate(_theImage,
                _dstMinX, _dstMinY, _dstWidth, _dstHeight,
                1, 1,
                _destinationBands);
    }

    @Override
    public BufferedImage read( int imageIndex, ImageReadParam param ) throws IOException {
        prepareRead(imageIndex, param);

        _theImage = getDestination(param, getImageTypes(imageIndex), _width, _height);

        _srcXSubsampling = _imageReadParam.getSourceXSubsampling();
        _srcYSubsampling = _imageReadParam.getSourceYSubsampling();

        Point point = _imageReadParam.getDestinationOffset();
        _dstXOffset = point.x;
        _dstYOffset = point.y;

        // This could probably be made more efficient...
        Rectangle srcRegion = new Rectangle(0, 0, 0, 0);
        Rectangle destRegion = new Rectangle(0, 0, 0, 0);

        computeRegions(_imageReadParam, _width, _height, _theImage, srcRegion, destRegion);

        // Initial source pixel, taking source region and source
        // subsamplimg offsets into account
        _sourceXOffset = srcRegion.x;
        _sourceYOffset = srcRegion.y;

        _pixelsToRead = destRegion.width * destRegion.height;
        _pixelsRead = 0;

        processImageStarted(imageIndex);
        processImageProgress(0.0f);

        _tilesAcross = (_width + _tileOrStripWidth - 1) / _tileOrStripWidth;
        _tilesDown = (_height + _tileOrStripHeight - 1) / _tileOrStripHeight;

        int compression = getCompression();
        // Attempt to get decompressor and color converted from the read param

        TIFFColorConverter colorConverter = null;
        if (_imageReadParam instanceof TIFFImageReadParam) {
            TIFFImageReadParam tparam = (TIFFImageReadParam) _imageReadParam;
            this._decompressor = tparam.getTIFFDecompressor();
            colorConverter = tparam.getColorConverter();
        }

        // If we didn't find one, use a standard decompressor
        if (this._decompressor == null) {
            switch (compression) {
                case BaselineTIFFTagSet.COMPRESSION_NONE:
                    // Get the fillOrder field.
                    TIFFField fillOrderField = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_FILL_ORDER);

                    // Set the decompressor based on the fill order.
                    if (fillOrderField != null && fillOrderField.getAsInt(0) == 2) {
                        this._decompressor = new TIFFLSBDecompressor();
                    } else {
                        this._decompressor = new TIFFNullDecompressor();
                    }
                    break;
                case BaselineTIFFTagSet.COMPRESSION_CCITT_T_6:
                    // Try to create the codecLib decompressor.
                    // Fall back to the Java decompressor.
                    this._decompressor = new TIFFFaxDecompressor();
                    break;
                case BaselineTIFFTagSet.COMPRESSION_CCITT_T_4:
                    // Fall back to the Java decompressor.
                    this._decompressor = new TIFFFaxDecompressor();
                    break;
                case BaselineTIFFTagSet.COMPRESSION_CCITT_RLE:
                    this._decompressor = new TIFFFaxDecompressor();
                    break;
                case BaselineTIFFTagSet.COMPRESSION_PACKBITS:
                    this._decompressor = new TIFFPackBitsDecompressor();
                    break;
                case BaselineTIFFTagSet.COMPRESSION_LZW: {
                    TIFFField predictorField = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_PREDICTOR);
                    int predictor = ((predictorField == null) ? BaselineTIFFTagSet.PREDICTOR_NONE : predictorField.getAsInt(0));
                    this._decompressor = new TIFFLZWDecompressor(predictor);
                    break;
                }
                case BaselineTIFFTagSet.COMPRESSION_JPEG:
                    this._decompressor = new TIFFJPEGDecompressor();
                    break;
                case BaselineTIFFTagSet.COMPRESSION_ZLIB:
                case BaselineTIFFTagSet.COMPRESSION_DEFLATE: {
                    TIFFField predictorField = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_PREDICTOR);
                    int predictor = ((predictorField == null) ? BaselineTIFFTagSet.PREDICTOR_NONE : predictorField.getAsInt(0));
                    this._decompressor = new TIFFDeflateDecompressor(predictor);
                    break;
                }
                case BaselineTIFFTagSet.COMPRESSION_OLD_JPEG:
                    TIFFField jPEGProcField = _imageMetadata.getTIFFField(BaselineTIFFTagSet.TAG_JPEG_PROC);
                    if (jPEGProcField == null) {
                        processWarningOccurred("JPEGProc field missing; assuming baseline sequential JPEG process.");
                    } else if (jPEGProcField.getAsInt(0) != BaselineTIFFTagSet.JPEG_PROC_BASELINE) {
                        throw new IIOException("Old-style JPEG supported for baseline sequential JPEG process only!");
                    }
                    _decompressor = new TIFFOldJPEGDecompressor();
                    //throw new IIOException("Old-style JPEG not supported!");
                    break;
                default:
                    throw new IIOException("Unsupported compression type (tag number = " + compression + ")!");
            }

            if (_photometricInterpretation == BaselineTIFFTagSet.PHOTOMETRIC_INTERPRETATION_Y_CB_CR
                    && compression != BaselineTIFFTagSet.COMPRESSION_JPEG
                    && compression != BaselineTIFFTagSet.COMPRESSION_OLD_JPEG) {
                boolean convertYCbCrToRGB = _theImage.getColorModel().getColorSpace().getType() == ColorSpace.TYPE_RGB;
                TIFFDecompressor wrappedDecompressor = this._decompressor instanceof TIFFNullDecompressor ? null : this._decompressor;
                this._decompressor = new TIFFYCbCrDecompressor(wrappedDecompressor, convertYCbCrToRGB);
            }
        }

        if (colorConverter == null) {
            if (_photometricInterpretation == BaselineTIFFTagSet.PHOTOMETRIC_INTERPRETATION_CIELAB
                    && _theImage.getColorModel().getColorSpace().getType() == ColorSpace.TYPE_RGB) {
                colorConverter = new TIFFCIELabColorConverter();
            } else if (_photometricInterpretation == BaselineTIFFTagSet.PHOTOMETRIC_INTERPRETATION_Y_CB_CR
                    && !(this._decompressor instanceof TIFFYCbCrDecompressor)
                    && compression != BaselineTIFFTagSet.COMPRESSION_JPEG
                    && compression != BaselineTIFFTagSet.COMPRESSION_OLD_JPEG) {
                colorConverter = new TIFFYCbCrColorConverter(_imageMetadata);
            }
        }

        _decompressor.setReader(this);
        _decompressor.setMetadata(_imageMetadata);
        _decompressor.setImage(_theImage);
        _decompressor.setPhotometricInterpretation(_photometricInterpretation);
        _decompressor.setCompression(compression);
        _decompressor.setSamplesPerPixel(_samplesPerPixel);
        _decompressor.setBitsPerSample(_bitsPerSample);
        _decompressor.setSampleFormat(_sampleFormat);
        _decompressor.setExtraSamples(_extraSamples);
        _decompressor.setColorMap(_colorMap);
        _decompressor.setColorConverter(colorConverter);

        _decompressor.setSubsampleX(_srcXSubsampling);
        _decompressor.setSubsampleY(_srcYSubsampling);

        _decompressor.setSourceBands(_sourceBands);
        _decompressor.setDestinationBands(_destinationBands);

        // Compute bounds on the tile indices for this source region.
        int minTileX = xToTileX(srcRegion.x, 0, _tileOrStripWidth);
        int minTileY = yToTileY(srcRegion.y, 0, _tileOrStripHeight);
        int maxTileX = xToTileX(srcRegion.x + srcRegion.width - 1, 0, _tileOrStripWidth);
        int maxTileY = yToTileY(srcRegion.y + srcRegion.height - 1, 0, _tileOrStripHeight);

        if (_planarConfiguration == BaselineTIFFTagSet.PLANAR_CONFIGURATION_PLANAR) {
            _decompressor.setPlanar(true);

            int[] sb = new int[1];
            int[] db = new int[1];
            for (int tj = minTileY; tj <= maxTileY; tj++) {
                for (int ti = minTileX; ti <= maxTileX; ti++) {
                    for (int band = 0; band < _numBands; band++) {
                        sb[0] = _sourceBands[band];
                        _decompressor.setSourceBands(sb);
                        db[0] = _destinationBands[band];
                        _decompressor.setDestinationBands(db);
                        decodeTile(ti, tj, band);
                    }
                    reportProgress();
                }
            }
        } else {
            for (int tj = minTileY; tj <= maxTileY; tj++) {
                for (int ti = minTileX; ti <= maxTileX; ti++) {
                    decodeTile(ti, tj, -1);
                    reportProgress();
                }
            }
        }

        if (abortRequested()) {
            processReadAborted();
        } else {
            processImageComplete();
        }

        return _theImage;
    }

    @Override
    public void reset() {
        super.reset();
        resetLocal();
    }

    private void resetLocal() {
        _stream = null;
        _gotHeader = false;
        _imageReadParam = getDefaultReadParam();
        _streamMetadata = null;
        _currIndex = -1;
        _imageMetadata = null;
        _imageStartPosition = new ArrayList<>();
        _numImages = -1;
        _imageTypeMap = new HashMap<>();
        _width = -1;
        _height = -1;
        _numBands = -1;
        _tileOrStripWidth = -1;
        _tileOrStripHeight = -1;
        _planarConfiguration = BaselineTIFFTagSet.PLANAR_CONFIGURATION_CHUNKY;
    }

    /**
     * Package scope method to allow decompressors, for example, to
     * emit warning messages.
     */
    void forwardWarningMessage( String warning ) {
        processWarningOccurred(warning);
    }
}
