/*
 * $RCSfile: TIFFLZWDecompressor.java,v $
 *
 * 
 * Copyright (c) 2005 Sun Microsystems, Inc. All  Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met: 
 * 
 * - Redistribution of source code must retain the above copyright 
 *   notice, this  list of conditions and the following disclaimer.
 * 
 * - Redistribution in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in 
 *   the documentation and/or other materials provided with the
 *   distribution.
 * 
 * Neither the name of Sun Microsystems, Inc. or the names of 
 * contributors may be used to endorse or promote products derived 
 * from this software without specific prior written permission.
 * 
 * This software is provided "AS IS," without a warranty of any 
 * kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND 
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
 * EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL 
 * NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF 
 * USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR 
 * ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
 * CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND
 * REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES. 
 * 
 * You acknowledge that this software is not designed or intended for 
 * use in the design, construction, operation or maintenance of any 
 * nuclear facility. 
 *
 * $Revision: 1.1 $
 * $Date: 2005/02/11 05:01:48 $
 * $State: Exp $
 */
package com.sun.media.imageioimpl.plugins.tiff;

import java.io.IOException;
import javax.imageio.IIOException;
import com.sun.media.imageio.plugins.tiff.BaselineTIFFTagSet;
import com.sun.media.imageio.plugins.tiff.TIFFDecompressor;

public final class TIFFLZWDecompressor extends TIFFDecompressor {
	private static final int[] AND_TABLE = {
		511, 
		1023,
		2047,
		4095
	};

	private final int _predictor;

	private byte[] _srcData;
	private byte[] _dstData;

	private int _srcIndex;
	private int _dstIndex;

	private byte[][] _stringTable;
	private int _tableIndex, _bitsToGet = 9;

	private int _nextData;
	private int _nextBits;

	public TIFFLZWDecompressor( int predictor ) throws IIOException {
		super();

		if (predictor != BaselineTIFFTagSet.PREDICTOR_NONE
				&& predictor != BaselineTIFFTagSet.PREDICTOR_HORIZONTAL_DIFFERENCING) {
			throw new IIOException("Illegal value for Predictor in "
				+ "TIFF file");
		}
		this._predictor = predictor;
	}

	@Override
	public void decodeRaw( byte[] bytes,
                           int dstOffset,
						   int bitsPerPixel,
						   int scanlineStride ) throws IOException {

		// Check bitsPerSample.
		if (_predictor == BaselineTIFFTagSet.PREDICTOR_HORIZONTAL_DIFFERENCING) {
			for (int aBitsPerSample : bitsPerSample) {
				if (aBitsPerSample != 8) {
					throw new IIOException(aBitsPerSample + "-bit samples "
							+ "are not supported for Horizontal "
							+ "differencing Predictor");
				}
			}
		}

		stream.seek(offset);

		byte[] sdata = new byte[byteCount];
		stream.readFully(sdata);

		int bytesPerRow = (srcWidth * bitsPerPixel + 7) / 8;
		byte[] buf;
		int bufOffset;
		if (bytesPerRow == scanlineStride) {
			buf = bytes;
			bufOffset = dstOffset;
		} else {
			buf = new byte[bytesPerRow * srcHeight];
			bufOffset = 0;
		}

		decode(sdata, 0, buf, bufOffset);

		if (bytesPerRow != scanlineStride) {
			int off = 0;
			for (int y = 0; y < srcHeight; y++) {
				System.arraycopy(buf, off, bytes, dstOffset, bytesPerRow);
				off += bytesPerRow;
				dstOffset += scanlineStride;
			}
		}
	}

	private int decode( byte[] sdata, int srcOffset, byte[] ddata, int dstOffset ) throws IOException {
		if (sdata[0] == (byte)0x00 && sdata[1] == (byte)0x01) {
			throw new IIOException("TIFF 5.0-style LZW compression is not supported!");
		}

		this._srcData = sdata;
		this._dstData = ddata;

		this._srcIndex = srcOffset;
		this._dstIndex = dstOffset;

		this._nextData = 0;
		this._nextBits = 0;

		initializeStringTable();

		int code, oldCode = 0;
		byte[] string;

		while ((code = getNextCode()) != 257) {
			if (code == 256) {
				initializeStringTable();
				code = getNextCode();
				if (code == 257) {
					break;
				}

				writeString(_stringTable[code]);
				oldCode = code;
			} else {
				if (code < _tableIndex) {
					string = _stringTable[code];

					writeString(string);
					addStringToTable(_stringTable[oldCode], string[0]); 
					oldCode = code;
				} else {
					string = _stringTable[oldCode];
					string = composeString(string, string[0]);
					writeString(string);
					addStringToTable(string);
					oldCode = code;
				}
			}
		}

		if (_predictor == BaselineTIFFTagSet.PREDICTOR_HORIZONTAL_DIFFERENCING) {

			for (int j = 0; j < srcHeight; j++) {

				int count = dstOffset + samplesPerPixel * (j * srcWidth + 1);

				for (int i = samplesPerPixel; i < srcWidth * samplesPerPixel; i++) {

					_dstData[count] += _dstData[count - samplesPerPixel];
					count++;
				}
			}
		}

		return _dstIndex - dstOffset;
	}

	/**
	 * Initialize the string table.
	 */
    private void initializeStringTable() {
		_stringTable = new byte[4096][];

		for (int i = 0; i < 256; i++) {
			_stringTable[i] = new byte[1];
			_stringTable[i][0] = (byte)i;
		}

		_tableIndex = 258;
		_bitsToGet = 9;
	}

	/**
	 * Write out the string just uncompressed.
	 */
    private void writeString( byte[] string ) {
		if (_dstIndex < _dstData.length) {
			int maxIndex = Math.min(string.length,
					_dstData.length - _dstIndex);

			for (int i = 0; i < maxIndex; i++) {
				_dstData[_dstIndex++] = string[i];
			}
		}
	}

	/**
	 * Add a new string to the string table.
	 */
    private void addStringToTable( byte[] oldString, byte newString ) {
		int length = oldString.length;
		byte[] string = new byte[length + 1];
		System.arraycopy(oldString, 0, string, 0, length);
		string[length] = newString;

		// Add this new String to the table
		_stringTable[_tableIndex++] = string;

		switch (_tableIndex) {
			case 511:
				_bitsToGet = 10;
				break;
			case 1023:
				_bitsToGet = 11;
				break;
			case 2047:
				_bitsToGet = 12;
				break;
		}
	}

	/**
	 * Add a new string to the string table.
	 */
    private void addStringToTable( byte[] string ) {
		// Add this new String to the table
		_stringTable[_tableIndex++] = string;

		switch (_tableIndex) {
			case 511:
				_bitsToGet = 10;
				break;
			case 1023:
				_bitsToGet = 11;
				break;
			case 2047:
				_bitsToGet = 12;
				break;
		}
	}

	/**
	 * Append <code>newString</code> to the end of <code>oldString</code>.
	 */
    private byte[] composeString( byte[] oldString, byte newString ) {
		int length = oldString.length;
		byte[] string = new byte[length + 1];
		System.arraycopy(oldString, 0, string, 0, length);
		string[length] = newString;

		return string;
	}

	// Returns the next 9, 10, 11 or 12 bits
    private int getNextCode() {
		// Attempt to get the next code. The exception is caught to make
		// this robust to cases wherein the EndOfInformation code has been
		// omitted from a strip. Examples of such cases have been observed
		// in practice.

		try {
			_nextData = (_nextData << 8) | (_srcData[_srcIndex++] & 0xff);
			_nextBits += 8;

			if (_nextBits < _bitsToGet) {
				_nextData = (_nextData << 8) | (_srcData[_srcIndex++] & 0xff);
				_nextBits += 8;
			}

			int code =
					(_nextData >> (_nextBits - _bitsToGet)) & AND_TABLE[_bitsToGet - 9];
			_nextBits -= _bitsToGet;

			return code;
		} catch (ArrayIndexOutOfBoundsException e) {
			// Strip not terminated as expected: return EndOfInformation code.
					return 257;
		}
	}
}
