/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.generator;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeBadGraphException;
import fr.greyc.pantheon.ariane.model.graph.element.CompositeItem;
import fr.greyc.pantheon.ariane.model.graph.element.ConstantItem;
import fr.greyc.pantheon.ariane.model.graph.element.ForControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.IfControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.MergeControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.SelectControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.EvalItem;
import fr.greyc.pantheon.ariane.model.graph.element.FileNameItem;
import fr.greyc.pantheon.ariane.model.graph.element.FormatItem;
import fr.greyc.pantheon.ariane.model.graph.element.GraphItemModel;
import fr.greyc.pantheon.ariane.model.graph.element.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.element.MacroOperatorItem;
import fr.greyc.pantheon.ariane.model.graph.element.NoteItem;
import fr.greyc.pantheon.ariane.model.graph.element.OperatorItem;
import fr.greyc.pantheon.ariane.model.graph.element.SaverItem;
import fr.greyc.pantheon.ariane.model.graph.element.ScriptItem;
import fr.greyc.pantheon.ariane.model.graph.element.SystemItem;
import fr.greyc.pantheon.ariane.model.graph.element.ViewerItem;
import fr.greyc.pantheon.ariane.model.graph.element.VirtualItem;
import fr.greyc.pantheon.ariane.model.graph.element.WhileControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.presenter.SubGraphPresenter;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * The <code>ShellGenerator</code> generates bash Shell code for the specified graph.
 * It operates in three steps:<ol>
 * <li>transforms the graph into a list of operators usable for the topological sort;
 * <li>executes the topological sort;
 * <li>generates the related code.
 * </ol>
 */
public final class BashGenerator extends CodeGenerator {
    private final List<String> _outputImages = new ArrayList<>();

    BashGenerator() {
        // Only created by the code generator.
    }

    @Override
    protected void generateScriptHeader( String filename, List<GraphItemModel> arguments, List<GraphItemModel> inputs, List<GraphItemModel> outputs ) {
        println("#!/bin/bash");
        println("#");
        println("# @(#)" + filename + "				" + getCurrentDate());
        println("#");
        println("# This file was automatically generated by Ariane.");
        println("# Ariane is part of the Pantheon Project.\n");

        println("IFS=$(echo -en \"\\n\\b\")");
        println("");

        println("# Check argument line consistency.");
        println("if [ $# -ge 1 -a \"$1\" = \"-h\" -o $# -ne " + (arguments.size() + inputs.size() + outputs.size()) + " ]");
        println("then");
        print("echo \"Usage: $0", 1);
        for (GraphItemModel argument : arguments) {
            print(" " + argument.getNickName());
        }
        for (int i = 0; i < inputs.size(); i++) {
            print(" input-file" + i);
        }
        for (int i = 0; i < outputs.size(); i++) {
            print(" output-file" + i);
        }
        println("\"");

        print("echo \"  example: $0", 1);
        for (GraphItemModel argument : arguments) {
            if (argument.getFormula() != null && !argument.getFormula().isEmpty()) {
                print(" " + argument.getFormula());
            } else {
                print(" " + argument.getNickName());
            }
        }
        for (GraphItemModel input : inputs) {
            print(" " + input.getNickName());
        }
        for (GraphItemModel output : outputs) {
            String name = output.getNickName();
            if ("saver".equals(name)) {
                name += output.getId();
            }
            print(" " + name);
        }
        println("\"");

        println("exit", 1);
        println("fi");
    }

    @Override
    protected void generateScriptFooter() {
        println("\nmain $*");
    }

    @Override
    protected void generateUtilitySubroutinesDefinition() {
        // Do nothing
    }

    @Override
    protected void generateMainPrologue( List<GraphItemModel> litems, List<GraphItemModel> arguments, List<GraphItemModel> inputs, List<GraphItemModel> outputs ) {
        println("\n#\n# The main program\n#");
        for (GraphItemModel item : litems) {
            if (item instanceof NoteItem) {
                println("# " + convertCommentIntoShell(item.getFormula()), 0);
            }
        }
        println("function main() {");
        println("rm -f " + UserPreferences.getTemporaryFilePrefix() + "?*" + UserPreferences.getTemporaryFileExtension(), 1);
        for (GraphItemModel item : arguments) {
            println(createVariableName(1, item.getId(), 0) + "=$1; shift", 1);
        }
        for (GraphItemModel item : inputs) {
            println("ln -sf $1 " + createOutputImageName(1, item.getId(), 0) + "; "
                    + createVariableName(1, item.getId(), 0)
                    + "=$1; shift", 1);
        }
    }

    @Override
    protected void generateMainEpilogue( List<GraphItemModel> arguments, List<GraphItemModel> inputs, List<GraphItemModel> outputs ) {
        println("}");
    }

    @Override
    protected void generateItem( LoaderItem item, int graphId, int indent ) {
        if (item.getInputParameterSlots().get(0).isConnected()) {
            String parameters = getActualParameterValue(item, 0, graphId);
            String imagename = createOutputImageName(graphId, item.getId(), 0);
            println("cp " + parameters + " " + imagename, indent);
            println(createVariableName(graphId, item.getId(), 1) + "=" + imagename, indent);
        }
    }

    @Override
    protected void generateItem( SaverItem item, int graphId, int indent ) {
        if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
            if (item.getInputParameterSlots().get(0).isConnected()) {
                String parameters = getActualParameterValue(item, 0, graphId);
                println("cp " + createInputImageName(item.getInputDataSlots().get(0), graphId)
                        + " " + parameters, indent);
            } else {
                println("cp " + createInputImageName(item.getInputDataSlots().get(0), graphId)
                        + " $1; shift", indent);
            }
            String imagename = createOutputImageName(graphId, item.getId(), 0);
            println("ln -sf "
                    + createInputImageName(item.getInputDataSlots().get(0), graphId)
                    + " " + imagename, indent);
            println(createVariableName(graphId, item.getId(), 0) + "=" + imagename, indent);
        }
    }

    @Override
    protected void generateItem( ViewerItem item, int graphId, int indent ) {
        if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
            final String commandShell = item.getViewerCommandName();
            println(commandShell + " " + createInputImageName(item.getInputDataSlots().get(0), graphId), indent);
            println("ln -sf "
                    + createInputImageName(item.getInputDataSlots().get(0), graphId)
                    + " "
                    + createOutputImageName(graphId, item.getId(), 0), indent);
        }
    }

    @Override
    protected void generateItem( OperatorItem item, int graphId, int indent ) throws ArianeBadGraphException {
        String codeName = OperatorItem.getCommandPart(item.getBinary());
        if (codeName == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error2") + item.getName());
        }

        // Get input images.
        List<String> inputLine = new LinkedList<>();
        for (InputDataSlot image : item.getInputDataSlots()) {
            if (image.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            inputLine.add(createInputImageName(image, graphId));
        }

        // Get parameters
        List<String> parameterLine = new LinkedList<>();
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameterLine.add(getActualParameterValue(item, i, graphId));
        }

        // Get mask
        List<String> maskLine = new LinkedList<>();
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, graphId))) {
            maskLine.add(createInputImageName(mask, graphId));
        }

        // Get output images.
        List<String> outputLine = new LinkedList<>();
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            outputLine.add(createOutputImageName(graphId, item.getId(), i));
        }

        StringBuilder code = new StringBuilder();
        List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
        for (String command : commandLine) {
            code.append(command).append(" ");
        }

        if (item.getOutputResultSlot().isConnected() || hasMergeItemInItsSuccessors(item)) {
            println(code + "; " + createVariableName(graphId, item.getId(), 0) + "=`" + UserPreferences.getCommandStatus() + "`", indent);
        } else {
            println(code.toString(), indent);
        }
    }

    @Override
    protected void generateItem( SystemItem item, int graphId, int indent ) throws ArianeBadGraphException {
        String codeName = OperatorItem.getCommandPart(item.getBinary());
        if (codeName == null || codeName.trim().isEmpty()) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error3") + item.getName() + "(" + item.getId() + ")");
        }

        // Get parameters
        List<String> parameterLine = new LinkedList<>();
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameterLine.add(getActualParameterValue(item, i, graphId));
        }

        // Get mask
        List<String> maskLine = new LinkedList<>();
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, graphId))) {
            maskLine.add(createInputImageName(mask, graphId));
        }

        // Get input images.
        List<String> inputLine = new LinkedList<>();
        for (InputDataSlot image : item.getInputDataSlots()) {
            if (image.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            inputLine.add(createInputImageName(image, graphId));
        }

        // Get output images.
        List<String> outputLine = new LinkedList<>();
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            outputLine.add(createOutputImageName(graphId, item.getId(), i));
        }

        StringBuilder code = new StringBuilder();

        List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
        for (String command : commandLine) {
            code.append(command).append(" ");
        }
        if (item.getOutputResultSlot().isConnected() || hasMergeItemInItsSuccessors(item)) {
            println(code + "; " + createVariableName(graphId, item.getId(), 0) + "=`" + UserPreferences.getCommandStatus() + "`", indent);
        } else {
            println(code.toString(), indent);
        }
    }

    @Override
    protected void generateItem( EvalItem item, int graphId, int indent ) {
        String[] parameters = new String[item.getInputParameterSlots().size()];
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameters[i] = getActualParameterValue(item, i, graphId);
        }
        println(createVariableName(graphId, item.getId(), 0) + "=" + "$(echo \""
                + convertArithmeticFunctionIntoShell(item.getFormula())
                + convertArithmeticFormulaIntoShell(item.getFormula(), parameters)
                + "\"|bc -l);", indent);
    }

    @Override
    protected void generateItem( FormatItem item, int graphId, int indent ) {
        String[] parameters = new String[item.getInputParameterSlots().size()];
        // Get parameters
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameters[i] = getActualParameterValue(item, i, graphId);
        }

        StringBuilder code = new StringBuilder();
        StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), " $()+'\"", true);
        String etape;
        while (stringTokenizer.hasMoreTokens()) {
            etape = stringTokenizer.nextToken();
            if ("+".equals(etape)) {
                code.append("");
            } else if (" ".equals(etape)) {
                code.append("");
            } else if ("quote".equals(etape)) {
                code.append("");
            } else if ("(".equals(etape)) {
                code.append("\\'");
            } else if (")".equals(etape)) {
                code.append("\\'");
            } else if ("$".equals(etape)) {
                int numberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
                code.append(parameters[numberOfParameters]);
            } else {
                code.append(etape);
            }
        }
        println(createVariableName(graphId, item.getId(), 0) + "=" + code, indent);
    }

    @Override
    protected void generateItem( SelectControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
        println("# Select", indent);

        String value = getActualParameterValue(item, 0, graphId);
        for (int i = 0; i < item.getInputDataSlots().size(); i++) {
            if (!item.getInputDataSlots().get(i).isConnected()) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            println("if [ " + value + " -eq " + i + " ]", indent);
            println("then", indent);

            println("ln -sf "
                    + createInputImageName(item.getInputDataSlots().get(i), graphId) + " "
                    + createOutputImageName(graphId, item.getId(), 0), indent + 1);
            println(createVariableName(graphId, item.getId(), 0) + "=" + i + ";", indent + 1);
            println("fi", indent);
        }
    }

    @Override
    protected void generateItem( IfControlStructureItem item, int graphId, int indent ) {
        String[] parameters = new String[2];
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameters[i] = getNumericalParameterValue(item, i, graphId);
        }
        String commandShell = convertLogicalFormulaIntoShell(item.getFormula(), parameters[0], parameters[1]);
        println("if [ $( echo \"" + commandShell + "\" |bc -l) -gt 0 ]", indent);
    }

    @Override
    protected void generateThenPart( IfControlStructureItem item, int graphId, int indent ) {
        println("then", indent);
        print("ln -sf " + createInputImageName(item.getInputDataSlots().get(0), graphId) + " "
                + createOutputImageName(graphId, item.getId(), 0), indent + 1);
        println("; " + createVariableName(graphId, item.getId(), 0) + "=0");
    }

    @Override
    protected void generateElsePart( IfControlStructureItem item, int graphId, int indent ) {
        println("else", indent);
        print("ln -sf " + createInputImageName(item.getInputDataSlots().get(0), graphId) + " "
                + createOutputImageName(graphId, item.getId(), 1), indent + 1);
        println("; " + createVariableName(graphId, item.getId(), 0) + "=1");
    }

    @Override
    protected void generateFiPart( IfControlStructureItem item, int graphId, int indent ) {
        println("fi", indent);
    }

    @Override
    protected void generateItem( MergeControlStructureItem item, int graphId, int indent ) {
        InputDataSlot in0;
        in0 = item.getInputDataSlots().get(0);
        int id0;
        id0 = in0.getConnectedSlot().getGraphItem().getId();

        print("if [ -n \"${" + createVariableName(graphId, id0, 0) + "}\" -a ", indent);
        println("-r " + createInputImageName(item.getInputDataSlots().get(0), graphId) + " ]");
        println("then", indent);
        print("cp "
                + createInputImageName(item.getInputDataSlots().get(0), graphId)
                + " "
                + createOutputImageName(graphId, item.getId(), 0), indent + 1);
        println("; " + createVariableName(graphId, item.getId(), 0) + "=0;");
        println("rm " + createInputImageName(item.getInputDataSlots().get(0), graphId), indent + 1);
        println("else", indent);
        print("cp "
                + createInputImageName(item.getInputDataSlots().get(1), graphId) + " "
                + createOutputImageName(graphId, item.getId(), 0), indent + 1);
        println("; " + createVariableName(graphId, item.getId(), 0) + "=1;");
        println("rm " + createInputImageName(item.getInputDataSlots().get(1), graphId), indent + 1);
        println("fi", indent);
    }

    @Override
    protected void generateItem( FileNameItem item, int graphId, int indent ) {
        if (!item.isArgumentOfRelatedScript()) {
            println(createVariableName(graphId, item.getId(), 0) + "='" + item.getFormula() + "'", indent);
        }
    }

    @Override
    protected void generateItem( ConstantItem item, int graphId, int indent ) {
        if (!item.isArgumentOfRelatedScript()) {
            String value = item.getFormula().replaceAll("\"", "\\\\\"");
            println(createVariableName(graphId, item.getId(), 0) + "=\"" + value + "\"", indent);
        }
    }

    @Override
    protected void generateItem( ScriptItem item, int graphId, int indent ) {
        String value = item.getFormula().replaceAll("\"", "\\\"");
        println(createVariableName(graphId, item.getId(), 0) + "=\"" + value + "\"", indent);
    }

    @Override
    protected void generateItem( CompositeItem item, int graphId, int indent ) throws ArianeBadGraphException {
        StringBuilder code = new StringBuilder("macro" + item.getId() + "_" + item.getGraph().getIdentifier());

        // Get parameters
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            code.append(" ").append(getActualParameterValue(item, i, graphId));
        }

        // Get mask
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, graphId))) {
            code.append(" ").append(createInputImageName(mask, graphId));
        }

        // Get input images.
        for (InputDataSlot image : item.getInputDataSlots()) {
            if (image.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            code.append(" ").append(createInputImageName(image, graphId));
        }

        // Get output images.
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            code.append(" ").append(createOutputImageName(graphId, item.getId(), i));
        }
        // Get output result.
        if (item.getOutputResultSlot().isConnected()) {
            println(code + "; " + createVariableName(graphId, item.getId(), 0) + "=`" + UserPreferences.getCommandStatus() + "`", indent);
        } else {
            println(code + "; ", indent);
        }
    }

    @Override
    protected void generateSubGraph( ForControlStructureItem item ) throws ArianeBadGraphException {
        SubGraphPresenter subGraph = item.getSubGraph();

        generateComment(item, subGraph.getListOfGraphItems());

        println("function macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "() {");

        // Get parameters
        println("begin=${1%.*}; shift", 1);
        println("end=${1%.*}; shift", 1);
        println("step=${1%.*}; shift", 1);

        int subgraphId = subGraph.getIdentifier();
        for (int i = 3; i < item.getInputParameterSlots().size(); i++) {
            println(createVariableName(subgraphId, 0, i - 2) + "=$1; shift", 1);
        }

        // Get mask
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, item.getGraph().getIdentifier()))) {
            int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
            println("ln -fs $1 " + createOutputImageName(subgraphId, 0, maskIndex) + "; shift", 1);
        }

        // Get input images
        List<InputDataSlot> inputImages = item.getInputDataSlots();
        for (int i = 0; i < inputImages.size(); i++) {
            InputDataSlot isi = inputImages.get(i);
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            println("cp $1 " + createOutputImageName(subgraphId, 0, i) + "; shift", 1);
        }

        println("");

        VirtualItem virtualItem = subGraph.getVirtualItem();

        // Create the output images in case of no iteration
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            println("touch " + createInputImageName(isi, subgraphId), 1);
        }

        // Initialize the return variable in case of no iteration
        {
            ParameterSlot isp = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
            println(createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "=SUCCESS", 1);
        }

        println("");
        println("local i", 1);
        println("for ((i=$begin; i<=$end; i+=$step ))", 1);
        println("do", 1);
        println(createVariableName(subgraphId, 0, 0) + "=$i;", 2);

        List<GraphItemModel> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());
        generateSubroutineBody(litems, subgraphId, 2);

        // Set intermediate images
        for (int i = 0; i < item.getInputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            String input = createInputImageName(isi, subgraphId);
            String output = createOutputImageName(subgraphId, 0, i);
            if (!input.equals(output)) {
                println("cp " + input + " " + output, 2);
            }
        }

        println("done", 1);

        println("");

        // Get output images
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
            println("ln -fs " + createInputImageName(isi, subgraphId) + " $1; shift", 1);
        }
        // Output result
        ParameterSlot isp = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
        if (isp.getConnectedSlot() == null) {
            println("psetstatus SUCCESS", 1);
        } else {
            println("psetstatus ${" + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "}", 1);
        }
        println("}");

        generateSubroutinesDefinition(litems);
    }

    @Override
    protected void generateSubGraph( WhileControlStructureItem item ) throws ArianeBadGraphException {
        SubGraphPresenter subGraph = item.getSubGraph();

        generateComment(item, subGraph.getListOfGraphItems());

        println("function macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "() {");

        // Get parameters
        int subgraphId = subGraph.getIdentifier();
        String[] parameters = new String[item.getInputParameterSlots().size()];
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            println(createVariableName(subgraphId, 0, i) + "=$1; shift", 1);
            parameters[i] = "${" + createVariableName(subgraphId, 0, i) + "}";
        }

        // Get mask.
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, item.getGraph().getIdentifier()))) {
            int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
            println("ln -fs $1 " + createOutputImageName(subgraphId, 0, maskIndex) + "; shift", 1);
        }

        // Get input images.
        List<InputDataSlot> inputImages = item.getInputDataSlots();
        for (int i = 0; i < inputImages.size(); i++) {
            InputDataSlot isi = inputImages.get(i);
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            println("cp $1 " + createOutputImageName(subgraphId, 0, i) + "; shift", 1);
        }

        // Parse the boolean expression.
        String testFormula = convertLogicalFormulaIntoShell(item.getFormula(), parameters[0], parameters[1]);

        println("");

        VirtualItem virtualItem = subGraph.getVirtualItem();

        // Create output images in case of no repetition
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
            println("touch " + createInputImageName(isi, subgraphId), 1);
        }

        // Initialize the return variable in case of no iteration
        {
            ParameterSlot isp = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
            println(createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "=SUCCESS", 1);
        }

        println("");
        println("while [ $( echo \"" + testFormula + "\" |bc -l) -gt 0 ]", 1);
        println("do", 1);

        List<GraphItemModel> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());
        generateSubroutineBody(litems, subgraphId, 2);

        // Get intermediate images
        for (int i = 0; i < item.getInputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            // Discard cp image image (it happens with the output = input)
            String input = createInputImageName(isi, subgraphId);
            String output = createOutputImageName(subgraphId, 0, i);
            if (!input.equals(output)) {
                println("cp " + input + " " + output, 2);
            }
        }

        // Get intermediate parameters
        for (int i = 0; i < 2; i++) {
            ParameterSlot isp = virtualItem.getVirtualIntermediateResultSlot(i);
            if (isp.getConnectedSlot() != null) {
                println(createVariableName(subgraphId, 0, i)
                        + "=${"
                        + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank())
                        + "}", 2);
            }
        }
        println("done", 1);

        // Get output images
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
            println("ln -sf " + createInputImageName(isi, subgraphId) + " $1; shift", 1);
        }

        // Output result.
        ParameterSlot isp = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
        if (isp.getConnectedSlot() == null) {
            println("psetstatus SUCCESS", 1);
        } else {
            println("psetstatus ${" + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "}", 1);
        }
        println("}");

        generateSubroutinesDefinition(litems);
    }

    @Override
    protected void generateSubGraph( MacroOperatorItem item ) throws ArianeBadGraphException {
        SubGraphPresenter subGraph = item.getSubGraph();

        generateComment(item, subGraph.getListOfGraphItems());

        println("function macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "() {");

        // Get parameters.
        final int subgraphId = subGraph.getIdentifier();
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            println(createVariableName(subgraphId, 0, i) + "=$1; shift", 1);
        }

        // Get mask.
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName(mask, item.getGraph().getIdentifier()))) {
            int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
            println("ln -fs $1 " + createOutputImageName(subgraphId, 0, maskIndex) + "; shift", 1);
        }

        // Get input images.
        List<InputDataSlot> inputImages = item.getInputDataSlots();
        for (int i = 0; i < inputImages.size(); i++) {
            InputDataSlot isi = inputImages.get(i);
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            println("ln -fs $1 " + createOutputImageName(subgraphId, 0, i) + "; shift", 1);
        }

        println("");

        List<GraphItemModel> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());
        generateSubroutineBody(litems, subgraphId, 1);

        // Get output images.
        VirtualItem virtualItem = subGraph.getVirtualItem();
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
            println("ln -fs " + createInputImageName(isi, subgraphId) + " $1; shift", 1);
        }

        println("");

        // Output result.
        ParameterSlot isp = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
        if (isp.getConnectedSlot() == null) {
            println("psetstatus SUCCESS", 1);
        } else {
            println("psetstatus ${" + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "}", 1);
        }
        println("}");

        generateSubroutinesDefinition(litems);
    }

    private void generateComment( GraphItemModel item, List<GraphItemModel> litems ) {
        println("\n#\n# " + item.getNickName());
        for (GraphItemModel note : litems) {
            if (note instanceof NoteItem) {
                println("# " + convertCommentIntoShell(note.getFormula()));
            }
        }
        println("#");
    }

    private String createVariableName( int graphId, int itemId, int slotRank ) {
        return "var" + graphId + "_" + itemId + "_" + slotRank;
    }

    private String createOutputImageName( int graphId, int itemId, int slotRank ) {
        String name = UserPreferences.createTemporaryName(graphId, itemId, slotRank);
        _outputImages.add(name);
        return name;
    }

    private String getActualParameterValue( GraphItemModel item, int slotRank, int graphId ) {
        ParameterSlot isp = (ParameterSlot) item.getSlot("InSlotParameter", slotRank);
        if (isp.getConnectedSlot() == null) {
            String expression = ((ParameterSlot) item.getSlot("InSlotParameter", slotRank)).getValue();
            if (expression.charAt(0) == '\'') {
                return expression;
            } else {
                return "'" + expression + "'";
            }
        } else {
            return "${" + createVariableName(graphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "}";
        }
    }

    private String getNumericalParameterValue( GraphItemModel item, int slotRank, int graphId ) {
        ParameterSlot isp = (ParameterSlot) item.getSlot("InSlotParameter", slotRank);
        if (isp.getConnectedSlot() == null) {
            return ((ParameterSlot) item.getSlot("InSlotParameter", slotRank)).getValue();
        } else {
            return "${" + createVariableName(graphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "}";
        }
    }

    private String convertCommentIntoShell( String comment ) {
        if (comment != null) {
            return comment.replaceAll("\n", "\n# ");
        }
        return "";
    }

    private String convertArithmeticFormulaIntoShell( String formula, String[] parameters ) {
        StringBuilder code = new StringBuilder();
        StringTokenizer stingTokenizer = new StringTokenizer(formula, "$ -+()*/^0123456789.E", true);
        String etape;
        while (stingTokenizer.hasMoreTokens()) {
            etape = stingTokenizer.nextToken();
            if ("e".equals(etape) || "E".equals(etape)) {
                code.append("*10^");
            } else if ("cos".equals(etape)) {
                code.append("c");
            } else if ("sin".equals(etape)) {
                code.append("s");
            } else if ("ln".equals(etape)) {
                code.append("l");
            } else if ("exp".equals(etape)) {
                code.append("e");
            } else if ("atan".equals(etape)) {
                code.append("a");
            } else if ("$".equals(etape)) {
                int numberOfParameters = Integer.parseInt(stingTokenizer.nextToken());
                code.append(parameters[numberOfParameters]);
            } else {
                code.append(etape);
            }
        }
        return code.toString();
    }

    private String convertArithmeticFunctionIntoShell( String formula ) {
        StringBuilder code = new StringBuilder();
        StringTokenizer stringTokenizer = new StringTokenizer(formula, "$ -+()*/^0123456789.E", true);
        String etape;
        boolean abs = false;
        boolean max = false;
        boolean min = false;
        boolean tan = false;
        boolean log2 = false;
        boolean log10 = false;
        boolean acos = false;
        boolean asin = false;

        while (stringTokenizer.hasMoreTokens()) {
            etape = stringTokenizer.nextToken();
            if ("abs".equals(etape) && !abs) {
                code.append(" define abs(x){; if (x>=0) return(x) else return (-x)} ");
                abs = true;
            } else if ("max".equals(etape) && !max) {
                code.append(" define max(x,y){; if (x>y) return(x) else return(y)} ");
                max = true;
            } else if ("min".equals(etape) && !min) {
                code.append(" define min(x,y){; if (x<y) return(x) else return(y)} ");
                min = true;
            } else if ("atan".equals(etape) && !tan) {
                code.append(" define tan(x){; return (s(x)/c(x)} ");
                tan = true;
            } else if ("acos".equals(etape) && !acos) {
                code.append(" define acos(x){; if (x==1) return(0) if (x==-1) return (4*a(1)) return (2*a(1)-a(x/sqrt(1-(x^2))))} ");
                acos = true;
            } else if ("asin".equals(etape) && !asin) {
                code.append(" define asin(x){; if (x==1) return(2*a(1)) if (x==-1) return (-2*a(1)) return (a(x/sqrt(1-(x^2))))} ");
                asin = true;
            } else if ("log".equals(etape)) {
                etape = stringTokenizer.nextToken();
                if ("2".equals(etape) && !log2) {
                    code.append(" define log2(x){; return l(x)/l(2)} ");
                    log2 = true;
                } else if ("1".equals(etape) && !log10) {
                    code.append(" define log10(x){; return l(x)/l(10)} ");
                    log10 = true;
                }
            }
        }
        return code.toString();
    }

    private String convertLogicalFormulaIntoShell( String formula, String parameter0, String parameter1 ) {
        // Parse the boolean expression into Shell.
        char[] formulaCharacters = formula.toCharArray();
        StringBuilder commandShell = new StringBuilder();
        int index = 0;
        while (index < formulaCharacters.length) {
            if (formulaCharacters[index] == '&') {
                commandShell.append(" && ");
                index++;
            } else if (formulaCharacters[index] == '|') {
                commandShell.append(" || ");
                index++;
            } else if (formulaCharacters[index] == '<') {
                if (formulaCharacters[index + 1] == '=') {
                    commandShell.append(" <= ");
                    index += 2;
                } else {
                    commandShell.append(" < ");
                    index++;
                }
            } else if (formulaCharacters[index] == '>') {
                if (formulaCharacters[index + 1] == '=') {
                    commandShell.append(" >= ");
                    index += 2;
                } else {
                    commandShell.append(" > ");
                    index++;
                }
            } else if (formulaCharacters[index] == '=') {
                if (formulaCharacters[index + 1] == '=') {
                    commandShell.append(" == ");
                    index += 2;
                } else {
                    index++; // degenerated case!
                }
            } else if (formulaCharacters[index] == '!') {
                if (formulaCharacters[index + 1] == '=') {
                    commandShell.append(" != ");
                    index += 2;
                } else {
                    commandShell.append(" ! ");
                    index++;
                }
            } else if (formulaCharacters[index] == '(') {
                commandShell.append(" ");
                index++;
            } else if (formulaCharacters[index] == ')') {
                commandShell.append(" ");
                index++;
            } else if (formulaCharacters[index] == '$') {
                if (formulaCharacters[index + 1] == '0') {
                    commandShell.append(parameter0);
                } else {
                    commandShell.append(parameter1);
                }
                index += 2;
            } else {
                commandShell.append(formulaCharacters[index++]);
            }
        }
        return commandShell.toString();
    }
}
