/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.generator;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeBadGraphException;
import fr.greyc.pantheon.ariane.model.graph.element.CompositeItem;
import fr.greyc.pantheon.ariane.model.graph.element.ConstantItem;
import fr.greyc.pantheon.ariane.model.graph.element.ForControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.IfControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.MergeControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.SelectControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.WhileControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.EvalItem;
import fr.greyc.pantheon.ariane.model.graph.element.FileNameItem;
import fr.greyc.pantheon.ariane.model.graph.element.FormatItem;
import fr.greyc.pantheon.ariane.model.graph.element.GraphItemModel;
import fr.greyc.pantheon.ariane.model.graph.element.IScriptArgument;
import fr.greyc.pantheon.ariane.model.graph.element.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.element.MacroOperatorItem;
import fr.greyc.pantheon.ariane.model.graph.element.OperatorItem;
import fr.greyc.pantheon.ariane.model.graph.element.SaverItem;
import fr.greyc.pantheon.ariane.model.graph.element.ScriptItem;
import fr.greyc.pantheon.ariane.model.graph.element.SystemItem;
import fr.greyc.pantheon.ariane.model.graph.element.ViewerItem;
import fr.greyc.pantheon.ariane.model.graph.element.VirtualItem;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.SlotModel;
import fr.greyc.pantheon.ariane.presenter.Presenter;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.text.DateFormat;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Deque;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Arrays;

/**
 * The <code>CodeGenerator</code> is the base class for all the language generators.
 * A <code>CodeGenerator</code> generates a script file that contains code of a specified
 * language for encoding a specified graph.
 * <p>
 * It implements the <b>Template Method</b> Design Pattern.
 */
public abstract class CodeGenerator {
    private static final String[][] AVAILABLE_LANGUAGES = {
            {"fr.greyc.pantheon.ariane.model.generator.BashGenerator", "Bash", "sh"},
            {"fr.greyc.pantheon.ariane.model.generator.PythonGenerator", "Python", "py"},
            {"fr.greyc.pantheon.ariane.model.generator.PerlGenerator", "Perl", "pl"}
//      ,{"fr.greyc.pantheon.ariane.model.generator.MsdosGenerator", "MsDos", "bat"},
    };
    private File _scriptFile;
    private FileOutputCode _file;

    CodeGenerator() {
        // empty
    }

    public static String[][] getAvailable_Languages() {
        if (UserPreferences.isPandore()) {
            String[][] newAvailableLanguagesList = new String[AVAILABLE_LANGUAGES.length + 1][];
            for (int i = 0; i < AVAILABLE_LANGUAGES.length; i++) {
                newAvailableLanguagesList[i] = Arrays.copyOf(AVAILABLE_LANGUAGES[i], AVAILABLE_LANGUAGES[i].length);
            }
            newAvailableLanguagesList[AVAILABLE_LANGUAGES.length] = new String[]{"fr.greyc.pantheon.ariane.model.generator.CppGenerator", "c++", "cpp"};
            return newAvailableLanguagesList;
        } else {
            return AVAILABLE_LANGUAGES;
        }
    }

    public static CodeGenerator createCodeGeneratorFromLanguage( File chosenFile, String language ) {
        for (String[] language_info : AVAILABLE_LANGUAGES) {
            if (language_info[1].equals(language)) {
                try {
                    CodeGenerator generator = (CodeGenerator) Class.forName(language_info[0]).getDeclaredConstructor().newInstance();
                    generator._scriptFile = chosenFile;
                    return generator;
                } catch (IllegalAccessException | InstantiationException | ClassNotFoundException | NoSuchMethodException | InvocationTargetException e) {
                    return null;
                }
            }
        }
        return null;
    }

    public static CodeGenerator createCodeGeneratorFromSuffix( File chosenFile, String suffix ) {
        for (String[] language_info : AVAILABLE_LANGUAGES) {
            if (language_info[2].equals(suffix)) {
                try {
                    CodeGenerator generator = (CodeGenerator) Class.forName(language_info[0]).getDeclaredConstructor().newInstance();

                    generator._scriptFile = chosenFile;
                    return generator;
                } catch (IllegalAccessException | InstantiationException | ClassNotFoundException | NoSuchMethodException | InvocationTargetException e) {
                    return null;
                }
            }
        }
        return null;
    }

    public final void generate( Presenter presenter, List<GraphItemModel> items ) {
        try {
            _file = new FileOutputCode(_scriptFile);
            generateCode(items, _scriptFile.getName());
            presenter.printSystemMessage(ArianeSettings.getString("Generator.success") + " " + _scriptFile);
        } catch (IOException | ArianeBadGraphException e) {
            presenter.printErrorMessage(ArianeSettings.getString("Generator.failure") + e.getMessage());
        } finally {
            if (_file != null) {
                _file.close();
            }
        }
    }

    private void generateCode( List<GraphItemModel> listOfGraphItems, String filename ) throws ArianeBadGraphException {
        List<GraphItemModel> orderedListOfGraphItems = buildOrderedItemsSequenceUsingTopologicalSort(listOfGraphItems);
        List<GraphItemModel> args = getListOfAllScriptArguments(orderedListOfGraphItems);
        List<GraphItemModel> inputs = getListOfAllLoaderItems(orderedListOfGraphItems);
        List<GraphItemModel> outputs = getListOfAllSaverItems(orderedListOfGraphItems);

        generateScriptHeader(filename, args, inputs, outputs);
        generateUtilitySubroutinesDefinition();
        generateSubroutinesDefinition(orderedListOfGraphItems);
        generateMainPrologue(orderedListOfGraphItems, args, inputs, outputs);
        generateSubroutineBody(orderedListOfGraphItems, 1, 1);
        generateMainEpilogue(args, inputs, outputs);
        generateScriptFooter();
    }

    final List<GraphItemModel> buildOrderedItemsSequenceUsingTopologicalSort( List<GraphItemModel> litems ) {
        List<GraphItemModel> graphItems = new ArrayList<>(litems);
        List<GraphItemModel> orderedItems = new ArrayList<>();
        Map<GraphItemModel, Integer> incomings = computeNumberOfIncomings(graphItems);
        Deque<GraphItemModel> stack = getAllNodesWithNoIncomings(incomings);
        // depth-first search: Do not change this order.
        while (!stack.isEmpty()) {
            GraphItemModel currentItem = stack.removeFirst();
            orderedItems.add(currentItem);
            graphItems.remove(currentItem);
            List<GraphItemModel> successors = getSuccessorsOfItem(currentItem);
            Collections.reverse(successors);
            for (GraphItemModel successorItem : successors) {
                int nbOfIncomings = incomings.get(successorItem);
                incomings.put(successorItem, --nbOfIncomings);
                if (nbOfIncomings == 0) {
                    stack.addFirst(successorItem);
                }
            }
        }
        return orderedItems;
    }

    private List<GraphItemModel> getSuccessorsOfItem( GraphItemModel item ) {
        List<GraphItemModel> successors = new ArrayList<>();
        for (OutputDataSlot outputSlots : item.getOutputDataSlots()) {
            for (InputSlot inputSlot : outputSlots.getConnectedInputSlots()) {
                GraphItemModel nextItem = inputSlot.getGraphItem();
                successors.add(nextItem);
            }
        }
        OutputResultSlot outputResult = item.getOutputResultSlot();
        if (outputResult != null) {
            for (InputSlot inputSlot : outputResult.getConnectedInputSlots()) {
                GraphItemModel nextItem = inputSlot.getGraphItem();
                successors.add(nextItem);
            }
        }
        return successors;
    }

    private Map<GraphItemModel, Integer> computeNumberOfIncomings( List<GraphItemModel> graphItems ) {
        Map<GraphItemModel, Integer> incomings = new HashMap<>();

        for (GraphItemModel item : graphItems) {
            incomings.put(item, 0);
        }

        for (GraphItemModel item : graphItems) {
            for (GraphItemModel nextItem : getSuccessorsOfItem(item)) {
                if (!(nextItem instanceof VirtualItem)) {
                    int numberOfIncomings = incomings.get(nextItem);
                    incomings.put(nextItem, ++numberOfIncomings);
                }
            }
        }
        return incomings;
    }

    private Deque<GraphItemModel> getAllNodesWithNoIncomings( Map<GraphItemModel, Integer> incomings ) {
        Deque<GraphItemModel> litems = new ArrayDeque<>();

        incomings.entrySet().stream()
                .filter(( entry ) -> (incomings.get(entry.getKey()) == 0))
                .forEach(( entry ) -> litems.add(entry.getKey()));
        return litems;
    }

    private List<GraphItemModel> getListOfAllLoaderItems( List<GraphItemModel> litems ) {
        List<GraphItemModel> loaderItems = new ArrayList<>();
        for (GraphItemModel item : litems) {
            if (item instanceof LoaderItem && !item.getInputParameterSlots().get(0).isConnected()) {
                loaderItems.add(item);
            }
        }
        return loaderItems;
    }

    private List<GraphItemModel> getListOfAllScriptArguments( List<GraphItemModel> litems ) {
        List<GraphItemModel> items = new ArrayList<>();
        for (GraphItemModel item : litems) {
            if (item instanceof IScriptArgument && ((IScriptArgument) item).isArgumentOfRelatedScript()) {
                items.add(item);
            }
        }
        return items;
    }

    private List<GraphItemModel> getListOfAllSaverItems( List<GraphItemModel> litems ) {
        ArrayList<GraphItemModel> items = new ArrayList<>();
        for (GraphItemModel item : litems) {
            if (item instanceof SaverItem && !item.getInputParameterSlots().get(0).isConnected()) {
                items.add(item);
            }
        }
        return items;
    }

    final void generateSubroutinesDefinition( List<GraphItemModel> litems ) throws ArianeBadGraphException {
        for (GraphItemModel graphItem : litems) {
            if (graphItem instanceof WhileControlStructureItem) {
                generateSubGraph((WhileControlStructureItem) graphItem);
            } else if (graphItem instanceof ForControlStructureItem) {
                generateSubGraph((ForControlStructureItem) graphItem);
            } else if (graphItem instanceof MacroOperatorItem) {
                generateSubGraph((MacroOperatorItem) graphItem);
            }
        }
    }

    final void generateSubroutineBody( List<GraphItemModel> orderedListOfGraphItems, int graphNumber, int indentation ) throws ArianeBadGraphException {
        // 'IF' generation is more complex that is seems
        // when there exists items that are shared by operators in
        // both IF and ELSE parts.
        List<GraphItemModel> ifItems = new ArrayList<>();
        List<List<GraphItemModel>> elseItems = new ArrayList<>();
        List<GraphItemModel> generatedItems = new ArrayList<>();
        int indent = indentation;

        for (GraphItemModel graphItem : orderedListOfGraphItems) {
            generatedItems.add(graphItem);

            updateIfItems(graphNumber, ifItems, elseItems, indent, graphItem);

            if (graphItem instanceof LoaderItem) {
                generateItem((LoaderItem) graphItem, graphNumber, indent);
            } else if (graphItem instanceof SaverItem) {
                generateItem((SaverItem) graphItem, graphNumber, indent);
            } else if (graphItem instanceof ViewerItem) {
                generateItem((ViewerItem) graphItem, graphNumber, indent);
            } else if (graphItem instanceof EvalItem) {
                generateItem((EvalItem) graphItem, graphNumber, indent);
            } else if (graphItem instanceof FileNameItem) {
                generateItem((FileNameItem) graphItem, graphNumber, indent);
            } else if (graphItem instanceof FormatItem) {
                generateItem((FormatItem) graphItem, graphNumber, indent);
            } else if (graphItem instanceof ConstantItem) {
                generateItem((ConstantItem) graphItem, graphNumber, indent);
            } else if (graphItem instanceof ScriptItem) {
                generateItem((ScriptItem) graphItem, graphNumber, indent);
            } else if (graphItem instanceof SystemItem) {
                generateItem((SystemItem) graphItem, graphNumber, indent);
            } else if (graphItem instanceof OperatorItem) {
                generateItem((OperatorItem) graphItem, graphNumber, indent);
            } else if (graphItem instanceof SelectControlStructureItem) {
                generateItem((SelectControlStructureItem) graphItem, graphNumber, indent);
            } else if (graphItem instanceof IfControlStructureItem) {
                ifItems.add(0, graphItem);
                generateItem((IfControlStructureItem) graphItem, graphNumber, indent);
                generateThenPart((IfControlStructureItem) graphItem, graphNumber, indent++);
                elseItems.add(0, searchForFirstItemsOfTheElsePart(graphItem, generatedItems));
            } else if (graphItem instanceof MergeControlStructureItem) {
                if (!ifItems.isEmpty()) {
                    generateElsePart((IfControlStructureItem) ifItems.get(0), graphNumber, indent - 1);
                    elseItems.remove(0);
                    ifItems.remove(0);
                }
                generateFiPart(null, graphNumber, --indent);
                generateItem((MergeControlStructureItem) graphItem, graphNumber, indent);
            } else if (graphItem instanceof MacroOperatorItem) {
                generateItem((MacroOperatorItem) graphItem, graphNumber, indent);
            } else if (graphItem instanceof WhileControlStructureItem) {
                generateItem((CompositeItem) graphItem, graphNumber, indent);
            } else if (graphItem instanceof ForControlStructureItem) {
                generateItem((CompositeItem) graphItem, graphNumber, indent);
            }
        }
        for (int i = 0; i < ifItems.size(); i++) {
            generateFiPart(null, graphNumber, indent);
        }
    }

    private void updateIfItems( int graphNumber, List<GraphItemModel> ifItems, List<List<GraphItemModel>> elseItems, int indent, GraphItemModel graphItem ) {
        if (!ifItems.isEmpty() && !elseItems.isEmpty()) {
            List<GraphItemModel> items = elseItems.get(0);
            if (items.contains(graphItem)) {
                generateElsePart((IfControlStructureItem) ifItems.get(0), graphNumber, indent - 1);
                ifItems.remove(0);
            }
        }
    }

    private List<GraphItemModel> searchForFirstItemsOfTheElsePart( GraphItemModel ifItem, List<GraphItemModel> generatedItems ) {
        OutputDataSlot elseSlot = ifItem.getOutputDataSlots().get(1);
        List<InputSlot> connectedSlots = elseSlot.getConnectedInputSlots();

        // add items that are directly connected to the else slot of the if item (if not merge).
        ArrayList<GraphItemModel> graphItems = new ArrayList<>();
        for (InputSlot slot : connectedSlots) {
            if (!(slot.getGraphItem() instanceof MergeControlStructureItem)) {
                graphItems.add(slot.getGraphItem());
            }
        }

        // add items that compute the parameters of connected items
        // and that are not already generated.
        ArrayList<GraphItemModel> graphItemsBis = new ArrayList<>();
        for (GraphItemModel item : graphItems) {
            for (ParameterSlot parameter : item.getInputParameterSlots()) {
                if (parameter.isConnected()) {
                    GraphItemModel connectedItem = parameter.getConnectedSlot().getGraphItem();
                    if (!generatedItems.contains(connectedItem)) {
                        graphItemsBis.add(connectedItem);
                    }
                }
            }
        }

        // add operators that computes the input images
        // that are not already generated.
        for (GraphItemModel item : graphItems) {
            for (InputDataSlot slot : item.getInputDataSlots()) {
                GraphItemModel connectedItem = slot.getConnectedSlot().getGraphItem();
                if (slot.isConnected() && !generatedItems.contains(connectedItem)) {
                    graphItemsBis.add(connectedItem);
                }
            }
        }
        graphItems.addAll(graphItemsBis);
        return graphItems;
    }

    protected abstract void generateScriptHeader( String filename, List<GraphItemModel> arguments, List<GraphItemModel> inputs, List<GraphItemModel> outputs );

    protected abstract void generateScriptFooter();

    protected abstract void generateUtilitySubroutinesDefinition();

    protected abstract void generateMainPrologue( List<GraphItemModel> items, List<GraphItemModel> arguments, List<GraphItemModel> inputs, List<GraphItemModel> outputs );

    protected abstract void generateMainEpilogue( List<GraphItemModel> arguments, List<GraphItemModel> inputs, List<GraphItemModel> outputs );

    protected abstract void generateItem( LoaderItem item, int graphNumber, int indent );

    protected abstract void generateItem( SaverItem item, int graphNumber, int indent );

    protected abstract void generateItem( ViewerItem item, int graphNumber, int indent );

    protected abstract void generateItem( OperatorItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

    protected abstract void generateItem( EvalItem item, int graphNumber, int indent );

    protected abstract void generateItem( FileNameItem item, int graphNumber, int indent );

    protected abstract void generateItem( FormatItem item, int graphNumber, int indent );

    protected abstract void generateItem( ConstantItem item, int graphNumber, int indent );

    protected abstract void generateItem( ScriptItem item, int graphNumber, int indent );

    protected abstract void generateItem( SystemItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

    protected abstract void generateItem( SelectControlStructureItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

    protected abstract void generateThenPart( IfControlStructureItem item, int graphNumber, int indent );

    protected abstract void generateElsePart( IfControlStructureItem item, int graphNumber, int indent );

    protected abstract void generateFiPart( IfControlStructureItem item, int graphNumber, int indent );

    protected abstract void generateItem( IfControlStructureItem item, int graphNumber, int indent );

    protected abstract void generateItem( MergeControlStructureItem item, int graphNumber, int indent );

    protected abstract void generateItem( CompositeItem item, int graphNumber, int indent ) throws ArianeBadGraphException;

    protected abstract void generateSubGraph( MacroOperatorItem item ) throws ArianeBadGraphException;

    protected abstract void generateSubGraph( ForControlStructureItem item ) throws ArianeBadGraphException;

    protected abstract void generateSubGraph( WhileControlStructureItem item ) throws ArianeBadGraphException;

    final String getCurrentDate() {
        final Date creationDate = new Date();
        final DateFormat dateFormat = DateFormat.getDateInstance(DateFormat.MEDIUM);
        return dateFormat.format(creationDate);
    }

    final GraphItemModel getVirtualOperator( List<GraphItemModel> litems ) {
        for (GraphItemModel graphItem : litems) {
            if (graphItem instanceof VirtualItem) {
                return graphItem;
            }
        }
        return null;
    }

    final String createInputImageName( InputSlot slot, int graphId ) {
        return UserPreferences.createTemporaryName(graphId, slot.getConnectedSlot().getGraphItem().getId(), slot.getConnectedSlot().getRank());
    }

    final boolean hasMergeItemInItsSuccessors( GraphItemModel item ) {
        for (OutputSlot slot : item.getOutputDataSlots()) {
            if (slot.isConnected()) {
                for (SlotModel connectedSlot : slot.getConnectedSlots()) {
                    if (connectedSlot.getGraphItem() instanceof MergeControlStructureItem) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    final void println( String text ) {
        _file.println(text);
    }

    final void println( String text, int numberOfWhiteSpaces ) {
        _file.println(text, numberOfWhiteSpaces);
    }

    final void print( String text, int numberOfWhiteSpaces ) {
        _file.print(text, numberOfWhiteSpaces);
    }

    final void print( String text ) {
        _file.print(text);
    }
}
