/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.generator;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeBadGraphException;
import fr.greyc.pantheon.ariane.model.graph.element.CompositeItem;
import fr.greyc.pantheon.ariane.model.graph.element.ConstantItem;
import fr.greyc.pantheon.ariane.model.graph.element.ForControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.IfControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.MergeControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.SelectControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.EvalItem;
import fr.greyc.pantheon.ariane.model.graph.element.FileNameItem;
import fr.greyc.pantheon.ariane.model.graph.element.FormatItem;
import fr.greyc.pantheon.ariane.model.graph.element.GraphItemModel;
import fr.greyc.pantheon.ariane.model.graph.element.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.element.MacroOperatorItem;
import fr.greyc.pantheon.ariane.model.graph.element.NoteItem;
import fr.greyc.pantheon.ariane.model.graph.element.OperatorItem;
import fr.greyc.pantheon.ariane.model.graph.element.SaverItem;
import fr.greyc.pantheon.ariane.model.graph.element.ScriptItem;
import fr.greyc.pantheon.ariane.model.graph.element.SystemItem;
import fr.greyc.pantheon.ariane.model.graph.element.ViewerItem;
import fr.greyc.pantheon.ariane.model.graph.element.VirtualItem;
import fr.greyc.pantheon.ariane.model.graph.element.WhileControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.presenter.SubGraphPresenter;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * The <code>PerlGenerator</code> generates Python code for the specified graph.
 * It operates in three steps:<ol>
 * <li>transforms the graph into a list of operators usable for the topological sort;
 * <li>executes the topological sort;
 * <li>generates the related code.
 * </ol>
 */
public final class CppGenerator extends CodeGenerator {
    /**
     * The list of generated output images.
     */
    private final List<String> _outputImages = new ArrayList<>();

    protected CppGenerator() {
        // Only created by the code generator.
    }

    @Override
    protected void generateScriptHeader( String filename, List<GraphItemModel> arguments, List<GraphItemModel> inputs, List<GraphItemModel> outputs ) {
        println("/*");
        println(" *");
        println(" * @(#)" + filename + "			" + getCurrentDate());
        println(" *");
        println(" * This file was automatically generated by Ariane.");
        println(" * Ariane is part of the Pantheon Project.");
        println(" */\n");

        println("#include <pandore.h>");
        println("using namespace pandore;\n");
        println("#undef MAIN\n");
    }

    @Override
    protected void generateScriptFooter() {
        // Do nothing
    }

    @Override
    protected void generateUtilitySubroutinesDefinition() {
    }

    @Override
    protected void generateMainPrologue( List<GraphItemModel> litems, List<GraphItemModel> arguments, List<GraphItemModel> inputs, List<GraphItemModel> outputs ) {
        println("\n/*\n * The main program");

        // Comments
        for (GraphItemModel item : litems) {
            if (item instanceof NoteItem) {
                println(convertCommentIntoPython(item.getFormula()), 1);
            }
        }
        println(" */");
        print("Errc Script( ");

        for (GraphItemModel item : inputs) {
            print("Img2duc &" + createOutputImageName1(1, item.getId(), 0) + ", ");
        }
        for (GraphItemModel item : arguments) {
            print(createVariableName(1, item.getId(), 0) + ", ");
        }
        println(") {");
    }

    @Override
    protected void generateMainEpilogue( List<GraphItemModel> arguments, List<GraphItemModel> inputs, List<GraphItemModel> outputs ) {
        println("\nreturn SUCCESS;", 1);
        println("}\n");
        print("#define USAGE	\"usage: %s");
        for (GraphItemModel argument : arguments) {
            print(" " + argument.getNickName());
        }
        for (int i = 0; i < inputs.size(); i++) {
            print(" input-file" + i);
        }
        for (int i = 0; i < outputs.size(); i++) {
            print(" output-file" + i);
        }
        println("\")");

        println("#define PARC	" + arguments.size());
        println("#define FINC	" + inputs.size());
        println("#define FOUTC	" + outputs.size());
        println("#define MASK	1\n");

        println("int main( int argc, char *argv[] ) {");
        println("Errc result;                // The result code of the execution.", 1);
        println("Pobject* mask;              // The region map.", 1);
        println("Pobject* objin[FINC + 1];   // The input objects;", 1);
        println("Pobject* objs[FINC + 1];    // The source objects masked.", 1);
        println("Pobject* objout[FOUTC + 1]; // The output object.", 1);
        println("Pobject* objd[FOUTC + 1];   // The result object of the execution.", 1);
        println("char* parv[PARC + 1];       // The input parameters.\n", 1);

        println("ReadArgs(argc, argv, PARC, FINC, FOUTC, &mask, objin, objs, objout, objd, parv, USAGE, MASK);", 1);
        println("if (objs[0]->Type() == Img2duc) {", 1);
        println("Img2duc* const rgs = (Img2duc*)objs[0];", 2);
        println("objd[0] = new Img2duc(rgs->Props());", 2);
        println("Img2duc* rgd = (Imgéduc*)objd[0];\n", 2);
        println("result = Script(*rgs, *rgd, atoi(parv[0]), atoi(parv[1]));", 2);
        println("} else {", 1);
        println("PrintErrorFormat(objin, FINC);", 2);
        println("result = FAILURE;", 2);
        println("}", 1);
        println("if (result) {", 1);
        println("WriteArgs(argc, argv, PARC, FINC, FOUTC, &mask, objin, objs, objout, objd, MASK);", 2);
        println("}", 1);
        println("Exit(result);", 1);
        println("return 0;", 1);
        println("}");
    }

    @Override
    protected void generateItem( LoaderItem item, int graphId, int indent ) {
        if (item.getInputParameterSlots().get(0).isConnected()) {
            String parameter = getActualParameterValue(item, 0, graphId);
            String imagename = createOutputImageName1(graphId, item.getId(), 0);
            println("shutil.copy(" + parameter + ", " + "\"" + imagename + "\")", indent);
            println(createVariableName(graphId, item.getId(), 1) + " = \"" + imagename + "\"", indent);
        }
    }

    @Override
    protected void generateItem( SaverItem item, int graphId, int indent ) {
        if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
            if (item.getInputParameterSlots().get(0).isConnected()) {
                String parameters = getActualParameterValue(item, 0, graphId);
                println("shutil.copy(\"" + createInputImageName1(item.getInputDataSlots().get(0), graphId)
                        + "\", " + parameters + ")", indent);
            } else {
                println("shutil.copy(\"" + createInputImageName1(item.getInputDataSlots().get(0), graphId)
                        + "\", args[argc]); argc += 1", indent);
            }
            String imagename = createOutputImageName1(graphId, item.getId(), 0);
            println("link(\""
                    + createInputImageName1(item.getInputDataSlots().get(0), graphId)
                    + "\", \"" + imagename + "\")", indent);
            println(createVariableName(graphId, item.getId(), 0) + " = \"" + imagename + "\"", indent);
        }
    }

    @Override
    protected void generateItem( ViewerItem item, int graphId, int indent ) {
        // TODO c'est quoi ?
//      if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
//         final String commandShell = item.getViewerCommandName();
//         println("subprocess.call([\"" + commandShell + "\", \"" + createInputImageName1(item.getInputDataSlots().get(0), graphId) + "\"])", indent);
//         println("link(\""
//             + createInputImageName1(item.getInputDataSlots().get(0), graphId)
//             + "\", \""
//             + createOutputImageName11(graphId, item.getId(), 0)
//             + "\")", indent);
//      }
    }

    private String createInputImageName1( InputSlot slot, int graphId ) {
        String filename = UserPreferences.createTemporaryName(graphId, slot.getConnectedSlot().getGraphItem().getId(), slot.getConnectedSlot().getRank());
        return filename.substring(0, filename.indexOf('.'));
    }

    private String createOutputImageName1( int graphId, int itemId, int slotRank ) {
        String name = UserPreferences.createTemporaryName(graphId, itemId, slotRank);
        name = name.substring(0, name.indexOf('.'));
        _outputImages.add(name);
        return name;
    }

    @Override
    protected void generateItem( OperatorItem item, int graphId, int indent ) throws ArianeBadGraphException {
        String codeName = OperatorItem.getCommandPart(item.getBinary());
        if (codeName == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error2") + item.getName());
        }

        // Get input images.
        List<String> inputLine = new LinkedList<>();
        for (InputDataSlot image : item.getInputDataSlots()) {
            if (image.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            inputLine.add(createInputImageName1(image, graphId));
        }

        // Get parameters
        List<String> parameterLine = new LinkedList<>();
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameterLine.add(getActualParameterValue(item, i, graphId));
        }

        // Get mask
        List<String> maskLine = new LinkedList<>();
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName1(mask, graphId))) {
            maskLine.add(createInputImageName1(mask, graphId));
        }

        // Get output images.
        List<String> outputLine = new LinkedList<>();
        List<OutputDataSlot> outputDataSlots = item.getOutputDataSlots();
        for (int i = 0; i < outputDataSlots.size(); i++) {
            outputLine.add(createOutputImageName1(graphId, item.getId(), i));
        }

        StringBuilder code = null;
        List<String> commandLine = item.buildAbstractCommandLine(maskLine, inputLine, outputLine, parameterLine);
        commandLine.remove(0);
        for (String command : commandLine) {
            if (code == null) {
                code = new StringBuilder(command);
            } else {
                code.append(", ").append(command);
            }
        }
        if (item.getOutputResultSlot().isConnected() || hasMergeItemInItsSuccessors(item)) {
            println(createVariableName(graphId, item.getId(), 0) + " = " + codeName + "(" + code + ");", indent);
        } else {
            println(codeName + "(" + code + ");", indent);
        }
    }

    @Override
    protected void generateItem( SystemItem item, int graphId, int indent ) throws ArianeBadGraphException {
        String codeName = OperatorItem.getCommandPart(item.getBinary());
        if (codeName == null || codeName.trim().isEmpty()) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error3") + item.getName() + "(" + item.getId() + ")");
        }

        // Get parameters
        List<String> parameterLine = new LinkedList<>();
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameterLine.add(getActualParameterValue(item, i, graphId));
        }

        // Get mask
        List<String> maskLine = new LinkedList<>();
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName1(mask, graphId))) {
            maskLine.add(createInputImageName1(mask, graphId));
        }

        // Get input images.
        List<String> inputLine = new LinkedList<>();
        for (InputDataSlot image : item.getInputDataSlots()) {
            if (image.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            inputLine.add("\"" + createInputImageName1(image, graphId) + "\"");
        }

        // Get output images.
        List<OutputDataSlot> outputDataSlots = item.getOutputDataSlots();
        List<String> outputLine = new LinkedList<>();
        for (int i = 0; i < outputDataSlots.size(); i++) {
            outputLine.add("\"" + createOutputImageName1(graphId, item.getId(), i) + "\"");
        }

        List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
        StringBuilder code = null;
        for (String command : commandLine) {
            if (code == null) {
                code = new StringBuilder("\"" + command + " ");
            } else if (command.startsWith("\"var")) {
                code.append("\\\"\" + ").append(command.substring(1, command.length() - 1)).append(" + \"\\\" ");
            } else if (command.startsWith("var")) {
                code.append("\" + ").append(command).append(" + \" ");
            } else if (command.startsWith("'")) {
                code.append(" ").append(command).append(" ");
            } else {
                code.append(command.replaceAll("\"", "\\\\\"")).append(" ");
            }
        }

        println(createVariableName(graphId, item.getId(), 0) + " = " + "os.system(" + code + "\")", indent);
    }

    @Override
    protected void generateItem( EvalItem item, int graphId, int indent ) {
        String[] parameters = new String[item.getInputParameterSlots().size()];
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameters[i] = getActualParameterValue(item, i, graphId);
        }
        println(createVariableName(graphId, item.getId(), 0)
                + " = "
                + convertArithmeticFormulaIntoPython(item.getFormula(), parameters), indent);
    }

    @Override
    protected void generateItem( FormatItem item, int graphId, int indent ) {
        // Get parameters
        String[] parameters = new String[item.getInputParameterSlots().size()];
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameters[i] = getActualParameterValue(item, i, graphId);
        }

        StringBuilder code = new StringBuilder();
        StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), " $()+'\"", true);
        String etape;
        while (stringTokenizer.hasMoreTokens()) {
            etape = stringTokenizer.nextToken();
            if ("+".equals(etape)) {
                code.append(" + ");
            } else if ("quote".equals(etape)) {
                code.append("");
            } else if ("\"".equals(etape)) {
                code.append("\"");
            } else if ("(".equals(etape)) {
                code.append("\"'\"+");
            } else if (")".equals(etape)) {
                code.append("+\"'\"");
            } else if ("$".equals(etape)) {
                int numberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
                code.append(parameters[numberOfParameters]);
            } else {
                code.append(etape);
            }
        }
        println(createVariableName(graphId, item.getId(), 0) + " = " + code, indent);
    }

    @Override
    protected void generateItem( SelectControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
        println("# Select", indent);

        final String value = getActualParameterValue(item, 0, graphId);
        for (int i = 0; i < item.getInputDataSlots().size(); i++) {
            if (!item.getInputDataSlots().get(i).isConnected()) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            println("if (int(" + value + ") == " + i + "):", indent);
            println("link(\""
                    + createInputImageName1(item.getInputDataSlots().get(i), graphId) + "\", \""
                    + createOutputImageName1(graphId, item.getId(), 0) + "\")", indent + 1);
            println(createVariableName(graphId, item.getId(), 0) + " = \"" + i + "\"", indent + 1);
        }
    }

    @Override
    protected void generateItem( IfControlStructureItem item, int graphId, int indent ) {
        String[] parameters = new String[2];
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameters[i] = getActualParameterValue(item, i, graphId);
        }

        // Parse the boolean expression.
        StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), "&|$^!0123456789.E><=() +*-/", true);
        StringBuilder code = new StringBuilder();
        while (stringTokenizer.hasMoreTokens()) {
            String token = stringTokenizer.nextToken();
            if ("&".equals(token)) {
                code.append(" and ");
            } else if ("|".equals(token)) {
                code.append(" or ");
            } else if ("$".equals(token)) {
                String token2 = stringTokenizer.nextToken();
                if (token2.charAt(0) == '0') {
                    code.append("float(").append(parameters[0]).append(")");
                } else {
                    code.append("float(").append(parameters[1]).append(")");
                }
            } else {
                code.append(token);
            }
        }
        println("if " + code + ":", indent);
    }

    @Override
    protected void generateThenPart( IfControlStructureItem item, int graphId, int indent ) {
        println("link(\""
                + createInputImageName1(item.getInputDataSlots().get(0), graphId)
                + "\", \""
                + createOutputImageName1(graphId, item.getId(), 0) + "\")", indent + 1);
        println(createVariableName(graphId, item.getId(), 0) + " = 0", indent + 1);
    }

    @Override
    protected void generateElsePart( IfControlStructureItem item, int graphId, int indent ) {
        println("else:", indent);
        println("link(\""
                + createInputImageName1(item.getInputDataSlots().get(0), graphId)
                + "\", \""
                + createOutputImageName1(graphId, item.getId(), 1) + "\")", indent + 1);
        println(createVariableName(graphId, item.getId(), 0) + " = 1", indent + 1);
    }

    @Override
    protected void generateFiPart( IfControlStructureItem item, int graphId, int indent ) {
        // Nothing to do
    }

    @Override
    protected void generateItem( MergeControlStructureItem item, int graphId, int indent ) {
        InputDataSlot in0 = item.getInputDataSlots().get(0);

        int id0 = in0.getConnectedSlot().getGraphItem().getId();

        println("# Merge", indent);
        println("try: " + createVariableName(graphId, id0, 0) + "; input = 0", indent);
        println("except UnboundLocalError: input = 1", indent);
        print("if input == 0 and ", indent);
        println("os.path.isfile(\"" + createInputImageName1(item.getInputDataSlots().get(0), graphId) + "\"):");
        println("link(\""
                + createInputImageName1(item.getInputDataSlots().get(0), graphId)
                + "\", \""
                + createOutputImageName1(graphId, item.getId(), 0) + "\")", indent + 1);
        println(createVariableName(graphId, item.getId(), 0) + " = 0", indent + 1);
        println("else:", indent);
        println("link(\""
                + createInputImageName1(item.getInputDataSlots().get(1), graphId) + "\", \""
                + createOutputImageName1(graphId, item.getId(), 0) + "\")", indent + 1);
        println(createVariableName(graphId, item.getId(), 0) + " = 1", indent + 1);
    }

    @Override
    protected void generateItem( FileNameItem item, int graphId, int indent ) {
        if (!item.isArgumentOfRelatedScript()) {
            println(createVariableName(graphId, item.getId(), 0) + " = '" + item.getFormula() + "'", indent);
        }
    }

    @Override
    protected void generateItem( ConstantItem item, int graphId, int indent ) {
        if (!item.isArgumentOfRelatedScript()) {
            String value = item.getFormula().replaceAll("\"", "\\\\\"");
            println(createVariableName(graphId, item.getId(), 0) + " = \"" + value + "\"", indent);
        }
    }

    @Override
    protected void generateItem( ScriptItem item, int graphId, int indent ) {
        String value = item.getFormula().replaceAll("\"", "\\\\\"");
        println(createVariableName(graphId, item.getId(), 0) + " = \"\"\"" + value + "\"\"\"", indent);
    }

    @Override
    protected void generateItem( CompositeItem item, int graphId, int indent ) throws ArianeBadGraphException {
        StringBuilder code = new StringBuilder("macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "(");
        boolean comma = false;

        // Get input images.
        for (InputDataSlot image : item.getInputDataSlots()) {
            if (image.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
            }
            if (comma) {
                code.append(", ");
            } else {
                comma = true;
            }
            code.append(createInputImageName1(image, graphId));
        }

        // Get output images.
        Collection<OutputDataSlot> outputDataSlots = item.getOutputDataSlots();
        for (int i = 0; i < outputDataSlots.size(); i++) {
            if (comma) {
                code.append(", ");
            } else {
                comma = true;
            }
            code.append(createOutputImageName1(graphId, item.getId(), i));
        }
        // Get parameters
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            if (comma) {
                code.append(", ");
            } else {
                comma = true;
            }
            code.append(getActualParameterValue(item, i, graphId));
        }

        // Get output result.
        if (item.getOutputResultSlot().isConnected()) {
            println(createVariableName(graphId, item.getId(), 0) + " = " + code + ");", indent);
        } else {
            println(code + ");", indent);
        }
    }

    @Override
    protected void generateSubGraph( ForControlStructureItem item ) throws ArianeBadGraphException {
        SubGraphPresenter subGraph = item.getSubGraph();

        generateComment(item, subGraph.getListOfGraphItems());

        println("Errc macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "( *args ) {");

        println("argc = 0", 1);

        // Get parameters
        int subgraphId = subGraph.getIdentifier();
        println("begin = int(args[argc]); argc += 1", 1);
        println("end = int(args[argc]); argc += 1", 1);
        println("step = int(args[argc]); argc += 1", 1);
        for (int i = 3; i < item.getInputParameterSlots().size(); i++) {
            println(createVariableName(subgraphId, 0, i - 2) + " = args[argc]; argc += 1", 1);
        }
        println(createVariableName(subgraphId, 0, 0) + " = str(begin)", 1);

        // Get mask.
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName1(mask, item.getGraph().getIdentifier()))) {
            int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
            println("link(args[argc], \"" + createOutputImageName1(subgraphId, 0, maskIndex) + "\"); argc += 1", 1);
        }

        // Get input files
        int rank = 0;
        for (InputDataSlot isi : item.getInputDataSlots()) {
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
            }
            println("shutil.copy(args[argc], \"" + createOutputImageName1(subgraphId, 0, rank++) + "\"); argc += 1", 1);
        }

        // Create the output images in case of no iteration
        VirtualItem virtualItem = subGraph.getVirtualItem();
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
            }
            println("if not os.path.exists(\"" + createInputImageName1(isi, subgraphId) + "\"): open( \"" + createInputImageName1(isi, subgraphId) + "\", 'w').close()", 1);
        }

        // Initialize the return variable in case of no iteration
        {
            ParameterSlot isp = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
            println(createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "=\"SUCCESS\"", 1);
        }

        println("");
        println("for i in range(begin, end+1, step):", 1);
        println(createVariableName(subgraphId, 0, 0) + " = str(i)", 2);

        List<GraphItemModel> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());
        generateSubroutineBody(litems, subgraphId, 2);

        // Get intermediate images
        for (int i = 0; i < item.getInputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
            }
            String input = createInputImageName1(isi, subgraphId);
            String output = createOutputImageName1(subgraphId, 0, i);
            if (!input.equals(output)) {
                println("shutil.copy(\"" + input + "\",\"" + output + "\")", 2);
            }
        }

        // Get output images.
        println("");

        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
            println("link(\"" + createInputImageName1(isi, subgraphId) + "\",args[argc]); argc += 1", 1);
        }

        // Output result
        ParameterSlot isp = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
        if (isp.getConnectedSlot() == null) {
            println("return 0;", 1);
        } else {
            println("return " + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + ";", 1);
        }
        println("");

        generateSubroutinesDefinition(litems);
        println("}");
    }

    @Override
    protected void generateSubGraph( WhileControlStructureItem item ) throws ArianeBadGraphException {
        SubGraphPresenter subGraph = item.getSubGraph();

        generateComment(item, subGraph.getListOfGraphItems());

        println("Errc macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "( *args ) {");

        println("argc = 0", 1);

        // Get parameters
        int subgraphId = subGraph.getIdentifier();
        String[] parameters = new String[item.getInputParameterSlots().size()];
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            println(createVariableName(subgraphId, 0, i) + " = args[argc]; argc += 1", 1);
            parameters[i] = createVariableName(subgraphId, 0, i);
        }

        // Get mask.
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(createInputImageName1(mask, item.getGraph().getIdentifier()))) {
            int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
            println("link(args[argc], \"" + createOutputImageName1(subgraphId, 0, maskIndex) + "\"); argc += 1", 1);
        }

        // Get input files
        int rank = 0;
        for (InputDataSlot isi : item.getInputDataSlots()) {
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
            }
            println("shutil.copy(args[argc], \"" + createOutputImageName1(subgraphId, 0, rank++) + "\"); argc += 1", 1);
        }

        // Parse the boolean expression.
        StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), "&|$^!0123456789.E><=() +*-/", true);
        String etape;
        StringBuilder testFormula = new StringBuilder();
        while (stringTokenizer.hasMoreTokens()) {
            etape = stringTokenizer.nextToken();
            if ("&".equals(etape)) {
                testFormula.append(" and ");
            } else if ("|".equals(etape)) {
                testFormula.append(" or ");
            } else if ("$".equals(etape)) {
                if (stringTokenizer.nextToken().charAt(0) == '0') {
                    testFormula.append("float(").append(parameters[0]).append(")");
                } else {
                    testFormula.append("float(").append(parameters[1]).append(")");
                }
            } else {
                testFormula.append(etape);
            }
        }

        println("");

        println("while (" + testFormula + "):", 1);

        List<GraphItemModel> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());
        generateSubroutineBody(litems, subgraphId, 2);

        // Get intermediate images
        VirtualItem virtualItem = subGraph.getVirtualItem();
        for (int i = 0; i < item.getInputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
            }
            String input = createInputImageName1(isi, subgraphId);
            String output = createOutputImageName1(subgraphId, 0, i);
            if (!input.equals(output)) {
                println("shutil.copy(\"" + input + "\",\"" + output + "\")", 2);
            }
        }

        // Get intermediate parameters
        for (int i = 0; i < 2; i++) {
            ParameterSlot isp = virtualItem.getVirtualIntermediateResultSlot(i);
            if (isp.getConnectedSlot() != null) {
                println(parameters[i] + " = " + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()), 2);
            }
        }

        // Get output images.
        println("");

        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
            println("link(\"" + createInputImageName1(isi, subgraphId) + "\",args[argc]); argc += 1", 1);
        }

        // Output result
        ParameterSlot isp = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
        if (isp.getConnectedSlot() == null) {
            println("return 0;", 1);
        } else {
            println("return " + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + ";", 1);
        }
        println("");

        generateSubroutinesDefinition(litems);
        println("}");
    }

    @Override
    protected void generateSubGraph( MacroOperatorItem item ) throws ArianeBadGraphException {
        SubGraphPresenter subGraph = item.getSubGraph();

        generateComment(item, subGraph.getListOfGraphItems());

        print("Errc macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "( ");
        int subgraphId = subGraph.getIdentifier();

        // Get input images.
        int rank = 0;
        for (InputDataSlot isi : item.getInputDataSlots()) {
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error.incomplete"));
            }
            print("Img2duc &" + createOutputImageName1(subgraphId, 0, rank++) + ", ");
        }

        // Get output images.
        VirtualItem virtualItem = subGraph.getVirtualItem();
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
            print("Img2duc &" + createInputImageName1(isi, subgraphId) + ", ");
        }

        // Get parameters.
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            print("float " + createVariableName(subgraphId, 0, i) + ", ");
        }

        println(" ) {");

        List<GraphItemModel> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());
        generateSubroutineBody(litems, subgraphId, 1);

        println("");

        // Output result.
        ParameterSlot isp = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
        if (isp.getConnectedSlot() == null) {
            println("return 0;", 1);
        } else {
            println("return " + createVariableName(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + ";", 1);
        }
        generateSubroutinesDefinition(litems);
        println("}");
    }

    private void generateComment( GraphItemModel item, List<GraphItemModel> litems ) {
        println("\n/*\n * " + item.getNickName() + "\n *");
        for (GraphItemModel note : litems) {
            if (note instanceof NoteItem) {
                println(" * " + note.getFormula());
            }
        }
        println(" */");
    }

    private String createVariableName( int graphId, int itemId, int slotRank ) {
        return "var" + graphId + "_" + itemId + "_" + slotRank;
    }

    private String getActualParameterValue( GraphItemModel item, int slotRank, int graphId ) {
        ParameterSlot isp = (ParameterSlot) item.getSlot("InSlotParameter", slotRank);
        if (isp.getConnectedSlot() == null) {
            return "\"" + ((ParameterSlot) item.getSlot("InSlotParameter", slotRank)).getValue() + "\"";
        } else {
            return createVariableName(graphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank());
        }
    }

    private String convertCommentIntoPython( String comment ) {
        if (comment != null) {
            return comment.replaceAll("\n", "\n   ");
        }
        return "";
    }

    private String convertArithmeticFormulaIntoPython( String formula, String[] parameters ) {
        StringTokenizer stringTokenizer = new StringTokenizer(formula, "$ -+()*/^0123456789.E", true);
        String etape;
        StringBuilder code = new StringBuilder();
        while (stringTokenizer.hasMoreTokens()) {
            etape = stringTokenizer.nextToken();
            if ("^".equals(etape)) {
                code.append("**");
            } else if ("ln".equals(etape)) {
                code.append("log");
            } else if ("atan".equals(etape)) {
                code.append("a");
            } else if ("$".equals(etape)) {
                int nnumberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
                code.append("float(").append(parameters[nnumberOfParameters]).append(")");
            } else {
                code.append(etape);
            }
        }
        return code.toString();
    }
}
