/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.generator;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeBadGraphException;
import fr.greyc.pantheon.ariane.model.graph.element.CompositeItem;
import fr.greyc.pantheon.ariane.model.graph.element.ConstantItem;
import fr.greyc.pantheon.ariane.model.graph.element.ForControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.IfControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.MergeControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.SelectControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.element.EvalItem;
import fr.greyc.pantheon.ariane.model.graph.element.FileNameItem;
import fr.greyc.pantheon.ariane.model.graph.element.FormatItem;
import fr.greyc.pantheon.ariane.model.graph.element.GraphItemModel;
import fr.greyc.pantheon.ariane.model.graph.element.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.element.MacroOperatorItem;
import fr.greyc.pantheon.ariane.model.graph.element.OperatorItem;
import fr.greyc.pantheon.ariane.model.graph.element.SaverItem;
import fr.greyc.pantheon.ariane.model.graph.element.ScriptItem;
import fr.greyc.pantheon.ariane.model.graph.element.SystemItem;
import fr.greyc.pantheon.ariane.model.graph.element.ViewerItem;
import fr.greyc.pantheon.ariane.model.graph.element.VirtualItem;
import fr.greyc.pantheon.ariane.model.graph.element.WhileControlStructureItem;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.presenter.SubGraphPresenter;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * The <code>MsdosGenerator</code> generates MsDos code for the specified graph.
 * It operates in three steps:<ol>
 * <li>transforms the graph into a list of operators usable for the topological sort;
 * <li>executes the topological sort;
 * <li>generates the related code.
 * </ol>
 */
public final class MsdosGenerator extends CodeGenerator {
    private final List<String> _outputImages = new ArrayList<>();

    protected MsdosGenerator() {
        // Only created by the code generator.
    }

    @Override
    protected void generateScriptHeader( String filename, List<GraphItemModel> arguments, List<GraphItemModel> inputs, List<GraphItemModel> outputs ) {
        println("@echo off");
        println("rem @(#)" + filename + "				" + getCurrentDate());
        println("rem");
        println("rem This file was automatically generated by Ariane.");
        println("rem Ariane is part of the Pantheon Project.");
        println("rem");
        println("rem GREYC - ENSICAEN & University of Caen Basse-Normandie, France");
        println("");

        println("rem Check argument line consistency.");
        println("if \"%1\"==\"-h\" goto USAGE");
        for (int i = 0; i < arguments.size() + inputs.size() + outputs.size(); i++) {
            println("if \"%1\"==\"\" goto USAGE");
        }
        println("goto SCRIPT");

        print("@echo \"Usage: $0 ");
        for (GraphItemModel argument : arguments) {
            print(" " + argument.getNickName());
        }
        for (int i = 0; i < inputs.size(); i++) {
            print(" input-file" + i);
        }
        for (int i = 0; i < outputs.size(); i++) {
            print(" output-file" + i);
        }
        println("\"");

        print("@echo \"  example: $0");
        for (GraphItemModel argument : arguments) {
            if (argument.getFormula() != null && !argument.getFormula().isEmpty()) {
                print(" " + argument.getFormula());
            } else {
                print(" " + argument.getNickName());
            }
        }
        for (GraphItemModel input : inputs) {
            print(" " + input.getNickName());
        }
        for (GraphItemModel output : outputs) {
            String name = output.getNickName();
            if ("saver".equals(name)) {
                name += output.getId();
            }
            print(" " + name);
        }
        println("\"");

        println("goto :END");
        println("\n:SCRIPT", 0);
    }

    @Override
    protected void generateScriptFooter() {
        println(":END");
    }

    @Override
    protected void generateUtilitySubroutinesDefinition() {
        // Do nothing
    }

    private String createImageName( int slotIndex, int itemIndex, int graphId ) {
        String name = UserPreferences.createTemporaryName(graphId, itemIndex, slotIndex);
        _outputImages.add(name);
        return name;
    }

    private String getInputImage( InputSlot slot, int graphId ) {
        return UserPreferences.createTemporaryName(graphId, slot.getConnectedSlot().getGraphItem().getId(), slot.getConnectedSlot().getRank());
    }

    private String variable( int graph, int item, int slot ) {
        return "var" + graph + "_" + item + "_" + slot;
    }

    private String getParameter( GraphItemModel item, int slotIndex, int graphId ) {
        ParameterSlot isp = (ParameterSlot) item.getSlot("InSlotParameter", slotIndex);
        if (isp.getConnectedSlot() == null) {
            return ((ParameterSlot) item.getSlot("InSlotParameter", slotIndex)).getValue();
        } else {
            return "%" + variable(graphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "%";
        }
    }

    private String getNumericalParameter( GraphItemModel item, int slotIndex, int graphId ) {
        ParameterSlot isp = (ParameterSlot) item.getSlot("InSlotParameter", slotIndex);
        if (isp.getConnectedSlot() == null) {
            return ((ParameterSlot) item.getSlot("InSlotParameter", slotIndex)).getValue();
        } else {
            return "%" + variable(graphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "%";
        }
    }

    private String translateLogicalFormula( String formula, String parameter0, String parameter1 ) {
        // Parse the boolean expression into Shell.
        char[] formulaCharacters = formula.toCharArray();
        StringBuilder commandShell = new StringBuilder();
        int index = 0;
        while (index < formulaCharacters.length) {
            if (formulaCharacters[index] == '&') {
                commandShell.append(" && ");
                index++;
            } else if (formulaCharacters[index] == '|') {
                commandShell.append(" || ");
                index++;
            } else if (formulaCharacters[index] == '<') {
                if (formulaCharacters[index + 1] == '=') {
                    commandShell.append(" <= ");
                    index += 2;
                } else {
                    commandShell.append(" < ");
                    index++;
                }
            } else if (formulaCharacters[index] == '>') {
                if (formulaCharacters[index + 1] == '=') {
                    commandShell.append(" >= ");
                    index += 2;
                } else {
                    commandShell.append(" > ");
                    index++;
                }
            } else if (formulaCharacters[index] == '=') {
                if (formulaCharacters[index + 1] == '=') {
                    commandShell.append(" == ");
                    index += 2;
                } else {
                    index++; // degenerated case!
                }
            } else if (formulaCharacters[index] == '!') {
                if (formulaCharacters[index + 1] == '=') {
                    commandShell.append(" != ");
                    index += 2;
                } else {
                    commandShell.append(" ! ");
                    index++;
                }
            } else if (formulaCharacters[index] == '(') {
                commandShell.append(" ");
                index++;
            } else if (formulaCharacters[index] == ')') {
                commandShell.append(" ");
                index++;
            } else if (formulaCharacters[index] == '$') {
                if (formulaCharacters[index + 1] == '0') {
                    commandShell.append(parameter0);
                } else {
                    commandShell.append(parameter1);
                }
                index += 2;
            } else {
                commandShell.append(formulaCharacters[index++]);
            }
        }
        return commandShell.toString();
    }

    @Override
    protected void generateMainPrologue( List<GraphItemModel> litems, List<GraphItemModel> arguments, List<GraphItemModel> inputs, List<GraphItemModel> outputs ) {
        println("\nrem\nrem The main program\n");
        println("del " + UserPreferences.getTemporaryFilePrefix() + "?*" + UserPreferences.getTemporaryFileExtension(), 1);
        for (GraphItemModel item : arguments) {
            println("set " + variable(1, item.getId(), 0) + "=%1", 1);
            println("shift", 1);
        }
        for (GraphItemModel item : inputs) {
            println("copy %1 " + createImageName(0, item.getId(), 1), 1);
            println("shift", 1);
        }
    }

    @Override
    protected void generateMainEpilogue( List<GraphItemModel> arguments, List<GraphItemModel> inputs, List<GraphItemModel> outputs ) {
    }

    @Override
    protected void generateItem( LoaderItem item, int graphId, int indent ) {
        if (item.getInputParameterSlots().size() > 0 && item.getInputParameterSlots().get(0).isConnected()) {
            String parameters = getParameter(item, 0, graphId);
            println("copy " + parameters + " " + createImageName(0, item.getId(), graphId), indent);
        }
    }

    @Override
    protected void generateItem( SaverItem item, int graphId, int indent ) {
        if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
            if (item.getInputParameterSlots().size() > 0 && item.getInputParameterSlots().get(0).isConnected()) {
                String parameters = getParameter(item, 0, graphId);
                println("copy " + getInputImage(item.getInputDataSlots().get(0), graphId) + " " + parameters, indent);
            } else {
                println("copy " + getInputImage(item.getInputDataSlots().get(0), graphId) + " %1", indent);
                println("shift", indent);
            }
            println("copy "
                    + getInputImage(item.getInputDataSlots().get(0), graphId)
                    + " "
                    + createImageName(0, item.getId(), graphId), indent);
        }
    }

    @Override
    protected void generateItem( ViewerItem item, int graphId, int indent ) {
        String commandShell = item.getViewerCommandName();
        if (item.getInputDataSlots().get(0).getConnectedSlot() != null) {
            println(commandShell + " " + getInputImage(item.getInputDataSlots().get(0), graphId), indent);
            println("copy "
                    + getInputImage(item.getInputDataSlots().get(0), graphId)
                    + " "
                    + createImageName(0, item.getId(), graphId), indent);
        }
    }

    @Override
    protected void generateItem( OperatorItem item, int graphId, int indent ) throws ArianeBadGraphException {
        String codeName = OperatorItem.getCommandPart(item.getBinary());
        if (codeName == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error2") + item.getName());
        }

        // Get parameters
        List<String> parameterLine = new LinkedList<>();
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameterLine.add(getParameter(item, i, graphId));
        }

        // Get mask
        List<String> maskLine = new LinkedList<>();
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, graphId))) {
            maskLine.add(getInputImage(mask, graphId));
        }

        // Get input images.
        List<String> inputLine = new LinkedList<>();
        for (InputDataSlot image : item.getInputDataSlots()) {
            if (image.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            inputLine.add(getInputImage(image, graphId));
        }

        // Get output images.
        List<String> outputLine = new LinkedList<>();
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            outputLine.add(createImageName(i, item.getId(), graphId));
        }

        StringBuilder code = new StringBuilder();
        List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
        for (String command : commandLine) {
            code.append(command).append(" ");
        }

        // Get output result.
        println(code.toString(), indent);
        println("call " + UserPreferences.getCommandStatus(), indent);
        println("call pset " + variable(graphId, item.getId(), 0), indent);
    }

    @Override
    protected void generateItem( SystemItem item, int graphId, int indent ) throws ArianeBadGraphException {
        String codeName = OperatorItem.getCommandPart(item.getBinary());
        if (codeName == null) {
            throw new ArianeBadGraphException(ArianeSettings.getString("Generator.error2") + item.getName());
        }

        // Get parameters
        LinkedList<String> parameterLine = new LinkedList<>();
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameterLine.add(getParameter(item, i, graphId));
        }

        // Get mask
        LinkedList<String> maskLine = new LinkedList<>();
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, graphId))) {
            maskLine.add(getInputImage(mask, graphId));
        }

        // Get input images.
        LinkedList<String> inputLine = new LinkedList<>();
        for (InputDataSlot image : item.getInputDataSlots()) {
            if (image.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            inputLine.add(getInputImage(image, graphId));
        }

        // Get output images.
        LinkedList<String> outputLine = new LinkedList<>();
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            outputLine.add(createImageName(i, item.getId(), graphId));
        }

        StringBuilder code = new StringBuilder();
        List<String> commandLine = item.buildAbstractCommandLine(maskLine, parameterLine, inputLine, outputLine);
        for (String command : commandLine) {
            code.append(command).append(" ");
        }

        // Get output result.
        println(code.toString(), indent);
        println("call " + UserPreferences.getCommandStatus(), indent);
        println("call pset " + variable(graphId, item.getId(), 0), indent);
    }

    private String parseFormula( String formula, String[] parameters ) {
        StringBuilder code = new StringBuilder();
        StringTokenizer stringTokenizer = new StringTokenizer(formula, "$ -+()*/^0123456789.E", true);
        String etape;
        while (stringTokenizer.hasMoreTokens()) {
            etape = stringTokenizer.nextToken();
            if ("e".equals(etape) || "E".equals(etape)) {
                code.append("*10^");
            } else if ("cos".equals(etape)) {
                code.append("c");
            } else if ("sin".equals(etape)) {
                code.append("s");
            } else if ("ln".equals(etape)) {
                code.append("l");
            } else if ("exp".equals(etape)) {
                code.append("e");
            } else if ("atan".equals(etape)) {
                code.append("a");
            } else if ("$".equals(etape)) {
                int numberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
                code.append(parameters[numberOfParameters]);
            } else {
                code.append(etape);
            }
        }
        return code.toString();
    }

    private String setFunction( String formula ) {
        StringBuilder code = new StringBuilder();
        StringTokenizer stringTokenizer = new StringTokenizer(formula, "$ -+()*/^0123456789.E", true);
        String etape;
        boolean abs = false;
        boolean max = false;
        boolean min = false;
        boolean tan = false;
        boolean log2 = false;
        boolean log10 = false;
        boolean acos = false;
        boolean asin = false;

        while (stringTokenizer.hasMoreTokens()) {
            etape = stringTokenizer.nextToken();
            if ("abs".equals(etape) && !abs) {
                code.append(" define abs(x){; if (x>=0) return(x) else return (-x)} ");
                abs = true;
            } else if ("max".equals(etape) && !max) {
                code.append(" define max(x,y){; if (x>y) return(x) else return(y)} ");
                max = true;
            } else if ("min".equals(etape) && !min) {
                code.append(" define min(x,y){; if (x<y) return(x) else return(y)} ");
                min = true;
            } else if ("tan".equals(etape) && !tan) {
                code.append(" define tan(x){; return (s(x)/c(x)} ");
                tan = true;
            } else if ("acos".equals(etape) && !acos) {
                code.append(" define acos(x){; if (x==1) return(0) if (x==-1) return (4*a(1)) return (2*a(1)-a(x/sqrt(1-(x^2))))} ");
                acos = true;
            } else if ("asin".equals(etape) && !asin) {
                code.append(" define asin(x){; if (x==1) return(2*a(1)) if (x==-1) return (-2*a(1)) return (a(x/sqrt(1-(x^2))))} ");
                asin = true;
            } else if ("log".equals(etape)) {
                etape = stringTokenizer.nextToken();
                if ("2".equals(etape) && !log2) {
                    code.append(" define log2(x){; return l(x)/l(2)} ");
                    log2 = true;
                } else if ("1".equals(etape) && !log10) {
                    code.append(" define log10(x){; return l(x)/l(10)} ");
                    log10 = true;
                }
            }
        }
        return code.toString();
    }

    @Override
    protected void generateItem( EvalItem item, int graphId, int indent ) {
        String[] parameters = new String[item.getInputParameterSlots().size()];
        // Get parameters
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameters[i] = getParameter(item, i, graphId);
        }

        println("set " + variable(graphId, item.getId(), 0) + "=" + "$(echo \""
                + setFunction(item.getFormula())
                + parseFormula(item.getFormula(), parameters)
                + "\"|bc -l);", indent);
    }

    @Override
    protected void generateItem( FormatItem item, int graphId, int indent ) {
        String[] parameters = new String[item.getInputParameterSlots().size()];
        // Get parameters
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameters[i] = getParameter(item, i, graphId);
        }

        StringBuilder code = new StringBuilder();
        StringTokenizer stringTokenizer = new StringTokenizer(item.getFormula(), " $()+'\"", true);
        String etape;
        while (stringTokenizer.hasMoreTokens()) {
            etape = stringTokenizer.nextToken();
            if ("+".equals(etape)) {
                code.append("");
            } else if ("quote".equals(etape)) {
                code.append("");
            } else if ("\"".equals(etape)) {
                code.append("");
            } else if ("(".equals(etape)) {
                code.append("'");
            } else if (")".equals(etape)) {
                code.append("'");
            } else if ("$".equals(etape)) {
                int numberOfParameters = Integer.parseInt(stringTokenizer.nextToken());
                code.append(parameters[numberOfParameters]);
            } else {
                code.append(etape);
            }
        }
        println(variable(graphId, item.getId(), 0) + "=\"" + code + "\"", indent);
    }

    @Override
    protected void generateItem( SelectControlStructureItem item, int graphId, int indent ) throws ArianeBadGraphException {
        String value = getParameter(item, 0, graphId);
        println("rem Select", indent);
        if (value == null) {
            value = variable(graphId, item.getId(), 0);
            println("read " + value, indent);
            value = "%" + value + "%";
        }
        for (int i = 0; i < item.getInputDataSlots().size(); i++) {
            println("if [ " + value + " -eq " + i + " ]", indent);
            println("then", indent);
            println("copy "
                    + getInputImage(item.getInputDataSlots().get(i), graphId) + " "
                    + createImageName(0, item.getId(), graphId), indent + 1);
            println("set " + variable(graphId, item.getId(), 0) + "=" + i + ";", indent + 1);
            println("fi", indent);
        }
    }

    @Override
    protected void generateItem( IfControlStructureItem item, int graphId, int indent ) {
        String[] parameters = new String[2];
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            parameters[i] = getNumericalParameter(item, i, graphId);
        }
        String commandShell = translateLogicalFormula(item.getFormula(), parameters[0], parameters[1]);
        println("if [ $( echo \"" + commandShell + "\" |bc -l) -gt 0 ]", indent);
    }

    @Override
    protected void generateThenPart( IfControlStructureItem item, int graphId, int indent ) {
        println("then", indent);
        print("copy " + getInputImage(item.getInputDataSlots().get(0), graphId) + " "
                + createImageName(0, item.getId(), graphId), indent + 1);
        println("; " + variable(graphId, item.getId(), 0) + "=0");
    }

    @Override
    protected void generateElsePart( IfControlStructureItem item, int graphId, int indent ) {
        println("else", indent);
        print("copy " + getInputImage(item.getInputDataSlots().get(0), graphId) + " "
                + createImageName(1, item.getId(), graphId), indent + 1);
        println("; " + variable(graphId, item.getId(), 0) + "=1");
    }

    @Override
    protected void generateFiPart( IfControlStructureItem item, int graphId, int indent ) {
        println("fi", indent);
    }

    @Override
    protected void generateItem( MergeControlStructureItem item, int graphId, int indent ) {
        // input image
        InputDataSlot in0;
        in0 = item.getInputDataSlots().get(0);
        int id0;
        id0 = in0.getConnectedSlot().getGraphItem().getId();
        print("if [ -n \"%" + variable(graphId, id0, 0) + "%\" -a ", indent);
        println("-r " + getInputImage(item.getInputDataSlots().get(0), graphId) + " ]");
        println("then", indent);
        print("copy "
                + getInputImage(item.getInputDataSlots().get(0), graphId)
                + " "
                + createImageName(0, item.getId(), graphId), indent + 1);
        println("; " + variable(graphId, item.getId(), 0) + "=0;");
        println("del " + getInputImage(item.getInputDataSlots().get(0), graphId), indent + 1);
        println("else", indent);
        print("copy "
                + getInputImage(item.getInputDataSlots().get(1), graphId) + " "
                + createImageName(0, item.getId(), graphId), indent + 1);
        println("; " + variable(graphId, item.getId(), 0) + "=1;");
        println("del " + getInputImage(item.getInputDataSlots().get(1), graphId), indent + 1);
        println("fi", indent);
    }

    @Override
    protected void generateItem( FileNameItem item, int graphId, int indent ) {
        if (!item.isArgumentOfRelatedScript()) {
            println("set " + variable(graphId, item.getId(), 0) + "='" + item.getFormula() + "'", indent);
        }
    }

    @Override
    protected void generateItem( ConstantItem item, int graphId, int indent ) {
        if (!item.isArgumentOfRelatedScript()) {
            String value = item.getFormula().replaceAll("\"", "\\\\\"");
            println("set " + variable(graphId, item.getId(), 0) + "=\"" + value + "\"", indent);
        }
    }

    @Override
    protected void generateItem( ScriptItem item, int graphId, int indent ) {
        String value = item.getFormula().replaceAll("\"", "\\\\\"").replaceAll("\n", "\\\\\n");
        println("set " + variable(graphId, item.getId(), 0) + "=\"" + value + "\"", indent);
    }

    @Override
    protected void generateItem( CompositeItem item, int graphId, int indent ) throws ArianeBadGraphException {
        StringBuilder code = new StringBuilder("macro" + item.getId() + "_" + item.getGraph().getIdentifier());

        // Get parameters
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            code.append(" ").append(getParameter(item, i, graphId));
        }

        // Get mask
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, graphId))) {
            code.append(" ").append(getInputImage(mask, graphId));
        }

        // Get input images.
        for (InputDataSlot image : item.getInputDataSlots()) {
            if (image.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            code.append(" ").append(getInputImage(image, graphId));
        }

        // Get output images.
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            code.append(" ").append(createImageName(i, item.getId(), graphId));
        }
        // Get output result.
        println(code.toString());
        println("call " + UserPreferences.getCommandStatus(), indent);
        println("call pset " + variable(graphId, item.getId(), 0), indent);
    }

    /**
     * Parses the comment in order to add a sharp after each newlines.
     *
     * @param comment a string that contains the comment.
     * @return a new string.
     */
    private String parseComment( String comment ) {
        return comment.replaceAll("\n", "\nrem ");
    }

    private void generateComment( GraphItemModel item, List<GraphItemModel> litems ) {
        println("\nrem\nrem " + item.getNickName());
        for (GraphItemModel note : litems) {
            println("rem " + parseComment(note.getFormula()));
        }
        println("rem");
    }

    @Override
    protected void generateSubGraph( ForControlStructureItem item ) throws ArianeBadGraphException {
        SubGraphPresenter subGraph = item.getSubGraph();
        List<GraphItemModel> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());

        // Get and discard the virtual operator
        GraphItemModel virtualOperator = getVirtualOperator(litems);
        litems.remove(virtualOperator);
        if (litems.isEmpty()) {
            throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
        }

        generateComment(item, subGraph.getListOfGraphItems());

        println("function macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "() {");

        // Get parameters
        int subgraphId = subGraph.getIdentifier();
        println("# Parameters", 1);
        println("begin=%1; shift", 1);
        println("end=%1; shift", 1);
        println("step=%1; shift", 1);
        for (int i = 3; i < item.getInputParameterSlots().size(); i++) {
            println("set " + variable(subgraphId, 0, i - 2) + "=%1; shift", 1);
        }

        // Get mask
        println("# Input data files", 1);
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
            int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
            println("ln -fs %1 " + createImageName(maskIndex, 0, subgraphId) + "; shift", 1);
        }

        // Get input images
        List<InputDataSlot> inputImages = item.getInputDataSlots();
        for (int i = 0; i < inputImages.size(); i++) {
            InputDataSlot isi = inputImages.get(i);
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            println("copy %1 " + createImageName(i, 0, subgraphId) + "; shift", 1);
        }
        println("");
        // Create the output images in case of no iteration
        VirtualItem virtualItem = (VirtualItem) virtualOperator;
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
            println("touch " + getInputImage(isi, subgraphId), 1);
        }

        // Initialize the return variable in case of no iteration
        {
            ParameterSlot isp = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
            println(variable(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "=SUCCESS", 1);
        }

        println("");
        println("local i", 1);
        println("for ((i=%begin%; i<=%end%; i+=%step% ))", 1);
        println("do", 1);
        println(variable(subgraphId, 0, 0) + "=%i;", 2);
        generateSubroutineBody(litems, subgraphId, 2);

        // Set intermediate images
        for (int i = 0; i < item.getInputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            String input = getInputImage(isi, subgraphId);
            String output = createImageName(i, 0, subgraphId);
            if (!input.equals(output)) {
                println("copy " + input + " " + output, 2);
            }
        }

        println("done", 1);

        println("");
        // Get output images
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
            println("copy " + getInputImage(isi, subgraphId) + " %1", 1);
            println("shift", 1);
        }
        // Output result
        ParameterSlot isp = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
        if (isp.getConnectedSlot() == null) {
            println("psetstatus SUCCESS", 1);
        } else {
            println("psetstatus %" + variable(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "%", 1);
        }
        println("}");

        generateSubroutinesDefinition(litems);
    }

    @Override
    protected void generateSubGraph( WhileControlStructureItem item ) throws ArianeBadGraphException {
        SubGraphPresenter subGraph = item.getSubGraph();
        List<GraphItemModel> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());

        // Get and discard the virtual operator
        GraphItemModel virtualOperator = getVirtualOperator(litems);
        litems.remove(virtualOperator);
        if (litems.isEmpty()) {
            throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
        }

        generateComment(item, subGraph.getListOfGraphItems());

        println("function macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "() {");

        // Get parameters
        int subgraphId = subGraph.getIdentifier();
        println("rem Parameters", 1);
        String[] parameters = new String[item.getInputParameterSlots().size()];
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            println("set " + variable(subgraphId, 0, i) + "=%1", 1);
            println("shift", 1);
            parameters[i] = "%" + variable(subgraphId, 0, i) + "%";
        }

        // Get mask.
        println("rem Input data files", 1);
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
            int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
            println("copy %1 " + createImageName(maskIndex, 0, subgraphId), 1);
            println("shift", 1);
        }

        // Get input images.
        List<InputDataSlot> inputImages = item.getInputDataSlots();
        for (int i = 0; i < inputImages.size(); i++) {
            InputDataSlot isi = inputImages.get(i);
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            println("copy %1 " + createImageName(i, 0, subgraphId), 1);
            println("shift", 1);
        }

        // Parse the boolean expression.
        String testFormula = translateLogicalFormula(item.getFormula(), parameters[0], parameters[1]);
        VirtualItem virtualItem = (VirtualItem) virtualOperator;

        println("");
        // Create output images in case of no repetition
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
            println("touch " + getInputImage(isi, subgraphId), 1);
        }

        // Initialize the return variable in case of no iteration
        {
            ParameterSlot isp = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
            println(variable(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "=SUCCESS", 1);
        }

        println("");
        println("while [ $( echo \"" + testFormula + "\" |bc -l) -gt 0 ]", 1);
        println("do", 1);

        generateSubroutineBody(litems, subgraphId, 2);

        // Get intermediate images
        for (int i = 0; i < item.getInputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualIntermediateDataSlots(i);
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            // Discard cp image image (it happens with the output = input)
            String input = getInputImage(isi, subgraphId);
            String output = createImageName(i, 0, subgraphId);
            if (!input.equals(output)) {
                println("copy " + input + " " + output, 2);
            }
        }

        // Get intermediate parameters
        for (int i = 0; i < 2; i++) {
            ParameterSlot isp = virtualItem.getVirtualIntermediateResultSlot(i);
            if (isp.getConnectedSlot() != null) {
                println(variable(subgraphId, 0, i) + "=%" + variable(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "%", 2);
            }
        }
        println("done", 1);
        // Get output images
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
            println("copy " + getInputImage(isi, subgraphId) + " %1", 1);
            println("shift", 1);
        }

        // Output result.
        ParameterSlot isp = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
        if (isp.getConnectedSlot() == null) {
            println("psetstatus SUCCESS", 1);
        } else {
            println("psetstatus %" + variable(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "%", 1);
        }
        println("}");

        generateSubroutinesDefinition(litems);
    }

    @Override
    protected void generateSubGraph( MacroOperatorItem item ) throws ArianeBadGraphException {
        SubGraphPresenter subGraph = item.getSubGraph();
        List<GraphItemModel> litems = buildOrderedItemsSequenceUsingTopologicalSort(subGraph.getListOfGraphItems());

        // Get and discard the virtual operator.
        GraphItemModel virtualOperator = getVirtualOperator(litems);
        litems.remove(virtualOperator);
        if (litems.isEmpty()) {
            throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
        }

        generateComment(item, subGraph.getListOfGraphItems());

        println("function macro" + item.getId() + "_" + item.getGraph().getIdentifier() + "() {");

        // Get parameters.
        int subgraphId = subGraph.getIdentifier();
        println("rem arguments", 1);
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            println("set " + variable(subgraphId, 0, i) + "=%1", 1);
            println("shift", 1);
        }

        // Get mask.
        println("rem Input data files", 1);
        InputMaskSlot mask = item.getInputMaskSlot();
        if (mask != null && mask.isConnected() && _outputImages.contains(getInputImage(mask, item.getGraph().getIdentifier()))) {
            int maskIndex = subGraph.getVirtualItem().getVirtualInputMaskSlot().getRank();
            println("copy %1 " + createImageName(maskIndex, 0, subgraphId), 1);
            println("shift", 1);
        }

        // Get input images.
        List<InputDataSlot> inputImages = item.getInputDataSlots();
        for (int i = 0; i < inputImages.size(); i++) {
            InputDataSlot isi = inputImages.get(i);
            // Unlinked input!
            if (isi.getConnectedSlot() == null) {
                throw new ArianeBadGraphException(item.getNickName() + ": " + ArianeSettings.getString("Generator.error.incomplete"));
            }
            println("copy %1 " + createImageName(i, 0, subgraphId), 1);
            println("shift", 1);
        }

        println("");

        generateSubroutineBody(litems, subgraphId, 1);

        VirtualItem virtualItem = (VirtualItem) virtualOperator;
        // Get output images.
        for (int i = 0; i < item.getOutputDataSlots().size(); i++) {
            InputDataSlot isi = virtualItem.getVirtualOutputDataSlots().get(i);
            println("copy " + getInputImage(isi, subgraphId) + " %1", 1);
            println("shift", 1);
        }

        println("");
        // Output result.
        ParameterSlot isp = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
        if (isp.getConnectedSlot() == null) {
            println("psetstatus SUCCESS", 1);
        } else {
            println("psetstatus %" + variable(subgraphId, isp.getConnectedSlot().getGraphItem().getId(), isp.getConnectedSlot().getRank()) + "%", 1);
        }
        println("}");

        generateSubroutinesDefinition(litems);
    }
}
