/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph;

import fr.greyc.pantheon.ariane.model.graph.element.*;
import fr.greyc.pantheon.ariane.model.graph.slot.SlotModel;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.model.xml.XMLFileOutputStream;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

/**
 * A graph is an executable program
 * represented as an oriented graph of items.
 */
public abstract class GraphModel {
    private static int _numberOfCreatedGraphs;
    private final int _identifier;
    private int _numberOfCreatedItemsInThisGraph;
    private File _storageFile;
    private final List<GraphItemModel> _graphItems = new LinkedList<>();

    GraphModel() {
        _identifier = ++_numberOfCreatedGraphs;
    }

    GraphModel( File absoluteFile ) {
        this();
        _storageFile = absoluteFile;
    }

    public int getIdentifier() {
        return _identifier;
    }

    public int getNextItemIdentifier() {
        return ++_numberOfCreatedItemsInThisGraph;
    }

    public void dispose() {
        for (GraphItemModel item : _graphItems) {
            item.dispose();
        }
        _graphItems.clear();
    }

    public List<GraphItemModel> getListOfGraphItems() {
        return _graphItems;
    }

    public File getStorageFile() {
        return _storageFile;
    }

    public void add( GraphItemModel item ) {
        _graphItems.add(item);
    }

    public void remove( GraphItemModel item ) {
        _graphItems.remove(item);
    }

    public void resetAll() {
        for (GraphItemModel item : _graphItems) {
            item.resetAll();
        }
    }

    public void reset() {
        for (GraphItemModel item : _graphItems) {
            item.reset();
        }
    }

    public GraphItemModel getGraphItem( int identifier ) {
        for (GraphItemModel item : _graphItems) {
            if (item.isInSameReadPasteSession() && item.getInternalId() == identifier) {
                return item;
            }
        }
        return null;
    }

    public void save( File file ) throws IOException {
        XMLFileOutputStream out = new XMLFileOutputStream(file, null);
        out.write(convertToXMLTree(file));
        out.close();
        _storageFile = file;
    }

    public XMLElement convertToXMLTree( File file ) throws IOException {
        XMLElement element;
        if (isRoot()) {
            element = new XMLElement("Graph");
        } else {
            element = new XMLElement("SubGraph");
        }
        XMLElement items = new XMLElement("ItemList");
        for (GraphItemModel item : _graphItems) {
            items.addChildren(item.convertToXML(file));
        }
        XMLElement connections = new XMLElement("ConnectionList");
        for (GraphItemModel item : _graphItems) {
            connections.addChildren(item.convertSlotsToXML());
        }
        element.addChildren(items);
        element.addChildren(connections);
        return element;
    }

    protected abstract boolean isRoot();

    public void createFromXMLTree( List<XMLElement> xmlTree, GraphPresenter graph ) throws IOException {
        List<XMLElement> xmlSubTree = null;
        boolean found = false;
        if (isRoot()) {
            xmlSubTree = xmlTree.get(0).getChildren();
            String tagName = xmlSubTree.get(0).getName();
            if ("Graph".equalsIgnoreCase(tagName) || "Graphe".equalsIgnoreCase(tagName)) {
                xmlSubTree = xmlSubTree.get(0).getChildren();
                found = true;
            }
        } else {
            for (XMLElement element : xmlTree) {
                String tagName = element.getName();
                if ("SubGraph".equalsIgnoreCase(tagName) || "SousGraphe".equalsIgnoreCase(tagName)) {
                    xmlSubTree = element.getChildren();
                    found = true;
                    break;
                }
            }
        }

        if (!found || xmlSubTree == null) {
            throw new IOException("Incorrect XML file.");
        }
        for (XMLElement element : xmlSubTree) {
            String tagName = element.getName();
            if ("ItemList".equalsIgnoreCase(tagName) || "ListeElements".equalsIgnoreCase(tagName)) {
                GraphModel.createFromXML(graph, element.getChildren());
            } else if ("ConnectionList".equalsIgnoreCase(tagName) || "ListeConnexions".equalsIgnoreCase(tagName)) {
                SlotModel.createFromXML(this, element.getChildren());
            }
        }
    }

    public void synchronizeItemFilesWithSystemFiles() {
        for (GraphItemModel item : _graphItems) {
            if (item instanceof LoaderItem) {
                ((LoaderItem) item).synchronizeWithFileSystem();
            } else if (item instanceof FileNameItem) {
                ((FileNameItem) item).synchronizeWithFileSystem();
            }
        }
    }

    public static List<GraphItemModel> createFromXML( GraphPresenter graphPresenter, List<XMLElement> xmlTree ) throws IOException {
        List<GraphItemModel> items = new ArrayList<>();
        if (xmlTree == null) {
            return items;
        }

        for (XMLElement element : xmlTree) {
            GraphItemModel item = null;
            String tagName = element.getName();
            if (OperatorItem.TAGNAME.equalsIgnoreCase(tagName) || "Operateur".equalsIgnoreCase(tagName)) {
//                item = ElementFactory().createOperator(graphPresenter, element.getChildren());
                item = new OperatorItem(graphPresenter, element.getChildren());
            } else if (NoteItem.TAGNAME.equalsIgnoreCase(tagName)) {
                item = new NoteItem(graphPresenter, element.getChildren());
            } else if ("Fichier".equalsIgnoreCase(tagName)) {
                // Kept for the sake of compatibility
                item = DataFile.getInstance(graphPresenter, element.getChildren());
            } else if (LoaderItem.TAGNAME.equalsIgnoreCase(tagName)) {
                item = new LoaderItem(graphPresenter, element.getChildren());
            } else if (SaverItem.TAGNAME.equalsIgnoreCase(tagName)) {
                item = new SaverItem(graphPresenter, element.getChildren());
            } else if (ViewerItem.TAGNAME.equalsIgnoreCase(tagName)) {
                item = new ViewerItem(graphPresenter, element.getChildren());
            } else if (FileNameItem.TAGNAME.equalsIgnoreCase(tagName) || "Datafile".equalsIgnoreCase(tagName)) {
                item = new FileNameItem(graphPresenter, element.getChildren());
            } else if (ConstantItem.TAGNAME.equalsIgnoreCase(tagName)) {
                item = new ConstantItem(graphPresenter, element.getChildren());
            } else if (SystemItem.TAGNAME.equalsIgnoreCase(tagName)) {
                item = new SystemItem(graphPresenter, element.getChildren());
            } else if (ScriptItem.TAGNAME.equalsIgnoreCase(tagName)) {
                item = new ScriptItem(graphPresenter, element.getChildren());
            } else if (MacroOperatorItem.TAGNAME.equalsIgnoreCase(tagName) || "MacroOperateur".equalsIgnoreCase(tagName)) {
                item = new MacroOperatorItem(graphPresenter, element.getChildren());
            } else if (MergeControlStructureItem.TAGNAME.equalsIgnoreCase(tagName)) {
                item = new MergeControlStructureItem(graphPresenter, element.getChildren());
            } else if (EvalItem.TAGNAME.equalsIgnoreCase(tagName)) {
                item = new EvalItem(graphPresenter, element.getChildren());
            } else if (FormatItem.TAGNAME.equalsIgnoreCase(tagName)) {
                item = new FormatItem(graphPresenter, element.getChildren());
            } else if (IfControlStructureItem.TAGNAME.equalsIgnoreCase(tagName)) {
                item = new IfControlStructureItem(graphPresenter, element.getChildren());
            } else if (SelectControlStructureItem.TAGNAME.equalsIgnoreCase(tagName)) {
                item = new SelectControlStructureItem(graphPresenter, element.getChildren());
            } else if (WhileControlStructureItem.TAGNAME.equalsIgnoreCase(tagName)) {
                item = new WhileControlStructureItem(graphPresenter, element.getChildren());
            } else if ("Until".equalsIgnoreCase(tagName)) { // Keep for compatibility
                item = new WhileControlStructureItem(graphPresenter, element.getChildren());
            } else if (ForControlStructureItem.TAGNAME.equalsIgnoreCase(tagName)) {
                item = new ForControlStructureItem(graphPresenter, element.getChildren());
            }
            if (item != null) {
                graphPresenter.addItemAtAbsolutePosition(item, item.getLocation(), true);
                items.add(item);
            }
        }
        return items;
    }

}
