/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.element;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingResultException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.tree.SubroutineTreeItem;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.presenter.SubGraphPresenter;
import fr.greyc.pantheon.ariane.view.graph.XYLayout;
import fr.greyc.pantheon.ariane.view.graph.element.CompositeItemView;
import fr.greyc.pantheon.ariane.view.graph.element.CompositeItemView.StateName;
import fr.greyc.pantheon.ariane.view.graph.element.ControlItemView;

import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;

/**
 * An <code>CompositeItem</code> represents an item with an inner subgraph.
 */
public abstract class CompositeItem extends ControlItem {
    private SubGraphPresenter _subGraph;

    CompositeItem( GraphPresenter graph, TreeItem treeItem ) {
        super(new CompositeItemView(), graph);
        setRelatedTreeItem(treeItem);
        createSlots(0, 0, 1, 0, 0);
        ((CompositeItemView) getView()).changeState(StateName.Uneditable);
    }

    CompositeItem( TreeItem treeItem, GraphPresenter graph, String name ) {
        super(new CompositeItemView(), treeItem, graph, name);
        ((CompositeItemView) getView()).changeState(StateName.Editable);
        createSlots(0, 0, 1, 0, 0);
    }

    CompositeItem( GraphPresenter graph ) {
        super(new CompositeItemView(), graph);
        ((CompositeItemView) getView()).changeState(StateName.Editable);
        createSlots(0, 0, 1, 0, 0);
    }

    @Override
    public final SubGraphPresenter getSubGraph() {
        return _subGraph;
    }

    public final void viewSubgraph() {
        _subGraph.showContent();
    }

    @Override
    public final void setNickName( String name ) {
        super.setNickName(name);
        if (_subGraph != null) {
            Ariane.getPresenter().renameGraph(_subGraph, getNickName());
        }
    }

    @Override
    public final void removeHook() {
        _subGraph.closeFrame();
    }

    /////
    ///// E X E C U T I O N
    /////
    @Override
    public final void resetAll() {
        super.resetAll();
        if (_subGraph != null) {
            _subGraph.resetAll();
        }
    }

    /////
    ///// I N P U T   /   O U T P U T
    /////
    @Override
    public final XMLElement convertToXML( File file ) throws IOException {
        XMLElement element = new XMLElement(getTagName());
        List<XMLElement> xmlTree = convertPropertiesToXML(file);
        if (getRelatedTreeItem() instanceof SubroutineTreeItem && ((CompositeItemView) getView()).isUneditable()) {
            xmlTree.add(new XMLElement("SubroutineReference", getRelatedTreeItem().getName()));
        }
        element.addChildren(xmlTree);
        element.addChildren(_subGraph.convertToXMLTree(file));
        return element;
    }

    public final void changeEditableStatus( boolean editable ) {
        ((CompositeItemView) getView()).changeEditableStatus(editable);
    }

    public abstract TreeItem restoreDefaultTreeItem();

    protected abstract int getNumberOfFixedParameters();

    public final void changePrototype() {
        ControlItemView.Prototype p = askForPrototype(getNickName(), getInputDataSlots().size(),
                getInputParameterSlots().size() - getNumberOfFixedParameters(), getOutputDataSlots().size());

        if (p != null) {
//         getGraph().startActionEdition();
            if (!p.getName().equals(getNickName())) {
                setNickName(p.getName(), true);
            }
            int difference;

            /// 1. Parameters
            difference = p.getParameters() - getInputParameterSlots().size() + getNumberOfFixedParameters();
            if (difference != 0) {
                if (difference > 0) { // Add parameters
                    for (int i = 0; i < difference; i++) {
                        int outer = getInputParameterSlots().size();
                        int inner = getVirtualItem().getVirtualInputParameterSlots().size();
                        // 1. Ajouter un slot dans le virtual operator
                        OutputResultSlot slot = new OutputResultSlot(getVirtualItem(), inner);
                        slot.getView().setToolTipText(getVirtualItem().getGraph().getSuperGraphItem().getNickName() + ArianeSettings.getString("VirtualOperator.parameter") + inner);
                        getVirtualItem().getVirtualInputParameterSlots().add(slot);
                        // 2. Ajouter le slot dans le sous-graphe
                        getSubGraph().add(slot, XYLayout.Orientation.NORTHWEST, inner);
                        // 3. Ajouter le slot sur l'item
                        getInputParameterSlots().add(new ParameterSlot(this, outer));
                    }
                } else { // Remove parameters
                    for (int i = 0; i < -difference; i++) {
                        int outer = getInputParameterSlots().size() - 1;
                        int inner = getVirtualItem().getVirtualInputParameterSlots().size() - 1;
                        // 1. disconnect outer parameter
                        getInputParameterSlots().get(outer).disconnect();
                        // 2. disconnect inner parameter
                        OutputResultSlot slot = getVirtualItem().getVirtualInputParameterSlots().get(inner);
                        slot.disconnectAll();
                        // 3. supprimer le slot dans le virtual operator.
                        getVirtualItem().getVirtualInputParameterSlots().remove(inner);
                        // 4. Supprimer le slot dans le sous-graphe.
                        getSubGraph().remove(slot);
                        // 5. Remove macro slot from the macro
                        getInputParameterSlots().remove(outer);
                    }
                }
            }
            /// 2. Inputs
            difference = p.getInputs() - getInputDataSlots().size();
            if (difference != 0) {
                if (difference > 0) { // Add new inputs
                    for (int i = 0; i < difference; i++) {
                        int outer = getInputDataSlots().size();

                        // 1. Ajouter un slot dans le virtual
                        OutputDataSlot slot1 = new OutputDataSlot(getVirtualItem(), outer);
                        slot1.getView().setToolTipText(getVirtualItem().getGraph().getSuperGraphItem().getNickName() + ArianeSettings.getString("VirtualOperator.input") + outer);
                        getVirtualItem().getVirtualInputDataSlots().add(outer, slot1);
                        getSubGraph().add(slot1, XYLayout.Orientation.WEST, outer);

                        // 2. Ajouter d'un slot d'entree intermediaire
                        int intermediate = getVirtualItem().getVirtualOutputDataSlots().size() - getOutputDataSlots().size();
                        if (intermediate > 0) {
                            InputDataSlot slot2 = new InputDataSlot(getVirtualItem(), intermediate);
                            slot2.getView().setToolTipText(getVirtualItem().getGraph().getSuperGraphItem().getNickName() + ArianeSettings.getString("VirtualOperator.reparameter") + intermediate);
                            getVirtualItem().getVirtualOutputDataSlots().add(slot2);
                            getSubGraph().add(slot2, XYLayout.Orientation.SOUTHEAST, intermediate);
                        }
                        // 3. Ajouter le slot sur le composant
                        getInputDataSlots().add(new InputDataSlot(this, outer));
                    }
                } else {
                    for (int i = 0; i < -difference; i++) {
                        int outer = getInputDataSlots().size() - 1;
                        int intermediate = getVirtualItem().getVirtualOutputDataSlots().size() - getOutputDataSlots().size();
                        // 1. disconnet outer output
                        getInputDataSlots().get(outer).disconnect();
                        // 2. disconnect inner output
                        OutputDataSlot slot1 = getVirtualItem().getVirtualInputDataSlots().get(outer);
                        slot1.disconnectAll();
                        // 3. supprimer le slot dans le virtual operator.
                        getVirtualItem().getVirtualInputDataSlots().remove(slot1);
                        // 4. Supprimer le slot dans le sous-graphe.
                        getSubGraph().remove(slot1);
                        if (intermediate > 0) {
                            InputDataSlot slot2 = getVirtualItem().getVirtualIntermediateDataSlots(outer);
                            slot2.disconnectAll();
                            getVirtualItem().getInputDataSlots().remove(slot2);
                            getSubGraph().remove(slot2);
                        }
                        // 5. Remove macro slot from the macro
                        getInputDataSlots().remove(outer);
                    }
                }
            }

            ////// 3. Outputs
            difference = p.getOutputs() - getOutputDataSlots().size();
            if (difference != 0) {
                if (difference > 0) { // Add outputs
                    for (int i = 0; i < difference; i++) {
                        // 1. Ajouter un slot dans le virtual operator
                        int outer = getOutputDataSlots().size();
                        int inner = getVirtualItem().getVirtualOutputDataSlots().size() - getInputDataSlots().size();

                        InputDataSlot slot = new InputDataSlot(getVirtualItem(), inner);
                        slot.getView().setToolTipText(getVirtualItem().getGraph().getSuperGraphItem().getNickName() + ArianeSettings.getString("VirtualOperator.output") + inner);
                        getVirtualItem().getVirtualOutputDataSlots().add(slot);
                        // 2. Ajouter le slot dans le sous-graphe
                        getSubGraph().add(slot, XYLayout.Orientation.EAST, inner); // - 1
                        // 3. Ajouter le slot sur le macro
                        getOutputDataSlots().add(new OutputDataSlot(this, outer));
                    }
                } else { // Remove outputs
                    for (int i = 0; i < -difference; i++) {
                        int outer = getOutputDataSlots().size() - 1;
                        // 1. disconnect outer output
                        getOutputDataSlots().get(outer).disconnectAll();
                        // 2. disconnect inner output
                        InputDataSlot slot = getVirtualItem().getVirtualOutputDataSlots().get(outer);
                        slot.disconnectAll();
                        // 3. supprimer le slot dans le virtual operator.
                        getVirtualItem().getVirtualOutputDataSlots().remove(outer);
                        // 4. Supprimer le slot dans le sous-graphe.
                        getSubGraph().remove(slot);
                        // 5. Remove macro slot from the macro
                        getOutputDataSlots().remove(outer);
                    }
                }
            }
//         getGraph().endActionEdition();
            getView().resetPropertyViewer();
            getView().draw();
        }
    }

    /////
    ///// A C C E S S O R S
    /////
    final void initComment() {
        setComment("\t");
    }

    @Override
    protected final void disposeHook() {
        _subGraph.dispose();
        _subGraph = null;
    }

    protected abstract boolean linkItemWithVirtualItem( VirtualItem virtualtem ) throws ArianeException;

    final void setPrototype( String name, int inputImageCount, int imageParameterCount, int outputImageCount, int interImageCount, int interParameterCount ) {
        setNickName(name);
        setSlotsNumber(inputImageCount, imageParameterCount, outputImageCount);
        createSubGraph(name, inputImageCount, imageParameterCount, outputImageCount);
    }

    final SubGraphPresenter createSubGraph( String name, int inputImageCount, int inputParameterCount, int outputImageCount, int interImageCount, int interParameterCount ) {
        _subGraph = new SubGraphPresenter(this, name, inputImageCount, inputParameterCount, outputImageCount, interImageCount, interParameterCount);
        return _subGraph;
    }

    protected abstract SubGraphPresenter createSubGraph( String name, int inputImageCount, int inputParameterCount, int outputImageCount );

    @Override
    protected final void resetHook() {
        _subGraph.reset();
        resetHookHook();
    }

    void resetHookHook() {
        // Empty
    }

    final VirtualItem getVirtualItem() {
        return _subGraph.getVirtualItem();
    }

    final boolean getInputDataFiles( VirtualItem virtualItem ) throws ArianeMissingInputException {
        Iterator<InputDataSlot> inputDataSlots = getInputDataSlots().iterator();
        Iterator<OutputDataSlot> outputDataSlots = virtualItem.getVirtualInputDataSlots().iterator();
        while (inputDataSlots.hasNext()) {
            InputDataSlot inputDataSlot = inputDataSlots.next();
            if (inputDataSlot.hasValue(true)) {
                OutputDataSlot outputdataSlot = outputDataSlots.next();
                outputdataSlot.setValue(inputDataSlot.getValue());
            } else {
                return false;
            }
        }
        return true;
    }

    final void getMask( VirtualItem virtualItem ) throws ArianeRuntimeException {
        InputMaskSlot inputMask = getInputMaskSlot();
        // Note: Mask is optional
        if (inputMask.isConnected() && inputMask.hasValue(false)) {
            OutputDataSlot outputDataSlot = (OutputDataSlot) virtualItem.getVirtualInputMaskSlot();
            outputDataSlot.setValue(inputMask.getValue());
        }
    }

    final boolean setOutputResult( VirtualItem virtualItem ) throws ArianeMissingResultException {
        ParameterSlot inputParameterSlot = (ParameterSlot) virtualItem.getVirtualOutputResultSlot();
        try {
            if (!inputParameterSlot.hasValue(true)) {
                return false;
            }
            OutputResultSlot outputResultSlot = getOutputResultSlot();
            outputResultSlot.setValue(inputParameterSlot.getValue());
            return true;
        } catch (ArianeMissingParameterException e) {
            throw changeExceptionType(e, "Operator.error8");
        }
    }

    final void setOutputData( VirtualItem virtualItem ) throws ArianeMissingResultException {
        Iterator<OutputDataSlot> inputDataSlots = getOutputDataSlots().iterator();
        Iterator<InputDataSlot> outputDataSlots = virtualItem.getVirtualOutputDataSlots().iterator();
        while (inputDataSlots.hasNext()) {
            InputDataSlot inputDataSlot = outputDataSlots.next();
            try {
                if (inputDataSlot.hasValue(true)) {
                    OutputDataSlot outputDataSlot = inputDataSlots.next();
                    outputDataSlot.setValue(inputDataSlot.getValue());
                } else {
                    throw new ArianeMissingResultException(ArianeSettings.getString("Operator.error9"));
                }
            } catch (ArianeMissingInputException e) {
                throw changeExceptionType(e, "Operator.error9");
            }
        }
    }

    private ArianeMissingResultException changeExceptionType( ArianeException exception, String message ) {
        ArianeMissingResultException newException = new ArianeMissingResultException(ArianeSettings.getString(message));
        newException.initCause(exception);
        return newException;
    }

    final ArianeRuntimeException displayError( Exception exception, ArianeSettings.ErrorCode error ) {
        if (GraphPresenter.isExecutionRunning()) {
            Ariane.getPresenter().printErrorMessage(error, getAbsoluteName(), getId(), exception.getLocalizedMessage());
            getGraph().stopExecution();
            changeStatus(Status.ERROR);
        }
        return new ArianeRuntimeException(exception);
    }

    @Override
    protected final void createAdditionalPropertyFromXML( List<XMLElement> xmlTree ) throws IOException {
        boolean subroutine = false;
        for (XMLElement element : xmlTree) {
            if ("SubroutineReference".equalsIgnoreCase(element.getName())) {
                subroutine = getRelatedTreeItem() instanceof SubroutineTreeItem;
                break;
            }
        }
        SubGraphPresenter subGraph = createSubGraph(getName(), getInputDataSlots().size(), getInputParameterSlots().size(), getOutputDataSlots().size());
        if (subroutine) {
            ((CompositeItemView) getView()).changeState(StateName.Uneditable);
            subGraph.createFromXMLTree(((SubroutineTreeItem) getRelatedTreeItem()).getBody());
        } else {
            subGraph.createFromXMLTree(xmlTree);
        }
    }

    final void executeBody( VirtualItem virtualItem ) throws ArianeMissingResultException {
        for (int k = 0; k < getInputDataSlots().size(); k++) {
            try {
                InputDataSlot intermediateInputSlot = virtualItem.getVirtualIntermediateDataSlots(k);
                intermediateInputSlot.hasValue(true);
            } catch (ArianeMissingInputException e) {
                throw changeExceptionType(e, "Operator.error10");
            }
        }
    }

    final void setIntermediateInputs( VirtualItem virtualItem ) throws ArianeMissingInputException, IOException {
        for (int k = 0; k < getInputDataSlots().size(); k++) {
            InputDataSlot intermediateInputSlot = virtualItem.getVirtualIntermediateDataSlots(k);
            if (intermediateInputSlot.hasValue(true)) {
                OutputDataSlot inputSlot = virtualItem.getVirtualInputDataSlot(k);
                DataFile.copyFile(intermediateInputSlot.getValue(), inputSlot.getTempName());
                inputSlot.setValue();
            }
        }
    }

    private void setSlotsNumber( int inputImageCount, int inputParameterCount, int outputImageCount ) {
        List<InputDataSlot> inputDataSlots = getInputDataSlots();
        if (inputImageCount > inputDataSlots.size()) {
            for (int i = inputDataSlots.size(); i < inputImageCount; i++) {
                inputDataSlots.add(new InputDataSlot(this, i));
            }
        }
        List<ParameterSlot> inputParameterSlots = getInputParameterSlots();
        if (inputParameterCount > inputParameterSlots.size()) {
            for (int i = inputParameterSlots.size(); i < inputParameterCount; i++) {
                inputParameterSlots.add(new ParameterSlot(this, i));
            }
        }
        List<OutputDataSlot> outputDataSlots = getOutputDataSlots();
        if (outputImageCount > outputDataSlots.size()) {
            for (int i = outputDataSlots.size(); i < outputImageCount; i++) {
                outputDataSlots.add(new OutputDataSlot(this, i));
            }
        }
    }
}
