/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.element;

import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.view.graph.element.GraphElementView;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.util.List;

/**
 * A <code>DataFile</code> is either a loader, a saver or a viewer of a data files.
 */
public abstract class DataFile extends GraphItemModel {
    private String _currentFilename;

    DataFile( GraphElementView graphItemView, TreeItem treeItem, GraphPresenter graph, String name, String binary ) {
        super(graphItemView, treeItem, graph, name, null);
        super.setBinary(binary);
    }

    DataFile( GraphElementView graphItemView, GraphPresenter graph ) {
        super(graphItemView, graph);
    }

    /**
     * Creates a <code>Loader</code> or a <code>Saver</code> on the given graph with
     * the given XML data.
     * Notice: Loader has 1 input image and 0 output image and saver has 0 input image
     * and 1 output image. !!!! This method is kept for sake of compatibility.
     *
     * @param graph the parent graph.
     * @param data  the XML tree of data.
     * @return the item.
     * @throws IOException XML errors.
     */
    public static DataFile getInstance( GraphPresenter graph, List<XMLElement> data ) throws IOException {
        int inputDataSlots = 0;
        for (XMLElement element : data) {
            if ("ImagesEntree".equalsIgnoreCase(element.getName())) {
                inputDataSlots = Integer.parseInt(element.getValue());
            }
        }
        if (inputDataSlots > 0) {
            return new SaverItem(graph, data);
        }
        return new LoaderItem(graph, data);
    }

    public static String getBaseFileName( String binary ) {
        return binary.substring(binary.lastIndexOf(File.separator) + 1);
    }

    public final void setCurrentFilename( String filename ) {
        _currentFilename = filename;
    }

    public final String getCurrentFilename() {
        return _currentFilename;
    }

    /**
     * Sets the name of the related physical file. If the name of the file is specified as a relative path than the absolute name is built with the directory of the project.
     *
     * @param filename         a file name (absolute or relative).
     * @param currentDirectory the name of the current directory or null if the related graph file has no name.
     */
    public static String setPathToFilename( String filename, String currentDirectory ) {
        if (filename == null || filename.length() == 1) {
            return "";
        }
        String binary = filename;
        // relative name
        if (binary.charAt(0) == '.') {
            binary = binary.substring(2);
            File file = new File(currentDirectory, binary.replaceAll("\\\\", "/"));
            return file.getAbsolutePath();
        } else if (binary.startsWith("$FILE_DIRECTORY")) { // TODO Garder pour les fichiers normaux hors binaires ?
            String fileDirectory = UserPreferences.getPreferredBinaryDirectory();
            binary = binary.substring("$FILE_DIRECTORY".length() + 1);
            File file = new File(fileDirectory, binary.replaceAll("\\\\", "/"));
            return file.getAbsolutePath();
        }
        // absolute name
        File binaryFilename = new File(binary);
        if (binaryFilename.exists()) {
            try {
                return binaryFilename.getCanonicalPath();
            } catch (IOException e) {
                // Nothing to do
            }
        }
        // Keep for the sake of compatibility
        File file = new File(currentDirectory, binary.replaceAll("\\\\", "/"));
        if (file.exists()) {
            return file.getAbsolutePath();
        }
        // default
        return binaryFilename.getAbsolutePath();
    }

    /**
     * Returns the name of the file either with relative path
     * if the file is in the filePath directory or with an absolute path otherwise.
     *
     * @param dataFileName       the name of a data file.
     * @param graphDirectoryPath the path to the graph directory.
     * @param baseDirectoryPath  the path to the specified base directory.
     * @return the name.
     */
    public static String getRelativePath( String dataFileName, String graphDirectoryPath, String baseDirectoryPath ) {
        // Case of simple cut-copy/paste or no filename or filename with relative path.
        if (dataFileName == null || graphDirectoryPath == null || dataFileName.trim().isEmpty() || dataFileName.trim().charAt(0) != File.separatorChar) {
            return dataFileName;
        }

        File file = new File(dataFileName);
        String filePath = file.getParent();

        if (baseDirectoryPath != null && filePath != null && filePath.startsWith(baseDirectoryPath)) {
            // Use base directory path.
            String subDirectory = filePath.substring(baseDirectoryPath.length());
            if (!subDirectory.isEmpty()) {
                subDirectory = File.separator + subDirectory;
            }
            return "$FILE_DIRECTORY" + subDirectory + File.separator + file.getName();
        } else {
            // Use related path if the file is in the directory, else use the absolute name.
            if (filePath != null && filePath.startsWith(graphDirectoryPath)) {
                return "." + filePath.substring(graphDirectoryPath.length()) + File.separator + file.getName();
            }
        }
        return file.getAbsolutePath();
    }

    @Override
    public final String toString() {
        if (getBinary() == null) {
            return "(" + getId() + "):" + getName();
        } else {
            return "(" + getId() + "):" + getBinary();
        }
    }

    public static void copyFile( String from, String to ) throws IOException {
        if (from.equals(to)) {
            return;
        }
        File src = new File(from);
        File dst = new File(to);
        Files.copy(src.toPath(), dst.toPath(), StandardCopyOption.REPLACE_EXISTING);
    }
}
