/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.element;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeArithmeticException;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingFormulaException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.element.etc.StringExpression;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.view.graph.element.DefaultControlElementView;

import java.io.IOException;
import java.util.List;

/**
 * A <code>FormatItem</code> implements a string manipulator.
 */
public final class FormatItem extends ControlItem {
    public static final String TAGNAME = "Format";
    private StringExpression _expression;

    public FormatItem( TreeItem treeItem, GraphPresenter graph ) throws InterruptedException {
        super(new DefaultControlElementView(), treeItem, graph, treeItem.getName());
        askAndSetPrototype(null, -1, 2, -1);
    }

    public FormatItem( GraphPresenter graph, List<XMLElement> elements ) throws IOException {
        super(new DefaultControlElementView(), graph);
        createFromXML(elements);
        // Compile formula if it exists.
        try {
            _expression = new StringExpression(getFormula(), getInputParameterSlots().size());
        } catch (ArianeArithmeticException e) {
            // Nothing to do
        }
        setComment(getFormula());
    }

    FormatItem( TreeItem treeItem, GraphPresenter graph, String name ) { // For test
        super(new DefaultControlElementView(), treeItem, graph, name);
        createSlots(treeItem.getNumberOfInputDataSlots(), treeItem.getNumberOfParameterSlots(), 0, treeItem.getNumberOfOutputDataSlots(), 1);
    }

    @Override
    protected void setPrototype( String name, int iIm, int iP, int oIm ) {
        setAbstractName(name);
        setFormula(""); // IMPERATIVE to signify that it uses formula.
        createSlots(0, iP, 0, 0, 0);
    }

    @Override
    public boolean changeFormula( String textFormula ) {
        if (getFormula() == null || !getFormula().equals(textFormula)) {
            try {
                _expression = new StringExpression(textFormula, getInputParameterSlots().size());
                setFormula(textFormula);
                setComment(textFormula);
            } catch (ArianeArithmeticException e) {
                Ariane.getPresenter().printErrorMessage(e.getLocalizedMessage());
                return false;
            }
        }
        return true;
    }

    /////
    ///// A C C E S S O R S
    /////

    /**
     * @return the tag name.
     */
    @Override
    public String getTagName() {
        return TAGNAME;
    }

    /////
    ///// E X E C U T I O N
    /////
    @Override
    protected boolean primitiveExecute() throws ArianeRuntimeException {
        try {
            if (getFormula().isEmpty()) {
                throw new ArianeMissingFormulaException(ArianeSettings.getString("arithmetic.error0"));
            }
            int numberOfParameters = getInputParameterSlots().size();

            String[] val = new String[numberOfParameters];
            for (int i = 0; i < numberOfParameters; i++) {
                ParameterSlot parameter = getInputParameterSlots().get(i);
                if (parameter.hasValue(true)) {
                    val[i] = parameter.getValue();
                } else {
                    return false;
                }
            }

            final String result = _expression.evaluate(val);
            getOutputResultSlot().setValue(result);
            Ariane.getPresenter().printSystemMessage("format(" + getFormula() + ")=" + result);
            changeStatus(Status.EXECUTED);
            return true;
        } catch (ArianeException e) {
            if (GraphPresenter.isExecutionRunning()) {
                Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
                getGraph().stopExecution();
                changeStatus(Status.ERROR);
            }
            throw new ArianeRuntimeException(e);
        }
    }
}
