/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.element;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.element.etc.ItemSignatureParser;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.SlotModel;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.presenter.Presenter;
import fr.greyc.pantheon.ariane.presenter.SubGraphPresenter;
import fr.greyc.pantheon.ariane.view.command.RenameItemEdit;
import fr.greyc.pantheon.ariane.presenter.GraphItemPresenter;
import fr.greyc.pantheon.ariane.view.graph.element.GraphElementView;

import java.awt.Point;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.MalformedParameterizedTypeException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

/**
 * A <code>GraphItem</code> defines an item that can be displayed in a graph.
 * It is graphically represented by a box with input and output slots.
 */
public abstract class GraphItemModel {
    private final int _globalId;
    private int _internalId = -1;
    private final int _currentReadOrPasteSessionId;
    private String _name;
    private String _nickName;
    private String _binary;
    private String _category;
    private String _formula;

    private final GraphItemPresenter _graphItemPresenter;
    private TreeItem _treeItem;
    private GraphPresenter _graph;

    private boolean _selected; // A deplacer dans presenter.

    private final List<InputDataSlot> _inputDataSlots = new ArrayList<>();
    private final List<OutputDataSlot> _outputDataSlots = new ArrayList<>();
    private final List<ParameterSlot> _inputParameterSlots = new ArrayList<>();
    private final List<OutputResultSlot> _outputResultSlots = new ArrayList<>();
    private InputMaskSlot _inputMaskSlot;

    private Status _executionStatus = Status.UNEXECUTED;

    public enum Status {
        UNEXECUTED,
        RUNNING,
        EXECUTED,
        ERROR
    }

    GraphItemModel( GraphElementView graphItemView, GraphPresenter graph ) {
        _graph = graph;
        // Set a unique identifier number for the whole application!
        _globalId = _graph.getNextGraphItemIdentifier();
        _currentReadOrPasteSessionId = Presenter.getCurrentReadOrPasteSessionId();
        _graphItemPresenter = new GraphItemPresenter(this, graphItemView);
    }

    GraphItemModel( GraphElementView graphItemView, TreeItem treeItem, GraphPresenter graph, String name, String category ) {
        this(graphItemView, graph);
        _treeItem = treeItem;
        _name = name;
        _category = category;
    }

    public int getId() {
        return _globalId;
    }

    final void setInternalId( int identifier ) {
        _internalId = identifier;
    }

    public final int getInternalId() {
        return _internalId;
    }

    public final boolean isInSameReadPasteSession() {
        return _currentReadOrPasteSessionId == Presenter.getCurrentReadOrPasteSessionId();
    }

    /////
    ///// P R O P E R T Y
    /////
    public final Point getLocation() {
        return _graphItemPresenter.getView().getLocation();
    }

    public final void setLocation( Point point ) {
        _graphItemPresenter.getView().setLocation(point);
    }

    public final GraphElementView getView() {
        return _graphItemPresenter.getView();
    }

    public final void setRelatedTreeItem( TreeItem relatedTreeItem ) {
        _treeItem = relatedTreeItem;
    }

    final TreeItem getRelatedTreeItem() {
        if (_treeItem == null) {
            return Ariane.getPresenter().getTree().getTreeItem(getName(), getCategory());
        } else {
            return _treeItem;
        }
    }

    public final void remove() {
        disconnectAllSlots();
        _graphItemPresenter.getView().dispose();
        removeHook();
    }

    void removeHook() {
        // empty
    }

    public final void dispose() {
        disposeHook();
        disconnectAllSlots();
        _graphItemPresenter.getView().dispose();
        for (SlotModel inputDataSlot : _inputDataSlots) {
            inputDataSlot.dispose();
        }
        _inputDataSlots.clear();

        for (SlotModel inputParameterSlot : _inputParameterSlots) {
            inputParameterSlot.dispose();
        }
        _inputParameterSlots.clear();

        if (_inputMaskSlot != null) {
            _inputMaskSlot.dispose();
        }
        _inputMaskSlot = null;

        for (SlotModel outputDataSlot : _outputDataSlots) {
            outputDataSlot.dispose();
        }
        _outputDataSlots.clear();

        for (SlotModel outputResultSlot : _outputResultSlots) {
            outputResultSlot.dispose();
        }
        _outputResultSlots.clear();
        _graph = null;
    }

    void disposeHook() {
        // Empty
    }

    public final void delete() { // TODO dans le presenter
        _graph.removeItem(this);
    }

    /////
    ///// P R O P E R T Y
    /////
    public final void setSignature( String signature ) {
        if (signature != null && !signature.trim().isEmpty()) {
            // Set the default value for the parameters
            try {
                String[][] values = ItemSignatureParser.parse(signature, _inputParameterSlots.size());
                for (int i = 0; i < _inputParameterSlots.size(); i++) {
                    _inputParameterSlots.get(i).setName(values[i][0]);
                    _inputParameterSlots.get(i).setValue(values[i][1]);
                }
            } catch (MalformedParameterizedTypeException | ArrayIndexOutOfBoundsException e) {
                Ariane.getPresenter().printErrorMessage(getName() + ": " + ArianeSettings.getString("Operator.error6"));
            }
        }
    }

    /////
    ///// E X E C U T I O N
    /////
    public final boolean execute() { // TODO passer la condition en parametre ?
        if (!GraphPresenter.isExecutionRunning()) {
            return false;
        }
        if (getExecutionStatus() == Status.EXECUTED) {
            return true;
        }
        try {
            if (primitiveExecute()) {
                _graphItemPresenter.getView().updatePropertyViewer();
                return true;
            }
        } catch (ArianeRuntimeException e) {
            // Nothing to do
        }
        return false;
    }

    protected abstract boolean primitiveExecute() throws ArianeRuntimeException;

    public final Status getExecutionStatus() {
        return _executionStatus;
    }

    final void changeStatus( Status value ) { // TODO Dans le presenter ?
        if (!_executionStatus.equals(value)) {
            _executionStatus = value;
            if (_executionStatus == Status.ERROR && _graph.getSuperGraphItem() != null) {
                _graph.getSuperGraphItem().changeStatus(value);
            }
            _graphItemPresenter.getView().displayStatus();
        }
    }

    public void resetAll() { // TODO dans presenter ?
        if (_executionStatus != Status.UNEXECUTED) {
            reset();
            if (_outputDataSlots != null) {
                for (OutputDataSlot slot : _outputDataSlots) {
                    slot.reset();
                    // Reset all connected items.
                    if (slot.isConnected()) {
                        for (InputSlot inputSlot : slot.getConnectedInputSlots()) {
                            inputSlot.getGraphItem().resetAll();
                        }
                    }
                }
            }

            if (!_outputResultSlots.isEmpty()) {
                getOutputResultSlot().reset();
                if (getOutputResultSlot().isConnected()) {
                    for (InputSlot inputSlot : getOutputResultSlot().getConnectedInputSlots()) {
                        inputSlot.getGraphItem().resetAll();
                    }
                }
            }
            // Its parent
            if (_graph.getSuperGraphItem() != null) {
                _graph.getSuperGraphItem().resetAll();
            }
        }
    }

    public final void reset() {
        if (_executionStatus != Status.UNEXECUTED) {
            changeStatus(Status.UNEXECUTED);
            _graphItemPresenter.getView().updatePropertyViewer();
        }
        resetHook();
    }

    void resetHook() {
        // Empty
    }

    /////
    ///// C O N N E C T I O N
    /////
    private void disconnectAllSlots() {
        for (SlotModel slot : _outputDataSlots) {
            slot.disconnectAll();
        }
        if (_inputParameterSlots != null) {
            for (InputSlot slot : _inputParameterSlots) {
                slot.disconnect();
            }
        }
        for (InputSlot slot : _inputDataSlots) {
            slot.disconnect();
        }
        if (_inputMaskSlot != null) {
            _inputMaskSlot.disconnect();
        }
        if (!_outputResultSlots.isEmpty()) {
            getOutputResultSlot().disconnectAll();
        }
    }

    public final void fireConnection( InputSlot slot, boolean connection ) {
        fireConnectionHook(slot, connection);
        _graphItemPresenter.getView().fireConnection();
    }

    void fireConnectionHook( InputSlot slot, boolean connection ) {
        // Empty
    }

    final void createSlots( int inputDataSlotNumber, int inputParameterSlotNumber, int inputMaskSlotNumber, int outputDataSlotNumber, int outputResultSlotNumber ) {
        for (int i = 0; i < inputDataSlotNumber; i++) {
            _inputDataSlots.add(new InputDataSlot(this, i));
        }
        if (inputMaskSlotNumber > 0) {
            _inputMaskSlot = new InputMaskSlot(this);
        }
        for (int i = 0; i < inputParameterSlotNumber; i++) {
            _inputParameterSlots.add(new ParameterSlot(this, i));
        }
        for (int i = 0; i < outputDataSlotNumber; i++) {
            _outputDataSlots.add(new OutputDataSlot(this, i));
        }
        for (int i = 0; i < outputResultSlotNumber; i++) {
            _outputResultSlots.add(new OutputResultSlot(this, i));
        }
    }

    public final List<InputDataSlot> getInputDataSlots() {
        return _inputDataSlots;
    }

    public final List<OutputDataSlot> getOutputDataSlots() {
        return _outputDataSlots;
    }

    public final InputMaskSlot getInputMaskSlot() {
        return _inputMaskSlot;
    }

    public final List<ParameterSlot> getInputParameterSlots() {
        return _inputParameterSlots;
    }

    public final List<OutputResultSlot> getOutputResultSlots() {
        return _outputResultSlots;
    }

    public final OutputResultSlot getOutputResultSlot() {
        if (_outputResultSlots.isEmpty()) {
            return null;
        } else {
            return _outputResultSlots.get(0);
        }
    }

    public final SlotModel getSlot( String typeSlot, int rank ) {
        if ("InSlotImage".equalsIgnoreCase(typeSlot)) {
            return getInputDataSlots().get(rank);
        } else if ("InSlotMask".equalsIgnoreCase(typeSlot)) {
            return getInputMaskSlot();
        } else if ("InSlotParameter".equalsIgnoreCase(typeSlot)) {
            return getInputParameterSlots().get(rank);
        } else if ("OutSlotImage".equalsIgnoreCase(typeSlot)) {
            if (rank >= getOutputDataSlots().size()) {
                return null;
            }
            return getOutputDataSlots().get(rank);
        } else if ("OutSlotResult".equalsIgnoreCase(typeSlot)) {
            // For sake of ascendent compatibility
            if (rank < getOutputResultSlots().size()) {
                return getOutputResultSlots().get(rank);
            } else {
                return null;
            }
        }
        return null;
    }

    /////
    ///// A C C E S S O R S
    /////
    public SubGraphPresenter getSubGraph() {
        return null;
    }

    final void setComment( String comment ) {
        _graphItemPresenter.getView().setComment(comment);
    }

    public final String getCategory() {
        if (_treeItem == null) {
            return _category;
        }
        return _treeItem.getCategory();
    }

    public final String getDocumentationManualURL() {
        if (_treeItem == null) {
            return "";
        }
        return _treeItem.getDocumentationURL();
    }

    public String getBinary() {
        return _binary;
    }

    String getBinary( String filePath ) {
        return _binary;
    }

    public void setBinary( String binary ) {
        _binary = binary;
    }

    String getFormula( String filepath ) {
        return _formula;
    }

    public boolean isDisplayableFormula() {
        return true;
    }

    public final String getFormula() {
        return _formula;
    }

    public boolean changeFormula( String textFormula ) {
        return true;
    }

    public void setFormula( String formula ) {
        _formula = formula.trim();
        resetAll();
        getGraph().setModified(true);
        _graphItemPresenter.getView().firePropertyName(getNickName());
        getGraph().getView().repaint(); // TODO rien a faire ici
    }

    public String getFormulaText() {
        return getFormula();
    }

    public String getFormulaLabel() {
        return ArianeSettings.getString("PropertyViewer.formula");
    }

    public final String getName() {
        return _name;
    }

    final String getAbsoluteName() {
        return getGraph().getName() + "/" + getNickName(); // TODO mettre le nom dans le model ?
    }

    public final void setAbstractName( String name ) {
        if (name.length() > 0) {
            _name = name;
        }
    }

    public final void setTemporaryNickName( String name ) {
        _nickName = name;
        updateName();
    }

    public void setNickName( String name ) {
        if (name != null && name.length() > 0) {
            _nickName = name;
            _graphItemPresenter.getView().firePropertyTextName(name);
        } else {
            _nickName = null;
            _graphItemPresenter.getView().firePropertyTextName("");
        }
        _graphItemPresenter.getView().updatePropertyViewer();
        _graph.setModified(true);
    }

    public final void setNickName( String newName, boolean addInHistory ) { // TODO
        String oldName = _nickName;
        if (!newName.equals(oldName)) {
            setNickName(newName);
            if (addInHistory) {
                _graph.validateOneAction(new RenameItemEdit(this, newName, oldName));
            }
            updateName();
            _graph.revalidate();
        }
    }

    public final String getNickName() {
        return (_nickName == null) ? getName() : _nickName;
    }

    public final GraphPresenter getGraph() {
        return _graph;
    }

    public abstract String getTagName();

    /**
     * Updates the item contents after modification through the properties dialog box.
     */
    public final void updateName() {
        _graphItemPresenter.getView().firePropertyName(getNickName());
        _graphItemPresenter.getView().setToolTipText(toString());
    }

    /////
    ///// S E L E C T I O N
    /////
    public final boolean isSelected() {
        return _selected;
    }

    public final void setSelected( boolean selected ) {
        if (_selected != selected) {
            for (SlotModel slot : _inputDataSlots) {
                slot.getPresenter().setSelected(selected);
            }
            for (SlotModel slot : _inputParameterSlots) {
                slot.getPresenter().setSelected(selected);
            }
            for (SlotModel slot : _outputDataSlots) {
                slot.getPresenter().setSelected(selected);
            }
            for (SlotModel slot : _outputResultSlots) {
                slot.getPresenter().setSelected(selected);
            }
            SlotModel slot = _inputMaskSlot;
            if (slot != null) {
                slot.getPresenter().setSelected(selected);
            }
            _selected = selected;
            _graphItemPresenter.getView().setSelected(selected);
        }
    }
    /////
    /////  I N P U T S   &   O U T P U T S
    /////

    public XMLElement convertToXML( File file ) throws IOException {
        XMLElement element = new XMLElement(getTagName());
        element.addChildren(convertPropertiesToXML(file));
        return element;
    }

    final List<XMLElement> convertPropertiesToXML( File file ) {
        LinkedList<XMLElement> xmlTree = new LinkedList<>();
        xmlTree.add(new XMLElement("NickName", _nickName));
        xmlTree.add(new XMLElement("Name", getName()));
        xmlTree.add(new XMLElement("Identifier", Integer.toString(getId())));
        xmlTree.add(new XMLElement("X-coordinate", Integer.toString(getLocation().x)));
        xmlTree.add(new XMLElement("Y-coordinate", Integer.toString(getLocation().y)));

        String currentPath;
        if (file != null) {
            currentPath = file.getParentFile().getAbsolutePath();
        } else {
            currentPath = null;
        }
        String tmp = getBinary(currentPath);
        if (tmp != null) {
            xmlTree.add(new XMLElement("Binary", tmp));
        }
        if (getCategory() != null) {
            xmlTree.add(new XMLElement("Category", getCategory()));
        }
        if (getFormula() != null) {
            xmlTree.add(new XMLElement("Formula", getFormula(currentPath)));
        }
        if (!_inputDataSlots.isEmpty()) {
            xmlTree.add(new XMLElement("InputFiles", Integer.toString(_inputDataSlots.size())));
        }
        if (!_inputParameterSlots.isEmpty()) {
            xmlTree.add(new XMLElement("Parameters", Integer.toString(_inputParameterSlots.size())));
        }
        if (!_outputDataSlots.isEmpty()) {
            xmlTree.add(new XMLElement("OutputFiles", Integer.toString(_outputDataSlots.size())));
        }
        if (this instanceof IScriptArgument) {
            IScriptArgument item = (IScriptArgument) this;
            xmlTree.add(new XMLElement("Is-Argument", Boolean.toString(item.isArgumentOfRelatedScript())));
        }
        return xmlTree;
    }

    public final List<XMLElement> convertSlotsToXML() {
        List<XMLElement> domTree = new ArrayList<>();
        for (InputSlot slot : _inputDataSlots) {
            XMLElement element = slot.convertToXML();
            if (element != null) {
                domTree.add(element);
            }
        }
        for (InputSlot slot : _inputParameterSlots) {
            XMLElement element = slot.convertToXML();
            if (element != null) {
                domTree.add(element);
            }
        }
        if (_inputMaskSlot != null) {
            XMLElement element = _inputMaskSlot.convertToXML();
            if (element != null) {
                domTree.add(element);
            }
        }
        return domTree;
    }

    final void createFromXML( List<XMLElement> xmlTree ) throws IOException {
        createPropertyFromXML(xmlTree);
        createAdditionalPropertyFromXML(xmlTree);
    }

    private void createPropertyFromXML( List<XMLElement> xmlTree ) {
        int inputParameterNumber = 0;
        int inputDataNumber = 0;
        int outputDataNumber = 0;
        Point coordinates = new Point(0, 0);
        for (XMLElement element : xmlTree) {
            final String tagName = element.getName();
            final String value = element.getValue();
            // French tags are kept for sake of compatibility.
            if ("NickName".equalsIgnoreCase(tagName)) {
                setNickName(value);
            } else if ("Name".equalsIgnoreCase(tagName) || "Nom".equalsIgnoreCase(tagName)) {
                setAbstractName(value);
            } else if ("Identifier".equalsIgnoreCase(tagName) || "Identifieur".equalsIgnoreCase(tagName)) {
                setInternalId(Integer.parseInt(value));
            } else if ("X-coordinate".equalsIgnoreCase(tagName) || "Abscisse".equalsIgnoreCase(tagName)) {
                coordinates.x = Integer.parseInt(value);
            } else if ("Y-coordinate".equalsIgnoreCase(tagName) || "Ordonnee".equalsIgnoreCase(tagName)) {
                coordinates.y = Integer.parseInt(value);
            } else if ("Binary".equalsIgnoreCase(tagName) || "Binaire".equalsIgnoreCase(tagName)) {
                setBinary(value);
            } else if ("Category".equalsIgnoreCase(tagName) || "Categorie".equalsIgnoreCase(tagName)) {
                _category = value;
            } else if ("Formula".equalsIgnoreCase(tagName)) {
                setFormula(value);
            } else if ("Parameters".equalsIgnoreCase(tagName) || "ParametresEntree".equalsIgnoreCase(tagName)) {
                inputParameterNumber = Integer.parseInt(value);
            } else if ("InputFiles".equalsIgnoreCase(tagName) || "InputImages".equalsIgnoreCase(tagName) || "ImagesEntree".equalsIgnoreCase(tagName)) {
                inputDataNumber = Integer.parseInt(value);
            } else if ("OutputFiles".equalsIgnoreCase(tagName) || "OutputImages".equalsIgnoreCase(tagName) || "ImagesSortie".equalsIgnoreCase(tagName)) {
                outputDataNumber = Integer.parseInt(value);
            } else if ("Is-Argument".equalsIgnoreCase(tagName) && this instanceof IScriptArgument) {
                IScriptArgument item = (IScriptArgument) this;
                item.setArgumentOfRelatedScript(Boolean.parseBoolean(value));
            }
        }
        for (int i = 0; i < inputDataNumber; i++) {
            _inputDataSlots.add(new InputDataSlot(this, i));
        }
        for (int i = 0; i < inputParameterNumber; i++) {
            _inputParameterSlots.add(new ParameterSlot(this, i));
        }
        for (int i = 0; i < outputDataNumber; i++) {
            _outputDataSlots.add(new OutputDataSlot(this, i));
        }
        if (_treeItem == null && Ariane.getPresenter() != null) {
            if ("File".equals(_category)) { // keep it for the sake of compatibility.
                if (this instanceof LoaderItem) {
                    _name = "loader";
                    createSlots(0, 1, 0, 0, 0);
                    _category = null;
                } else if (this instanceof SaverItem) {
                    _name = "saver";
                    createSlots(0, 1, 0, 0, 0);
                    _category = null;
                }
            } else if ("Viewer".equals(_category) && this instanceof ViewerItem) {
                _name = "viewer";
                _category = null;
            } else if ("Comment".equals(_category) && this instanceof NoteItem) {
                _category = null;
            } else if ("Function".equals(_category) && this instanceof EvalItem) {
                _category = null;
            }
            _treeItem = getRelatedTreeItem();
        }
        setLocation(coordinates);
    }

    void createAdditionalPropertyFromXML( List<XMLElement> xmlTree ) throws IOException {
        // Empty
    }
}
