/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.element;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeDataFileIOException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.view.graph.element.DataFileView;
import fr.greyc.pantheon.ariane.view.graph.element.LoaderItemView;

import java.io.File;
import java.io.IOException;
import java.util.List;

/**
 * A <code>LoaderItem</code> manages data files loading within a graph.
 */
public final class LoaderItem extends DataFile {

   public static final String TAGNAME = "Loader";

   private long _dateOfLastModification;

   public LoaderItem( TreeItem relatedTreeItem, GraphPresenter graph ) {
      super(new LoaderItemView(), relatedTreeItem, graph, relatedTreeItem.getName(), null);
      createSlots(0, 1, 0, 1, 1);
   }

   public LoaderItem( GraphPresenter graph, List<XMLElement> data ) throws IOException {
      super(new LoaderItemView(), graph);
      createSlots(0, 0, 0, 0, 1);
      createFromXML(data);
      if (getInputParameterSlots().isEmpty()) { // For the sake of compatibility
         createSlots(0, 1, 0, 0, 0);
      }
   }

   @Override
   public String getTagName() {
      return TAGNAME;
   }

   /*
    * Returns the name of the file with either
    * - relative path if the file is in the specified directory
    * - or with an absolute path.
    */
   @Override
   public String getBinary( String directory ) {
      return getRelativePath(super.getBinary(), directory, UserPreferences.getPreferredFileDirectory());
   }

   /**
    * Sets the name of the related physical file. If the name of the file is specified as a relative path than the absolute name is built with the directory name of the project.
    *
    * @param filename a file name (absolute or relative)
    */
   @Override
   public void setBinary( String filename ) {
      if (filename != null && !filename.trim().isEmpty()) {
         String currentDirectory = getGraph().getDirectoryName();
         super.setBinary(setPathToFilename(filename, currentDirectory));
      }
   }

   /**
    * Resynchronizes the item with the related file. if the related file has been changed, resets the item.
    */
   public void synchronizeWithFileSystem() {
      if (getBinary() != null) {
         File file = new File(getBinary());
         if (_dateOfLastModification < file.lastModified()) {
            resetAll();
            _dateOfLastModification = file.lastModified();
         }
      }
   }

    @Override
   protected boolean primitiveExecute() throws ArianeRuntimeException {
      try {
         setCurrentFilename(null);
         final ParameterSlot parameter = getInputParameterSlots().isEmpty() ? null : getInputParameterSlots().get(0);
         final File file;
         if (parameter != null && parameter.hasValue(false)) {
            ((DataFileView)getView()).changeBinaryName(getView().getNameEditor(), null, false);
            file = new File(parameter.getValue());
         } else if (getBinary() == null || getBinary().isEmpty()) {
            throw new ArianeDataFileIOException(ArianeSettings.ErrorCode.MISSING_FILE, ArianeSettings.getString("ImageFile.error3"));
         } else {
            file = new File(getBinary());
            _dateOfLastModification = file.lastModified();
         }
         if (!file.exists()) {
            throw new ArianeDataFileIOException(ArianeSettings.ErrorCode.UNKNOWN_FILE, ArianeSettings.getString("ImageFile.error4")
                + " \""
                + file.getAbsolutePath()
                + "\"");
         }
         final String filename = file.getAbsolutePath();
         getOutputDataSlots().get(0).setValue(filename);
         setCurrentFilename(filename);
         if (getOutputResultSlot() != null) { // Keep this test for sake of compatibility
            getOutputResultSlot().setValue(filename);
         }
      } catch (ArianeDataFileIOException | ArianeMissingParameterException e) {
         if (GraphPresenter.isExecutionRunning()) {
            Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
            getGraph().stopExecution();
            changeStatus(Status.ERROR);
         }
         throw new ArianeRuntimeException(e);
      }

      // Validate the result.
      changeStatus(Status.EXECUTED);
      return true;
   }

   @Override
   public void fireConnectionHook( InputSlot slot, boolean connection ) {
      if (slot == getInputParameterSlots().get(0)) {
         ((LoaderItemView) getView()).setButtonEnabled(!connection);
      }
   }
}
