/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.element;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.element.etc.ArgumentLineParser;
import fr.greyc.pantheon.ariane.model.graph.element.etc.OperatorProcessBuilder;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.view.graph.element.OperatorElementView;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.LinkedList;
import java.util.List;

/**
 * An <code>OperatorItem</code> is a graph element that encapsulates a Shell command.
 * It uses inputs to produce outputs with specific parameter values.
 */
public final class OperatorItem extends GraphItemModel {
    public static final String TAGNAME = "Operator";
    public static final String SEPARATOR = "&";

    public OperatorItem( TreeItem treeItem, GraphPresenter graph ) {
        super(new OperatorElementView(), treeItem, graph, treeItem.getName(), treeItem.getCategory());
        createSlots(treeItem.getNumberOfInputDataSlots(), treeItem.getNumberOfParameterSlots(), 1, treeItem.getNumberOfOutputDataSlots(), 1);
        setSignature(treeItem.getSignature());
    }

    public OperatorItem( GraphPresenter graph, List<XMLElement> xmlTree ) throws IOException {
        super(new OperatorElementView(), graph);
        // Creates mask and result because they are not saved in the xmlTree.
        createSlots(0, 0, 1, 0, 1);
        createFromXML(xmlTree);
        if (getRelatedTreeItem() != null) {
            setSignature(getRelatedTreeItem().getSignature());
        } else {
            Ariane.getPresenter().printErrorMessage(ArianeSettings.ErrorCode.RUNTIME, getAbsoluteName(), getId(), ArianeSettings.getString("Operator.error7"));
        }
    }

    @Override
    public String toString() {
        StringBuilder commandLine = new StringBuilder("(" + getId() + "): " + getName());

        for (ParameterSlot isp : getInputParameterSlots()) {
            if (isp.getValue() != null) {
                commandLine.append(" ").append(isp.getValue());
            }
        }
        return commandLine.toString();
    }

    @Override
    public String getTagName() {
        return TAGNAME;
    }

    /////
    ///// E X E C U T I O N
    /////
    private String getRelativeFilePathName( String filename ) {
        String directory = UserPreferences.getTemporaryDirectory();
        if (filename.startsWith(directory)) {
            return filename.substring(directory.length());
        }
        return filename;
    }

    public static String getCommandPart( String binary ) {
        if (binary == null) {
            return null;
        }
        String[] format = binary.split(OperatorItem.SEPARATOR);
        if (format.length > 0) {
            if (format[0].startsWith("$FILE_DIRECTORY")) {
                String newBinary = format[0].substring("$FILE_DIRECTORY".length() + 1);
                String fileDirectory = UserPreferences.getPreferredBinaryDirectory();
                File file = new File(fileDirectory, newBinary.replaceAll("\\\\", "/"));
                if (fileDirectory == null) {
                    return file.getName();
                } else {
                    return file.getAbsolutePath();
                }
            } else {
                return format[0];
            }
        }
        return null;
    }

    public static String getFormatPart( String binary ) {
        if (binary != null) {
            int index = binary.indexOf(OperatorItem.SEPARATOR);
            if (index > 0 && index + 1 < binary.length()) {
                return binary.substring(index + 1);
            } else {
                return "[$p] [-m $m] [$i] [$o]";
            }
        }
        return null;
    }

    @Override
    public String getBinary() {
        return (getRelatedTreeItem() == null) ? super.getBinary() : getRelatedTreeItem().getBinaryPath();
    }

    /**
     * Builds the command line as specified in the command format and with the arguments.
     * The binary name of the program is returns with the absolute pathname.
     * Syntax: format string uses: $p for parameter; $p0 for the first parameter $m for the mask
     * $i for input; $i0 for the first input data $o for output; $o0 for the first output data.
     *
     * @param maskLine      the mask value.
     * @param parameterLine the parameter values.
     * @param outputLine    the output values.
     * @return the command line.
     */
    public List<String> buildAbstractCommandLine( List<String> maskLine, List<String> parameterLine, List<String> inputLine, List<String> outputLine ) {
        return ArgumentLineParser.parse(getCommandPart(getBinary()), getFormatPart(getBinary()), maskLine, parameterLine, inputLine, outputLine);
    }

    /**
     * Builds the command line as specified in the command format and with the arguments.
     * The binary name of the program is returns with the absolute pathname. %p for parameter;
     * %p0 for the first parameter %m for the mask %i for input; %i0 for the first input data file.
     * %o for output; %o0 for the first output data file.
     *
     * @param maskLine      the mask value.
     * @param parameterLine the parameter values.
     * @param outputLine    the output values.
     * @return the command line.
     */
    private List<String> buildCommandLine( List<String> maskLine, List<String> parameterLine, List<String> inputLine, List<String> outputLine ) {
        return ArgumentLineParser.parse(UserPreferences.getAbsoluteBinary(getCommandPart(getBinary())), getFormatPart(getBinary()), maskLine, parameterLine, inputLine, outputLine);
    }

    @Override
    protected boolean primitiveExecute() throws ArianeRuntimeException {
        try {
            if (getBinary() == null) {
                throw new ArianeRuntimeException(ArianeSettings.getString("Operator.error7"));
            }

            List<String> parameterLine = new LinkedList<>();
            if (!getParameters(parameterLine)) {
                return false;
            }

            List<String> maskLine = new LinkedList<>();
            getOptionalMask(maskLine);

            List<String> inputLine = new LinkedList<>();
            if (!getInputDataFile(inputLine)) {
                return false;
            }

            List<String> outputLine = new LinkedList<>();
            createOutputDatFiles(outputLine);

            List<String> commandLine = buildCommandLine(maskLine, parameterLine, inputLine, outputLine);
            StringBuilder consoleMessage = new StringBuilder(getName());
            for (int i = 1; i < commandLine.size(); i++) {
                consoleMessage.append(" ").append(getRelativeFilePathName(commandLine.get(i)));
            }
            Ariane.getPresenter().printSystemMessage(consoleMessage.toString());

            changeStatus(Status.RUNNING);
            executeBinary(commandLine);
            changeStatus(Status.EXECUTED);
            return true;
        } catch (ArianeException e) {
            throw displayError(e, e.getErrorCode());
        } catch (IOException e) {
            throw displayError(e, ArianeSettings.ErrorCode.INPUT_OUTPUT);
        } catch (InterruptedException e) {
            throw displayError(e, ArianeSettings.ErrorCode.INTERRUPTED_EXECUTION);
        }
    }

    private void createOutputDatFiles( List<String> outputLine ) {
        for (OutputDataSlot slot : getOutputDataSlots()) {
            outputLine.add(slot.getTempName());
        }
    }

    private void executeBinary( List<String> commandLine ) throws IOException, InterruptedException, ArianeRuntimeException {
        OperatorProcessBuilder processBuilder = new OperatorProcessBuilder(commandLine);
        long start = System.currentTimeMillis();
        Process proc1 = processBuilder.start();
        getGraph().setCurrentProcess(proc1);

        try (BufferedReader reader = new BufferedReader(new InputStreamReader(proc1.getInputStream()))) {
            String line;
            while ((line = reader.readLine()) != null) {
                Ariane.getPresenter().printOutputMessage(line);
            }
        }
        int exitValue = proc1.waitFor();
        getGraph().setCurrentProcess(null);

        if (exitValue == 0) {
            // Get the status value
            if (!UserPreferences.getCommandStatus().isEmpty()) {
                processBuilder = new OperatorProcessBuilder(UserPreferences.getStatusCommandPath());
                Process proc2 = processBuilder.start();
                InputStream resultStream = proc2.getInputStream();
                proc2.waitFor();

                // Exploit the results
                byte[] resultByte = new byte[resultStream.available() + 1];
                if (resultStream.read(resultByte) > 0) {
                    String resultString = new String(resultByte);
                    getOutputResultSlot().setValue(resultString.trim());
                }
            } else {
                getOutputResultSlot().setValue("SUCCESS");
            }
        }

        // I mean if an execution fails than the error code !=0.
        if (exitValue != 0) {
            throw new ArianeRuntimeException(ArianeSettings.getString("Operator.error2"));
        }
        // Everything is OK -> validate the result of all outputs.
        for (OutputDataSlot slot : getOutputDataSlots()) {
            slot.setValue();
        }
        long end = System.currentTimeMillis();
        Ariane.getPresenter().printSystemMessage("(" + (end - start) / 1000.0f + "s)");
    }

    private boolean getInputDataFile( List<String> inputLine ) throws ArianeMissingInputException {
        for (InputDataSlot slot : getInputDataSlots()) {
            if (slot.hasValue(true)) {
                inputLine.add(slot.getValue());
            } else {
                return false;
            }
        }
        return true;
    }

    private void getOptionalMask( List<String> maskLine ) throws ArianeRuntimeException {
        if (getInputMaskSlot().isConnected() && getInputMaskSlot().hasValue(false)) {
            String temp = getInputMaskSlot().getValue();
            maskLine.add(temp);
        }
    }

    private boolean getParameters( List<String> parameterLine ) throws ArianeMissingParameterException {
        // Get the parameter values
        for (ParameterSlot slot : getInputParameterSlots()) {
            if (slot.hasValue(true)) {
                parameterLine.add(slot.getValue());
            } else {
                return false;
            }
        }
        return true;
    }

    private ArianeRuntimeException displayError( Exception exception, ArianeSettings.ErrorCode inputoutput ) {
        if (GraphPresenter.isExecutionRunning()) {
            Ariane.getPresenter().printErrorMessage(inputoutput, getAbsoluteName(), getId(), exception.getLocalizedMessage());
            getGraph().stopExecution();
            changeStatus(Status.ERROR);
        } else {
            changeStatus(Status.UNEXECUTED);
        }
        return new ArianeRuntimeException(exception);
    }
}
