/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.element;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.view.graph.element.DataFileView;
import fr.greyc.pantheon.ariane.view.graph.element.SaverItemView;

import java.io.File;
import java.io.IOException;
import java.util.List;

/**
 * A <code>saverItem</code> manages saving of a data file in a graph.
 */
public final class SaverItem extends DataFile {
    public static final String TAGNAME = "Saver";

    public SaverItem( TreeItem relatedTreeItem, GraphPresenter graph ) {
        super(new SaverItemView(), relatedTreeItem, graph, relatedTreeItem.getName(), null);
        createSlots(1, 1, 0, 1, 1);
    }

    public SaverItem( GraphPresenter graph, List<XMLElement> data ) throws IOException {
        super(new SaverItemView(), graph);
        createSlots(0, 0, 0, 0, 1);
        createFromXML(data);
        if (getOutputDataSlots().isEmpty()) { // For the sake of compatibility
            createSlots(0, 0, 0, 1, 0);
        }
    }

    @Override
    public String getTagName() {
        return TAGNAME;
    }

    /*
     * Sets the name of the related physical file.
     * If the name of the file is specified as a relative path
     * than the absolute name is built with the directory of the project.
     */
    @Override
    public void setBinary( String binary ) {
        if (binary != null && !binary.trim().isEmpty()) {
            String currentDirectory = getGraph().getDirectoryName();
            super.setBinary(setPathToFilename(binary, currentDirectory));
        }
    }

    /*
     * Returns the name of the file with either
     * - relative path if the file is in the specified directory
     * - or with an absolute path.
     */
    @Override
    public String getBinary( String directory ) {
        return getRelativePath(super.getBinary(), directory, UserPreferences.getPreferredFileDirectory());
    }

    @Override
    protected boolean primitiveExecute() throws ArianeRuntimeException {
        try {
            setCurrentFilename(null);
            final InputDataSlot isi = getInputDataSlots().get(0);
            if (isi.hasValue(true)) {
                File file = getFilename();
                copyFile(isi.getValue(), file.getAbsolutePath());
                String exec = ArianeSettings.getString("ImageFile.save") + ": " + file.getAbsolutePath();
                Ariane.getPresenter().printSystemMessage(exec);
                final String filename = file.getAbsolutePath();
                setCurrentFilename(filename);
                getOutputDataSlots().get(0).setValue(filename);
                if (getOutputResultSlot() != null) { // Keep this test for sake of compatibility
                    getOutputResultSlot().setValue(filename);
                }
            } else {
                return false;
            }
        } catch (ArianeException e) {
            if (GraphPresenter.isExecutionRunning()) {
                Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), getId(), e.getLocalizedMessage());
                getGraph().stopExecution();
                changeStatus(Status.ERROR);
            }
            throw new ArianeRuntimeException(e);
        } catch (IOException e) {
            if (GraphPresenter.isExecutionRunning()) {
                Ariane.getPresenter().printErrorMessage(ArianeSettings.ErrorCode.INPUT_OUTPUT, getAbsoluteName(), getId(), ArianeSettings.getString("ImageFile.error6") + e.getLocalizedMessage());
                getGraph().stopExecution();
                changeStatus(Status.ERROR);
            }
            throw new ArianeRuntimeException(e);
        }
        // Validate the result.
        changeStatus(Status.EXECUTED);
        return true;
    }

    private File getFilename() throws ArianeMissingParameterException, ArianeMissingInputException {
        File file;
        ParameterSlot parameter = getInputParameterSlots().isEmpty() ? null : getInputParameterSlots().get(0);
        if (parameter != null && parameter.hasValue(false)) {
            ((DataFileView)getView()).changeBinaryName(getView().getNameEditor(), null, false);
            file = new File(parameter.getValue());
        } else if (getBinary() == null || getBinary().isEmpty()) {
            throw new ArianeMissingInputException(ArianeSettings.getString("ImageFile.error7"));
        } else {
            file = new File(getBinary());
        }
        return file;
    }

    @Override
    public void fireConnectionHook( InputSlot slot, boolean connection ) {
        if (!getInputParameterSlots().isEmpty() && slot == getInputParameterSlots().get(0)) {
            ((SaverItemView) getView()).setButtonEnabled(!connection);
        }
    }
}
