/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.element;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.PlatformSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingInputException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.element.etc.ArgumentLineParser;
import fr.greyc.pantheon.ariane.model.graph.element.etc.OperatorProcessBuilder;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.view.graph.element.SystemItemView;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.LinkedList;
import java.util.List;

/**
 * A <code>SystemItem</code> is a graph item that executes an operating system command.
 * It stores the command line and gets its parameters, inputs and outputs from slots.
 */
public final class SystemItem extends ControlItem {
    public static final String TAGNAME = "System";

    public SystemItem( TreeItem treeItem, GraphPresenter graph ) throws InterruptedException {
        super(new SystemItemView(), treeItem, graph, treeItem.getName());
        askAndSetPrototype(null, 1, 1, 1);
        super.setBinary("");
    }

    public SystemItem( GraphPresenter graph, List<XMLElement> elements ) throws IOException {
        super(new SystemItemView(), graph);
        createFromXML(elements);
    }

    /*
     * For tests
     */
    public SystemItem( TreeItem treeItem, GraphPresenter graph, int inputs, int parameters, int outputs ) {
        super(new SystemItemView(), treeItem, graph, treeItem.getName());
        setPrototype(treeItem.getName(), inputs, parameters, outputs);
        super.setBinary("");
    }

    @Override
    protected void setPrototype( String name, int inputImageNumber, int inputParameterNumber, int outputImageNumber ) {
        setAbstractName(name);
        createSlots(inputImageNumber, inputParameterNumber, 0, outputImageNumber, 1);
    }

    /////
    ///// A C C E S S O R S
    /////
    @Override
    public String getTagName() {
        return TAGNAME;
    }

    /////
    ///// E X E C U T I O N
    /////

    /*
     * Builds the command line as specified in the command format
     * and with the arguments. The binary name of the program
     * is returns with the absolute pathname.
     * Syntax: format string uses:
     * $p for parameter; $p0 for the first parameter
     * $m for the mask
     * $i for input; $i0 for the first input image
     * $o for output; $o0 for the first output image.
     */
    public List<String> buildAbstractCommandLine( List<String> maskLine, List<String> parameterLine, List<String> inputLine, List<String> outputLine ) {
        return ArgumentLineParser.parse("", getBinary(), maskLine, parameterLine, inputLine, outputLine);
    }

    @Override
    protected boolean primitiveExecute() throws ArianeRuntimeException {
        try {
            List<String> parameterLine = new LinkedList<>();
            if (!getParameterValues(parameterLine)) {
                return false;
            }

            List<String> inputLine = new LinkedList<>();
            if (!getInputData(inputLine)) {
                return false;
            }

            List<String> outputLine = new LinkedList<>();
            createOutputDataAsTemporaryFiles(outputLine);

            List<String> maskLine = new LinkedList<>();
            List<String> commandLine = ArgumentLineParser.parse("", getBinary(), maskLine, parameterLine, inputLine, outputLine);

            StringBuilder consoleMessage = new StringBuilder();
            for (String aCommandLine : commandLine) {
                consoleMessage.append(" ").append(aCommandLine);
            }
            Ariane.getPresenter().printSystemMessage(consoleMessage.toString());

            commandLine.clear();
            if (PlatformSettings.isWindows()) {
                commandLine.add("cmd");
                commandLine.add("/C");
                commandLine.add(consoleMessage.toString());
            } else {
                String shell = System.getenv("SHELL");
                if (shell.length() == 0) {
                    shell = "/bin/sh";
                }
                commandLine.add(shell);
                commandLine.add("-c");
                commandLine.add(consoleMessage.toString());
            }

            // Execution of the operator binary
            changeStatus(Status.RUNNING);
            OperatorProcessBuilder pb = new OperatorProcessBuilder(commandLine);
            Process process1 = pb.start();
            getGraph().setCurrentProcess(process1);
            try (BufferedReader br = new BufferedReader(new InputStreamReader(process1.getInputStream()))) {
                String line;
                while ((line = br.readLine()) != null) {
                    Ariane.getPresenter().printOutputMessage(line);
                }
            }
            int exitValue = process1.waitFor();
            getGraph().setCurrentProcess(null);

            if (exitValue == 0) {
                getOutputResultSlot().setValue("SUCCESS");
            } else {
                throw new ArianeRuntimeException(ArianeSettings.getString("Operator.error2"));
            }
            // Everything is OK -> validate the result of all outputs.
            for (OutputDataSlot slot : getOutputDataSlots()) {
                slot.setValue();
            }
            changeStatus(Status.EXECUTED);
            return true;
        } catch (ArianeException e) {
            errorMessage(e, e.getErrorCode());
            throw new ArianeRuntimeException(e);
        } catch (IOException e) {
            errorMessage(e, ArianeSettings.ErrorCode.INPUT_OUTPUT);
            throw new ArianeRuntimeException(e);
        } catch (InterruptedException e) {
            errorMessage(e, ArianeSettings.ErrorCode.INTERRUPTED_EXECUTION);
            throw new ArianeRuntimeException(e);
        }
    }

    private void errorMessage( Exception error, ArianeSettings.ErrorCode errorCode ) {
        if (GraphPresenter.isExecutionRunning()) {
            Ariane.getPresenter().printErrorMessage(errorCode, getAbsoluteName(), getId(), error.getLocalizedMessage());
            getGraph().stopExecution();
            changeStatus(Status.ERROR);
        } else {
            changeStatus(Status.UNEXECUTED);
        }
    }

    private void createOutputDataAsTemporaryFiles( List<String> outputLine ) {
        for (OutputDataSlot slot : getOutputDataSlots()) {
            outputLine.add(slot.getTempName());
        }
    }

    private boolean getInputData( List<String> inputLine ) throws ArianeMissingInputException {
        for (InputDataSlot slot : getInputDataSlots()) {
            if (slot.hasValue(true)) {
                inputLine.add(slot.getValue());
            } else {
                return false;
            }
        }
        return true;
    }

    private boolean getParameterValues( List<String> parameterLine ) throws ArianeMissingParameterException {
        // Get the parameter values
        for (ParameterSlot slot : getInputParameterSlots()) {
            if (slot.hasValue(true)) {
                parameterLine.add(slot.getValue());
            } else {
                return false;
            }
        }
        return true;
    }
}
