/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.element;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeArithmeticException;
import fr.greyc.pantheon.ariane.model.exception.ArianeException;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.element.etc.LogicalExpression;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.presenter.SubGraphPresenter;

import java.io.IOException;
import java.util.List;

/**
 * A <code>WhileControlStructureItem</code> implements the repetitive structure
 * control by a graph item. The sequence of Items to has be repeated is enclosed
 * inside a subgraph.
 */
public final class WhileControlStructureItem extends CompositeItem {
    public static final String TAGNAME = "While";
    private static final String ABSTRACT_NAME = "while";

    // From item tree
    public WhileControlStructureItem( TreeItem treeItem, GraphPresenter graph ) throws InterruptedException {
        super(treeItem, graph, ABSTRACT_NAME);
        askAndSetPrototype(ABSTRACT_NAME + super.getId(), 1, 1, 1);
        initComment();
        setSignature(getRelatedTreeItem().getSignature());
    }

    // From file
    public WhileControlStructureItem( GraphPresenter graph, List<XMLElement> xmlTree ) throws IOException {
        super(graph);
        createFromXML(xmlTree);
        String expression;
        if (getFormula() == null && super.getBinary() != null) { // Keep this for sake of compatibility
            expression = super.getBinary();
            super.setBinary(null);
        } else {
            expression = getFormula();
        }

        try {
            checkExpression(expression);
            super.setFormula(expression);
        } catch (ArianeArithmeticException e) {
            Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), super.getId(), e.getLocalizedMessage());
            super.setFormula("");
        }
        if (getRelatedTreeItem() == null) {
            setRelatedTreeItem(restoreDefaultTreeItem());
        }
        initComment();
        setSignature(getRelatedTreeItem().getSignature());
    }

    // From subroutine
    public WhileControlStructureItem( GraphPresenter graph, List<XMLElement> xmlTree, TreeItem treeItem ) throws IOException {
        super(graph, treeItem);
        createFromXML(xmlTree);
        String expression;
        if (getFormula() == null && super.getBinary() != null) { // Keep this for sake of compatibility
            expression = super.getBinary();
            super.setBinary(null);
        } else {
            expression = getFormula();
        }
        try {
            checkExpression(expression);
            super.setFormula(expression);
        } catch (ArianeArithmeticException e) {
            Ariane.getPresenter().printErrorMessage(e.getErrorCode(), getAbsoluteName(), super.getId(), e.getLocalizedMessage());
            super.setFormula("");
        }
        setAbstractName(treeItem.getName());
        setNickName(treeItem.getName());
        initComment();
        setSignature(getRelatedTreeItem().getSignature());
    }

    // For tests
    WhileControlStructureItem( TreeItem treeItem, GraphPresenter graph, int inputDataSlotNumber, int inputParameterSlotNumber, int outputDataSlotNumber ) { // For test
        super(treeItem, graph, ABSTRACT_NAME);
        createSlots(inputDataSlotNumber, inputParameterSlotNumber + getNumberOfFixedParameters() + 1, 0, outputDataSlotNumber, 2);
        initComment();
        setSignature(getRelatedTreeItem().getSignature());
    }

    @Override
    protected int getNumberOfFixedParameters() {
        return 2;
    }

    @Override
    protected void setPrototype( String name, int inputDataCount, int parameterCount, int outputDataCount ) {
        super.setPrototype(name, inputDataCount, parameterCount + getNumberOfFixedParameters(), outputDataCount, inputDataCount, 2);
        setFormula(""); // IMPERATIVE to signify that it uses formula.
    }

    @Override
    protected SubGraphPresenter createSubGraph( String name, int inputDataCount, int parameterCount, int outputDataCount ) {
        return createSubGraph(getNickName(), inputDataCount, parameterCount, outputDataCount, inputDataCount, 2);
    }

    @Override
    public String getTagName() {
        return TAGNAME;
    }

    @Override
    public TreeItem restoreDefaultTreeItem() {
        setAbstractName(ABSTRACT_NAME);
        return Ariane.getPresenter().getTree().getTreeItem(ABSTRACT_NAME, null);
    }

    /////
    ///// E X E C U T I O N
    /////
    @Override
    public void resetHookHook() {
        initComment();
    }

    @Override
    protected boolean linkItemWithVirtualItem( VirtualItem virtualItem ) throws ArianeException {
        // 1. Get the parameter values
        for (int i = 0; i < getInputParameterSlots().size(); i++) {
            ParameterSlot parameter = getInputParameterSlots().get(i);
            if (parameter.hasValue(true)) {
                OutputResultSlot result = virtualItem.getVirtualInputParameterSlots().get(i);
                result.setValue(parameter.getValue());
            } else {
                return false;
            }
        }

        getMask(virtualItem);

        return getInputDataFiles(virtualItem);
    }

    @Override
    protected boolean primitiveExecute() throws ArianeRuntimeException {
        try {
            if (getFormula().isEmpty()) {
                throw new ArianeArithmeticException(ArianeSettings.getString("arithmetic.error0"));
            }

            VirtualItem virtualItem = getVirtualItem();
            if (!linkItemWithVirtualItem(virtualItem)) {
                return false;
            }

            double[] whileVariables = new double[2];
            if (!getLoopParameters(whileVariables)) {
                return false;
            }

            changeStatus(Status.RUNNING);

            // Repetition
            int iteration = 0;
            LogicalExpression condition = getCompiledExpression();

            Ariane.getPresenter().printSystemMessage("> While loop (" + condition.toHtml() + ") do");
            while (condition.evaluate(whileVariables[0], whileVariables[1]) && GraphPresenter.isExecutionRunning()) {
                iteration++;

                reset();
                virtualItem.changeStatus(Status.EXECUTED); // Prevent the re-execution of the virtual operator.

                Ariane.getPresenter().printSystemMessage("> cycle " + iteration);
                setComment(Integer.toString(iteration));
                changeStatus(Status.RUNNING);

                // Set the value of the 2 inner parameters.
                for (int i = 0; i < 2; i++) {
                    virtualItem.getVirtualInputParameterSlot(i).setValue(Float.toString((float) whileVariables[i]));
                }

                // The body execution is split in 2 steps
                // to avoid problems when two operators use the same entry.
                executeBody(virtualItem);
                setIntermediateInputs(virtualItem);

                setLoopParameterValues(virtualItem, whileVariables);
            }
            Ariane.getPresenter().printSystemMessage("end while loop");

            setOutputData(virtualItem);

            if (!setOutputResult(virtualItem)) {
                return false;
            }
            changeStatus(Status.EXECUTED);
            return true;
        } catch (ArianeException e) {
            throw displayError(e, e.getErrorCode());
        } catch (IOException e) {
            throw displayError(e, ArianeSettings.ErrorCode.INPUT_OUTPUT);
        }
    }

    private boolean getLoopParameters( double[] whileVariables )
            throws ArianeMissingParameterException {
        for (int i = 0; i < 2; i++) {
            ParameterSlot parameter = getInputParameterSlots().get(i);
            if (parameter.hasNumericalValue(true)) {
                whileVariables[i] = (int) parameter.getNumericalValue();
            } else {
                return false;
            }
        }
        return true;
    }

    private void setLoopParameterValues( VirtualItem virtualItem, double[] whileVariables ) throws ArianeMissingParameterException {
        for (int i = 0; i < 2; i++) {
            ParameterSlot parameter = virtualItem.getVirtualIntermediateResultSlot(i);
            if (parameter.isConnected() && parameter.hasNumericalValue(true)) {
                whileVariables[i] = (int) parameter.getNumericalValue();
            }
        }
    }

    /////
    ///// P R O P E R T I E S
    /////
    @Override
    public boolean changeFormula( String condition ) {
        if (getFormula() != null && getFormula().equals(condition)) {
            return true;
        }
        try {
            checkExpression(condition);
            setFormula(condition);
            return true;
        } catch (ArianeArithmeticException e) {
            Ariane.getPresenter().printErrorMessage(e.getLocalizedMessage());
        }
        return false;
    }

    private LogicalExpression getCompiledExpression() {
        try {
            return new LogicalExpression().compile(getFormula());
        } catch (ArianeArithmeticException e) {
            return null;
        }
    }

    private void checkExpression( String textFormula ) throws ArianeArithmeticException {
        new LogicalExpression().compile(textFormula);
    }
}
