/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.slot;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.element.GraphItemModel;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.SlotPresenter;
import fr.greyc.pantheon.ariane.view.graph.slot.InputMaskSlotView;
import fr.greyc.pantheon.ariane.view.graph.slot.LinkView;

/**
 * The class <code>InputMaskSlot</code> defines an input mask connection slot.
 * It can only be connected to an <code>OutputDataSlot</code>.
 */
public final class InputMaskSlot extends InputSlot {

    public InputMaskSlot( GraphItemModel item ) {
        super(new SlotPresenter(new InputMaskSlotView()), item, 0);
    }

    @Override
    protected LinkModel createLink() {
        return new MaskLinkModel(this, new LinkView());
    }

    @Override
    public String toString() {
        return "InputMaskSlot";
    }

    /////
    /////  R E S U L T
    /////

    /**
     * Tries to get the data of the connected slot. Result exists if:
     * <ul>
     * <li>the slot is connected;
     * <li>the result of the connected slot is validate.
     * </ul>
     *
     * @param isEmptyFatalError is an empty value a fatal error?
     * @return true if the result is validate.
     */
    public boolean hasValue( boolean isEmptyFatalError ) throws ArianeRuntimeException {
        if (isConnected()) {
            return getConnectedSlot().hasValue();
        } else if (isEmptyFatalError) {
            throw new ArianeRuntimeException(ArianeSettings.getString("Operator.error3") + getRank());
        }
        return false;
    }

    public String getValue() {
        return ((OutputDataSlot) getConnectedSlot()).getValue();
    }

    @Override
    public boolean isAcceptableDropSite( SlotModel slot ) {
        return slot.getGraphItem() != this.getGraphItem() && slot instanceof OutputDataSlot;
    }

    @Override
    public XMLElement convertToXML() {
        if (isConnected()) {
            XMLElement element = new XMLElement("Connection");
            element.addChildren(new XMLElement("EndItemId", Integer.toString(getGraphItem().getId())));
            element.addChildren(new XMLElement("EndSlotType", "InSlotMask"));
            element.addChildren(new XMLElement("EndSlotIndex", Integer.toString(getRank())));
            getConnectedSlot().convertToXML(element);
            if (!getLink().getPath().isEmpty()) {
                element.addChildren(new XMLElement("UserDefined", Boolean.toString(getLink().isUserDefined())));
                XMLElement pa = new XMLElement("LinkPath");
                for (int i = 0; i < getLink().getPath().size(); i++) {
                    XMLElement po = new XMLElement("LinkPoint");
                    po.addChildren(new XMLElement("IdPoint", Integer.toString(i)));
                    po.addChildren(new XMLElement("X-coordinate", Integer.toString((int) getLink().getPath().get(i).getX())));
                    po.addChildren(new XMLElement("Y-coordinate", Integer.toString((int) getLink().getPath().get(i).getY())));
                    pa.addChildren(po);
                }
                element.addChildren(pa);
            }
            return element;
        }
        return null;
    }
}
