/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.slot;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeSlotException;
import fr.greyc.pantheon.ariane.model.graph.element.GraphItemModel;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.SlotPresenter;
import fr.greyc.pantheon.ariane.view.command.ConnectItemEdit;
import fr.greyc.pantheon.ariane.view.command.DisconnectItemEdit;

import java.util.ArrayList;
import java.util.List;

/**
 * The class <code>InputSlot</code> defines input connection slot.
 * Only one <code>OutputSlot</code> can be connected to one <code>InputSlot</code>.
 * It is a drop site for an <code>OutputSlot</code>.
 * Three types of input slots are available:
 * <ul>
 * <li>@see InputdataSlot
 * <li>@see InputMaskSlot
 * <li>@see InputParameterSlot
 * </ul>
 */
public abstract class InputSlot extends SlotModel {
    private transient OutputSlot _connectedSlot;
    private transient LinkModel _link;

    InputSlot( SlotPresenter slotPresenter, GraphItemModel item, int rank ) {
        super(slotPresenter, item, rank);
    }

    public final LinkModel getLink() {
        return _link;
    }

    final void setLink( LinkModel link ) {
        _link = link;
    }

    /////
    ///// C O N N E C T I O N   /   D I S C O N N E C T I O N
    /////
    @Override
    public final void connectWith( SlotModel from ) {
        try {
            OutputSlot outputSlot = (OutputSlot) from;
            // check if the slot are in the same graph!
            if (getGraph() != outputSlot.getGraph()) {
                throw new ArianeSlotException(ArianeSettings.getString("Slot.error3"));
            }
            // check slots consistency.
            boolean isConnectable = this instanceof ParameterSlot && outputSlot instanceof OutputResultSlot
                    || this instanceof InputDataSlot && outputSlot instanceof OutputDataSlot
                    || this instanceof InputMaskSlot && outputSlot instanceof OutputDataSlot;
            if (isConnectable) {
                getGraph().startActionEdition();
                outputSlot.connectTo(this, null, true);
                getGraph().endActionEdition();
            } else {
                throw new ArianeSlotException(ArianeSettings.getString("Slot.error2"));
            }
        } catch (ArianeSlotException e) {
            Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Slot.error1") + e.getMessage());
        } catch (ClassCastException e) {
            Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Slot.error1") + ArianeSettings.getString("Slot.error2"));
        }
    }

    final void connectFrom( OutputSlot outputSlot ) {
        if (!outputSlot.equals(_connectedSlot)) {
            disconnect();
            outputSlot.connectTo(this, null, true);
        }
    }

    public final OutputSlot getConnectedSlot() {
        return _connectedSlot;
    }

    @Override
    public final List<? extends SlotModel> getConnectedSlots() {
        List<SlotModel> slots = new ArrayList<>();
        slots.add(_connectedSlot);
        return slots;
    }

    @Override
    public final boolean isConnected() {
        return _connectedSlot != null;
    }

    final void setConnectedSlot( OutputSlot slot, LinkModel link, boolean isAddedToHistory ) {
        getGraphItem().resetAll();
        getGraphItem().getGraph().setModified(true);
        boolean isDisconnected = slot == null;
        if (isDisconnected) {
            getView().setDisconnectedRendering();
            if (isAddedToHistory) {
                getGraph().validateAction(new DisconnectItemEdit(getGraph(), this, _connectedSlot));
            }
            _link.dispose();
            getGraphItem().fireConnection(this, false);
        } else {
            getView().setConnectedRendering();
            if (link != null) {
                _link = link;
            } else {
                _link = createLink();
            }
            if (isAddedToHistory) {
                getGraph().validateAction(new ConnectItemEdit(getGraph(), this, slot));
            }
            getGraphItem().fireConnection(this, true);
        }
        _connectedSlot = slot;
    }

    public final void setConnectedSlot( OutputSlot slot ) { // For tests
        _connectedSlot = slot;
    }

    @Override
    public final void disconnectAll() {
        disconnect();
    }

    public final void disconnect() {
        if (_connectedSlot != null) {
            _connectedSlot.disconnectFrom(this, true);
            disposeLink();
        }
    }

    protected abstract LinkModel createLink();

    public void disposeLink() {
      /*
       * empty
       */
    }

    public abstract XMLElement convertToXML();
}
