/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.slot;

import fr.greyc.pantheon.ariane.view.graph.slot.LinkView;

import java.awt.Point;
import java.awt.Shape;
import java.awt.geom.Line2D;
import java.awt.geom.Path2D;
import java.util.ArrayList;
import java.util.List;

/**
 * A <code>Link</code> is a connection between two slots.
 * It is represented as a polyline, namely a series of straight lines.
 */
public abstract class LinkModel {
    static final int GAP = 10;
    private static final int MIN_DISTANCE_BETWEEN_LINK_AND_SLOT = 10;
    private final InputSlot _targetSlot;
    private final LinkView _linkView;
    private Line2D.Float[] _polyline;
    private int _selectedLinkPartNumber = -1;
    private boolean _userDefined;
    private boolean _selected;

    LinkModel( InputSlot targetSlot, LinkView linkView ) {
        _targetSlot = targetSlot;
        _linkView = linkView;
        linkView.setModel(this);
    }

    public LinkView getView() {
        return _linkView;
    }

    final void dispose() {
        _targetSlot.getPresenter().setSelected(false);
    }

    final Line2D.Float[] getPolyline() {
        return _polyline.clone();
    }

    final void setPath( List<Point> path, boolean isUserDefined ) {
        if (path.isEmpty()) {
            _userDefined = false;
        } else {
            _userDefined = isUserDefined;
            buildPolyline(path);
        }
    }

    /**
     * Returns the current link as a series of intermediate points if this line is reshaped by the user.
     * If the link is the default link, namely not reshaped by the user,
     * the method returns an empty array of size 0.
     *
     * @return the current link as a series of intermediate points.
     */
    final List<Point> getPath() {
        ArrayList<Point> path = new ArrayList<>();
        if (_userDefined) {
            for (Line2D.Float line : _polyline) {
                path.add(new Point((int) line.x1, (int) line.y1));
            }
            path.add(new Point((int) _polyline[_polyline.length - 1].x2, (int) _polyline[_polyline.length - 1].y2));
        }
        return path;
    }

    final boolean isUserDefined() {
        return _userDefined;
    }

    final void setUserDefined() {
        _userDefined = true;
    }

    private boolean hasMoved() {
        int dx1 = (int) (_targetSlot.getConnectedSlot().getView().getAnchorPoint().getX() - _targetSlot.getView().getAnchorPoint().getX());
        int dx2 = (int) (_polyline[0].x1 - _polyline[_polyline.length - 1].x2);
        int dy1 = (int) (_targetSlot.getConnectedSlot().getView().getAnchorPoint().getY() - _targetSlot.getView().getAnchorPoint().getY());
        int dy2 = (int) (_polyline[0].y1 - _polyline[_polyline.length - 1].y2);
        return dx1 == dx2 && dy1 == dy2;
    }

    private boolean isModified() {
        return (int) _polyline[0].x1 != (int) _targetSlot.getConnectedSlot().getView().getAnchorPoint().getX()
                || (int) _polyline[0].y1 != (int) _targetSlot.getConnectedSlot().getView().getAnchorPoint().getY()
                || (int) _polyline[_polyline.length - 1].x2 != (int) _targetSlot.getView().getAnchorPoint().getX()
                || (int) _polyline[_polyline.length - 1].y2 != (int) _targetSlot.getView().getAnchorPoint().getY();
    }

    private void movePolyline() {
        final double dx = _targetSlot.getConnectedSlot().getView().getAnchorPoint().getX() - _polyline[0].x1;
        final double dy = _targetSlot.getConnectedSlot().getView().getAnchorPoint().getY() - _polyline[0].y1;
        for (Line2D.Float line : _polyline) {
            line.x1 += dx;
            line.x2 += dx;
            line.y1 += dy;
            line.y2 += dy;
        }
    }

    /**
     * Modifies the link because one the source or target item has been moved.
     *
     * @return false if the link cannot be modified because it becomes too complex.
     */
    protected abstract boolean modifyPolyline();

    public void buildPolyline() {
        if (_polyline == null) {
            buildPolyline(createPolyline());
        } else if (isModified()) {
            if (hasMoved()) {
                movePolyline();
            } else if (!_userDefined || !modifyPolyline()) {
                buildPolyline(createPolyline());
                _userDefined = false;
            }
        }
    }

    private void buildPolyline( List<Point> path ) {
        _polyline = new Line2D.Float[path.size() - 1];
        // create the lines from the intermediate points
        for (int i = 0; i < path.size() - 1; i++) {
            final Point point1 = path.get(i);
            final Point point2 = path.get(i + 1);
            _polyline[i] = new Line2D.Float((float) point1.getX(), (float) point1.getY(), (float) point2.getX(), (float) point2.getY());
        }
    }

    /**
     * Builds the polyline as a series of straight lines.
     */
    protected abstract List<Point> createPolyline();

    private int getLineRankAtSpecifiedCoordinates( Point coordinates ) {
        int rank = 0;
        double distance = Double.MAX_VALUE;
        for (int i = 0; i < _polyline.length; i++) {
            if (_polyline[i].ptSegDist(coordinates.x, coordinates.y) < distance) {
                distance = _polyline[i].ptSegDist(coordinates.x, coordinates.y);
                rank = i;
            }
        }
        return rank;
    }

    public final boolean isCloseTo( double x, double y ) {
        if (_polyline != null) {
            Shape shape = buildLinkShape();
            return shape.intersects(x - MIN_DISTANCE_BETWEEN_LINK_AND_SLOT, y - MIN_DISTANCE_BETWEEN_LINK_AND_SLOT, 2 * MIN_DISTANCE_BETWEEN_LINK_AND_SLOT, 2 * MIN_DISTANCE_BETWEEN_LINK_AND_SLOT) && !shape.contains(x, y);
        }
        return false;
    }

    public Path2D buildLinkShape() {
        if (_polyline.length == 1) {
            Path2D line = new Path2D.Double();
            line.moveTo(_polyline[0].getP1().getX(), _polyline[0].getP1().getY());
            line.lineTo(_polyline[_polyline.length - 1].getP2().getX(), _polyline[_polyline.length - 1].getP2().getY());
            return line;
        } else if (_polyline.length == 3) {
            Path2D path = new Path2D.Double();
            path.moveTo(_polyline[0].getP1().getX(), _polyline[0].getP1().getY());
            path.curveTo(_polyline[1].getP1().getX(), _polyline[1].getP1().getY(),
                    _polyline[2].getP1().getX(), _polyline[2].getP1().getY(),
                    _polyline[2].getP2().getX(), _polyline[2].getP2().getY());
            return path;
        } else if (_polyline.length == 5) {
            Path2D path = new Path2D.Double();
            path.moveTo(_polyline[0].getP1().getX(), _polyline[0].getP1().getY());
            path.curveTo(_polyline[1].getP1().getX(), _polyline[1].getP1().getY(),
                    _polyline[1].getP2().getX(), _polyline[1].getP2().getY(),
                    (_polyline[2].getP1().getX() + _polyline[2].getP2().getX()) / 2, (_polyline[2].getP1().getY() + _polyline[2].getP2().getY()) / 2);
            path.curveTo(_polyline[3].getP1().getX(), _polyline[3].getP1().getY(),
                    _polyline[3].getP2().getX(), _polyline[3].getP2().getY(),
                    _polyline[4].getP2().getX(), _polyline[4].getP2().getY());
            return path;
        } else {
            Path2D path = new Path2D.Double();
            path.moveTo(_polyline[0].getP1().getX(), _polyline[0].getP1().getY());
            path.curveTo((_polyline[1].getP1().getX() + _polyline[1].getP2().getX()) / 2, (_polyline[1].getP1().getY() + _polyline[1].getP2().getY()) / 2,
                    (_polyline[2].getP1().getX() + _polyline[2].getP2().getX()) / 2, (_polyline[2].getP1().getY() + _polyline[2].getP2().getY()) / 2,
                    _polyline[3].getP2().getX(), _polyline[3].getP2().getY());
            return path;
        }
    }

    public final void setSelected( boolean selected, Point xycoord ) {
        if (selected) {
            _selectedLinkPartNumber = getLineRankAtSpecifiedCoordinates(xycoord);
            if (_selectedLinkPartNumber == 0) {
                _selectedLinkPartNumber = 1;
            } else if (_selectedLinkPartNumber == _polyline.length - 1) {
                _selectedLinkPartNumber = _polyline.length - 2;
            }
        } else {
            _selectedLinkPartNumber = -1;
        }
        _selected = selected;
        _targetSlot.getPresenter().setSelected(selected);
    }

    /**
     * Moves the current selected intermediate line of the link according to the new coordinates.
     *
     * @param xycoord the new coordinates.
     */
    public abstract void move( Point xycoord );

    public final int getSelectedLineId() {
        return _selectedLinkPartNumber;
    }

    public final void move( int lineRank, Point xycoord ) {
        _selectedLinkPartNumber = lineRank;
        move(xycoord);
        _selectedLinkPartNumber = -1;
    }

    final Point getSourceCoordinate() {
        return _targetSlot.getConnectedSlot().getView().getAnchorPoint();
    }

    final Point getTargetCoordinate() {
        return _targetSlot.getView().getAnchorPoint();
    }

    public Line2D.Float getSelectedLinkPart() {
        if (_selectedLinkPartNumber > 0) {
            return _polyline[_selectedLinkPartNumber];
        } else {
            return null;
        }
    }

    public boolean isSelected() {
        return _selected;
    }
}
