/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.slot;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeSlotException;
import fr.greyc.pantheon.ariane.model.graph.element.GraphItemModel;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.SlotPresenter;

import java.util.ArrayList;
import java.util.List;

/**
 * The class <code>OutputSlot</code> defines an output connection slot.
 * Several <code>InputSlot</code> can be connected to an output slot.
 * Two types of output slots are available:
 * <ul>
 * <li>@see OutputDataSlot
 * <li>@see outputResultSlot
 * </ul>
 */
public abstract class OutputSlot extends SlotModel {
    private final transient List<InputSlot> _connectedSlots = new ArrayList<>();
    private boolean _validate;

    OutputSlot( SlotPresenter slotPresenter, GraphItemModel item, int rank ) {
        super(slotPresenter, item, rank);
    }

    /////
    ///// C O N N E C T I O N   /   D I S C O N N E C T I O N
    /////
    @Override
    public final void connectWith( SlotModel from ) {
        try {
            InputSlot inputSlot = (InputSlot) from;
            // check if the slot are in the same graph!
            if (getGraph() != inputSlot.getGraph()) {
                throw new ArianeSlotException(ArianeSettings.getString("Slot.error3"));
            }
            // check slots consistency.
            boolean connect = this instanceof OutputDataSlot && inputSlot instanceof InputDataSlot
                    || this instanceof OutputDataSlot && inputSlot instanceof InputMaskSlot
                    || this instanceof OutputResultSlot && inputSlot instanceof ParameterSlot;
            if (connect) {
                getGraph().startActionEdition();
                inputSlot.connectFrom(this);
                getGraph().endActionEdition();
            } else {
                throw new ArianeSlotException(ArianeSettings.getString("Slot.error2"));
            }
        } catch (ArianeSlotException e) {
            Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Slot.error1") + e.getMessage());
        } catch (ClassCastException e) {
            Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Slot.error1") + ArianeSettings.getString("Slot.error2"));
        }
    }

    /**
     * Connects the given slot with the slot.
     * <br>
     * If the given slot is already connected to the slot does nothing. Else, disconnects the given slot if needed and adds the slot to the list of connected slots.
     *
     * @param toSlot           the input slot to be connected.
     * @param link             an existing link or null
     * @param isAddedToHistory true if the action must be stored in the action history (for use with undo/redo).
     */
    public final void connectTo( InputSlot toSlot, LinkModel link, boolean isAddedToHistory ) {
        if (!_connectedSlots.contains(toSlot)) {
            toSlot.disconnectAll();
            _connectedSlots.add(toSlot);
            getView().setConnectedRendering();
            toSlot.setConnectedSlot(this, link, isAddedToHistory);
        }
    }

    @Override
    public final List<? extends SlotModel> getConnectedSlots() {
        return _connectedSlots;
    }

    public final List<InputSlot> getConnectedInputSlots() {
        return _connectedSlots;
    }

    @Override
    public final boolean isConnected() {
        return !_connectedSlots.isEmpty();
    }

    public final void disconnectFrom( InputSlot inputSlot, boolean isAddedToHistory ) {
        inputSlot.setConnectedSlot(null, null, isAddedToHistory);
        inputSlot.disposeLink();
        if (_connectedSlots.remove(inputSlot) && _connectedSlots.isEmpty()) {
            getView().setDisconnectedRendering();
        }
    }

    /*
     * Implementation: calls the disconnection method of each slot
     * since disconnection remove slot from the list.
     */
    @Override
    public final void disconnectAll() {
        while (!_connectedSlots.isEmpty()) {
            disconnectFrom(_connectedSlots.get(0), true);
        }
    }

    public final boolean isValidate() {
        return _validate;
    }

    final void setValidated() {
        _validate = true;
    }

    public final void reset() {
        _validate = false;
    }

    final boolean hasValue() {
        return getGraphItem().execute() && isValidate();
    }

    protected abstract void convertToXML( XMLElement element );
}
