/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.slot;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.graph.element.GraphItemModel;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.SlotPresenter;
import fr.greyc.pantheon.ariane.view.command.ChangeParameterValueEdit;
import fr.greyc.pantheon.ariane.view.graph.slot.InputParameterSlotView;
import fr.greyc.pantheon.ariane.view.graph.slot.LinkView;

/**
 * The class <code>InputParameterSlot</code> defines an input parameter connection slot.
 * An input parameter slot can be connected to a OutputResultSlot or can hold its own value
 * (either float or String value).
 */
public final class ParameterSlot extends InputSlot {
    private String _value;
    private String _name;

    public ParameterSlot( GraphItemModel item, int rank ) {
        super(new SlotPresenter(new InputParameterSlotView()), item, rank);
    }

    @Override
    protected LinkModel createLink() {
        ((InputParameterSlotView) getView()).setEnabledTextfield(false);
        return new ValueLinkModel(this, new LinkView());
    }

    @Override
    public void disposeLink() {
        _value = null;
        ((InputParameterSlotView) getView()).setEnabledTextfield(true);
    }

    @Override
    public String toString() {
        return "Parameter $" + getRank();
    }

    /////
    /////  V A L U E
    /////

    /**
     * Tests if the parameter has a validate value. If it is connected, tests if the connected slot has a validate value.
     *
     * @param isEmptyFatalError is an empty value a fatal error?
     * @return true if the parameter has a validate value.
     */
    public boolean hasValue( boolean isEmptyFatalError ) throws ArianeMissingParameterException {
        if (isConnected()) {
            // Gets the value of the connected slot.
            OutputResultSlot outputSlot = (OutputResultSlot) getConnectedSlot();
            if (outputSlot.hasValue()) {
                setValue(outputSlot.getValue());
            } else {
                return false;
            }
        } else if (isEmptyFatalError && _value == null) {
            throw new ArianeMissingParameterException(ArianeSettings.getString("Operator.error1") + getRank());
        }
        return _value != null;
    }

    /**
     * Tests if the parameter has a float value and is validated. If it is connected tests if the connected slot has a float value and is validated.
     *
     * @param isEmptyFatalError is an empty value a fatal error?
     * @return true if the parameter has a validate value and is a float.
     */
    public boolean hasNumericalValue( boolean isEmptyFatalError ) throws ArianeMissingParameterException {
        if (hasValue(isEmptyFatalError)) {
            try {
                Float.valueOf(getValue());
                return true;
            } catch (NumberFormatException e) {
                return false;
            }
        } else {
            return false;
        }
    }

    public void setName( String name ) {
        _name = name;
    }

    public String getName() {
        if (_name == null) {
            return ArianeSettings.getString("PropertyViewer.parameter") + " $" + getRank();
        } else {
            return _name;
        }
    }

    public boolean hasName() {
        return _name != null;
    }

    public void setValue( String value ) {
        if (value != null && !value.isEmpty()) {
            _value = value.trim();
        } else {
            _value = null;
        }
        ((InputParameterSlotView) getView()).updateTextfield(_value);
    }

    public String getValue() {
        return _value;
    }

    public float getNumericalValue() {
        try {
            return Float.parseFloat(getValue());
        } catch (NumberFormatException e) {
            Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Slot.error4"));
            return 0.0f;
        }
    }

    public void changeValue( String value, boolean addInHistory ) {
        if (value == null && getValue() == null) {
            return;
        }
        if (value == null || !value.equals(getValue())) {
            final String oldValue = getValue();
            setValue(value);
            getGraphItem().resetAll();
            getGraphItem().getView().updatePropertyViewer();
            getGraph().setModified(true);
            if (addInHistory) {
                getGraph().validateOneAction(new ChangeParameterValueEdit(this, value, oldValue));
            }
        }
    }

    @Override
    public boolean isAcceptableDropSite( SlotModel slot ) {
        return slot instanceof OutputResultSlot;
    }

    @Override
    public XMLElement convertToXML() {
        XMLElement element = new XMLElement("Connection");
        element.addChildren(new XMLElement("EndItemId", Integer.toString(getGraphItem().getId())));
        element.addChildren(new XMLElement("EndSlotType", "InSlotParameter"));
        element.addChildren(new XMLElement("EndSlotIndex", Integer.toString(getRank())));
        if (isConnected()) {
            getConnectedSlot().convertToXML(element);
            if (!getLink().getPath().isEmpty()) {
                element.addChildren(new XMLElement("UserDefined", Boolean.toString(getLink().isUserDefined())));
                XMLElement pa = new XMLElement("LinkPath");
                for (int i = 0; i < getLink().getPath().size(); i++) {
                    XMLElement po = new XMLElement("LinkPoint");
                    po.addChildren(new XMLElement("IdPoint", Integer.toString(i)));
                    po.addChildren(new XMLElement("X-coordinate", Integer.toString((int) getLink().getPath().get(i).getX())));
                    po.addChildren(new XMLElement("Y-coordinate", Integer.toString((int) getLink().getPath().get(i).getY())));
                    pa.addChildren(po);
                }
                element.addChildren(pa);
            }
        } else {
            element.addChildren(new XMLElement("ParameterValue", getValue()));
        }
        return element;
    }
}
