/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.tree;

import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.model.graph.element.*;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;

import java.awt.datatransfer.DataFlavor;
import java.util.List;

/**
 * An <code>OperatorTreeItem</code> represents an operator
 * that can be inserted in an Ariane graph.
 */
public final class OperatorTreeItem extends TreeItem {
    public static final DataFlavor[] MIME_TYPE = {new DataFlavor("application/x-java-serialized-operator-item-tree", "Operator Tree Item")};

    public OperatorTreeItem( ItemType typeItem, String name, String docURL, int inputDataSlots, int inputParameterSlots, int outputDataSlots, String signature ) {
        super(typeItem, name, docURL, inputDataSlots, inputParameterSlots, outputDataSlots, signature);
    }

    OperatorTreeItem( List<XMLElement> xmlTree ) {
        super(xmlTree);
    }

    @Override
    public DataFlavor[] getTransferDataFlavors() {
        return MIME_TYPE;
    }

    public void setBinaryPath( String binary, String synopsis ) {
        setBinaryPath(binary + OperatorItem.SEPARATOR + synopsis);
    }

    @Override
    final void setBinaryPath( String path ) {
        String preferredDirectory = UserPreferences.getPreferredBinaryDirectory();
        if (preferredDirectory != null && !preferredDirectory.trim().isEmpty()) {
            String[] commandLine = path.split(OperatorItem.SEPARATOR);
            String binary = DataFile.getRelativePath(commandLine[0], preferredDirectory, UserPreferences.getPreferredFileDirectory());
            if (commandLine.length > 1) {
                super.setBinaryPath(binary + OperatorItem.SEPARATOR + commandLine[1]);
            } else {
                super.setBinaryPath(binary);
            }
        } else {
            super.setBinaryPath(path);
        }
    }

    @Override
    public GraphItemModel createGraphItem( GraphPresenter graph ) { // TODO deplacer dans une factory?.
        try {
            switch (getType()) {
                case OPERATOR:
//                    item = GraphElementFactory.createOperator(this, graph);
                    return new OperatorItem(this, graph);
                case LOADER:
                    return new LoaderItem(this, graph);
                case SAVER:
                    return new SaverItem(this, graph);
                case VIEWER:
                    return new ViewerItem(this, graph);
                case FILENAME:
                    return new FileNameItem(this, graph);
                case CONSTANT:
                    return new ConstantItem(this, graph);
                case MACRO_OPERATOR:
                    return new MacroOperatorItem(this, graph);
                case EVAL:
                    return new EvalItem(this, graph);
                case FORMAT:
                    return new FormatItem(this, graph);
                case IF_CONTROL:
                    return new IfControlStructureItem(this, graph);
                case SELECT_CONTROL:
                    return new SelectControlStructureItem(this, graph);
                case FOR_CONTROL:
                    return new ForControlStructureItem(this, graph);
                case WHILE_CONTROL:
                    return new WhileControlStructureItem(this, graph);
                case MERGE_CONTROL:
                    return new MergeControlStructureItem(this, graph);
                case NOTE:
                    return new NoteItem(this, graph);
                case SYSTEM:
                    return new SystemItem(this, graph);
                case SCRIPT:
                    return new ScriptItem(this, graph);
                default:
                    return null;
            }
        } catch (InterruptedException e) {
            return null;
        }
    }

    @Override
    public XMLElement convertToXMLTree() {
        if (getType() == ItemType.OPERATOR) {
            XMLElement element = new XMLElement("Operator");
            element.addChildren(new XMLElement("Category", getCategory()));
            element.addChildren(new XMLElement("Name", getName()));
            element.addChildren(new XMLElement("Documentation", getDocumentationURL()));
            element.addChildren(new XMLElement("Binary", getBinaryPath()));
            element.addChildren(new XMLElement("Parameters", Integer.toString(getNumberOfParameterSlots())));
            element.addChildren(new XMLElement("Signature", getSignature()));
            element.addChildren(new XMLElement("Inputs", Integer.toString(getNumberOfInputDataSlots())));
            element.addChildren(new XMLElement("Outputs", Integer.toString(getNumberOfOutputDataSlots())));
            return element;
        } else {
            return null;
        }
    }

    @Override
    protected void loadItemFromXMLTree( List<XMLElement> xmlelements ) {
        setType(ItemType.OPERATOR);
        for (XMLElement item : xmlelements) {
            if ("Category".equalsIgnoreCase(item.getName()) || "Categorie".equalsIgnoreCase(item.getName())) {
                setCategory(item.getValue());
            } else if ("Name".equalsIgnoreCase(item.getName()) || "Nom".equalsIgnoreCase(item.getName())) {
                setName(item.getValue());
            } else if ("Documentation".equalsIgnoreCase(item.getName())) {
                setDocumentationURL(item.getValue());
            } else if ("Binary".equalsIgnoreCase(item.getName()) || "Binaire".equalsIgnoreCase(item.getName())) {
                setBinaryPath(item.getValue());
            } else if ("Parameters".equalsIgnoreCase(item.getName()) || "Parametre".equalsIgnoreCase(item.getName())) {
                setNumberOfParameterSlots(Integer.parseInt(item.getValue()));
            } else if ("Signature".equalsIgnoreCase(item.getName())) {
                setSignature(item.getValue());
            } else if ("Inputs".equalsIgnoreCase(item.getName()) || "Entree".equalsIgnoreCase(item.getName())) {
                setNumberOfInputDataSlots(Integer.parseInt(item.getValue()));
            } else if ("Outputs".equalsIgnoreCase(item.getName()) || "Sortie".equalsIgnoreCase(item.getName())) {
                setNumberOfOutputDataSlots(Integer.parseInt(item.getValue()));
            }
        }
    }
}
