/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.tree;

import fr.greyc.pantheon.ariane.model.graph.element.GraphItemModel;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.util.List;

/**
 * A <code>TreeItem</code> is an item in the tree of the explorer.
 * It represents an element that can be inserted in a graph.
 */
public abstract class TreeItem implements Transferable, Cloneable {
    private ItemType _typeItem;
    private String _name;
    private String _documentationURL;
    private String _binaryPath;
    private String _categoryName;
    private String _signature;
    private int _numberOfInputDataSlots;
    private int _numberOfInputParameterSlots;
    private int _numberOfOutputDataSlots;

    TreeItem( ItemType typeItem, String name, String docURL, int inputDataSlots, int inputParameterSlots, int outputDataSlots, String signature ) {
        _typeItem = typeItem;
        _name = name;
        _documentationURL = docURL;
        _numberOfInputParameterSlots = inputParameterSlots;
        _numberOfInputDataSlots = inputDataSlots;
        _numberOfOutputDataSlots = outputDataSlots;
        _signature = signature;
    }

    TreeItem( List<XMLElement> xmlTree ) {
        loadItemFromXMLTree(xmlTree);
    }

    public final String getName() {
        return _name;
    }

    public final void setName( String text ) {
        _name = text;
    }

    public final String getDocumentationURL() {
        return _documentationURL;
    }

    public final void setDocumentationURL( String url ) {
        _documentationURL = url;
    }

    public final String getCategory() {
        return _categoryName;
    }

    public final void setCategory( String categoryName ) {
        _categoryName = categoryName;
    }

    public final String getBinaryPath() {
        return _binaryPath;
    }

    void setBinaryPath( String path ) {
        _binaryPath = path;
    }

    public final int getNumberOfParameterSlots() {
        return _numberOfInputParameterSlots;
    }

    public final void setNumberOfParameterSlots( int count ) {
        _numberOfInputParameterSlots = count;
    }

    public final String getSignature() {
        return _signature;
    }

    /**
     * @param signature (eg. $p0:updateNoteContents=0; $p1:titi=1;).
     */
    public final void setSignature( String signature ) {
        _signature = signature;
    }

    public final int getNumberOfInputDataSlots() {
        return _numberOfInputDataSlots;
    }

    public final void setNumberOfInputDataSlots( int count ) {
        _numberOfInputDataSlots = count;
    }

    public final int getNumberOfOutputDataSlots() {
        return _numberOfOutputDataSlots;
    }

    public final void setNumberOfOutputDataSlots( int count ) {
        _numberOfOutputDataSlots = count;
    }

    /**
     * Returns the name of the item. Warning! this method is used to display the name in the tree.
     */
    @Override
    public final String toString() {
        return _name;
    }

    @Override
    public abstract DataFlavor[] getTransferDataFlavors();

    @Override
    public final Object getTransferData( DataFlavor flavor ) throws UnsupportedFlavorException {
        if (isDataFlavorSupported(flavor)) {
            return this;
        } else {
            throw new UnsupportedFlavorException(flavor);
        }
    }

    @Override
    public final boolean isDataFlavorSupported( DataFlavor flavor ) {
        return flavor.equals(getTransferDataFlavors()[0]);
    }

    public abstract GraphItemModel createGraphItem( GraphPresenter graph );

    public abstract XMLElement convertToXMLTree();

    public Object clone() throws CloneNotSupportedException {
        return super.clone();
    }

    final ItemType getType() {
        return _typeItem;
    }

    final void setType( ItemType type ) {
        _typeItem = type;
    }

    protected abstract void loadItemFromXMLTree( List<XMLElement> xmlelements );

    public enum ItemType {
        CONSTANT,
        EVAL,
        FILENAME,
        FOR_CONTROL,
        FORMAT,
        IF_CONTROL,
        LOADER,
        MACRO_OPERATOR,
        MERGE_CONTROL,
        NOTE,
        OPERATOR,
        SAVER,
        SCRIPT,
        SELECT_CONTROL,
        SYSTEM,
        VIEWER,
        WHILE_CONTROL,
    }
}