/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.xml;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.charset.StandardCharsets;

/**
 * A <code>XMLFileOutputStream</code> defines an output stream for XML files.
 */
public final class XMLFileOutputStream {
    private final OutputStream _outputStream;

    public XMLFileOutputStream( final File file, final String dtdFilename ) throws IOException {
        _outputStream = new BufferedOutputStream(new FileOutputStream(file));
        if (dtdFilename == null) {
            write("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n", 0);
        } else {
            write("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n<!DOCTYPE ARIANE SYSTEM '" + dtdFilename + "'>\n", 0);
        }
        write("<ARIANE version=\"1.1\">\n", 0);
    }

    private void write( final String string, final int depth ) throws IOException {
        for (int i = 0; i < depth * 2; i++) {
            _outputStream.write(' ');
        }
        _outputStream.write(string.getBytes());
    }

    private void writeEncoded( final String string ) throws IOException {
        if (string != null) {
            // Replace character '&' by &amp; '<' by &lt;
            final byte[] encoded = string.getBytes(StandardCharsets.ISO_8859_1);
            for (byte anEncoded : encoded) {
                switch (anEncoded) {
                    case '&':
                        _outputStream.write("&amp;".getBytes());
                        break;
                    case '<':
                        _outputStream.write("&lt;".getBytes());
                        break;
                    default:
                        _outputStream.write(anEncoded);
                        break;
                }
            }
        }
    }

    private void write( final XMLElement xmlRoot, final int indentationSize ) throws IOException {
        if (xmlRoot != null) {
            if (xmlRoot.getChildren() == null) { // A LEAF.
                write("<" + xmlRoot.getName() + ">", indentationSize);
                writeEncoded(xmlRoot.getValue());
                write("</" + xmlRoot.getName() + ">\n", 0);
            } else {                            // A NODE.
                write("<" + xmlRoot.getName() + ">\n", indentationSize);
                for (XMLElement element : xmlRoot.getChildren()) {
                    write(element, indentationSize + 1);
                }
                write("</" + xmlRoot.getName() + ">\n", indentationSize);
            }
        }
    }

    public void write( final XMLElement element ) throws IOException {
        write(element, 0);
    }

    public void close() throws IOException {
        write("</ARIANE>", 0);
        _outputStream.close();
    }
}
