/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.presenter;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.graph.GraphModel;
import fr.greyc.pantheon.ariane.model.graph.ThreadDirectoryExecution;
import fr.greyc.pantheon.ariane.model.graph.ThreadGraphExecution;
import fr.greyc.pantheon.ariane.model.graph.element.GraphItemModel;
import fr.greyc.pantheon.ariane.model.graph.element.LoaderItem;
import fr.greyc.pantheon.ariane.model.graph.element.SaverItem;
import fr.greyc.pantheon.ariane.model.graph.element.VirtualItem;
import fr.greyc.pantheon.ariane.model.graph.slot.*;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.view.command.AddItemEdit;
import fr.greyc.pantheon.ariane.view.command.MoveItemEdit;
import fr.greyc.pantheon.ariane.view.command.MoveLinkEdit;
import fr.greyc.pantheon.ariane.view.command.RemoveItemEdit;
import fr.greyc.pantheon.ariane.view.graph.GraphView;
import fr.greyc.pantheon.ariane.view.graph.XYConstraint;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import javax.swing.undo.UndoableEdit;

public abstract class GraphPresenter {
    private DraggingSelection _draggingSelection;
    private final UndoRedoManager _undoRedoManager = new UndoRedoManager();
    private boolean _isAlive = true;
    private String _name;
    private static boolean _isRunning;
    private Process _currentProcess;

    private final GraphView _graphView;
    private final GraphModel _graphModel;
    private final List<GraphItemModel> _selectedItems = new LinkedList<>();

    GraphPresenter( GraphModel graphModel ) {
        _graphModel = graphModel;
        _graphView = new GraphView(this);
    }

    public final void dispose() {
        _isAlive = false;
        _graphModel.dispose();
        _undoRedoManager.discardAllEdits();
        unSelectAll();
    }

    /////
    ///// F R A M E S
    /////
    public final void closeFrame() {
        for (GraphItemModel item : _graphModel.getListOfGraphItems()) {
            if (item.getSubGraph() != null) {
                item.getSubGraph().closeFrame();
            }
        }
        Ariane.getPresenter().removeGraph(this);
    }

    final void renameFrame() {
        for (GraphItemModel item : _graphModel.getListOfGraphItems()) {
            if (item.getSubGraph() != null) {
                item.getSubGraph().renameFrame();
            }
        }
        Ariane.getPresenter().renameGraph(this, getName());
    }

    /////
    ///// A C C E S S O R S
    /////
    final GraphModel getModel() {
        return _graphModel;
    }

    public final GraphView getView() {
        return _graphView;
    }

    public final int getIdentifier() {
        return _graphModel.getIdentifier();
    }

    public final void setName( String name ) {
        _name = name;
    }

    public final String getName() {
        return _name;
    }

    final boolean isCloseConfirmed() {
        return _graphView.isCloseConfirmed();
    }

    final boolean isAlive() {
        return _graphView.getParent() != null && _isAlive;
    }

    public abstract GraphItemModel getSuperGraphItem();

    public final String getDirectoryName() {
        File file = getRoot()._graphModel.getStorageFile();
        if (file != null) {
            return file.getParent();
        }
        return null;
    }

    public final boolean isAssociatedToFile( File file ) {
        File currentFile = getRoot()._graphModel.getStorageFile();
        if (currentFile != null) {
            try {
                return file.getCanonicalPath().equals(currentFile.getCanonicalPath());
            } catch (IOException e) {
                return false;
            }
        }
        return false;
    }

    public final int getNextGraphItemIdentifier() {
        return _graphModel.getNextItemIdentifier();
    }

    public abstract boolean isModified();

    public abstract void setModified( boolean value );

    public abstract boolean isRoot();

    public abstract GraphPresenter getRoot();

    public abstract boolean close();

    public final boolean isSelectedItemsEmpty() {
        return _selectedItems.isEmpty();
    }

    public final List<GraphItemModel> getListOfGraphItems() {
        return _graphModel.getListOfGraphItems();
    }

    public final GraphItemModel getGraphItem( int identifier ) {
        return _graphModel.getGraphItem(identifier);
    }

    /////
    ///// ACTIONS ON ELEMENTS
    /////
    final void disconnectSlot( SlotModel incoming ) {
        startActionEdition();
        incoming.disconnectAll();
        endActionEdition();
    }

    public final void addItemAtRelativePosition( TreeItem incoming, Point coordinates ) {
        Point viewPosition = _graphView.getViewPosition();
        Point absolutePosition = new Point(viewPosition.x + coordinates.x, viewPosition.y + coordinates.y);
        addItemAtAbsolutePosition(incoming, absolutePosition);
    }

    public final void addItemAtAbsolutePosition( TreeItem incoming, Point coordinates ) {
        GraphItemModel newItem = incoming.createGraphItem(this);
        if (newItem != null) {
            startActionEdition();
            addItemAtAbsolutePosition(newItem, coordinates, true);
            endActionEdition();
            unSelectAll();
            addInSelection(newItem, false);
            newItem.getView().requestFocusInWindow();
        }
    }

    public final void addItemAtAbsolutePosition( GraphItemModel item, Point xycoord, boolean addInHistory ) {
        _graphModel.add(item);
        item.getView().draw();
        _graphView.addElement(item.getView(), xycoord);

        if (addInHistory) {
            validateAction(new AddItemEdit(this, item));
        }
    }

    public final void removeItem( GraphItemModel item ) {
        startActionEdition();
        removeItem(item, true);
        revalidate();
        endActionEdition();
    }

    public final void removeItem( GraphItemModel item, boolean addInHistory ) {
        item.remove();
        _graphModel.remove(item);
        _graphView.remove(item.getView());
        if (addInHistory) {
            validateAction(new RemoveItemEdit(this, item));
        }
    }

    public final void moveItem( GraphItemModel item, Point location, boolean addInHistory ) {
        if (addInHistory) {
            validateAction(new MoveItemEdit(this, item, item.getLocation(), location));
        }
        item.setLocation(location);
        _graphView.revalidate();
    }

    private void moveItems( List<GraphItemModel> litems, Point translation ) {
        for (GraphItemModel item : litems) {
            Point xy = item.getLocation();
            xy.translate(-translation.x, -translation.y);
            moveItem(item, xy, true);
        }
        _graphView.repaint();
    }

    private void moveItems( List<GraphItemModel> litems, Point from, Point to ) {
        final Point translation = new Point(from.x - to.x, from.y - to.y);
        for (GraphItemModel item : litems) {
            Point origin = new Point(item.getLocation().x + translation.x, item.getLocation().y + translation.y);
            validateAction(new MoveItemEdit(this, item, origin, item.getLocation()));
        }
        _graphView.repaint();
    }

    public final void moveLink( LinkModel link, Point from, Point to ) {
        validateOneAction(new MoveLinkEdit(this, link, from, to));
    }

    /////
    ///// E X E C U T I O N
    /////
    public final void revalidate() {
        _graphView.revalidate();
        _graphView.repaint();
    }

    public final void resetAll() {
        _graphModel.resetAll();
    }

    public final void reset() {
        _graphModel.reset();
    }

    public final void resetSelectedItems() {
        _selectedItems.forEach(GraphItemModel::resetAll);
    }

    public final synchronized void setCurrentProcess( Process process ) {
        _currentProcess = process;
    }

    public final synchronized void stopExecution() {
        _isRunning = false;
        Ariane.getPresenter().setExecution(false);
        if (_currentProcess != null) {
            _currentProcess.destroy();
        }
    }

    public static boolean isExecutionRunning() {
        return _isRunning;
    }

    public final void startTestExecution() {
        _isRunning = true;
    }

    public final void startExecution() {
        if (_selectedItems.isEmpty()) {
            Ariane.getPresenter().printWarning(ArianeSettings.getString("Graph.execution.warning1"));
        } else {
            _isRunning = true;
            // Creates a thread and executes the graph with it.
            ThreadGraphExecution thread = new ThreadGraphExecution(_selectedItems);
            thread.start();
        }
    }

    public final void executeFolder( LoaderItem loader ) {
        if (loader.getBinary() == null) {
            Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Graph.execution.error1") + loader.getName() + "(" + loader.getId() + ")");
            return;
        }
        File loaderFile = new File(loader.getBinary());
        if (!loaderFile.exists()) {
            Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Graph.execution.error1") + loader.getName() + "(" + loader.getId() + ")");
            return;
        }

        List<GraphItemModel> terminalItems = getTerminalItems(loader);

        // Get the savers to store data in the appropriate files.
        HashMap<SaverItem, File> saverDirectories = new HashMap<>();
        for (GraphItemModel item : terminalItems) {
            if (item instanceof SaverItem) {
                File saverFile = new File(item.getBinary());
                if (saverFile.getParentFile() == null || !saverFile.getParentFile().exists()) {
                    Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Graph.execution.error1") + item.getName() + "(" + item.getId() + ")");
                    return;
                }
                saverDirectories.put((SaverItem) item, saverFile);
            }
        }

        int suffixPosition = loaderFile.getName().lastIndexOf('.');
        final String loaderSuffix = suffixPosition > 0 ? loaderFile.getName().substring(suffixPosition) : "";
        final File[] files = loaderFile.getParentFile().listFiles(( dir, name ) -> {
            File file = new File(dir, name);
            if (file.isDirectory() || file.isHidden()) {
                return false;
            }
            String suffix = "";
            int suffixPosition1 = name.lastIndexOf('.');
            if (suffixPosition1 > 0) {
                suffix = name.substring(suffixPosition1);
            }
            return suffix.equals(loaderSuffix);
        });

        if (files == null) {
            Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Graph.execution.error2") + loaderFile.getParent());
            return;
        }
        Arrays.sort(files);

        _isRunning = true;
        ThreadDirectoryExecution thread = new ThreadDirectoryExecution(terminalItems, loader, loaderFile, files, saverDirectories);
        thread.start();
    }

    private List<GraphItemModel> getTerminalItems( LoaderItem loader ) {
        // Get the list of items to be executed starting from the loader.
        LinkedList<GraphItemModel> terminalItems = new LinkedList<>();
        LinkedList<GraphItemModel> visitedItems = new LinkedList<>();
        LinkedList<GraphItemModel> stack = new LinkedList<>();
        stack.add(loader);

        while (!stack.isEmpty()) {
            GraphItemModel item = stack.removeLast();
            if (visitedItems.contains(item)) {
                continue;
            }
            visitedItems.add(item);

            for (ParameterSlot slot : item.getInputParameterSlots()) {
                if (slot.isConnected()) {
                    GraphItemModel neighbour = slot.getConnectedSlot().getGraphItem();
                    stack.add(neighbour);
                }
            }

            for (InputDataSlot slot : item.getInputDataSlots()) {
                if (slot.isConnected()) {
                    GraphItemModel neighbour = slot.getConnectedSlot().getGraphItem();
                    stack.add(neighbour);
                }
            }

            InputMaskSlot maskSlot = item.getInputMaskSlot();
            if (maskSlot != null && maskSlot.isConnected()) {
                GraphItemModel neighbour = maskSlot.getConnectedSlot().getGraphItem();
                stack.add(neighbour);
            }

            boolean lastItem = true;
            for (OutputDataSlot slot : item.getOutputDataSlots()) {
                if (slot.isConnected()) {
                    lastItem = false;
                    for (InputSlot neighbourSlot : slot.getConnectedInputSlots()) {
                        if (neighbourSlot.isConnected()) {
                            stack.add(neighbourSlot.getGraphItem());
                        }
                    }
                }
            }

            if (item.getOutputResultSlot() != null) {
                OutputResultSlot slot = item.getOutputResultSlot();
                if (slot.isConnected()) {
                    lastItem = false;
                    for (InputSlot neighbourSlot : slot.getConnectedInputSlots()) {
                        if (neighbourSlot.isConnected()) {
                            stack.add(neighbourSlot.getGraphItem());
                        }
                    }
                }
            }

            if (lastItem) {
                terminalItems.add(item);
            }
        }
        return terminalItems;
    }

    public final void displayOutputs( int output ) {
        for (GraphItemModel x : _selectedItems) {
            if (x.getOutputDataSlots().size() > output) {
                x.getOutputDataSlots().get(output).getPresenter().displayDataFileContents();
            }
        }
    }

    public final void selectAll() {
        unSelectAll();
        for (GraphItemModel item : _graphModel.getListOfGraphItems()) {
            if (!(item instanceof VirtualItem)) {
                addInSelection(item, true);
            }
        }
    }

    public final void unSelectAll() {
        for (GraphItemModel gi : _selectedItems) {
            gi.setSelected(false);
        }
        _selectedItems.clear();
        _graphView.resetCurrentSelection();
    }

    private void selectItem( GraphItemModel item ) {
        unSelectAll();
        _selectedItems.add(item);
        item.setSelected(true);
    }

    /**
     * Adds the given item in the current selection according to the specified added value:
     * <ul>
     * <li>Shift pressed: adds the item to the current selection.
     * <li>Shift pressed and item already in the selection: removes it from the selection.
     * <li>Shift not pressed : builds a new selection with this item.
     * </ul>
     *
     * @param item         the item to be added.
     * @param shiftPressed key shift pressed?
     */
    public final void addInSelection( GraphItemModel item, boolean shiftPressed ) {
        if (shiftPressed) {
            if (item.isSelected()) {
                _selectedItems.remove(item);
                item.setSelected(false);
            } else {
                item.setSelected(true);
                _selectedItems.add(item);
            }
        } else {
            selectItem(item);
        }
    }

    private Rectangle getRectangleAroundSelectedItems() {
        int xs = 100000;
        int ys = 100000;
        int xe = 0;
        int ye = 0;
        for (GraphItemModel item : _selectedItems) {
            Point itemcoord = item.getLocation();
            if (itemcoord.x < xs) {
                xs = itemcoord.x;
            }
            if (itemcoord.y < ys) {
                ys = itemcoord.y;
            }
            if (item.getView().getWidth() + itemcoord.x > xe) {
                xe = item.getView().getWidth() + itemcoord.x;
            }
            if (item.getView().getHeight() + itemcoord.y > ye) {
                ye = item.getView().getHeight() + itemcoord.y;
            }
        }
        return new Rectangle(xs, ys, xe - xs, ye - ys);
    }

    public final void dragSelection( MouseEvent event ) {
        Point xycoord = event.getPoint();
        if (_draggingSelection == null) {    // start dragging.
            _draggingSelection = new DraggingSelection(getRectangleAroundSelectedItems(), xycoord);
        } else {                            // dragging the current selection...
            Point shift = _draggingSelection.getOffset();
            Point newOrigin = new Point(xycoord.x + shift.x, xycoord.y + shift.y);
            if (newOrigin.x < 0) {
                newOrigin.x = 0;
            }
            if (newOrigin.y < 0) {
                newOrigin.y = 0;
            }
            Point translation = _draggingSelection.getTranslation(newOrigin);
            _graphView.scrollRectToVisible(new Rectangle(xycoord.x, xycoord.y, 10, 10));
            dragSelection(translation);
        }
    }

    public final boolean dropSelection() {
        if (_draggingSelection != null) {
            moveSelection(_draggingSelection.getOriginLocation(), _draggingSelection.getLocation());
            _draggingSelection = null;
            return true;
        }
        return false;
    }

    private void moveSelection( Point from, Point to ) {
        startActionEdition();
        moveItems(_selectedItems, from, to);
        endActionEdition();
    }

    public final void moveSelection( Point translation ) {
        for (GraphItemModel item : _selectedItems) {
            if (item.getLocation().x - translation.x < 0 || item.getLocation().y - translation.y < 0) {
                return;
            }
        }
        startActionEdition();
        moveItems(_selectedItems, translation);
        endActionEdition();
    }

    private void dragSelection( Point translation ) {
        for (GraphItemModel item : _selectedItems) {
            Point xy = item.getLocation();
            xy.translate(-translation.x, -translation.y);
            moveItem(item, xy, false);
        }
        _graphView.repaint();
    }

    public final void deleteSelectedItems() {
        startActionEdition();
        while (!_selectedItems.isEmpty()) {
            GraphItemModel item = _selectedItems.remove(0);
            removeItem(item, true);
        }
        endActionEdition();
        unSelectAll();
    }

    public final XMLElement copySelectedItems() {
        try {
            XMLElement element = new XMLElement("ClipBoard");
            XMLElement items = new XMLElement("ItemList");
            for (GraphItemModel item : _selectedItems) {
                items.addChildren(item.convertToXML(null));
            }
            XMLElement connections = new XMLElement("ConnectionList");
            for (GraphItemModel item : _selectedItems) {
                connections.addChildren(item.convertSlotsToXML());
            }
            element.addChildren(items);
            element.addChildren(connections);
            return element;
        } catch (IOException ex) {
            return null;
        }
    }

    final void pasteSelectedItems( XMLElement clipBoard ) {
        startActionEdition();
        try {
            Point viewPosition = _graphView.getViewPosition();
            List<GraphItemModel> items = GraphModel.createFromXML(this, clipBoard.getChildren().get(0).getChildren());
            SlotModel.createFromXML(getModel(), clipBoard.getChildren().get(1).getChildren());

            if (!items.isEmpty()) {
                Point min = items.get(0).getLocation();
                for (GraphItemModel item : items) {
                    Point point = item.getLocation();
                    if (point.y < min.y) {
                        min.y = point.y;
                    }
                    if (point.x < min.x) {
                        min.x = point.x;
                    }
                }
                final Point translation = new Point(viewPosition.x + 5 - min.x, viewPosition.y + 5 - min.y);
                // translate all item to be visible into the current viewport.
                for (GraphItemModel item : items) {
                    Point coordinates = item.getLocation();
                    coordinates.translate(translation.x, translation.y);
                    _graphView.add(item.getView(), new XYConstraint(coordinates)); //TODO dans vue
                    _graphView.setComponentZOrder(item.getView(), 0); // Put it in front
                }
            }
            // start the new selection
            unSelectAll();
            for (GraphItemModel item : _graphModel.getListOfGraphItems()) {
                // Discard virtual operators.
                if (!(item instanceof VirtualItem) && item.isInSameReadPasteSession()) {
                    addInSelection(item, true);
                }
            }
            _graphView.setViewPosition(viewPosition);
            revalidate();
        } catch (IOException ex) {
            // Nothing to do
        }
        endActionEdition();
    }

    public final LinkModel detectLink( double x, double y ) {
        for (GraphItemModel item : _graphModel.getListOfGraphItems()) {
            LinkedList<InputSlot> allSlots = new LinkedList<>();
            allSlots.addAll(item.getInputDataSlots());
            allSlots.addAll(item.getInputParameterSlots());
            if (item.getInputMaskSlot() != null) {
                allSlots.add(item.getInputMaskSlot());
            }
            for (InputSlot slot : allSlots) {
                if (slot.getConnectedSlot() != null) {
                    LinkModel link = slot.getLink();
                    if (link.isCloseTo(x, y)) {
                        return link;
                    }
                }
            }
        }
        return null;
    }

    /////
    ///// I N P U T   /   O U T P U T
    /////
    public abstract void save();

    public abstract void saveAs();

    final File askForFilename( File file ) {
        return _graphView.askForFilename(file);
    }

    final File getStorageFile() {
        return _graphModel.getStorageFile();
    }

    final void save( File file ) {
        try {
            _graphModel.save(file);
            Ariane.getPresenter().printSystemMessage(_graphModel.getStorageFile() + ArianeSettings.getString("Workspace.saving.success"));
            setModified(false);
        } catch (IOException ex) {
            Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("Workspace.saving.error1") + file.getName() + " " + ex.getLocalizedMessage());
        }
    }

    public final XMLElement convertToXMLTree( File file ) throws IOException {
        return _graphModel.convertToXMLTree(file);
    }

    public final void createFromXMLTree( List<XMLElement> xmlTree ) throws IOException {
        _graphModel.createFromXMLTree(xmlTree, this);
        _undoRedoManager.clearCurrentListOfActions();
    }

    /////
    ///// U N D O / R E D O (H I S T O R Y)
    /////
    public final void validateOneAction( UndoableEdit action ) {
        startActionEdition();
        validateAction(action);
        endActionEdition();
    }

    public final void validateAction( UndoableEdit action ) {
        _undoRedoManager.validateAction(action);
    }

    public final void startActionEdition() {
        _undoRedoManager.startActionEdition();
    }

    public final void endActionEdition() {
        _undoRedoManager.endActionEdition(this);
    }

    public final boolean canUndo() {
        return _undoRedoManager.canUndo();
    }

    public final boolean canRedo() {
        return _undoRedoManager.canRedo();
    }

    public final void undo() {
        setModified(true);
        _undoRedoManager.undo();
    }

    public final void redo() {
        setModified(true);
        _undoRedoManager.redo();
    }

    public void synchronizeItemFilesWithSystemFiles() {
        _graphModel.synchronizeItemFilesWithSystemFiles();
    }

    ///////
    /////// S E L E C T I O N
    ///////
    private static final class DraggingSelection {

        private final Rectangle _selectionRectangle;
        private final Point _offsetBrweenMousePointerAndRectanglel;
        private final Point _rectangleLocation;

        DraggingSelection( Rectangle rectangle, Point xycoord ) {
            _selectionRectangle = rectangle;
            _offsetBrweenMousePointerAndRectanglel = new Point(rectangle.x - xycoord.x, rectangle.y - xycoord.y);
            _rectangleLocation = rectangle.getLocation();
        }

        Point getOriginLocation() {
            return _selectionRectangle.getLocation();
        }

        Point getOffset() {
            return _offsetBrweenMousePointerAndRectanglel;
        }

        Point getLocation() {
            return _rectangleLocation;
        }

        Point getTranslation( Point newOrigin ) {
            Point translation = new Point();
            translation.x = _rectangleLocation.x;
            translation.y = _rectangleLocation.y;
            _rectangleLocation.x = newOrigin.x;
            _rectangleLocation.y = newOrigin.y;
            translation.x -= newOrigin.x;
            translation.y -= newOrigin.y;
            return translation;
        }
    }
}
