/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.presenter;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.model.generator.CodeGenerator;
import fr.greyc.pantheon.ariane.view.explorer.*;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.model.xml.XMLFileInputStream;
import fr.greyc.pantheon.ariane.view.OperatorLibraryPathChooser;
import fr.greyc.pantheon.ariane.view.PropertyPanel;
import fr.greyc.pantheon.ariane.view.UserPreferencesChooser;
import fr.greyc.pantheon.ariane.view.Workspace;
import fr.greyc.pantheon.ariane.view.utils.ScriptNameChooser;
import org.xml.sax.SAXException;

import javax.xml.parsers.ParserConfigurationException;
import java.awt.Point;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

public final class Presenter {
    private static int _currentReadPasteSession;
    private final Workspace _workspace;
    private final ItemTreeModel _itemTreeModel;

    public Presenter( ItemTreeModel itemTreeModel ) {
        _itemTreeModel = itemTreeModel;
        _workspace = new Workspace(this, new ItemTreeView(this, itemTreeModel));
    }

    public static int getCurrentReadOrPasteSessionId() {
        return _currentReadPasteSession;
    }

    /////
    ///// W O R K S P A C E
    /////
    public void generatesScriptCodeFromGraph( GraphPresenter graph ) {
        String[] chosenFile = ScriptNameChooser.chooseFilename(graph.getName(), CodeGenerator.getAvailable_Languages());
        if (chosenFile.length > 0) {
            CodeGenerator generator = CodeGenerator.createCodeGeneratorFromLanguage(new File(chosenFile[0]), chosenFile[1]);
            generator.generate(this, graph.getListOfGraphItems());
        }
    }

    public boolean generatesScriptCodeFromCurrentGraph( String scriptFilename, List<String> graphFilenames ) {
        int suffixPosition = scriptFilename.lastIndexOf('.');
        String suffix = suffixPosition > 0 ? scriptFilename.substring(suffixPosition + 1) : "";
        GraphPresenter graph = loadGraph(new File(graphFilenames.get(0)));
        CodeGenerator generator = CodeGenerator.createCodeGeneratorFromSuffix(new File(scriptFilename), suffix);
        if (generator != null && graph != null) {
            generator.generate(this, graph.getListOfGraphItems());
            return true;
        }
        return false;
    }

    public void chooseOperatorLibrary() {
        OperatorLibraryPathChooser libraryChooser = new OperatorLibraryPathChooser(UserPreferences.isOperatorLibraryUsed(), ArianeSettings.getLibraryPath());
        libraryChooser.setVisible(true);
        if (libraryChooser.isModificationsValidatedByUser()) {
            changeOperatorLibraryPath(libraryChooser.isTheLibraryUsed(), libraryChooser.getOperatorLibraryPath());
            if (UserPreferences.updateUserPreferenceFile()) {
                _workspace.repaintAllGraphs();
            }
        }
    }

    public void openPreferencesChooser() {
        UserPreferencesChooser preferenceChooser = new UserPreferencesChooser();
        preferenceChooser.setVisible(true);
        if (preferenceChooser.isValidatedByUser()) {
            _workspace.repaintAllGraphs();
            UserPreferences.updateUserPreferenceFile();
        }
    }

    public void restoreSplitPanesLocation() {
        _workspace.restoreSplitPanesLocation();
    }

    public void display( List<String> graphFilenames ) {
        List<GraphPresenter> graphs = new ArrayList<>();
        for (String graphFilename : graphFilenames) {
            graphs.add(loadGraph(new File(graphFilename)));
        }
        _workspace.display(graphs);
    }

    public void initContent() {
        _workspace.initContent();
    }

    public Workspace getWorkspace() {
        return _workspace;
    }

    /////
    ///// G R A P H
    /////
    public void openFile( String filename ) {
        if (!_workspace.isGraphAlreadyLoaded(filename)) {
            GraphPresenter graph = loadGraph(new File(filename));
            if (graph != null) {
                _workspace.addGraph(graph);
                Ariane.getPresenter().printSystemMessage(filename + ArianeSettings.getString("Workspace.load.message"));
            }
        }
    }

    public void createItemsFromClipboard( GraphPresenter graph, Transferable contents ) {
        try {
            _currentReadPasteSession++;
            XMLElement result = (XMLElement) contents.getTransferData(XMLElement.FLAVOR);
            graph.pasteSelectedItems(result);
        } catch (UnsupportedFlavorException | IOException e) {
            // Nothing to do
        }
    }

    public void addNewItemInGraphAtRelativePosition( Object treeItem, Point coordinate ) {
        _workspace.addNewItemAtRelativePosition((TreeItem) treeItem, coordinate);
    }

    /////
    ///// C O N S O L E
    /////
    public void printOutputMessage( String message ) {
        _workspace.printOutputMessage(message);
    }

    public void printSystemMessage( String mesg ) {
        _workspace.printSystemMessage(mesg);
    }

    public void printWarning( String mesg ) {
        _workspace.printWarning(mesg);
    }

    public void printErrorMessage( String mesg ) {
        _workspace.printErrorMessage(mesg);
    }

    public void printErrorMessage( ArianeSettings.ErrorCode errorCode, String item, int itemId, String mesg ) {
        _workspace.printErrorMessage(errorCode, item, itemId, mesg);
    }

    public void displayLibraryDocumentationPage( TreeItem item ) {
        String page = item.getDocumentationURL();
        displayLibraryDocumentationPage(item.getName(), page);
    }

    public void displayLibraryDocumentationPage( String item, String page ) {
        if (page != null) {
            displayDocumentationPage(ArianeSettings.getString("HTMLViewer.manual") + " " + item, page, null);
        }
    }

    public void displayDocumentationPage( String title, String page, String index ) {
        _workspace.displayHTMLManualPage(title, page, index);
    }

    public void copy() {
        _workspace.copy();
    }

    public void cut() {
        _workspace.cut();
    }

    public void paste() {
        _workspace.paste();
    }

    public ItemTreeModel getTree() {
        return _itemTreeModel;
    }

    public void displayPropertyPanel( GraphPresenter graph, PropertyPanel propertyPanel ) {
        _workspace.displayPropertyPanel(propertyPanel);
        _workspace.setCurrentGraph(graph);
    }

    public void renameGraph( GraphPresenter graph, String name ) {
        _workspace.renameGraph(graph, name);
    }

    public void setExecution( boolean isRunning ) {
        _workspace.setExecution(isRunning);
    }

    public boolean isPastable() {
        return _workspace.isPastable();
    }

    public void deleteSelectedItems() {
        _workspace.deleteSelectedItems();
    }

    public void changeLanguage() {
        _workspace.changeLanguage();
    }

    public int[] getSplitPaneRatios() {
        return _workspace.getSplitPaneRatios();
    }

    public void displayLibraryIndex() {
        displayDocumentationPage("Index", _itemTreeModel.getPathToLibraryDocumentationFile(), null);
    }

    public TreeItem getTreeItemWithName( String name ) {
        return _itemTreeModel.getItemEqualsTo(name);
    }

    public List<String> getTreeItemsWithNameStartingWith( String s ) {
        return _itemTreeModel.getItemsWithNameStartingWith(s);
    }

    public void openOperatorTreeEditor() {
        new OperatorTreeEditor(new TreeEditorPresenter(getTree()), getTree().getUserOperatorTree()).setVisible(true);
    }

    public void openSubroutineTreeEditor() {
        new SubroutineTreeEditor(new TreeEditorPresenter(getTree()), getTree().getSubroutineTree()).setVisible(true);
    }

    public boolean saveSubroutineTree() {
        return TreeEditorPresenter.saveSubroutineTree(getTree().getSubroutineTreeAsList());
    }

    void addGraph( SubGraphPresenter subGraph ) {
        _workspace.addGraph(subGraph);
    }

    void updateMenu() {
        _workspace.updateMenu();
    }

    void updateRecentFilesListMenu() {
        _workspace.updateRecentlyClosedWorkspacesListMenu();
    }

    void removeGraph( GraphPresenter graph ) {
        _workspace.removeGraph(graph);
    }

    void addRecentProjects( String absolutePath ) {
        _workspace.addRecentlyClosedWorkspace(absolutePath);
    }

    void setModifiedSaveB( boolean value ) {
        _workspace.setModifiedSaveB(value);
    }

    private void changeOperatorLibraryPath( boolean isUsed, String pathName ) {
        UserPreferences.setOperatorLibraryState(isUsed);
        if (!isUsed) {
            return;
        }
        String libraryFilename = pathName + ArianeSettings.getLibraryConfigFilename();
        try {
            List<XMLElement> configHandler = ArianeSettings.readConfigurationFile(new File(libraryFilename));
            _itemTreeModel.setCommonOperators(configHandler);
            ArianeSettings.setLibraryPath(pathName);
        } catch (IOException e) {
            UserPreferences.setOperatorLibraryState(false);
            Ariane.getPresenter().printErrorMessage(e.getLocalizedMessage());
        }
    }

    private GraphPresenter loadGraph( File file ) {
        _currentReadPasteSession++;
        try {
            XMLFileInputStream convert = new XMLFileInputStream(file);
            convert.parse();
            convert.close();
            GraphPresenter graph = new RootGraphPresenter(file);
            graph.createFromXMLTree(convert.getElements());
            addRecentProjects(file.getAbsolutePath());
            return graph;
        } catch (IOException | SAXException | ParserConfigurationException | IndexOutOfBoundsException ex) {
            printErrorMessage(ArianeSettings.getString("Workspace.loading.error1") + file.getName() + " " + ex.getLocalizedMessage());
            return null;
        }
    }
}
