/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.presenter.Presenter;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.*;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.AbstractAction;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.KeyStroke;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;

/**
 * The Terminal is an interface in which users can type in the name
 * of the operators to add into the current graph.
 */
final class Terminal extends JPanel {
    private static final String PROMPT_STRING = "$ ";
    private static final String EOL = "\n";
    private static final Point _coordinates = new Point(10, 10);
    private final JTextArea _textArea;
    private final CommandHistory _commandHistory = new CommandHistory();
    private final Presenter _presenter;
    private StringBuilder _commandLine = new StringBuilder();
    private int _commandPointer;

    public Terminal() { // For tests
        _presenter = null;
        _textArea = new JTextArea();
    }

    public Terminal( Presenter presenter ) {
        _presenter = presenter;
        _textArea = initContents();
        defineControlKeys();
    }

    void moveCaretPosition( int increment ) {
        if (_commandPointer + increment >= 0 && _commandPointer + increment <= _commandLine.length()) {
            setCaretPosition(_commandPointer + increment);
        }
    }

    void addCharacter( char character ) {
        if (Character.isLetterOrDigit(character) || character == '-' || character == '_') {
            int length = _textArea.getDocument().getLength();
            _textArea.insert(String.valueOf(character), length - _commandLine.length() + _commandPointer);
            _commandLine.insert(_commandPointer, character);
            moveCaretPosition(1);
        }
    }

    void removePreviousCharacter() {
        if (_commandPointer > 0) {
            removeCharacters(1);
            _commandLine.deleteCharAt(_commandPointer - 1);
            moveCaretPosition(-1);
        }
    }

    void clear() {
        _textArea.setText(PROMPT_STRING);
        _commandLine = new StringBuilder();
        setCaretPosition(0);
    }

    String commonBaseName( String name1, String name2 ) {
        byte[] v1 = name1.getBytes();
        byte[] v2 = name2.getBytes();
        int minSize = Math.min(v1.length, v2.length);
        for (int i = 0; i < minSize; i++) {
            if (v1[i] != v2[i]) {
                return name1.substring(0, i);
            }
        }
        return name1.substring(0, minSize);
    }

    String getCommandLine() {
        return _commandLine.toString();
    }

    private JTextArea initContents() {
        JTextArea textArea = new JTextArea();
        textArea.setFont(UserPreferences.getDefaultFont());
        textArea.setEditable(true);
        textArea.setWrapStyleWord(true);
        textArea.setLineWrap(true);
        textArea.setForeground(Color.LIGHT_GRAY);
        textArea.setMargin(new Insets(6, 4, 6, 4));
        textArea.setBackground(Color.BLACK);
        textArea.addKeyListener(new KeyboardController());
        textArea.setCaretColor(Color.WHITE);
        textArea.requestFocus();
        textArea.append(PROMPT_STRING);
        JScrollPane scrollPane = new JScrollPane(textArea);
        setLayout(new BorderLayout());
        add(scrollPane, BorderLayout.CENTER);
        return textArea;
    }

    private void defineControlKeys() {
        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_BACK_SPACE, Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()), "none");
        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_A, Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()), "none");
        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0), "enter");
        _textArea.getActionMap().put("enter", new AbstractAction() {
            @Override
            public void actionPerformed( ActionEvent event ) {
                evalCommandLine();
            }
        });

        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_G, Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()), "abort");
        _textArea.getActionMap().put("abort", new AbstractAction() {
            @Override
            public void actionPerformed( ActionEvent event ) {
                clearCurrentLine();
            }
        });

        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_TAB, 0, false), "completion");
        _textArea.getActionMap().put("completion", new AbstractAction() {
            @Override
            public void actionPerformed( ActionEvent event ) {
                commandNameCompletion();
            }
        });

        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_BACK_SPACE, 0), "backspace");
        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0), "backspace");
        _textArea.getActionMap().put("backspace", new AbstractAction() {
            @Override
            public void actionPerformed( ActionEvent event ) {
                removePreviousCharacter();
            }
        });

        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, 0), "left");
        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_KP_LEFT, 0), "left");
        _textArea.getActionMap().put("left", new AbstractAction() {
            @Override
            public void actionPerformed( ActionEvent event ) {
                moveCaretPosition(-1);
            }
        });

        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, 0), "right");
        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_KP_RIGHT, 0), "right");
        _textArea.getActionMap().put("right", new AbstractAction() {
            @Override
            public void actionPerformed( ActionEvent event ) {
                moveCaretPosition(+1);
            }
        });

        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_UP, 0, false), "up");
        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_KP_UP, 0), "up");
        _textArea.getActionMap().put("up", new AbstractAction() {
            @Override
            public void actionPerformed( ActionEvent event ) {
                moveHistory(-1);
            }
        });

        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0, false), "down");
        _textArea.getInputMap(WHEN_FOCUSED).put(KeyStroke.getKeyStroke(KeyEvent.VK_KP_DOWN, 0), "down");
        _textArea.getActionMap().put("down", new AbstractAction() {
            @Override
            public void actionPerformed( ActionEvent event ) {
                moveHistory(+1);
            }
        });
    }

    private void moveHistory( int increment ) {
        StringBuilder command = _commandHistory.getHistory(increment);
        if (command != null) {
            removeCharacters(_commandLine.length());
            _commandLine = new StringBuilder(command);
            _textArea.append(_commandLine.toString());
            setCaretPosition(_commandLine.length());
        }
    }

    private void setCaretPosition( int position ) {
        _commandPointer = position;
        _textArea.setCaretPosition(_textArea.getDocument().getLength() - _commandLine.length() + _commandPointer);
    }

    private void removeCharacters( int number ) {
        try {
            Document doc = _textArea.getDocument();
            int start = doc.getLength() - _commandLine.length() + _commandPointer - number;
            doc.remove(start, number);
        } catch (BadLocationException e) {
            Logger.getLogger(Ariane.class.getName()).log(Level.INFO, "Impossible error!", e);
        }
    }

    private void clearCurrentLine() {
        removeCharacters(_commandLine.length());
        _commandLine.setLength(0);
        setCaretPosition(0);
    }

    private void evalCommandLine() {
        if ("clear".equals(_commandLine.toString())) {
            clear();
        } else if (_commandLine.length() > 0) {
            Object treeItem = _presenter.getTreeItemWithName(_commandLine.toString());
            if (treeItem == null) {
                _textArea.append(EOL + ArianeSettings.getString("error") + ArianeSettings.getString("Terminal.error") + _commandLine);
            } else {
                _coordinates.translate(10, 10);
                _presenter.addNewItemInGraphAtRelativePosition(treeItem, _coordinates);
            }
            _commandHistory.add(_commandLine);
        }
        newCommandLine();
    }

    private void commandNameCompletion() {
        List<String> names = _presenter.getTreeItemsWithNameStartingWith(_commandLine.toString());
        if (names.size() == 1) {
            String complement = names.get(0).substring(_commandLine.length());
            _textArea.append(complement);
            _commandLine.append(complement);
            setCaretPosition(_commandLine.length());
        } else if (!names.isEmpty()) {
            // Find common base name
            String basename = names.get(0);
            for (String name : names) {
                basename = commonBaseName(basename, name);
            }
            basename = basename.substring(_commandLine.length());
            _textArea.append(basename);
            _commandLine.append(basename);
            newLine();
            names.forEach(( name ) -> _textArea.append(name + "\t"));
            _textArea.append(EOL + PROMPT_STRING + _commandLine);
            setCaretPosition(_commandLine.length());
        }
    }

    private void newCommandLine() {
        _textArea.append(EOL + PROMPT_STRING);
        _commandLine = new StringBuilder();
        setCaretPosition(0);
    }

    private void newLine() {
        _textArea.append(EOL);
    }

    private class CommandHistory {
        private final List<StringBuilder> _history = new ArrayList<>();
        private int _historyPointer;

        void add( StringBuilder command ) {
            _history.add(new StringBuilder(command));
            _historyPointer = _history.size();
        }

        StringBuilder getHistory( int increment ) {
            int newHistoryPointer = _historyPointer + increment;
            if (newHistoryPointer >= 0 && newHistoryPointer < _commandHistory._history.size()) {
                _historyPointer = newHistoryPointer;
                return _history.get(_historyPointer);
            } else if (newHistoryPointer == _commandHistory._history.size()) {
                _historyPointer = newHistoryPointer;
                return new StringBuilder();
            } else {
                return null;
            }
        }
    }

    private class KeyboardController implements KeyListener {

        @Override
        public void keyPressed( KeyEvent event ) {
            // empty
        }

        @Override
        public void keyReleased( KeyEvent event ) {
            // Empty
        }

        @Override
        public void keyTyped( KeyEvent event ) {
            addCharacter(event.getKeyChar());
            event.consume();        // make sure the text area doesn't handle this
        }
    }
}
