/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.model.graph.element.DataFile;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.presenter.Presenter;
import fr.greyc.pantheon.ariane.presenter.RootGraphPresenter;
import fr.greyc.pantheon.ariane.view.explorer.Explorer;
import fr.greyc.pantheon.ariane.view.explorer.ItemTreeView;
import fr.greyc.pantheon.ariane.view.graph.GraphEditor;
import fr.greyc.pantheon.ariane.view.graph.GraphView;
import fr.greyc.pantheon.swingext.FadingEffects;
import fr.greyc.pantheon.swingext.HTMLDocumentViewer;
import fr.greyc.pantheon.swingext.MessageBox;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.Transferable;
import java.awt.event.ActionEvent;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.JobName;
import javax.print.attribute.standard.MediaSizeName;
import javax.print.attribute.standard.OrientationRequested;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.JColorChooser;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.KeyStroke;

/**
 * The Workspace is the main window of the software.
 * It holds an editor of graphs, an explorer, a console and a terminal.
 */
public final class Workspace extends JPanel implements ClipboardOwner {
    private static final int MAXIMUM_RECENT_FILES = 10;

    private final Presenter _presenter;
    private final Explorer _itemExplorer;
    private final GraphEditor _graphEditor;
    private final Console _console;
    private final JPanel _itemPropertyPanel;

    private static List<String> _recentlyClosedWorkspaceFilenames = new ArrayList<>();

    private GraphPresenter _currentGraph;

    private final Toolbar _toolBar;
    private final MenuBar _menuBar;

    private final Map<String, Action> _actions = new HashMap<>();

    /**
     * The vertical split pane ratio (splitPaneHV2 + splitPaneV2).
     */
    private JSplitPane _splitPaneV1;
    /**
     * The vertical split pane ratio (splitpaneH property).
     */
    private JSplitPane _splitPaneV2;
    /**
     * The current horizontal split pane ratio (explorer + terminal).
     */
    private JSplitPane _splitPaneH1;
    /**
     * The current horizontal split pane ratio (graph + console).
     */
    private JSplitPane _splitPaneH2;

    public Workspace( Presenter presenter, ItemTreeView operatorsTree ) {
        _presenter = presenter;
        buildActions();
        _itemPropertyPanel = new JPanel(new BorderLayout());
        _itemPropertyPanel.setMinimumSize(new Dimension(70, 60));
        _console = new Console();
        _graphEditor = new GraphEditor(this);
        _itemExplorer = new Explorer(_presenter, operatorsTree);
        _toolBar = new Toolbar(_actions);
        _menuBar = new MenuBar(_actions, _itemExplorer.getMenu(), _recentlyClosedWorkspaceFilenames, presenter);
    }

    public void initContent() {
        JPanel contentPane = new JPanel();

        // Terminal + Explorer
        _splitPaneH1 = new JSplitPane(JSplitPane.VERTICAL_SPLIT, _itemExplorer, new Terminal(_presenter));
        _splitPaneH1.setResizeWeight(1.0);
        _splitPaneH1.setDividerSize(5);

        // Graph + console
        _splitPaneH2 = new JSplitPane(JSplitPane.VERTICAL_SPLIT, _graphEditor, new JScrollPane(_console));
        _splitPaneH2.setResizeWeight(1.0);
        _splitPaneH2.setDividerSize(5);

        // V2: tree | (editor + property)
        _splitPaneV2 = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, _splitPaneH2, _itemPropertyPanel);
        _splitPaneV2.setDividerSize(5);
        _splitPaneV2.setResizeWeight(1.0);

        // V1: tree | ((terminal explorer) + (editor+console+property)
        _splitPaneV1 = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, _splitPaneH1, _splitPaneV2);
        _splitPaneV1.setDividerSize(5);

        contentPane.setLayout(new BorderLayout());

        contentPane.add(_toolBar, BorderLayout.NORTH);
        contentPane.add(_splitPaneV1, BorderLayout.CENTER);

        setLayout(new BorderLayout());
        add(contentPane, BorderLayout.CENTER);

        changeLanguage();
        setCurrentGraph(null);

        InputMap inputMap = getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
        ActionMap actionMap = getActionMap();
        displayShortcutForOutputDisplay(inputMap, actionMap, 0);
        displayShortcutForOutputDisplay(inputMap, actionMap, 1);
        displayShortcutForOutputDisplay(inputMap, actionMap, 2);
        displayShortcutForOutputDisplay(inputMap, actionMap, 3);
        displayShortcutForOutputDisplay(inputMap, actionMap, 4);
        displayShortcutForOutputDisplay(inputMap, actionMap, 5);
    }

    public GraphEditor getGraphEditor() { // For test only
        return _graphEditor;
    }

    public int[] getSplitPaneRatios() {
        int[] ratios = UserPreferences.getSplitPaneH1Ratio();
        return new int[]{
                getSplitPaneRatio(_splitPaneH1, ratios[0]),
                getSplitPaneRatio(_splitPaneV1, ratios[1]),
                getSplitPaneRatio(_splitPaneH2, ratios[2]),
                getSplitPaneRatio(_splitPaneV2, ratios[3])
        };
    }

    public void restoreSplitPanesLocation() {
        int[] ratios = UserPreferences.getSplitPaneH1Ratio();
        _splitPaneH1.setDividerLocation(ratios[0]);
        _splitPaneV1.setDividerLocation(ratios[1]);
        _splitPaneH2.setDividerLocation(ratios[2]);
        _splitPaneV2.setDividerLocation(ratios[3]);
    }

    public static void setRecentProjectFilenames( List<String> filenames ) {
        _recentlyClosedWorkspaceFilenames = filenames;
    }

    public static List<String> getRecentProjectFilenames() {
        return _recentlyClosedWorkspaceFilenames;
    }

    public void addRecentlyClosedWorkspace( String filename ) {
        if (_recentlyClosedWorkspaceFilenames.contains(filename)) {
            _recentlyClosedWorkspaceFilenames.remove(filename);
            _recentlyClosedWorkspaceFilenames.add(0, filename);
        } else {
            if (_recentlyClosedWorkspaceFilenames.size() >= MAXIMUM_RECENT_FILES) {
                _recentlyClosedWorkspaceFilenames.remove(MAXIMUM_RECENT_FILES - 1);
            }
            _recentlyClosedWorkspaceFilenames.add(0, filename);
        }
    }

    public void updateRecentlyClosedWorkspacesListMenu() {
        _menuBar.updateRecentlyClosedWorkspacesListMenu(_recentlyClosedWorkspaceFilenames, _presenter);
    }

    private void displayShortcutForOutputDisplay( InputMap inputMap, ActionMap actionMap, final int outputIndex ) {
        inputMap.put(KeyStroke.getKeyStroke(Character.valueOf((char) ('0' + outputIndex)), 0), "Display" + outputIndex);
        actionMap.put("Display" + outputIndex, new AbstractAction(null) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                if (_currentGraph != null) {
                    _currentGraph.displayOutputs(outputIndex);
                }
            }
        });
    }

    private int getSplitPaneRatio( JSplitPane splitPane, int oldRatio ) {
        if (Ariane.getFrame().getExtendedState() == Frame.MAXIMIZED_BOTH) {
            if (splitPane.getLastDividerLocation() < 0) {
                return oldRatio;
            }
            return splitPane.getLastDividerLocation();
        }
        return splitPane.getDividerLocation();
    }

    /////
    ///// M E N U  &  T O O L B A R
    /////
    private void buildFileActions() {
        _actions.put("new", new AbstractAction(null, ArianeSettings.getIcon("new.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                createNewGraph();
            }
        });
        _actions.put("open", new AbstractAction(null, ArianeSettings.getIcon("open.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                openNewGraph();
            }
        });
        _actions.put("save", new AbstractAction(null, ArianeSettings.getIcon("save.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                saveFileAction();
            }
        });
        _actions.put("saveAs", new AbstractAction(null, ArianeSettings.getIcon("saveas.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                saveAsFileAction();
            }
        });
        _actions.put("print", new AbstractAction(null, ArianeSettings.getIcon("print.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                print();
            }
        });
        _actions.put("importSubroutine", new AbstractAction() {
            @Override
            public void actionPerformed( ActionEvent event ) {
                importSubroutinesAction();
            }
        });
        _actions.put("exportSubroutine", new AbstractAction() {
            @Override
            public void actionPerformed( ActionEvent event ) {
                exportSubroutinesAction();
            }
        });
        _actions.put("close", new AbstractAction() {
            @Override
            public void actionPerformed( ActionEvent event ) {
                close();
            }
        });
        _actions.put("closeAll", new AbstractAction() {
            @Override
            public void actionPerformed( ActionEvent event ) {
                closeAll();
            }
        });
        _actions.put("quit", new AbstractAction(null) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                quitAction();
            }
        });
    }

    private void buildEditActions() {
        _actions.put("undo", new AbstractAction(null, ArianeSettings.getIcon("undo.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                undo();
            }
        });
        _actions.put("redo", new AbstractAction(null, ArianeSettings.getIcon("redo.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                redo();
            }
        });
        _actions.put("copy", new AbstractAction(null, ArianeSettings.getIcon("copy.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                copy();
            }
        });
        _actions.put("cut", new AbstractAction(null, ArianeSettings.getIcon("cut.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                cut();
            }
        });
        _actions.put("paste", new AbstractAction(null, ArianeSettings.getIcon("paste.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                paste();
            }
        });
        _actions.put("delete", new AbstractAction(null, ArianeSettings.getIcon("delete.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                deleteSelectedItems();
            }
        });
        _actions.put("select", new AbstractAction(null) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                selectAll();
            }
        });
    }

    private void buildSearchActions() {
        _actions.put("configuration", new AbstractAction(null) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                FadingEffects.fadeIn(Ariane.getFrame());
                _presenter.chooseOperatorLibrary();
                FadingEffects.fadeOut(Ariane.getFrame());
            }
        });
        _actions.put("preferences", new AbstractAction(null, ArianeSettings.getIcon("preferences.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                FadingEffects.fadeIn(Ariane.getFrame());
                _presenter.openPreferencesChooser();
                FadingEffects.fadeOut(Ariane.getFrame());
            }
        });
        _actions.put("operatorTree", new AbstractAction(null, ArianeSettings.getIcon("operatortree.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                FadingEffects.fadeIn(Ariane.getFrame());
                _presenter.openOperatorTreeEditor();
                FadingEffects.fadeOut(Ariane.getFrame());
            }
        });
        _actions.put("subroutineTree", new AbstractAction(null, ArianeSettings.getIcon("subroutinetree.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                FadingEffects.fadeIn(Ariane.getFrame());
                _presenter.openSubroutineTreeEditor();
                FadingEffects.fadeOut(Ariane.getFrame());
            }
        });
    }

    private void buildGraphActions() {
        _actions.put("stop", new AbstractAction(null, ArianeSettings.getIcon("stop.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                stopExecution();
            }
        });
        _actions.get("stop").setEnabled(false);
        _actions.put("exec", new AbstractAction(null, ArianeSettings.getIcon("execute.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                execute();
            }
        });
        _actions.get("exec").setEnabled(false);
        _actions.put("reset", new AbstractAction(null, null) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                reset();
            }
        });
        _actions.put("generate", new AbstractAction(null, ArianeSettings.getIcon("generation.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                FadingEffects.fadeIn(Ariane.getFrame());
                _presenter.generatesScriptCodeFromGraph(_currentGraph.getRoot());
                FadingEffects.fadeOut(Ariane.getFrame());
            }
        });
    }

    private void buildHelpActions() {
        _actions.put("about", new AbstractAction(null) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                FadingEffects.fadeIn(Ariane.getFrame());
                new AboutBox(Ariane.getFrame()).setVisible(true);
                FadingEffects.fadeOut(Ariane.getFrame());
            }
        });
        _actions.put("help", new AbstractAction(null, ArianeSettings.getIcon("help.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                displayHTMLManualPage("Ariane", "manual.html", null);
            }
        });
    }

    private void buildActions() {
        buildFileActions();
        buildEditActions();
        buildSearchActions();
        buildGraphActions();
        buildHelpActions();
    }

    public JMenuBar buildMenuBar() {
        return _menuBar;
    }

    private void setInternationalText() {
        JColorChooser.setDefaultLocale(Locale.getDefault());
        JFileChooser.setDefaultLocale(Locale.getDefault());
        JOptionPane.setDefaultLocale(Locale.getDefault());

        _toolBar.setInternationalText();
        _menuBar.setInternationalText();
        _graphEditor.setInternationalText();
    }

    public void changeLanguage() {
        setInternationalText();
        _itemExplorer.changeLocale();
    }

    public boolean isGraphAlreadyLoaded( String filename ) {
        return _graphEditor.contains(filename);
    }

    public void display( List<GraphPresenter> graphs ) {
        graphs.stream().filter(graph -> (graph != null)).forEach(this::addGraph);
    }

    public void displayPropertyPanel( PropertyPanel propertyPanel ) {
        _itemPropertyPanel.removeAll();
        if (propertyPanel != null) {
            _itemPropertyPanel.add(propertyPanel, BorderLayout.CENTER);
        }
        _itemPropertyPanel.revalidate();
        _itemPropertyPanel.repaint();
    }

    /////
    ///// C O N S O L E
    /////
    public void printOutputMessage( String mesg ) {
        _console.displayOutputMessage(mesg);
    }

    public void printSystemMessage( String mesg ) {
        _console.displaySystemMessage(mesg);
    }

    public void printWarning( String mesg ) {
        _console.displayErrorMessage(ArianeSettings.getString("warning") + mesg);
    }

    public void printErrorMessage( String mesg ) {
        _console.displayErrorMessage(ArianeSettings.getString("error") + mesg);
    }

    public void printErrorMessage( ArianeSettings.ErrorCode errorCode, String item, int itemId, String mesg ) {
        _console.displayErrorMessage(item + " (" + itemId + "): " + ArianeSettings.getString("error") + "#" + errorCode.ordinal() + ": " + mesg);
    }

    /////
    /////  A C T I O N S
    /////

    /**
     * Displays the specified HTML 'page' as a manual page for the word 'title'. Pages are stored either
     * <ol>
     * <li>in the operators library package.
     * <li>in the Ariane jar file.
     * </ol>
     *
     * @param title the title of the manual.
     * @param page  the URL address of the manual content.
     * @param index index in the page as "#name".
     */
    public void displayHTMLManualPage( String title, String page, String index ) {
        URI uri = HTMLDocumentViewer.getURI(page, index);
        if (uri == null) {
            return;
        }
        String command = UserPreferences.getWebBrowserPath();
        if (command != null && command.length() > 0) {
            try {
                Runtime.getRuntime().exec(command + " " + uri);
                return;
            } catch (IOException e1) {
                printErrorMessage(ArianeSettings.getString("Workspace.browser.error") + e1.getMessage());
            }
        }
        new HTMLDocumentViewer(Ariane.getFrame(), title, uri).setVisible(true);
    }

    /////
    /////  H I S T O R Y
    /////
    private void undo() {
        if (_currentGraph != null) {
            _currentGraph.undo();
        }
    }

    private void redo() {
        if (_currentGraph != null) {
            _currentGraph.redo();
        }
    }

    public void updateMenu() {
        if (_currentGraph != null) {
            _actions.get("save").setEnabled(_currentGraph.isModified());
            _actions.get("undo").setEnabled(_currentGraph.canUndo());
            _actions.get("redo").setEnabled(_currentGraph.canRedo());
        }
    }

    public void setModifiedSaveB( boolean status ) {
        _actions.get("save").setEnabled(status);
    }

    public GraphPresenter getCurrentGraph() {
        return _currentGraph;
    }

    public void setCurrentGraph( GraphPresenter graph ) {
        _currentGraph = graph;
        if (_currentGraph == null) {
            _actions.get("close").setEnabled(false);
            _actions.get("closeAll").setEnabled(false);
            _actions.get("save").setEnabled(false);
            _actions.get("undo").setEnabled(false);
            _actions.get("redo").setEnabled(false);
            _actions.get("saveAs").setEnabled(false);
            _actions.get("print").setEnabled(false);
            _actions.get("copy").setEnabled(false);
            _actions.get("cut").setEnabled(false);
            _actions.get("paste").setEnabled(false);
            _actions.get("delete").setEnabled(false);
            _actions.get("generate").setEnabled(false);
            _actions.get("select").setEnabled(false);
        } else {
            _actions.get("close").setEnabled(true);
            _actions.get("closeAll").setEnabled(true);
            _actions.get("saveAs").setEnabled(true);
            if (!_actions.get("stop").isEnabled()) {
                _actions.get("exec").setEnabled(true);
            }
            _actions.get("print").setEnabled(true);
            _actions.get("copy").setEnabled(true);
            _actions.get("cut").setEnabled(true);
            _actions.get("paste").setEnabled(true);
            _actions.get("delete").setEnabled(true);
            _actions.get("generate").setEnabled(true);
            _actions.get("select").setEnabled(true);
            updateMenu();
        }
    }

    private void selectAll() {
        if (_currentGraph != null) {
            _currentGraph.selectAll();
        }
    }

    /////
    ///// C O P Y  /  P A S T E
    /////
    @Override
    public void lostOwnership( Clipboard clipboard, Transferable contents ) {
    }

    /**
     * Copies the selected items in the clipboard. Each item are copied in the clipboard.
     * <b>Warning</b>: connections between copied item are kept but connection between other items are deleted.
     */
    public void copy() {
        if (_currentGraph != null) {
            Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
            clipboard.setContents(_currentGraph.copySelectedItems(), this);
        }
    }

    /*
     * Warning: connections between copied items are kept
     * but connections between other items are removed.
     */
    public void cut() {
        copy();
        deleteSelectedItems();
    }

    public void paste() {
        Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
        Transferable contents = clipboard.getContents(null);
        if (_currentGraph != null && contents != null && contents.isDataFlavorSupported(XMLElement.FLAVOR)) {
            _presenter.createItemsFromClipboard(_currentGraph, contents);
        }
    }

    public boolean isPastable() {
        Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
        Transferable contents = clipboard.getContents(null);
        return _currentGraph != null && contents != null && contents.isDataFlavorSupported(XMLElement.FLAVOR);
    }

    public void deleteSelectedItems() {
        if (_currentGraph != null) {
            _currentGraph.deleteSelectedItems();
        }
    }

    public void addNewItemAtRelativePosition( TreeItem treeItem, Point coordinates ) {
        if (_currentGraph == null) {
            createNewGraph();
        }
        _currentGraph.addItemAtRelativePosition(treeItem, coordinates);
    }

    /////
    /////  P R I NT I N G
    /////
    private void print() {
        if (_currentGraph != null) {
            // Unselect all items.
            _currentGraph.unSelectAll();

            // Set up defaults preferences.
            PrintRequestAttributeSet aset = new HashPrintRequestAttributeSet();
            aset.add(new JobName("Ariane", UserPreferences.getCurrentLocale()));
            aset.add(OrientationRequested.LANDSCAPE);
            aset.add(MediaSizeName.ISO_A4);

            // Print it
            PrinterJob printJob = PrinterJob.getPrinterJob();
            printJob.setPrintable(_currentGraph.getView());
            if (printJob.printDialog(aset)) {
                try {
                    printJob.print(aset);
                } catch (PrinterException e) {
                    printErrorMessage(ArianeSettings.getString("Workspace.printing.error1") + e);
                }
            }
        }
    }

    public void repaintAllGraphs() {
        _graphEditor.repaintAllGraphs();
    }

    public void addGraph( GraphPresenter graph ) {
        _graphEditor.add(graph);
    }

    public void renameGraph( GraphPresenter graph, String name ) {
        graph.setName(name);
        _graphEditor.setTitle(graph, name);
    }

    public void removeGraph( GraphPresenter graph ) {
        if (_graphEditor.isOpen(graph)) {
            GraphPresenter newSelectedGraph = _graphEditor.remove(graph);
            setCurrentGraph(newSelectedGraph);
        }
    }

    public GraphView createNewGraph() {
        GraphPresenter graph = new RootGraphPresenter(ArianeSettings.getString("defaultfilename"));
        addGraph(graph);
        return graph.getView();
    }

    private void close() {
        if (_currentGraph != null) {
            _currentGraph.close();
            updateRecentlyClosedWorkspacesListMenu();
        }
    }

    private void closeAll() {
        _graphEditor.closeAllGraphs();
    }

    private void quitAction() {
        if (isQuitConfirmed()) {
            System.exit(0);
        }
    }

    private void execute() {
        if (_currentGraph != null) {
            _currentGraph.startExecution();
        }
    }

    private void reset() {
        if (_currentGraph != null) {
            _currentGraph.resetSelectedItems();
        }
    }

    private void stopExecution() {
        if (_currentGraph != null && GraphPresenter.isExecutionRunning()) {
            _currentGraph.stopExecution();
            printSystemMessage(ArianeSettings.getString("Workspace.execution.stop"));
        }
    }

    public void setExecution( boolean isRunning ) {
        if (isRunning) {
            _console.clear();
            _actions.get("exec").setEnabled(false);
            _actions.get("stop").setEnabled(true);
        } else {
            _actions.get("exec").setEnabled(true);
            _actions.get("stop").setEnabled(false);
        }
    }

    boolean isQuitConfirmed() {
        boolean quit = _graphEditor.closeAllGraphs();
        if (quit) {
            UserPreferences.deleteTemporaryFiles();
            UserPreferences.updateUserPreferenceFile();
        }
        return quit;
    }

    private void openNewGraph() {
        File currentFileName = new File(UserPreferences.getCurrentDirectory());

        JFileChooser fileChooser = new JFileChooser(currentFileName);
        fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
        fileChooser.setMultiSelectionEnabled(true);
        fileChooser.setFileFilter(new javax.swing.filechooser.FileFilter() {
            @Override
            public boolean accept( File file ) {
                String filename = file.getName();
                return filename.endsWith(ArianeSettings.DEFAULT_ARIANE_FILE_EXT) || file.isDirectory();
            }

            @Override
            public String getDescription() {
                return "Ariane (*" + ArianeSettings.DEFAULT_ARIANE_FILE_EXT + ")";
            }
        });

        FadingEffects.fadeIn(Ariane.getFrame());
        if (fileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
            for (File file : fileChooser.getSelectedFiles()) {
                _presenter.openFile(file.getAbsolutePath());
            }
            UserPreferences.setCurrentDirectory(fileChooser.getSelectedFiles()[0].getParent());
        }
        FadingEffects.fadeOut(Ariane.getFrame());
    }

    private void saveFileAction() {
        if (_currentGraph != null) {
            _currentGraph.save();
        }
    }

    private void saveAsFileAction() {
        if (_currentGraph != null) {
            _currentGraph.saveAs();
        }
    }

    private void importSubroutinesAction() {
        JFileChooser fileChooser = new JFileChooser();
        fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
        fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
        fileChooser.setCurrentDirectory(new File(ArianeSettings.getLibraryPath()));
        int returnVal = fileChooser.showOpenDialog(this);
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            File file = fileChooser.getSelectedFile();
            try { // TODO deplacer dans Presenter pour etre testé.
                _presenter.getTree().setSubroutineTree(ArianeSettings.readConfigurationFile(file));
                if (_presenter.saveSubroutineTree()) {
                    printSystemMessage(ArianeSettings.getString("Workspace.addsubroutine.message"));
                }
            } catch (IOException e) {
                printErrorMessage(ArianeSettings.getString("Workspace.addsubroutine.error") + " " + file.getName() + ". " + e.getLocalizedMessage());
            }
        }
    }

    private void exportSubroutinesAction() {
        JFileChooser fileChooser = new JFileChooser();
        fileChooser.setDialogType(JFileChooser.SAVE_DIALOG);
        fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
        fileChooser.setSelectedFile(new File(ArianeSettings.getLibraryPath()));
        if (fileChooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
            File file = fileChooser.getSelectedFile();
            String[] options = {ArianeSettings.getString("button.overwrite.text"), ArianeSettings.getString("button.nooverwrite.text")};
            if (file.exists() && MessageBox.showConfirmDialog(this,
                    ArianeSettings.getString("Workspace.dialog.message.part1")
                            + file.getName()
                            + ArianeSettings.getString("Workspace.dialog.message.part2"),
                    ArianeSettings.getString("Workspace.dialog.title"),
                    options, options[0]) != 0) {
                return;
            }
            try { // TODO deplacer dans Presenter pour etre testé.
                DataFile.copyFile(ArianeSettings.getSubroutineFilename(), file.getAbsolutePath());
                printSystemMessage(file.getName() + " " + ArianeSettings.getString("Workspace.saving.success"));
            } catch (IOException e) {
                printErrorMessage(ArianeSettings.getString("PreferenceChooser.save.error2") + " " + ArianeSettings.getPreferencesFilename() + " " + e.getLocalizedMessage());
            }
        }
    }
}
