/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.explorer;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.presenter.TreeEditorPresenter;
import fr.greyc.pantheon.swingext.GDialog;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.AbstractAction;
import javax.swing.Box;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JButton;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.border.EmptyBorder;
import javax.swing.border.LineBorder;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

/**
 * The class ItemTreeEditor defines a dialog box where users
 * can modify the tree of items (subroutine or operator subtree).
 */
abstract class ItemTreeEditor extends GDialog {
    private final MutableTreeItem _rootNode;
    private final JTree _jtree;
    private final DefaultTreeModel _jtreeModel;
    private final TreeEditorPresenter _presenter;
    private JButton _newButton;
    private JButton _editButton;
    private JButton _deleteButton;

    ItemTreeEditor( String title, TreeEditorPresenter presenter, MutableTreeItem rootNode ) {
        super((JFrame) null, title, true);
        super.setResizable(false);
        _presenter = presenter;
        _rootNode = (MutableTreeItem) rootNode.clone();
        _jtreeModel = new DefaultTreeModel(_rootNode);
        super.addWindowListener(new WindowAdapter() {
            @Override
            public final void windowClosing( WindowEvent event ) {
                cancelEdition();
            }
        });

        _jtree = new JTree(_jtreeModel);
        _jtree.getInputMap().put(KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_DELETE, 0), "Delete");
        _jtree.getActionMap().put("Delete", new AbstractAction(null) {
            @Override
            public final void actionPerformed( ActionEvent event ) {
                deleteSelectedItem();
            }
        });
        displayContent();
        TreePath path = new TreePath(_jtreeModel.getRoot());
        _jtree.setSelectionPath(path);
        selectNode();
        super.setLocationRelativeTo(Ariane.getFrame());
    }

    final void addNewItemBelowCurrentSelection() {
        TreePath currentSelection = _jtree.getSelectionPath();
        MutableTreeItem currentNode;
        if (currentSelection == null) {
            currentNode = _rootNode;
        } else {
            currentNode = (MutableTreeItem) currentSelection.getLastPathComponent();
        }
        if (currentNode.equals(_rootNode)) {
            addCategory(currentSelection, currentNode);
        } else {
            addOperator(currentSelection, currentNode);
        }
    }

    protected abstract void addSpecificMenuItem( JPopupMenu popupMenu );

    protected abstract boolean isNewActionEnable( TreePath path );

    protected abstract TreeItem createItem( String categoryName );

    protected abstract JDialog editItem( GDialog dialog, MutableTreeItem currentNode, DefaultTreeModel treeModel );

    protected abstract void storeInSettingFile( TreeEditorPresenter presenter, MutableTreeItem node );

    private void addOperator( TreePath currentSelection, MutableTreeItem currentNode ) {
        MutableTreeItem categoryNode;
        if (currentNode.getAllowsChildren()) {
            categoryNode = currentNode;
        } else {
            categoryNode = (MutableTreeItem) currentNode.getParent();
        }
        TreeItem newItem = createItem((String) categoryNode.getUserObject());
        if (newItem != null) {
            MutableTreeItem node = new MutableTreeItem(newItem, false);
            _jtreeModel.insertNodeInto(node, categoryNode, categoryNode.getChildCount());
            _jtree.expandPath(currentSelection);
            editItem(this, node, _jtreeModel).setVisible(true);
        }
    }

    private void addCategory( TreePath currentSelection, MutableTreeItem currentNode ) {
        MutableTreeItem node = new MutableTreeItem("category", true);
        _jtreeModel.insertNodeInto(node, currentNode, currentNode.getChildCount());
        _jtree.expandPath(currentSelection);
        _jtree.setExpandsSelectedPaths(true);
        startEditingCategory(node);
    }

    private void editCurrentItem() {
        TreePath currentSelection = _jtree.getSelectionPath();
        if (currentSelection != null) {
            MutableTreeItem selectedNode = (MutableTreeItem) currentSelection.getLastPathComponent();
            if (selectedNode.getLevel() > 0) {
                if (selectedNode.getUserObject() instanceof TreeItem) {
                    editItem(this, selectedNode, _jtreeModel).setVisible(true);
                } else {
                    startEditingCategory(selectedNode);
                }
            }
        }
    }

    private void deleteSelectedItem() {
        TreePath currentSelection = _jtree.getSelectionPath();
        if (currentSelection != null) {
            MutableTreeItem selectedNode = (MutableTreeItem) currentSelection.getLastPathComponent();
            MutableTreeItem parentNode = (MutableTreeItem) selectedNode.getParent();
            if (!selectedNode.equals(_rootNode) && parentNode != null) {
                _jtreeModel.removeNodeFromParent(selectedNode);
            }
            selectNode();
        }
    }

    private void openPopupMenuCategory( final JTree tree, int x, int y ) {
        JPopupMenu popupMenu = new JPopupMenu();
        JMenuItem item;
        addSpecificMenuItem(popupMenu);
        item = new JMenuItem(ArianeSettings.getString("ItemTreeEditor.editCategory"));
        item.addActionListener(event -> editCurrentItem());
        popupMenu.add(item);
        item = new JMenuItem(ArianeSettings.getString("ItemTreeEditor.deleteCategory"));
        item.addActionListener(event -> deleteSelectedItem());
        popupMenu.add(item);
        popupMenu.show(tree, x, y);
    }

    private void openPopupMenuItem( final JTree tree, int x, int y ) {
        JPopupMenu popupMenu = new JPopupMenu();
        JMenuItem item;
        item = new JMenuItem(ArianeSettings.getString("ItemTreeEditor.editItem"));
        item.addActionListener(event -> editCurrentItem());
        popupMenu.add(item);
        item = new JMenuItem(ArianeSettings.getString("ItemTreeEditor.deleteItem"));
        item.addActionListener(event -> deleteSelectedItem());
        popupMenu.add(item);
        popupMenu.show(tree, x, y);
    }

    private void displayContent() {
        setSize(300, 400);
        _jtree.setEditable(true);
        _jtree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
        _jtree.setShowsRootHandles(true);
        _jtree.addMouseListener(new InteractionController());
        _jtree.setBorder(new LineBorder(Color.BLACK));

        JPanel scrollPane = new JPanel(new BorderLayout());
        scrollPane.add(new JScrollPane(_jtree), BorderLayout.CENTER);
        scrollPane.setBorder(new EmptyBorder(10, 10, 10, 10));

        _editButton = new JButton(ArianeSettings.getString("ItemTreeEditor.editItem"));
        _editButton.addActionListener(event -> editCurrentItem());
        _editButton.setEnabled(false);

        _newButton = new JButton(ArianeSettings.getString("ItemTreeEditor.addItem"));
        _newButton.addActionListener(event -> addNewItemBelowCurrentSelection());
        _newButton.setEnabled(false);

        _deleteButton = new JButton(ArianeSettings.getString("ItemTreeEditor.deleteItem"));
        _deleteButton.addActionListener(event -> deleteSelectedItem());
        _deleteButton.setEnabled(false);

        JPanel buttonPanel1 = new JPanel();
        buttonPanel1.add(_editButton);
        buttonPanel1.add(_newButton);
        buttonPanel1.add(_deleteButton);

        JButton saveButton = new JButton(ArianeSettings.getString("ItemTreeEditor.save"), ArianeSettings.getIcon("gtk-apply.png"));
        saveButton.addActionListener(event -> saveCurrentTree());

        JButton cancelButton = new JButton();
        cancelButton.setAction(new AbstractAction(ArianeSettings.getString("ItemTreeEditor.cancel"), ArianeSettings.getIcon("gtk-cancel.png")) {
            @Override
            public final void actionPerformed( ActionEvent event ) {
                cancelEdition();
            }
        });
        Box buttonPanel = Box.createVerticalBox();
        buttonPanel.add(buttonPanel1);
        JPanel buttonPanel2 = new JPanel();
        buttonPanel2.add(saveButton);
        buttonPanel2.add(cancelButton);
        buttonPanel.add(buttonPanel2);
        setDefaultButton(saveButton);
        setCancelButton(cancelButton);

        getContentPane().setLayout(new BorderLayout());
        getContentPane().add(scrollPane, BorderLayout.CENTER);
        getContentPane().add(buttonPanel, BorderLayout.SOUTH);
    }

    private void openPopupMenu( int x, int y ) {
        MutableTreeItem selectedNode = (MutableTreeItem) _jtree.getLastSelectedPathComponent();
        if (selectedNode != null) {
            switch (selectedNode.getLevel()) {
                case 0:
                    openPopupMenuRoot(x, y);
                    break;
                case 1:
                    openPopupMenuCategory(_jtree, x, y);
                    break;
                case 2:
                    openPopupMenuItem(_jtree, x, y);
                    break;
                default:
                    break;
            }
        }
    }

    private void openPopupMenuRoot( int x, int y ) {
        JPopupMenu popupMenu = new JPopupMenu();
        JMenuItem item = new JMenuItem(ArianeSettings.getString("ItemTreeEditor.addCategory"));
        item.addActionListener(event -> addNewItemBelowCurrentSelection());
        popupMenu.add(item);
        popupMenu.show(_jtree, x, y);
    }

    private boolean isEditActionEnable( TreePath path ) {
        return path != null && path.getPathCount() > 1;
    }

    private boolean isDeleteActionEnable( TreePath path ) {
        return isEditActionEnable(path);
    }

    private void selectNode() {
        TreePath path = _jtree.getSelectionPath();
        _newButton.setEnabled(isNewActionEnable(path));
        _editButton.setEnabled(isEditActionEnable(path));
        _deleteButton.setEnabled(isDeleteActionEnable(path));
    }

    private void startEditingCategory( MutableTreeItem subtree ) {
        TreePath path = new TreePath(_jtreeModel.getPathToRoot(subtree));
        _jtree.setSelectionPath(path);
        _jtree.startEditingAtPath(path);
    }

    private void saveCurrentTree() {
        storeInSettingFile(_presenter, _rootNode);
        setVisible(false);
        dispose();
    }

    private void cancelEdition() {
        setVisible(false);
        dispose();
    }

    private class InteractionController implements MouseListener {

        @Override
        public final void mouseClicked( MouseEvent event ) {
            if (event.getClickCount() > 1) {
                editCurrentItem();
            } else {
                selectNode();
            }
        }

        @Override
        public final void mouseEntered( MouseEvent event ) {
        }

        @Override
        public final void mouseExited( MouseEvent event ) {
        }

        @Override
        public final void mousePressed( MouseEvent event ) {
            if (event.isPopupTrigger()) {
                openPopupMenu(event.getX(), event.getY());
            }
        }

        @Override
        public final void mouseReleased( MouseEvent event ) {
            if (event.isPopupTrigger()) {
                openPopupMenu(event.getX(), event.getY());
            }
        }
    }
}
