/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.explorer;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.tree.ItemLibrary;
import fr.greyc.pantheon.ariane.model.tree.ItemLibraryFactory;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.tree.TreeItemFactory;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

/**
 * The class ItemTreeModel contains all the available graph items
 * organized in a tree. This tree is displayed in the explorer from where
 * the user can drag and drop items into the graph editor.
 */
public final class ItemTreeModel extends DefaultTreeModel {
    private final MutableTreeItem _builtinSubtree = new MutableTreeItem();
    private final MutableTreeItem _commonOperatorSubtree = new MutableTreeItem();
    private MutableTreeItem _userOperatorSubtree = new MutableTreeItem();
    private MutableTreeItem _subRoutineSubtree = new MutableTreeItem();
    private String _pathToHTMLLibraryDocumentationFile;

    public ItemTreeModel() {
        super(new MutableTreeItem("Library", true));
        initContents();
    }

    public TreeItem getTreeItem( String name, String category ) {
        return getNode((MutableTreeItem) getRoot(), name, category);
    }

    public String getPathToLibraryDocumentationFile() {
        return _pathToHTMLLibraryDocumentationFile;
    }

    public MutableTreeItem getUserOperatorTree() {
        return _userOperatorSubtree;
    }

    public void setUserOperatorTree( MutableTreeItem newTree ) {
        _userOperatorSubtree = changeSubtree(_userOperatorSubtree, newTree);
    }

    public void setUserOperatorTree( List<XMLElement> itemTree ) {
        ItemLibrary library = ItemLibraryFactory.createItemsFromXMLTree(itemTree);
        addItemsInSubtree(library.treeItems, _userOperatorSubtree);
    }

    public MutableTreeItem getSubroutineTree() {
        return _subRoutineSubtree;
    }

    public void setSubroutineTree( MutableTreeItem newTree ) {
        _subRoutineSubtree = changeSubtree(_subRoutineSubtree, newTree);
    }

    public void setSubroutineTree( List<XMLElement> itemTree ) {
        ItemLibrary library = ItemLibraryFactory.createItemsFromXMLTree(itemTree);
        addItemsInSubtree(library.treeItems, _subRoutineSubtree);
    }

    public TreeItem getItemEqualsTo( String name ) {
        for (MutableTreeItem subtree : getSubTrees()) {
            List<TreePath> paths = getItemsWithTheSpecifiedNameInSpecifiedSubtree(subtree, name);
            for (TreePath treePath : paths) {
                MutableTreeItem leaf = (MutableTreeItem) treePath.getLastPathComponent();
                TreeItem item = leaf.getTreeItemObject();
                // Ensure that the operator BEGINS with the pattern.
                if (item.getName().equals(name)) {
                    return leaf.getTreeItemObject();
                }
            }
        }
        return null;
    }

    public List<String> getItemsWithNameStartingWith( String pattern ) {
        List<String> result = new ArrayList<>();
        for (MutableTreeItem subtree : getSubTrees()) {
            for (TreePath path : getItemsWithTheSpecifiedNameInSpecifiedSubtree(subtree, pattern)) {
                MutableTreeItem leaf = (MutableTreeItem) path.getLastPathComponent();
                TreeItem item = leaf.getTreeItemObject();
                // Ensure that the operator BEGINS with the pattern.
                if (item.getName().startsWith(pattern)) {
                    result.add(item.getName());
                }
            }
        }
        return result;
    }

    public void setCommonOperators( List<XMLElement> itemTree ) {
        ItemLibrary library = ItemLibraryFactory.createItemsFromXMLTree(itemTree);
        _commonOperatorSubtree.removeAllChildren();
        ((MutableTreeItem) getRoot()).setUserObject(library.version);
        reload();
        _pathToHTMLLibraryDocumentationFile = library.pathToHTMLLibraryDocumentationFile;
        addItemsInSubtree(library.treeItems, _commonOperatorSubtree);
    }

    void changeLocale() {
        _subRoutineSubtree.setUserObject(ArianeSettings.getString("Tree.subroutines.node"));
        _userOperatorSubtree.setUserObject(ArianeSettings.getString("Tree.personalOperators.node"));
        _commonOperatorSubtree.setUserObject(ArianeSettings.getString("Tree.libraryOperators.node"));
        _builtinSubtree.setUserObject(ArianeSettings.getString("Tree.builtinItems.node"));
        reload();
    }

    public List<TreeItem> getUserOperatorTreeAsList() {
        return subtreeToList(_userOperatorSubtree);
    }

    public List<TreeItem> getSubroutineTreeAsList() {
        return subtreeToList(_subRoutineSubtree);
    }

    void addSubroutine( TreeItem item ) {
        ArrayList<TreeItem> treeItems = new ArrayList<>();
        treeItems.add(item);
        addItemsInSubtree(treeItems, _subRoutineSubtree);
    }

    String[] getSubroutineCategories() {
        String[] categories = new String[_subRoutineSubtree.getChildCount()];
        for (int i = 0; i < _subRoutineSubtree.getChildCount(); i++) {
            categories[i] = _subRoutineSubtree.getChildAt(i).toString();
        }
        return categories;
    }

    List<TreePath> getItemsWithTheSpecifiedNameInTree( final String name ) {
        return getItemsWithTheSpecifiedNameInSpecifiedSubtree((MutableTreeItem) getRoot(), name);
    }

    List<TreePath> getItemsWithTheSpecifiedNameInSpecifiedSubtree( MutableTreeItem subtree, String name ) {
        ArrayList<TreePath> paths = new ArrayList<>();
        for (final Enumeration<?> e = subtree.children(); e.hasMoreElements(); ) {
            MutableTreeItem node = (MutableTreeItem) e.nextElement();
            if (node.isLeaf()) {
                String operatorName = node.toString();
                if (operatorName.contains(name)) {
                    paths.add(new TreePath(node.getPath()));
                }
            } else {
                paths.addAll(getItemsWithTheSpecifiedNameInSpecifiedSubtree(node, name));
            }
        }
        return paths;
    }

    void updateItemCategory( MutableTreeItem tree ) {
        for (final Enumeration<?> e1 = tree.children(); e1.hasMoreElements(); ) {
            MutableTreeItem node = (MutableTreeItem) e1.nextElement();
            String categoryName = (String) node.getUserObject();
            for (final Enumeration<?> e2 = node.children(); e2.hasMoreElements(); ) {
                MutableTreeItem node2 = (MutableTreeItem) e2.nextElement();
                TreeItem item = (TreeItem) node2.getUserObject();
                item.setCategory(categoryName);
            }
        }
    }

    private MutableTreeItem getCommonOperatorTree() {
        return _commonOperatorSubtree;
    }

    private void initContents() {
        createBuiltinItems();
        MutableTreeItem rootNode = (MutableTreeItem) getRoot();
        rootNode.add(_builtinSubtree);
        rootNode.add(_commonOperatorSubtree);
        rootNode.add(_subRoutineSubtree);
        rootNode.add(_userOperatorSubtree);
    }

    private void createBuiltinItems() {
        for (Object item : TreeItemFactory.createAllBuiltinItems()) {
            MutableTreeItem treeItem = new MutableTreeItem(item, false);
            _builtinSubtree.add(treeItem);
        }
    }

    private MutableTreeItem[] getSubTrees() {
        return new MutableTreeItem[]{
                getCommonOperatorTree(),
                getUserOperatorTree(),
                getSubroutineTree()};
    }

    private TreeItem getNode( MutableTreeItem subtree, String name, String category ) {
        for (final Enumeration<?> e = subtree.children(); e.hasMoreElements(); ) {
            MutableTreeItem node = (MutableTreeItem) e.nextElement();
            if (node.isLeaf()) {
                TreeItem treeItem = node.getTreeItemObject();
                if (name.equals(treeItem.getName()) && (category == null || category.equals(treeItem.getCategory()))) {
                    return treeItem;
                }
            } else {
                TreeItem item = getNode(node, name, category);
                if (item != null) {
                    return item;
                }
            }
        }
        return null;
    }

    private List<TreeItem> subtreeToList( MutableTreeItem subtree ) {
        ArrayList<TreeItem> treeList = new ArrayList<>();
        Enumeration<?> iterator = subtree.depthFirstEnumeration();
        while (iterator.hasMoreElements()) {
            MutableTreeItem node = (MutableTreeItem) iterator.nextElement();
            if (node.getUserObject() instanceof TreeItem) {
                treeList.add(node.getTreeItemObject());
            }
        }
        return treeList;
    }

    private MutableTreeItem changeSubtree( MutableTreeItem oldSubtree, MutableTreeItem newSubtree ) {
        updateItemCategory(newSubtree);
        MutableTreeItem rootNode = (MutableTreeItem) getRoot();
        int index = getIndexOfChild(rootNode, oldSubtree);
        removeNodeFromParent(oldSubtree);
        insertNodeInto(newSubtree, rootNode, index);
        return newSubtree;
    }

    private void addItemsInSubtree( List<TreeItem> treeItems, MutableTreeItem subtree ) {
        for (TreeItem treeItem : treeItems) {
            String categoryName = treeItem.getCategory();
            MutableTreeItem categoryItem = subtree.getNodeWithSpecifiedName(categoryName);
            if (categoryItem == null) {
                categoryItem = new MutableTreeItem(categoryName, true);
                subtree.add(categoryItem);
            } else {
                MutableTreeItem oldTreeItem = categoryItem.getNodeWithSpecifiedName(treeItem.getName());
                if (oldTreeItem != null) {
                    categoryItem.remove(oldTreeItem);
                }
            }
            MutableTreeItem item = new MutableTreeItem(treeItem, false);
            categoryItem.add(item);
        }
        reload(subtree);
    }
}
