/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.explorer;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.presenter.Presenter;

import java.awt.Cursor;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragSource;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragSourceListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.List;

import javax.swing.JPanel;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

/**
 * A ItemTreeView displays a tree of operator and subroutine items.
 */
public final class ItemTreeView extends JPanel implements DragSourceListener, DragGestureListener {
    private static final Cursor DRAG_AND_DROP_CURSOR;

    static {
        Point hotSpot = new Point(15, 15);
        DRAG_AND_DROP_CURSOR = Toolkit.getDefaultToolkit().createCustomCursor(ArianeSettings.getIcon("itemcursor.png").getImage(), hotSpot, "itemCursor");
    }

    private final ItemTreeModel _itemTreeModel;
    private final DragSource _dragSource;
    private final JTree _jtree;

    public ItemTreeView( final Presenter presenter, ItemTreeModel itemTreeModel ) {
        _itemTreeModel = itemTreeModel;
        _itemTreeModel.changeLocale();
        _jtree = new JTree(itemTreeModel);
        _jtree.setFont(UserPreferences.getDefaultFont());
        _jtree.getSelectionModel().setSelectionMode(TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
        _jtree.putClientProperty("JTree.lineStyle", "Angled");
        final DefaultTreeCellRenderer renderer = new DefaultTreeCellRenderer();
        renderer.setLeafIcon(ArianeSettings.getIcon("item.png"));
        renderer.setOpenIcon(null);
        renderer.setClosedIcon(null);
        _jtree.setCellRenderer(renderer);
        _jtree.addMouseListener(new MouseAdapter() {
            // Case of double-click -> display manual.
            @Override
            public void mousePressed( MouseEvent event ) {
                if (event.getClickCount() > 1) {
                    TreePath selectedPath = _jtree.getPathForLocation(event.getX(), event.getY());
                    TreeItem selectedTreeItem = getTreeItemAtTheSpecifiedPath(selectedPath);
                    if (selectedTreeItem != null) {
                        presenter.displayLibraryDocumentationPage(selectedTreeItem);
                    }
                }
            }
        });
        remove_CTRLA_FromShortCutsBecauseItIsUsedByGraphEditor();
        setLayout(new GridLayout());
        setBackground(_jtree.getBackground());
        add(_jtree);
        _dragSource = DragSource.getDefaultDragSource();
        _dragSource.createDefaultDragGestureRecognizer(_jtree, DnDConstants.ACTION_MOVE, this);
    }

    public void changeLocale() {
        _itemTreeModel.changeLocale();
    }

    /////
    ///// D R A G   &   D R O P
    /////
    @Override
    public void dragDropEnd( DragSourceDropEvent event ) {
        // Do nothing
    }

    @Override
    public void dragEnter( DragSourceDragEvent event ) {
        // Do nothing
    }

    @Override
    public void dragExit( DragSourceEvent event ) {
        // Do nothing
    }

    @Override
    public void dragOver( DragSourceDragEvent event ) {
        // Do nothing
    }

    @Override
    public void dropActionChanged( DragSourceDragEvent event ) {
        // Do nothing
    }

    @Override
    public void dragGestureRecognized( DragGestureEvent event ) {
        TreePath selectedPathInTheTree = _jtree.getSelectionPath();
        TreeItem selectedTreeItem = getTreeItemAtTheSpecifiedPath(selectedPathInTheTree);
        if (selectedTreeItem != null) {
            _dragSource.startDrag(event, DRAG_AND_DROP_CURSOR, selectedTreeItem, this);
        }
    }

    void highlightItemsThatContainTheSpecifiedPattern( final String pattern ) {
        List<TreePath> paths = _itemTreeModel.getItemsWithTheSpecifiedNameInTree(pattern);
        if (!paths.isEmpty()) {
            TreePath[] pathsTab = paths.toArray(new TreePath[paths.size()]);
            _jtree.setSelectionPaths(pathsTab);
        }
    }

    private void remove_CTRLA_FromShortCutsBecauseItIsUsedByGraphEditor() {
        _jtree.getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_A, Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()), "none");
    }

    private TreeItem getTreeItemAtTheSpecifiedPath( TreePath selectedPathInTheTree ) {
        if (selectedPathInTheTree != null) {
            MutableTreeItem selectedNode = (MutableTreeItem) selectedPathInTheTree.getLastPathComponent();
            if (selectedNode.isLeaf() && selectedNode.getUserObject() instanceof TreeItem) {
                return (TreeItem) selectedNode.getUserObject();
            }
        }
        return null;
    }
}
