/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.explorer;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.model.exception.ArianeMissingParameterException;
import fr.greyc.pantheon.ariane.model.graph.element.CompositeItem;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.model.tree.SubroutineTreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.swingext.GDialog;
import fr.greyc.pantheon.swingext.HTMLDocumentEditor;
import fr.greyc.pantheon.swingext.MessageBox;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

/**
 * The class SubroutineEditor defines a dialog box where
 * the user can add new subroutine in the explorer.
 */
public final class SubroutineEditor extends GDialog {
    private static String _currentDirectory = UserPreferences.getPreferredBinaryDirectory();
    private SubroutineTreeItem _currentItem;
    private MutableTreeItem _currentNode;
    private DefaultTreeModel _treeModel;
    private JComboBox<String> _categoryField;
    private JTextField _nameField;
    private JTextField _signatureField;
    private JTextField _docURLField;
    private boolean _inserted;
    private boolean _validated;

    public SubroutineEditor( CompositeItem item ) {
        // TODO pas ici... dans le presenter, c'est de la logique.
        super((JDialog) null, ArianeSettings.getString("SubroutineEditor.window.title"), true);
        addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing( WindowEvent event ) {
                cancelAction();
            }
        });

        XMLElement element;
        element = new XMLElement("Subroutine");
        element.addChildren(new XMLElement("Category", ""));
        element.addChildren(new XMLElement("Name", item.getNickName()));
        element.addChildren(new XMLElement("Documentation", ""));
        // Build the default parameter values list
        StringBuilder list = new StringBuilder();
        for (int i = 0; i < item.getInputParameterSlots().size(); i++) {
            ParameterSlot parameter = item.getInputParameterSlots().get(i);
            String name;
            if (parameter.hasName()) {
                name = parameter.getName();
            } else {
                name = "";
            }
            String defaultValue;
            try {
                boolean hasValue = parameter.hasValue(false);
                if (!parameter.isConnected() && hasValue) {
                    defaultValue = "=" + parameter.getValue();
                } else {
                    defaultValue = "";
                }
            } catch (ArianeMissingParameterException e) {
                defaultValue = "";
            }
            if (name.length() > 0 || !defaultValue.isEmpty()) {
                list.append("$").append(i).append(":").append(name).append(defaultValue).append(";");
            }
        }
        element.addChildren(new XMLElement("Signature", list.toString()));
        try {
            element.addChildren(item.convertToXML(null));
            _currentItem = new SubroutineTreeItem(element.getChildren());
            _inserted = true;
        } catch (IOException e) {
            _inserted = false;
        }
        displayContent();
        setLocationRelativeTo(null);
    }

    SubroutineEditor( JDialog parent, MutableTreeItem currentNode, DefaultTreeModel treeModel ) {
        super(parent, ArianeSettings.getString("SubroutineEditor.window.title"), true);
        _currentNode = currentNode;
        _currentItem = (SubroutineTreeItem) _currentNode.getUserObject();
        _treeModel = treeModel;
        _inserted = false;
        displayContent();
        setLocationRelativeTo(parent);
    }

    public boolean isValidated() {
        return _validated;
    }

    public SubroutineTreeItem getTreeItem() {
        return _currentItem;
    }

    private void displayContent() {
        setResizable(false);
        JPanel textPanel = new JPanel();
        textPanel.setLayout(new GridBagLayout());
        textPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), ArianeSettings.getString("SubroutineEditor.section.title")));
        GridBagConstraints cst = new GridBagConstraints();

        cst.fill = GridBagConstraints.BOTH;
        textPanel.add(new JLabel(ArianeSettings.getString("SubroutineEditor.category")), cst);
        String[] categories = Ariane.getPresenter().getTree().getSubroutineCategories();
        // TODO probleme ne voit pas les nouvelles categries créees
        // pire, elle ne prend pas en compte si on deplace la sousroutine.
        _categoryField = new JComboBox<>(categories);
        _categoryField.setEditable(true);
        _categoryField.setSelectedItem(_currentItem.getCategory());
        cst.gridwidth = GridBagConstraints.REMAINDER; //end row
        textPanel.add(_categoryField, cst);

        cst.gridwidth = 1;
        textPanel.add(new JLabel(ArianeSettings.getString("SubroutineEditor.name")), cst);
        _nameField = new JTextField(30);
        _nameField.setText(_currentItem.getName());
        cst.gridwidth = GridBagConstraints.REMAINDER; //end row
        textPanel.add(_nameField, cst);

        // Signature
        cst.gridwidth = 1;
        textPanel.add(new JLabel(ArianeSettings.getString("SubroutineEditor.signature")), cst);
        _signatureField = new JTextField(_currentItem.getSignature());
        cst.gridwidth = GridBagConstraints.REMAINDER; //end row
        textPanel.add(_signatureField, cst);
        textPanel.add(new JLabel(ArianeSettings.getString("SubroutineEditor.signatureExample")), cst);

        cst.gridwidth = 1;
        textPanel.add(new JLabel(ArianeSettings.getString("SubroutineEditor.documentation")), cst);
        _docURLField = new JTextField(30);
        _docURLField.setText(_currentItem.getDocumentationURL());
        textPanel.add(_docURLField, cst);
        JButton button2 = new JButton(ArianeSettings.getString("SubroutineEditor.load"));
        button2.addActionListener(event -> {
            JFileChooser fileChooser = new JFileChooser(_currentDirectory);
            fileChooser.setDialogType(JFileChooser.OPEN_DIALOG);
            fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
            final String prefix = "file://";
            if (_docURLField.getText().length() > prefix.length()) {
                fileChooser.setSelectedFile(new File(_docURLField.getText().substring(prefix.length())));
            }
            int returnVal = fileChooser.showOpenDialog(SubroutineEditor.this);
            if (returnVal == JFileChooser.APPROVE_OPTION) {
                _docURLField.setText(prefix + fileChooser.getSelectedFile().getAbsolutePath());
                _docURLField.setCaretPosition(0);
                _currentDirectory = fileChooser.getSelectedFile().getParentFile().getAbsolutePath();
            }
        });
        JButton button3 = new JButton(ArianeSettings.getString("OperatorEditor.edit"));
        button3.addActionListener(event -> {
            HTMLDocumentEditor editor = new HTMLDocumentEditor(SubroutineEditor.this);
            if (_docURLField.getText().length() > 6) {
                editor.setSelectedFile(new File(_docURLField.getText().substring(7)));
            }
            if (editor.edit() == HTMLDocumentEditor.OK_OPTION && editor.getSelectedFile() != null) {
                _docURLField.setText("file://" + editor.getSelectedFile().getAbsolutePath());
                _docURLField.setCaretPosition(0);
            }
        });
        textPanel.add(button2, cst);
        cst.gridwidth = GridBagConstraints.REMAINDER; //end row
        textPanel.add(button3, cst);

        // cancel button
        JButton cancelButton = new JButton();
        cancelButton.setAction(new AbstractAction(ArianeSettings.getString("SubroutineEditor.cancel"),
                ArianeSettings.getIcon("gtk-cancel.png")) {
            @Override
            public void actionPerformed( ActionEvent event ) {
                cancelAction();
            }
        });

        // Apply button
        JButton applyButton = new JButton(ArianeSettings.getString("SubroutineEditor.apply"), ArianeSettings.getIcon("gtk-apply.png"));

        applyButton.addActionListener(event -> applyAction());

        JPanel buttonPanel = new JPanel();
        buttonPanel.add(applyButton);
        buttonPanel.add(cancelButton);

        setDefaultButton(applyButton);
        setCancelButton(cancelButton);

        getContentPane().setLayout(new BorderLayout());
        getContentPane().add(textPanel, BorderLayout.NORTH);
        getContentPane().add(buttonPanel, BorderLayout.SOUTH);
        pack();
    }

    private void cancelAction() {
        _validated = false;
        setVisible(false);
        dispose();
    }

    private void applyAction() {
        String category = "" + _categoryField.getSelectedItem();
        if (category.trim().isEmpty()) {
            Ariane.getPresenter().printErrorMessage(ArianeSettings.getString("SubroutineEditor.error.message1"));
            return;
        }
        _currentItem.setCategory(category);
        _currentItem.setName(_nameField.getText());
        _currentItem.setSignature(_signatureField.getText());
        _currentItem.setDocumentationURL(_docURLField.getText());
        if (_inserted) {
            ItemTreeModel treeModel = Ariane.getPresenter().getTree();
            List<TreePath> litems = treeModel.getItemsWithTheSpecifiedNameInSpecifiedSubtree(treeModel.getSubroutineTree(), _currentItem.getName());
            if (!litems.isEmpty()) {
                String[] options = {ArianeSettings.getString("button.overwrite.text"), ArianeSettings.getString("button.nooverwrite.text")};
                if (MessageBox.showConfirmDialog(null,
                        _currentItem.getName() + " " + ArianeSettings.getString("Tree.dialog.overwriteMessage"),
                        ArianeSettings.getString("Tree.dialog.overWriteTitle"),
                        options, options[0]) != 0) {
                    return;
                }
                Ariane.getPresenter().printWarning(ArianeSettings.getString("CompositeItem.message2"));
            }
            treeModel.addSubroutine(_currentItem);
            _validated = true;
        } else {
            _treeModel.nodeChanged(_currentNode);
            _validated = false;
        }
        setVisible(false);
        dispose();
    }
}
