/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.graph;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;

final class ElasticLink {
    private final Point _connectionStartPoint; // Use graph absolute coordinates
    private Point _connectionEndPoint;

    ElasticLink( GraphView graph, int x, int y ) {
        _connectionEndPoint = screenToGraph(graph, x, y);
        _connectionStartPoint = screenToGraph(graph, x, y);
    }

    void moveElastic( GraphView graph, Point coordinate ) {
        _connectionEndPoint = screenToGraph(graph, coordinate.x, coordinate.y);
        Rectangle rectangle = new Rectangle(_connectionEndPoint.x, _connectionEndPoint.y, 1, 1);
        graph.scrollRectToVisible(rectangle);
    }

    void paint( Graphics graphics ) {
        draw(graphics, _connectionStartPoint, _connectionEndPoint);
    }

    private void draw( Graphics graphics, Point p1, Point p2 ) {
        graphics.setColor(Color.RED);
        drawLine(graphics, p1.x, p1.y, p2.x, p2.y);
    }

    private Point screenToGraph( GraphView graph, int x, int y ) {
        Point p1 = graph.getParent().getLocationOnScreen();
        Point p2 = graph.getLocation();
        return new Point(x - p1.x - p2.x, y - p1.y - p2.y);
    }

    // Bresenham's line drawing algorithm, taken from:
    // http://graphics.lcs.mit.edu/~mcmillan/comp136/Lecture6/Lines.html
    // The line-drawing routine in the AWT is actually buggy.  When I
    // draw lines using graphics contexts from different sources (e.g.
    // passed in to the repaint() routine, or gotten with getGraphics()),
    // I get different lines, which leaves artifacts when redrawing
    // rubber-bands.
    private void drawLine( Graphics graphicContext, int x1, int y1, int x2, int y2 ) {
        int dy = y2 - y1;
        int dx = x2 - x1;
        int stepx;
        int stepy;

        if (dy < 0) {
            dy = -dy;
            stepy = -1;
        } else {
            stepy = 1;
        }
        if (dx < 0) {
            dx = -dx;
            stepx = -1;
        } else {
            stepx = 1;
        }
        dy <<= 1; // dy is now 2*dy
        dx <<= 1; // dx is now 2*dx

        int x = x1;
        int y = y1;
        graphicContext.drawLine(x, y, x, y);
        if (dx > dy) {
            int fraction = dy - (dx >> 1);    // same as 2*dy - dx
            while (x != x2) {
                if (fraction >= 0) {
                    y += stepy;
                    fraction -= dx;           // same as fraction -= 2*dx
                }
                x += stepx;
                fraction += dy;               // same as fraction -= 2*dy
                graphicContext.drawLine(x, y, x, y);
            }
        } else {
            int fraction = dx - (dy >> 1);
            while (y != y2) {
                if (fraction >= 0) {
                    x += stepx;
                    fraction -= dy;
                }
                y += stepy;
                fraction += dx;
                graphicContext.drawLine(x, y, x, y);
            }
        }
    }
}
