/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.graph;

import java.awt.Component;
import java.awt.Font;
import java.awt.event.KeyEvent;
import java.io.File;
import java.util.List;
import java.util.Map;

import javax.swing.JScrollPane;
import javax.swing.KeyStroke;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.view.Workspace;
import fr.greyc.pantheon.swingext.multivieweditor.GMultiviewEditor;

import java.util.HashMap;

/**
 * The <code>GraphEditor</code> is a multiview editor of graph.
 * In each editor, each graph is displayed in a <code>JTabbedPane</code>.
 * One graph can be dragged from one editor to the other part.
 */
public final class GraphEditor extends GMultiviewEditor {
    private final Workspace _workspace;
    private final Map<Component, GraphPresenter> _indexTableOfCurrentGraphs = new HashMap<>();

    public GraphEditor( Workspace workspace ) {
        _workspace = workspace;
        Font font = workspace.getFont();
        super.setFont(font.deriveFont(Font.PLAIN, 11f));
        setCloseIcon(ArianeSettings.getIcon("window-close.png"));
    }

    public void add( GraphPresenter graph ) {
        JScrollPane scrollpane = getScrollPaneThatHolds(graph);
        boolean isExistingGraph = scrollpane != null;
        if (isExistingGraph) {
            selectTab(scrollpane);
        } else {
            scrollpane = createNewScrollpaneForGraph(graph);
            String name = createTabTitleFromGraphName(graph.getName(), graph.isRoot());
            addComponent(scrollpane, name);
            setSelectedComponent(scrollpane);
        }
    }

    public boolean contains( String filename ) {
        final File file = new File(filename);
        for (Component component : getAllComponents()) {
            GraphPresenter graph = _indexTableOfCurrentGraphs.get(component);
            if (graph.isAssociatedToFile(file)) {
                selectTab(component);
                return true;
            }
        }
        return false;
    }

    public void repaintAllGraphs() {
        for (Component scrollpane : getAllComponents()) {
            GraphPresenter graph = _indexTableOfCurrentGraphs.get(scrollpane);
            graph.getView().repaintAll();
        }
    }

    public boolean isOpen( GraphPresenter graph ) {
        return getScrollPaneThatHolds(graph) != null;
    }

    public GraphPresenter remove( GraphPresenter graph ) {
        final JScrollPane scrollpane = getScrollPaneThatHolds(graph);
        JScrollPane selectedScrollPane = (JScrollPane) removeComponent(scrollpane);
        if (selectedScrollPane != null) {
            return _indexTableOfCurrentGraphs.get(selectedScrollPane);
        } else {
            return null;
        }
    }

    public boolean closeAllGraphs() {
        List<Component> components = getAllComponents();
        while (!components.isEmpty()) {
            for (Component component : components) {
                GraphPresenter graph = _indexTableOfCurrentGraphs.get(component);
                // Closes only the root graph
                // because root graph deletes all subgraphs.
                if (graph != null && graph.isRoot()) {
                    if (!graph.close()) {
                        return false;
                    }
                    break;
                }
            }
            components = getAllComponents();
        }
        return true;
    }

    public void setTitle( GraphPresenter graph, String name ) {
        Component component = getScrollPaneThatHolds(graph);
        if (component != null) {
            setTabTitle(component, createTabTitleFromGraphName(name, graph.isRoot()));
        }
    }

    public void setInternationalText() {
        setMenuText(ArianeSettings.getString("MultiviewEditor.menuitem.split"), ArianeSettings.getString("MultiviewEditor.menuitem.unsplit"));
    }

    @Override
    public void close( Component component ) {
        GraphPresenter graph = _indexTableOfCurrentGraphs.get(component);
        if (graph != null && graph.close()) {
            _indexTableOfCurrentGraphs.remove(component);
        }
    }

    @Override
    public Component getSelectedComponent() {
        GraphPresenter graph = _workspace.getCurrentGraph();
        if (graph == null) {
            return null;
        } else {
            return getScrollPaneThatHolds(graph);
        }
    }

    @Override
    public void setSelectedComponent( Component component ) {
        _workspace.setCurrentGraph(_indexTableOfCurrentGraphs.get(component));
    }

    private JScrollPane getScrollPaneThatHolds( Object document ) {
        for (Component scrollpane : getAllComponents()) {
            if (_indexTableOfCurrentGraphs.get(scrollpane) == document) {
                return (JScrollPane) scrollpane;
            }
        }
        return null;
    }

    private JScrollPane createNewScrollpaneForGraph( GraphPresenter graph ) {
        JScrollPane scrollpane;
        scrollpane = new JScrollPane(graph.getView());
        // Discard the arrows events.
        scrollpane.setInputMap(WHEN_IN_FOCUSED_WINDOW, null);
        scrollpane.setInputMap(WHEN_FOCUSED, null);
        scrollpane.setInputMap(WHEN_ANCESTOR_OF_FOCUSED_COMPONENT, null);
        scrollpane.getVerticalScrollBar().setUnitIncrement(10);
        scrollpane.getHorizontalScrollBar().setUnitIncrement(10);
        scrollpane.unregisterKeyboardAction(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0));
        _indexTableOfCurrentGraphs.put(scrollpane, graph);

        return scrollpane;
    }

    private String createTabTitleFromGraphName( String title, boolean isRoot ) {
        if (isRoot) {
            return "<html><b><font color=FF0000>" + title + "</font></b></html>";
        } else {
            return title;
        }
    }
}
