/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.graph.element;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.graph.element.CompositeItem;
import fr.greyc.pantheon.ariane.model.tree.SubroutineTreeItem;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.view.command.ChangeCompositeStatusEdit;
import fr.greyc.pantheon.ariane.view.explorer.SubroutineEditor;
import fr.greyc.pantheon.ariane.view.graph.element.etc.ControlStructureColor;
import fr.greyc.pantheon.swingext.FadingEffects;

import java.awt.Color;
import java.awt.event.MouseEvent;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;

public final class CompositeItemView extends ControlItemView {
    private transient State _state = new Editable();

    @Override
    protected JPopupMenu initPopup() {
        return _state.initPopup();
    }

    @Override
    public Color getBackgroundColor() {
        return _state.getBackgroundColor();
    }

    @Override
    protected void onMouseDoubleClicked( MouseEvent event ) {
        _state.onMouseDoubleClicked(event);
    }

    public void changeEditableStatus( boolean editable ) {
        if (editable) {
            _state = new Editable();
            TreeItem relatedTreeItem = ((CompositeItem) getModel()).restoreDefaultTreeItem();
            getModel().setRelatedTreeItem(relatedTreeItem);
        } else {
            _state = new Uneditable();
        }
        draw();
    }

    public void changeState( StateName state ) {
        switch (state) {
            case Editable:
                _state = new Editable();
                break;
            case Uneditable:
                _state = new Uneditable();
                break;
            default:
                break;
        }
    }

    public boolean isUneditable() {
        return _state instanceof Uneditable;
    }

    /////
    ///// S T A T E
    /////
    public enum StateName {
        Editable, Uneditable
    }

    /**
     * The state design pattern to manage the composite
     * as an uneditable subroutine or an editable macro.
     */
    interface State {

        JPopupMenu initPopup();

        void onMouseDoubleClicked( MouseEvent event );

        Color getBackgroundColor();
    }

    class Uneditable implements State {

        @Override
        public final JPopupMenu initPopup() {
            JPopupMenu popupMenu = new JPopupMenu();

            popupMenu.add(createExecuteMenuItem());
            popupMenu.add(createResetMenuItem());
            popupMenu.addSeparator();
            popupMenu.add(createCopyMenuItem());
            popupMenu.add(createCutMenuItem());
            popupMenu.add(createDeleteMenuItem());
            popupMenu.addSeparator();
            popupMenu.add(createConversionItem());

            return popupMenu;
        }

        private JMenuItem createConversionItem() {
            JMenuItem item;
            item = new JMenuItem(ArianeSettings.getString("PopupMenu.convertAsItem"));
            item.addActionListener(event -> {
                removePropertyViewer();
                changeEditableStatus(true);
                getGraph().setModified(true);
                getGraph().validateOneAction(new ChangeCompositeStatusEdit((CompositeItem) getModel(), true));
            });
            return item;
        }

        @Override
        public void onMouseDoubleClicked( MouseEvent event ) {
            Ariane.getPresenter().printWarning(getModel().getNickName() + ": " + ArianeSettings.getString("CompositeItem.message1"));
        }

        @Override
        public Color getBackgroundColor() {
            return ControlStructureColor.getColor().darker();
        }
    }

    class Editable implements State {

        @Override
        public void onMouseDoubleClicked( MouseEvent event ) {
            ((CompositeItem) getModel()).viewSubgraph();
        }

        @Override
        public final JPopupMenu initPopup() {
            JPopupMenu popupMenu = new JPopupMenu();

            popupMenu.add(createExecuteMenuItem());
            popupMenu.add(createResetMenuItem());

            JMenuItem item;

            popupMenu.addSeparator();
            item = new JMenuItem(ArianeSettings.getString("PopupMenu.openSubgraph"));
            item.addActionListener(event -> ((CompositeItem) getModel()).viewSubgraph());
            popupMenu.add(item);
            popupMenu.addSeparator();
            popupMenu.add(createCopyMenuItem());
            popupMenu.add(createCutMenuItem());
            popupMenu.add(createDeleteMenuItem());
            popupMenu.addSeparator();
            popupMenu.add(createChangePrototypeMenuItem());
            popupMenu.addSeparator();

            item = new JMenuItem(ArianeSettings.getString("PopupMenu.store"));
            item.addActionListener(event -> storeAsRoutine());
            popupMenu.add(item);

            return popupMenu;
        }

        JMenuItem createChangePrototypeMenuItem() {
            JMenuItem item = new JMenuItem(ArianeSettings.getString("PopupMenu.changePrototype"));
            item.addActionListener(event -> ((CompositeItem) getModel()).changePrototype());
            return item;
        }

        final void storeAsRoutine() {
            SubroutineEditor subroutineEditor = new SubroutineEditor((CompositeItem) getModel());
            FadingEffects.fadeIn(Ariane.getFrame());
            subroutineEditor.setVisible(true);
            FadingEffects.fadeOut(Ariane.getFrame());
            if (subroutineEditor.isValidated()) {
                SubroutineTreeItem relatedTreeItem = subroutineEditor.getTreeItem();
                getModel().setSignature(relatedTreeItem.getSignature());
                getModel().setNickName(relatedTreeItem.getName());
                getModel().setAbstractName(relatedTreeItem.getName());
                getModel().setRelatedTreeItem(relatedTreeItem);
                removePropertyViewer();
                changeEditableStatus(false);
                getGraph().setModified(true);
                getGraph().validateOneAction(new ChangeCompositeStatusEdit((CompositeItem) getModel(), false));
            }
        }

        @Override
        public Color getBackgroundColor() {
            return ControlStructureColor.getColor();
        }
    }
}
