/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.graph.element;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.UserPreferences;
import fr.greyc.pantheon.ariane.model.graph.element.GraphItemModel;
import fr.greyc.pantheon.ariane.model.graph.slot.InputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.InputMaskSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputDataSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.OutputResultSlot;
import fr.greyc.pantheon.ariane.model.graph.slot.SlotModel;
import fr.greyc.pantheon.ariane.presenter.GraphItemPresenter;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.view.command.ChangeBinaryEdit;
import fr.greyc.pantheon.ariane.view.command.ChangeFormulaEdit;
import fr.greyc.pantheon.ariane.view.PropertyPanel;
import fr.greyc.pantheon.swingext.RoundedBorder;
import fr.greyc.pantheon.swingext.RoundedPanel;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.List;

import javax.swing.*;
import javax.swing.border.Border;
import javax.swing.border.EtchedBorder;

import java.awt.Toolkit;

/**
 * A <code>GraphElementView</code> defines the view of <code>GraphElement</code>.
 */
public abstract class GraphElementView extends RoundedPanel implements MouseListener, MouseMotionListener, FocusListener, KeyListener {
    private static final int ITEM_HEADER_SIZE = 12;
    private static final Color SELECTED_HIGHLIGHT_COLOR = new Color(255, 0, 0);
    private static final Border UNSELECTED_BORDER = new RoundedBorder(EtchedBorder.RAISED);
    private static final Border SELECTED_BORDER = new RoundedBorder(EtchedBorder.LOWERED, SELECTED_HIGHLIGHT_COLOR);
    private static final ImageIcon UNEXECUTED_STATUS_ICON;
    private static final ImageIcon RUNNING_STATUS_ICON;
    private static final ImageIcon EXECUTED_STATUS_ICON;
    private static final ImageIcon ERROR_STATUS_ICON;
    private static JPopupMenu _popupMenu;

    static {
        UNEXECUTED_STATUS_ICON = ArianeSettings.getIcon("StatusIcon0.png");
        RUNNING_STATUS_ICON = ArianeSettings.getIcon("StatusIcon1.png");
        EXECUTED_STATUS_ICON = ArianeSettings.getIcon("StatusIcon2.png");
        ERROR_STATUS_ICON = ArianeSettings.getIcon("StatusIcon3.png");
    }

    private final JLabel _labelStatus;
    private final JTextField _textFieldName;
    private final JLabel _labelName;
    private PropertyPanel _propertyPanel;
    private JComponent _commentPanel;
    private GraphItemPresenter _graphElementPresenter;

    GraphElementView() {
        _labelName = new JLabel();
        _textFieldName = new JTextField();
        _labelStatus = new JLabel();
        initContents();
    }

    private void initContents() {
        _labelName.setFont(UserPreferences.getDefaultFont());
        _textFieldName.addFocusListener(new FocusListener() {
            @Override
            public final void focusLost( FocusEvent event ) {
                _graphElementPresenter.setNickName(((JTextField) (event.getSource())).getText());
            }

            @Override
            public final void focusGained( FocusEvent event ) {
            }
        });
        _textFieldName.addActionListener(event -> _graphElementPresenter.setNickName(((JTextField) (event.getSource())).getText()));
        setLayout(new GridBagLayout());
        addMouseListener(this);
        addMouseMotionListener(this);
        addFocusListener(this);
        addKeyListener(this);
    }

    public final void dispose() {
        if (_propertyPanel != null) {
            _propertyPanel.setVisible(false);
        }
    }

    public final void setGraphElementPresenter( GraphItemPresenter graphElementPresenter ) {
        _graphElementPresenter = graphElementPresenter;
        _commentPanel = createCommentEditor(); // Mis ici pour ne pas être dans le constructeur parce que la méthode est surchargée ailleurs.
    }

    final GraphItemPresenter getPresenter() {
        return _graphElementPresenter;
    }

    final GraphItemModel getModel() {  // TODO
        return getPresenter().getModel();
    }

    final GraphPresenter getGraph() {  // TODO
        return getPresenter().getModel().getGraph();
    }

    public final JTextField getNameEditor() {
        return _textFieldName;
    }

    boolean isSelected() {
        return getPresenter().isSelected();
    }

    public final void setSelected( boolean selected ) {
        setBorder(selected ? SELECTED_BORDER : UNSELECTED_BORDER);
    }

    final String getFormula() {  // TODO
        final String formula = getModel().getFormula();
        return formula == null ? "" : formula;
    }

    final void setFormula( String formula ) {  // TODO
        getModel().setFormula(formula);
    }

    @Override
    public final String toString() { // TODO Utile ???
        return _graphElementPresenter.toString();
    }

    @Override
    public final String getName() {
        if (_graphElementPresenter == null) {
            return null;
        } else {
            return getModel().getName();
        }
    }

    JComponent createCommentEditor() { // TODO Surchargée...
        JLabel label = new JLabel();
        label.setMaximumSize(new Dimension(150, 10000));

        int size = UserPreferences.getDefaultFont().getSize();
        label.setFont(UserPreferences.getDefaultFont().deriveFont(size - 1.0f));
        label.setBorder(BorderFactory.createEmptyBorder(2, 2, 2, 2));
        return label;
    }

    public final void setComment( String comment ) {
        if (_commentPanel instanceof JLabel) {
            ((JLabel) _commentPanel).setText(comment);
        } else if (_commentPanel instanceof JTextArea) {
            ((JTextArea) _commentPanel).setText(comment);
        }
    }

    /////
    ///// E V E N T
    /////
    @Override
    public final void mouseClicked( MouseEvent event ) {
        if (event.getClickCount() > 1) {
            onMouseDoubleClicked(event);
        }
    }

    void onMouseDoubleClicked( MouseEvent event ) {
    }

    @Override
    public final void mousePressed( MouseEvent event ) {
        if (!event.isShiftDown() && !event.isControlDown()) {
            requestFocusInWindow();
        }
        getParent().setComponentZOrder(this, 0); // Put it in front
        if (event.isPopupTrigger()) { // Unix version -> open the popup menu
            _graphElementPresenter.addInCurrentSelection(false);
            drawPopup(event);
        }
    }

    @Override
    public final void mouseReleased( MouseEvent event ) {
        if (event.isShiftDown() || event.isControlDown()) {
            hidePropertyPanel();
        } else {
            showPropertyPanel();
        }
        onMouseReleased(event);
    }

    final void onMouseReleased( MouseEvent event ) {
        if (event.isPopupTrigger()) { // Window version -> open the popup menu
            _graphElementPresenter.addInCurrentSelection(false);
            drawPopup(event);
        } else if ((_popupMenu == null || !_popupMenu.isVisible())
                && !getGraph().dropSelection()
                && (event.isShiftDown() || event.isControlDown())) {
            _graphElementPresenter.addInCurrentSelection(true);
        }
    }

    @Override
    public final void mouseDragged( MouseEvent event ) {
        if (_popupMenu == null || !_popupMenu.isVisible()) {
            if (!_graphElementPresenter.isSelected()) {
                _graphElementPresenter.addInCurrentSelection(event.isShiftDown() || event.isControlDown());
            }
            event.translatePoint(getLocation().x, getLocation().y);
            getGraph().dragSelection(event);
        }
    }

    @Override
    public final void mouseMoved( MouseEvent event ) {
    }

    @Override
    public final void mouseEntered( MouseEvent event ) {
    }

    @Override
    public final void mouseExited( MouseEvent event ) {
    }

    @Override
    public void focusLost( FocusEvent event ) {
    }

    void focusGainedHook() {
    }

    @Override
    public final void focusGained( FocusEvent event ) {
        showPropertyPanel();
        if (_graphElementPresenter.isSelected()) { // TODO Deplacer dans le presenter
            focusGainedHook();
        } else {
            _graphElementPresenter.addInCurrentSelection(false);
        }
    }

    @Override
    public final void keyPressed( KeyEvent event ) {
        getGraph().getView().onKeyPressed(event);
    }

    @Override
    public final void keyReleased( KeyEvent event ) {
    }

    @Override
    public final void keyTyped( KeyEvent event ) {
    }

    /////
    ///// D I S P L A Y I N G    &    P R I N T I N G
    /////
    protected abstract Color getBackgroundColor();

    public void draw() {
        removeAll();
        setBackground(getBackgroundColor());
        setBorder(UNSELECTED_BORDER);

        _labelName.setText(getModel().getNickName()); // TODO danss le presenter.
        setToolTipText(toString());

        // layout
        GridBagConstraints constraints = new GridBagConstraints();
        constraints.gridheight = 1;

        // 1. parameters slots
        constraints.anchor = GridBagConstraints.NORTHWEST;
        constraints.gridy = 0;
        constraints.insets = new Insets(1, 5, 1, 0);

        constraints.gridx = 0;
        add(Box.createRigidArea(new Dimension(0, ITEM_HEADER_SIZE)), constraints);

        List<ParameterSlot> inputParameterSlots = getModel().getInputParameterSlots();
        // TODO retouner une liste de slotview..
        final Insets insets = new Insets(1, 0, 1, 0);
        for (int i = 0; i < inputParameterSlots.size(); i++) {
            SlotModel slot = inputParameterSlots.get(i);
            constraints.gridx = i;
            add(slot.getView(), constraints);
            constraints.insets = insets;
        }

        // 2. The mask slot
        int numberOfParamaters = inputParameterSlots.size();
        if (numberOfParamaters == 0) {
            numberOfParamaters = 1;
        }

        constraints.insets = new Insets(0, 0, 0, 5);
        constraints.anchor = GridBagConstraints.NORTHEAST;
        constraints.gridx = numberOfParamaters;
        InputMaskSlot inputMaskSlot = getModel().getInputMaskSlot(); // TODO dans le presenter

        if (inputMaskSlot != null) {
            add(inputMaskSlot.getView(), constraints);
        }
        constraints.insets = new Insets(0, 0, 0, 0);

        // 3. The name
        constraints.gridwidth = numberOfParamaters + 1;
        constraints.gridy++;
        constraints.gridx = 0;
        constraints.anchor = GridBagConstraints.CENTER;
        add(_labelName, constraints);

        // 4. The lights ramp that shows the current status.
        constraints.gridwidth = GridBagConstraints.REMAINDER;
        constraints.gridy++;
        constraints.gridx = 0;
        displayStatus();
        add(_labelStatus, constraints);

        // 5. Optional Comments
        constraints.gridwidth = GridBagConstraints.REMAINDER;
        constraints.gridy++;
        constraints.gridx = 0;
        if (_commentPanel.getPreferredSize().width > _commentPanel.getMaximumSize().width) {
            Dimension dim = _commentPanel.getPreferredSize();
            dim.width = _commentPanel.getMaximumSize().width;
            _commentPanel.setPreferredSize(dim);
        }
        add(_commentPanel, constraints);

        // 6. The input + output slots
        constraints.insets = new Insets(0, 0, 1, 1);
        constraints.gridy++;
        constraints.gridx = 0;
        List<InputDataSlot> inputDataSlots = getModel().getInputDataSlots(); // TODO recuperer les vues correspondantes
        List<OutputDataSlot> outputDataSlots = getModel().getOutputDataSlots(); // TODO

        int maxSlot = Math.max(inputDataSlots.size(), outputDataSlots.size());
        for (int i = 0; i < maxSlot; i++) {
            if (i < inputDataSlots.size()) {
                constraints.gridx = 0;
                constraints.anchor = GridBagConstraints.WEST;
                SlotModel slot = inputDataSlots.get(i);
                //				slot.updateBackgroundColor();
                add(slot.getView(), constraints);
            }
            if (i < outputDataSlots.size()) {
                constraints.gridx = numberOfParamaters;
                constraints.anchor = GridBagConstraints.EAST;
                SlotModel slot = outputDataSlots.get(i);
                add(slot.getView(), constraints);
            }
            constraints.gridy++;
        }

        // 8. The result
        int max = Math.max(inputDataSlots.size(), outputDataSlots.size());
        constraints.gridy += max;
        constraints.insets = new Insets(0, 0, 2, 10);
        List<OutputResultSlot> outputResultSlots = getModel().getOutputResultSlots();

        if (!outputResultSlots.isEmpty()) {
            constraints.gridx = numberOfParamaters;
            constraints.gridwidth = 1;
            constraints.gridheight = 1;
            constraints.anchor = GridBagConstraints.SOUTHEAST;
            add(getModel().getOutputResultSlot().getView(), constraints);
        }
    }

    public final void displayStatus() {
        switch (_graphElementPresenter.getExecutionStatus()) {
            case UNEXECUTED:
                _labelStatus.setIcon(UNEXECUTED_STATUS_ICON);
                break;
            case RUNNING:
                _labelStatus.setIcon(RUNNING_STATUS_ICON);
                break;
            case EXECUTED:
                _labelStatus.setIcon(EXECUTED_STATUS_ICON);
                break;
            case ERROR:
                _labelStatus.setIcon(ERROR_STATUS_ICON);
                break;
        }
    }

    final void centerText( String text, Graphics graphics, int row, boolean multiline ) {
        if (text == null) {
            return;
        }
        int line = row;
        FontMetrics fontMetrics;
        fontMetrics = graphics.getFontMetrics();
        Dimension size = getSize();
        if (fontMetrics.stringWidth(text) > size.width) {
            if (multiline) {
                String subtext = text;
                // Use several lines and do not center text.
                String tmp;
                while (subtext.length() > 0) {
                    int index = subtext.length();
                    tmp = subtext;
                    while (fontMetrics.stringWidth(tmp) > size.width - 4) {
                        tmp = subtext.substring(0, --index);
                    }
                    graphics.drawString(tmp, 2, line);
                    line += fontMetrics.getHeight();
                    subtext = subtext.substring(index);
                }
            }
            graphics.drawString(text, 2, line);
        } else { // Centers the text
            int column = (size.width - fontMetrics.stringWidth(text)) / 2;
            graphics.drawString(text, column, line);
        }
    }

    @Override
    public final void printChildren( Graphics gc ) {
    }

    @Override
    public final void printBorder( Graphics graphics ) {
        Dimension size;
        size = new Dimension(getSize());
        Graphics2D g2d = (Graphics2D) graphics;
        g2d.setColor(Color.black);
        BasicStroke lineStroke = new BasicStroke(2.0F, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
        g2d.setStroke(lineStroke);
        g2d.drawRect(0, 0, size.width, size.height);
    }

    @Override
    public void printComponent( Graphics gc ) {
        Dimension size;
        int row;
        size = new Dimension(getSize());
        gc.setColor(Color.white);
        gc.fillRect(0, 0, size.width, size.height);

        gc.setColor(Color.black);
        gc.setFont(UserPreferences.getDefaultFont());
        FontMetrics fontMetrics;
        fontMetrics = gc.getFontMetrics();
        row = fontMetrics.getHeight();
        // For aestheticism purpose, split title and parameters.
        gc.drawLine(0, row + 4, size.width, row + 4);
    }

    public final void displayDocumentationManual() {
        Ariane.getPresenter().displayLibraryDocumentationPage(getModel().getName(), getModel().getDocumentationManualURL());
    }

    /////
    ///// P R O P E R T Y    P A N E L
    /////

    /**
     * @return the list of specific properties (other than common properties).
     */
    public List<PropertyPanel.PropertyCouple> getSpecificProperties() {
        return null;
    }

    public JComponent getBinaryEditor() { // TODO
        String binary = getModel().getBinary();
        if (binary != null) {
            JTextField textField = new JTextField(binary);
            textField.setEnabled(false);
            textField.setEditable(false);
            return textField;
        }
        return null;
    }

    public JComponent getFormulaEditor() { // TODO
        JTextField formulaText = new JTextField(getModel().getFormulaText());
        formulaText.addFocusListener(new FocusListener() {
            @Override
            public final void focusLost( FocusEvent event ) {
                changeFormula((JTextField) event.getSource(), ((JTextField) event.getSource()).getText(), true);
            }

            @Override
            public final void focusGained( FocusEvent event ) {
            }
        });
        formulaText.addActionListener(( ActionEvent event ) -> changeFormula((JTextField) event.getSource(), ((JTextField) event.getSource()).getText(), true));
        return formulaText;
    }

    public void changeFormula( JComponent component, String newValue, boolean storeInHistory ) { // TODO
//        getModel().changeFormula(source, text, storeInHistory);
        final String oldFormula = getModel().getFormula();
        getModel().changeFormula(newValue);
        if (storeInHistory) {
            getGraph().validateOneAction(new ChangeFormulaEdit(getModel(), component, getModel().getFormula(), oldFormula));
        }

        if (component instanceof JTextField) {
            ((JTextField) component).setText(getModel().getFormulaText());
        } else if (component instanceof JButton) {
            ((JButton) component).setText(getModel().getFormulaText());
        }
        updatePropertyViewer();

    }

    public void changeBinary( JComponent component, String newValue, boolean storeInHistory ) {
        // TODO A mettre dans le grahpresenter.
//        getModel().changeBinary(source, text, storeInHistory);
        getModel().setBinary(newValue);
        if (component instanceof JTextField) {
            ((JTextField) component).setText(getModel().getBinary());
        } else if (component instanceof JButton) {
            ((JButton) component).setText(getModel().getBinary());
        }
        updatePropertyViewer();
        if (storeInHistory) {
            String oldBinary = getModel().getBinary();
            getGraph().validateOneAction(new ChangeBinaryEdit(getModel(), component, getModel().getBinary(), oldBinary));
        }
    }

    PropertyPanel createPropertyPanel() {
        return new PropertyPanel(ArianeSettings.getLanguage(), getModel());
    }

    public final void updatePropertyViewer() {
        if (_propertyPanel != null) {
            _propertyPanel.update(getModel());
        }
    }

    public void resetPropertyViewer() {
        removePropertyViewer();
        showPropertyPanel();
    }

    final void removePropertyViewer() {
        if (_propertyPanel != null) {
            _propertyPanel.setVisible(false);
        }
        _propertyPanel = null;
    }

    private void showPropertyPanel() {
        if (_propertyPanel == null || !_propertyPanel.isCorrectlyLocalized()) {
            Thread thread = new Thread(() -> {
                _propertyPanel = createPropertyPanel();
                _propertyPanel.setVisible(true);
                Ariane.getPresenter().displayPropertyPanel(getGraph(), _propertyPanel);
            });
            thread.start();
        } else {
            _propertyPanel.setVisible(true);
            Ariane.getPresenter().displayPropertyPanel(getGraph(), _propertyPanel);
        }
    }

    private void hidePropertyPanel() {
        Ariane.getPresenter().displayPropertyPanel(getGraph(), null);
    }

    public final void fireConnection() {
        if (_propertyPanel != null) {
            _propertyPanel.validate();
            _propertyPanel.repaint();
        }
    }

    public final void firePropertyName( String name ) {
        _labelName.setText(name);
    }

    public final void firePropertyTextName( String name ) {
        _textFieldName.setText(name);
    }

    /////
    ///// M E N U
    /////
    protected abstract JPopupMenu initPopup();

    private void drawPopup( MouseEvent event ) {
        _popupMenu = initPopup();
        _popupMenu.show(this, event.getX(), event.getY());
    }

    final JMenuItem createExecuteMenuItem() {
        JMenuItem item;

        item = new JMenuItem(ArianeSettings.getString("PopupMenu.execute"));
        item.addActionListener(event -> getGraph().startExecution());
        item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
        return item;
    }

    final JMenuItem createDeleteMenuItem() {
        JMenuItem item;
        item = new JMenuItem(ArianeSettings.getString("PopupMenu.remove"));
        item.addActionListener(event -> getModel().delete());
        item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
        return item;
    }

    final JMenuItem createCutMenuItem() {
        JMenuItem item;
        item = new JMenuItem(ArianeSettings.getString("Graph.Menu.cut"));
        item.addActionListener(event -> Ariane.getPresenter().cut());
        item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
        return item;
    }

    final JMenuItem createCopyMenuItem() {
        JMenuItem item;
        item = new JMenuItem(ArianeSettings.getString("Graph.Menu.copy"));
        item.addActionListener(event -> Ariane.getPresenter().copy());
        item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
        return item;
    }

    final JMenuItem createResetMenuItem() {
        JMenuItem item;
        item = new JMenuItem(ArianeSettings.getString("PopupMenu.reset"));
        item.addActionListener(event -> getModel().resetAll()); // TODO
        item.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
        return item;
    }
}
