/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.imageio;

import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.Toolkit;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;

/**
 * The class <code>ImageDecoder</code> provides facilities to decode image files.
 * Recognized image formats are gif, png, jpg and Pandore. ?? tif, bmp, ppm, pgm, pbm
 */
public final class ImageDecoder {

    private ImageDecoder() {
    }

    public static ImageIcon getThumbnail( String filename, int maxSize ) {
        Image image;

        // may be a regular image file (bmp, png, ...)
        try {
            image = ImageIO.read(new File(filename));
        } catch (IOException e) {
            image = null;
        }

        // May be a Pandore image??
        if (image == null) {
            PandoreImageDecoder pan = new PandoreImageDecoder(maxSize);
            if (pan.readImage(filename)) {
                image = Toolkit.getDefaultToolkit().createImage(pan);
            }
        }

        // Not a regular image
        if (image == null) {
            return null;
        }

        // Resize image.
        if (image.getWidth(null) > image.getHeight(null)) {
            if (image.getWidth(null) > maxSize) {
                // Optimization for custom images (otherwise resampling is very very slow
                if (image instanceof BufferedImage && ((BufferedImage) image).getType() == BufferedImage.TYPE_CUSTOM) {
                    BufferedImage image1 = new BufferedImage(image.getWidth(null), image.getHeight(null), BufferedImage.TYPE_INT_RGB);
                    image1.setData(((BufferedImage) image).getData());
                    image = scaleImage(image1, maxSize, -1);
                } else {
                    image = scaleImage(image, maxSize, -1);
                }
            }
        } else if (image.getHeight(null) > maxSize) {
            // Optimization for custom images (otherwise resampling is very very slow
            if (image instanceof BufferedImage && ((BufferedImage) image).getType() == BufferedImage.TYPE_CUSTOM) {
                BufferedImage image1 = new BufferedImage(image.getWidth(null), image.getHeight(null), BufferedImage.TYPE_INT_RGB);
                image1.setData(((BufferedImage) image).getData());
                image = scaleImage(image1, -1, maxSize);
            } else {
                image = scaleImage(image, -1, maxSize);
            }
        }

        return new ImageIcon(image);
    }

    private static Image scaleImage( Image image, int width, int height ) {
        int requiredWidth;
        int requiredHeight;

        if (width < 0) {
            requiredWidth = (int) ((double) height / image.getHeight(null) * image.getWidth(null));
            requiredHeight = height;
        } else if (height < 0) {
            requiredHeight = (int) ((double) width / image.getWidth(null) * image.getHeight(null));
            requiredWidth = width;
        } else {
            requiredWidth = width;
            requiredHeight = height;
        }
        double scaleFactor = (double) requiredHeight / image.getHeight(null);

        BufferedImage image2 = new BufferedImage(requiredWidth, requiredHeight, BufferedImage.TYPE_INT_RGB);
        Graphics2D imgG2d = (Graphics2D) image2.getGraphics();
        AffineTransform scaleTransform = AffineTransform.getScaleInstance(scaleFactor, scaleFactor);
        imgG2d.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
        imgG2d.setTransform(scaleTransform);
        imgG2d.drawImage(image, 0, 0, null);
        imgG2d.dispose();

        return image2;
    }
}
