/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.swingext;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.LayoutManager;

/**
 * The <code>AttributeValueLayout</code> class is a layout manager
 * that lays out a two column container's components.
 * The first column is the name of the attributes and the second one
 * is the value of the attributes.
 */
public final class AttributeValueLayout implements LayoutManager {
    private static final int NUMBER_OF_COLUMNS = 2;

    @Override
    public void addLayoutComponent( String name, Component comp ) {
        // Empty
    }

    @Override
    public void removeLayoutComponent( Component comp ) {
        // Empty
    }

    @Override
    public Dimension preferredLayoutSize( Container parent ) {
        synchronized (parent.getTreeLock()) {
            Insets insets = parent.getInsets();
            int ncomponents = parent.getComponentCount();
            int nrows = ncomponents / NUMBER_OF_COLUMNS;

            int width = 0;
            int height = 0;
            for (int i = 0; i < ncomponents; i++) {
                Component comp = parent.getComponent(i);
                Dimension dimension = comp.getPreferredSize();
                if (width < dimension.width) {
                    width = dimension.width;
                }
                if (height < dimension.height) {
                    height = dimension.height;
                }
            }
            return new Dimension(insets.left + insets.right + NUMBER_OF_COLUMNS * width + (NUMBER_OF_COLUMNS - 1),
                    insets.top + insets.bottom + nrows * height + (nrows - 1));
        }
    }

    @Override
    public Dimension minimumLayoutSize( Container parent ) {
        synchronized (parent.getTreeLock()) {
            Insets insets = parent.getInsets();
            int ncomponents = parent.getComponentCount();

            int nrows = ncomponents / NUMBER_OF_COLUMNS;
            int width = 0;
            int height = 0;
            for (int i = 0; i < ncomponents; i++) {
                Component comp = parent.getComponent(i);
                Dimension dimension = comp.getMinimumSize();
                if (width < dimension.width) {
                    width = dimension.width;
                }
                if (height < dimension.height) {
                    height = dimension.height;
                }
            }
            return new Dimension(insets.left + insets.right + NUMBER_OF_COLUMNS * width + (NUMBER_OF_COLUMNS - 1),
                    insets.top + insets.bottom + nrows * height + (nrows - 1));
        }
    }

    @Override
    public void layoutContainer( Container parent ) {
        synchronized (parent.getTreeLock()) {
            int ncomponents = parent.getComponentCount();
            if (ncomponents == 0) {
                return;
            }

            Insets insets = parent.getInsets();
            int nrows = ncomponents / NUMBER_OF_COLUMNS;

            int widthWOInsets = parent.getWidth() - (insets.left + insets.right);
            // Get the length of the first column.
            int widthOfFirstColumn = 0;
            for (int r = 0; r < nrows; r++) {
                Component comp = parent.getComponent(r * NUMBER_OF_COLUMNS);
                Dimension dimension = comp.getMinimumSize();
                if (dimension.width > widthOfFirstColumn) {
                    widthOfFirstColumn = dimension.width;
                }
            }
            int widthOfSecondColumn = widthWOInsets - widthOfFirstColumn - insets.left - insets.right;

            int heightWOInsets = parent.getHeight() - (insets.top + insets.bottom);
            int heightOnComponent = heightWOInsets / nrows;
            int extraHeightAvailable = (heightWOInsets - (heightOnComponent * nrows)) / 2;

            for (int r = 0, y = insets.top + extraHeightAvailable; r < nrows; r++, y += heightOnComponent) {
                int component = r * NUMBER_OF_COLUMNS;
                if (component < ncomponents) {
                    parent.getComponent(component).setBounds(insets.left, y, widthOfFirstColumn, heightOnComponent);
                }
                component++;
                if (component < ncomponents) {
                    parent.getComponent(component).setBounds(insets.left + widthOfFirstColumn, y, widthOfSecondColumn, heightOnComponent);
                }
            }
        }
    }
}
