/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.swingext;

import fr.greyc.pantheon.ariane.ArianeSettings;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;

import javax.swing.JEditorPane;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.WindowConstants;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;

/**
 * A <code>HTMLDocumentViewers</code> displays documentation as HTML pages.
 */
public final class HTMLDocumentViewer extends JFrame implements HyperlinkListener {
    private static final String ERROR_PAGE = "item.html";
    private final JEditorPane _webEditorPane;

    public HTMLDocumentViewer( JFrame parent, String title, URI page ) {
        _webEditorPane = new JEditorPane();
        _webEditorPane.addHyperlinkListener(this);
        super.getContentPane().add(new JScrollPane(_webEditorPane));
        super.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
        super.setBounds(200, 25, 800, 600);
        setPage(title, page);
        super.setLocationRelativeTo(parent);
    }

    /**
     * Finds the absolute address of the given document either in the operators library
     * documentation or in the Ariane manual pages.
     */
    public static URI getURI( String document, String fragment ) {
        if (document.isEmpty()) {
            try {
                return ArianeSettings.getDocumentationURI(ERROR_PAGE).toURI();
            } catch (URISyntaxException e) {
                return null;
            }
        }

        try {
            return directURI(document, fragment);
        } catch (IOException | URISyntaxException | IllegalArgumentException e) {
            // Nothing to do
        }

        try {
            return relativeURIToLibrary(document, fragment);
        } catch (IOException | URISyntaxException e) {
            // Nothing to do
        }

        try {
            return deprecatedURI1(document, fragment); // Kept for the sake of compatibility
        } catch (IOException | URISyntaxException e) {
            // Nothing to do
        }

        try {
            return deprecatedURI2(document, fragment);  // Kept for the sake of compatibility
        } catch (URISyntaxException e) {
            // Nothing to do
        }
        return null;
    }

    private static URI deprecatedURI2( String document, String fragment )
            throws URISyntaxException {
        URI page;
        // Ariane manual pages
        URL address = ArianeSettings.getDocumentationURI(document);
        // The last chance: the operator page.
        if (address == null) {
            return ArianeSettings.getDocumentationURI(ERROR_PAGE).toURI();
        } else {
            page = ArianeSettings.getDocumentationURI(document).toURI();
            return new URI(page.toString() + ((fragment == null) ? "" : fragment));
        }
    }

    private static URI deprecatedURI1( String document, String fragment ) throws IOException, URISyntaxException {
        URI page;
        File file = new File(ArianeSettings.getLibraryPath() + File.separator + document);
        // Check whether the file exists.
        String address = file.getCanonicalPath();
        page = new URI("file", address, fragment);
        InputStream stream = page.toURL().openStream(); // try to open it.
        stream.close();
        return page;
    }

    private static URI relativeURIToLibrary( String document, String fragment ) throws IOException, URISyntaxException {
        URI page;
        File file = new File(ArianeSettings.getLibraryPath() + ArianeSettings.getLibraryDocumentationPath() + File.separator + document);
        String address = file.getCanonicalPath();
        // Library documentation?
        page = new URI("file", address, fragment);
        // Check whether the file exists.
        InputStream stream = page.toURL().openStream(); // try to open it.
        stream.close();
        return page;
    }

    private static URI directURI( String document, String fragment ) throws URISyntaxException, IOException {
        URI page;
        if (fragment == null) {
            page = new URI(document);
        } else {
            page = new URI(document + fragment);
        }
        // Check whether the file exists.
        InputStream stream = page.toURL().openStream(); // try to open it.
        stream.close();
        return page;
    }

    private void setPage( String title, URI page ) {
        setTitle(title);
        if (page == null) {
            displayErrorPage(title);
        } else {
            try {
                _webEditorPane.setPage(page.toURL());
                _webEditorPane.setEditable(false);
            } catch (IOException e) {
                displayErrorPage(title);
            }
        }
    }

    @Override
    public void hyperlinkUpdate( HyperlinkEvent event ) {
        HyperlinkEvent.EventType eventtype = event.getEventType();
        if (eventtype == HyperlinkEvent.EventType.ACTIVATED) {
            URL url = event.getURL();
            try {
                _webEditorPane.setPage(url);
            } catch (IOException e) {
                displayErrorPage(url.toString());
            }
        }
    }

    private void displayErrorPage( String title ) {
        _webEditorPane.setContentType("text/html");
        _webEditorPane.setText("<html><body>" + ArianeSettings.getString("HTMLViewer.unknownpage") + title + "</body></html>");
    }
}
