/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.swingext;

import fr.greyc.pantheon.ariane.ImageFileUtils;
import fr.greyc.pantheon.imageio.ImageDecoder;

import java.awt.BorderLayout;
import java.awt.Cursor;
import java.awt.Dimension;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextArea;
import javax.swing.border.BevelBorder;
import javax.swing.border.EmptyBorder;

/**
 * An implementation of image previewer that is used in FileChooser.
 * Allows the visualization in thumbnail image for supported image format,
 * such as GIF or JPEG or PANDORE.
 */
public final class ImageFilePreviewer extends JPanel implements PropertyChangeListener, ImageFileUtils {
    private static final int MAXSIZE = 120;
    private static final int PREVIEW_LENGTH = 300;
    private ThumbnailReader _reader;

    public ImageFilePreviewer( JFileChooser fileChooser ) {
        super(new BorderLayout());
        super.setPreferredSize(new Dimension(200, 50));
        super.setBorder(new BevelBorder(BevelBorder.LOWERED));
        fileChooser.addPropertyChangeListener(this);
    }

    private void displayTextPreview( File selectedFile ) {
        add(loadText(selectedFile), BorderLayout.CENTER);
        revalidate();
        repaint();
    }

    private void displayImagePreview( File selectedFile ) {
        if (_reader != null) {
            _reader.kill();
        }
        _reader = new ThumbnailReader(this, selectedFile);
        _reader.start();
    }

    @Override
    public void propertyChange( PropertyChangeEvent event ) {
        if (event.getPropertyName().equals(JFileChooser.SELECTED_FILE_CHANGED_PROPERTY)) {
            File selectedFile = (File) event.getNewValue();
            removeAll();
            if (selectedFile != null) {
                if (ImageFileUtils.isText(selectedFile, false)) {
                    displayTextPreview(selectedFile);
                } else {
                    displayImagePreview(selectedFile);
                }
            }
        }
    }

    private final class ThumbnailReader extends Thread {

        private final JComponent _component;
        private final File _selectedFile;
        private Boolean _isKilled = false;

        ThumbnailReader( JComponent component, File selectedFile ) {
            _component = component;
            _selectedFile = selectedFile;
        }

        void kill() {
            _isKilled = true;
        }

        private boolean isKilled() {
            return _isKilled;
        }

        @Override
        public void run() {
            _component.getParent().getParent().setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
            ImageIcon imageIcon = loadImage(_selectedFile);
            if (isKilled()) {
                _component.getParent().getParent().setCursor(null);
                return;
            }
            JLabel label = new JLabel(imageIcon);
            if (isKilled()) {
                _component.getParent().getParent().setCursor(null);
                return;
            }
            _component.add(label, BorderLayout.CENTER);
            if (isKilled()) {
                _component.getParent().getParent().setCursor(null);
                return;
            }
            _component.getParent().getParent().setCursor(null);
            revalidate();
            repaint();
        }
    }

    private ImageIcon loadImage( File selectedFile ) {
        if (selectedFile != null) {
            return ImageDecoder.getThumbnail(selectedFile.getPath(), MAXSIZE);
        }
        return null;
    }

    private JTextArea loadText( File selectedFile ) {
        if (selectedFile != null) {
            try {
                String preview;
                try (FileInputStream fin = new FileInputStream(selectedFile)) {
                    final byte[] data = new byte[PREVIEW_LENGTH];
                    int bytesRead = 0;
                    for (int i = 0; i < PREVIEW_LENGTH; i++) {
                        int character = fin.read();
                        if (character == -1) {
                            break;
                        }
                        bytesRead++;
                        data[i] = (byte) character;
                    }
                    preview = new String(data, 0, bytesRead);
                }
                JTextArea textArea = new JTextArea(preview);
                textArea.setEditable(false);
                textArea.setBorder(new EmptyBorder(5, 5, 5, 5));
                textArea.setFont(textArea.getFont().deriveFont(8.0f));
                return textArea;
            } catch (IOException ex) {
                return null;
            }
        }
        return null;
    }

}
