/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */

package fr.greyc.pantheon.swingext.multivieweditor;

import java.awt.Component;
import java.util.ArrayList;
import java.util.List;

import javax.swing.Icon;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JSplitPane;
import javax.swing.TransferHandler;

/**
 * Multiview editor. A view holds several tabbed panes.
 */
public abstract class GMultiviewEditor extends JSplitPane {
    private static Icon _closeIcon;
    private JMenuItem _splitMenuItem;
    private JMenuItem _unsplitMenuItem;

    protected GMultiviewEditor() {
        TransferHandler handler = new TabTransferHandler();
        super.setOrientation(JSplitPane.HORIZONTAL_SPLIT);
        DnDTabbedPane tab1 = new DnDTabbedPane(handler);
        DnDTabbedPane tab2 = new DnDTabbedPane(handler);
        super.setLeftComponent(tab1);
        super.setRightComponent(tab2);
        JPopupMenu popup = createTabPopupMenu();
        tab1.setComponentPopupMenu(popup);
        tab2.setComponentPopupMenu(popup);
        super.setOneTouchExpandable(true);
        super.setDividerLocation(0.0);
    }

    protected final void setCloseIcon( Icon icon ) {
        _closeIcon = icon;
    }

    final Icon getCloseIcon() {
        return _closeIcon;
    }

    public abstract void close( Component component );

    public abstract void setSelectedComponent( Component component );

    protected abstract Component getSelectedComponent();

    public final List<Component> getAllComponents() {
        List<Component> objects = new ArrayList<>();

        for (DnDTabbedPane tabbedPane : getTabbedPanes()) {
            for (int i = 0; i < tabbedPane.getTabCount(); i++) {
                Component scrollpane = tabbedPane.getComponentAt(i);
                objects.add(scrollpane);
            }
        }
        return objects;
    }

    protected final void addComponent( Component scrollpane, String name ) {
        if (((DnDTabbedPane) getLeftComponent()).getTabCount() == 0 && ((DnDTabbedPane) getRightComponent()).getTabCount() == 0) {
            setSplittingEnabled(true);
        }
        DnDTabbedPane tab = getSelectedPane();
        tab.add(scrollpane);
        tab.setTabComponentAt(tab.indexOfComponent(scrollpane), new TabTitle(this, scrollpane, name));
        tab.setSelectedComponent(scrollpane);
    }

    protected final Component removeComponent( Component scrollpane ) {
        if (scrollpane != null) {
            DnDTabbedPane tabbedpane = (DnDTabbedPane) scrollpane.getParent();
            tabbedpane.remove(scrollpane);
            if (tabbedpane.getSelectedComponent() == null) {
                unsplit();
                tabbedpane = (DnDTabbedPane) getRightComponent();
                return tabbedpane.getSelectedComponent();
            } else {
                return tabbedpane.getSelectedComponent();
            }
        } else {
            return null;
        }
    }

    protected final void selectTab( Component component ) {
        for (DnDTabbedPane tab : getTabbedPanes()) {
            for (int i = 0; i < tab.getTabCount(); i++) {
                Component scrollpane = tab.getComponentAt(i);
                if (scrollpane.equals(component)) {
                    tab.setSelectedComponent(scrollpane);
                    setSelectedComponent(component);
                }
            }
        }
    }

    protected final void setTabTitle( Component component, String name ) {
        for (DnDTabbedPane tabbedpane : getTabbedPanes()) {
            int index = tabbedpane.indexOfComponent(component);
            if (index >= 0) {
                TabTitle tabTitle = (TabTitle) tabbedpane.getTabComponentAt(index);
                tabTitle.setTitle(name);
                return;
            }
        }
    }

    protected final void setMenuText( String text1, String text2 ) {
        _splitMenuItem.setText(text1);
        _unsplitMenuItem.setText(text2);
    }

    final void splitVertically() {
        setDividerSize(5);
        setDividerLocation(0.5);
        setSplittingEnabled(false);
    }

    final void unsplit() {
        DnDTabbedPane right = (DnDTabbedPane) getRightComponent();
        DnDTabbedPane left = (DnDTabbedPane) getLeftComponent();
        final int index = right.getTabCount();
        final int count = left.getTabCount();
        for (int i = 0; i < count; i++) {
            left.exportTab(0, right, index + i);
        }
        setDividerSize(0);
        setDividerLocation(0);
        if (right.getTabCount() == 0) {
            resetSplitting();
        } else {
            setSplittingEnabled(true);
        }
    }

    private DnDTabbedPane[] getTabbedPanes() {
        return new DnDTabbedPane[]{(DnDTabbedPane) getLeftComponent(), (DnDTabbedPane) getRightComponent()};
    }

    private DnDTabbedPane getSelectedPane() {
        Component component = getSelectedComponent();
        if (component == null) {
            return (DnDTabbedPane) getRightComponent();
        } else {
            return (DnDTabbedPane) component.getParent();
        }
    }

    private JPopupMenu createTabPopupMenu() {
        JPopupMenu popup = new JPopupMenu();
        _splitMenuItem = new JMenuItem();
        _splitMenuItem.addActionListener(event -> splitVertically());
        _unsplitMenuItem = new JMenuItem();
        _unsplitMenuItem.addActionListener(event -> unsplit());
        resetSplitting();
        popup.add(_splitMenuItem);
        popup.add(_unsplitMenuItem);
        return popup;
    }

    private void resetSplitting() {
        setDividerSize(0);
        _splitMenuItem.setEnabled(false);
        _unsplitMenuItem.setEnabled(false);
    }

    private void setSplittingEnabled( boolean state ) {
        _splitMenuItem.setEnabled(state);
        _unsplitMenuItem.setEnabled(!state);
    }
}
