/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.element;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.GraphModel;
import fr.greyc.pantheon.ariane.model.graph.RootGraph;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.tree.TreeItemFactory;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.presenter.RootGraphPresenter;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.awt.*;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;

public final class LoaderItemTest {
    private static final String TEST_GRAPH1TEST_AWS = "/src/test/resources/graph1test.aws";
    private GraphPresenter _graph;

    @BeforeAll
    public static void setUpClass() {
        Ariane.initForTest();
    }

    @BeforeEach
    public void setUp() {
        _graph = new RootGraphPresenter("graph");
    }

    @Test
    public void testWhenCreatingNewLoaderThenBinaryIsNull() {
        LoaderItem item = new LoaderItem(TreeItemFactory.createItem(TreeItem.ItemType.LOADER), _graph);
        assertNull(item.getBinary());
    }

    @Test
    public void testWhenSavingAndLoadingWithNewItemThenBinaryIsNull() throws IOException {
        LoaderItem item = new LoaderItem(TreeItemFactory.createItem(TreeItem.ItemType.LOADER), _graph);

        XMLElement xmlTree = item.convertToXML(new File("test/graph1test.aws"));
        LoaderItem item1 = new LoaderItem(_graph, xmlTree.getChildren());
        assertNull(item1.getBinary());
    }

    @Test
    public void testWhenChangingBinaryWithAbsoluteFilenameThenBinaryIsTheSame() throws IOException {
        LoaderItem item = new LoaderItem(TreeItemFactory.createItem(TreeItem.ItemType.LOADER), _graph);
        String directory = new File(".").getAbsoluteFile().getParent();
        String absoluteFilename = new File(directory + TEST_GRAPH1TEST_AWS).getAbsolutePath();
        item.setBinary(absoluteFilename);
        assertEquals(absoluteFilename, item.getBinary());
    }

    @Test
    public void testWhenChangingBinaryWithRelativeFilenameThenBinaryIsAbsoluteFilename() throws IOException {
        LoaderItem item = new LoaderItem(TreeItemFactory.createItem(TreeItem.ItemType.LOADER), _graph);
        String directory = new File(".").getAbsoluteFile().getParent();
        String absoluteFilename = new File(directory + TEST_GRAPH1TEST_AWS).getAbsolutePath();
        item.setBinary("src/test/resources/graph1test.aws");
        assertEquals(absoluteFilename, item.getBinary());
    }

    @Test
    public void testWhenSavingAndLoadingWithSpecifiedBinaryThenBinaryIsTheSame() throws IOException {
        LoaderItem item = new LoaderItem(TreeItemFactory.createItem(TreeItem.ItemType.LOADER), _graph);
        String directory = new File(".").getAbsoluteFile().getParent();
        File file = new File(directory + TEST_GRAPH1TEST_AWS);
        String absoluteFilename = file.getAbsolutePath();
        item.setBinary(absoluteFilename);

        XMLElement xmlTree = item.convertToXML(null);
        LoaderItem item1 = new LoaderItem(_graph, xmlTree.getChildren());
        assertEquals(absoluteFilename, item1.getBinary());
    }

    @Test
    public void testWhenSetBinaryWithRegularFileAndExecuteThenResultIsBinaryValue() throws IOException, ArianeRuntimeException {
        LoaderItem item = new LoaderItem(TreeItemFactory.createItem(TreeItem.ItemType.LOADER), _graph);
        String directory = new File(".").getAbsoluteFile().getParent();
        File file = new File(directory + TEST_GRAPH1TEST_AWS);
        String filename = file.getAbsolutePath();
        item.setBinary(filename);
        item.primitiveExecute();
        String result = item.getOutputDataSlots().get(0).getValue();
        assertEquals(filename, result);
    }

    @Test
    public void testWhenSetBinaryWithWrongFileAndExecuteThenResultIsBinaryValue() throws IOException, ArianeRuntimeException {
        LoaderItem item = new LoaderItem(TreeItemFactory.createItem(TreeItem.ItemType.LOADER), _graph);
        String directory = new File(".").getAbsoluteFile().getParent();
        File file = new File(directory + "/test/XXXX.aws");
        String filename = file.getAbsolutePath();
        item.setBinary(filename);
        assertThrows(ArianeRuntimeException.class, () -> item.primitiveExecute());
    }

    @Test
    void testWhenSetParameterWithRegularFileAndExecuteThenResultIsParametervalue() throws IOException, ArianeRuntimeException {
        LoaderItem item = new LoaderItem(TreeItemFactory.createItem(TreeItem.ItemType.LOADER), _graph);
        String directory = new File(".").getAbsoluteFile().getParent();
        File file = new File(directory + TEST_GRAPH1TEST_AWS);
        String parameterValue = file.getAbsolutePath();
        item.getInputParameterSlots().get(0).setValue(parameterValue);
        item.primitiveExecute();
        String result = item.getOutputDataSlots().get(0).getValue();
        assertEquals(parameterValue, result);
    }

    @Test
    void testWhenSetParameterWithWrongFileAndExecuteThenBinaryIs() throws IOException, ArianeRuntimeException {
        LoaderItem item = new LoaderItem(TreeItemFactory.createItem(TreeItem.ItemType.LOADER), _graph);
        String directory = new File(".").getAbsoluteFile().getParent();
        File file = new File(directory + "/test/XXXX.aws");
        String parameterValue = file.getAbsolutePath();
        item.getInputParameterSlots().get(0).setValue(parameterValue);
        assertThrows(ArianeRuntimeException.class, () -> item.primitiveExecute());
    }

    @Test
    public void testItemCreationFromXML() throws IOException {
        List<XMLElement> xml = new ArrayList<>();
        xml.add(new XMLElement("NickName", "Equipe"));
        xml.add(new XMLElement("Name", "loader"));
        xml.add(new XMLElement("Identifier", "30"));
        xml.add(new XMLElement("X-coordinate", "36"));
        xml.add(new XMLElement("Y-coordinate", "15"));
        xml.add(new XMLElement("Parameters", "1"));
        xml.add(new XMLElement("OutputFiles", "1"));
        String filename = "images/equipe2005.bmp";
        xml.add(new XMLElement("Binary", "./" + filename));

        String directory = new File(".").getAbsoluteFile().getParent();
        String absoluteFilename = new File(directory + File.separator + filename).getCanonicalPath();

        LoaderItem item = new LoaderItem(_graph, xml);
        assertTrue("Equipe".equals(item.getNickName())
                && "loader".equals(item.getName())
                && 30 == item.getInternalId()
                && item.getInputParameterSlots().size() == 1
                && item.getOutputDataSlots().size() == 1
                && item.getLocation().equals(new Point(36, 15))
                && absoluteFilename.equals(item.getBinary()));
    }

    @Test
    public void testWhenSavingAndLoadingWithParameterValueThenParameterIsSaved() throws IOException {
        GraphModel graphModel = new RootGraph();
        GraphPresenter graph = new RootGraphPresenter("graph", graphModel);

        LoaderItem item = new LoaderItem(TreeItemFactory.createItem(TreeItem.ItemType.LOADER), graph);
        String directory = new File(".").getAbsoluteFile().getParent();
        File file = new File(directory + TEST_GRAPH1TEST_AWS);
        String absoluteFilename = file.getAbsolutePath();
        item.setBinary(absoluteFilename);

        String parameterValue = file.getAbsolutePath();
        item.getInputParameterSlots().get(0).setValue(parameterValue);

        graphModel.add(item);
        XMLElement xmlTree = graphModel.convertToXMLTree(null);
        ArrayList<XMLElement> litems = new ArrayList<>();
        XMLElement xmlTree1 = new XMLElement("Ariane");
        xmlTree1.addChildren(xmlTree);
        litems.add(xmlTree1);
        graphModel.createFromXMLTree(litems, graph);

        GraphItemModel item1 = graphModel.getGraphItem(1);
        String result = item1.getInputParameterSlots().get(0).getValue();
        assertEquals(absoluteFilename, result);
    }
}
