/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.element;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.model.exception.ArianeRuntimeException;
import fr.greyc.pantheon.ariane.model.graph.slot.ParameterSlot;
import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.tree.TreeItemFactory;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.presenter.RootGraphPresenter;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.awt.*;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;

public final class SystemItemTest {
    private GraphPresenter _graph;

    @BeforeAll
    public static void setUpClass() {
        Ariane.initForTest();
    }

    @BeforeEach
    public void setUp() {
        _graph = new RootGraphPresenter("graph");
    }

    @Test
    void testNewSystemItem() throws InterruptedException {
        SystemItem item = new SystemItem(TreeItemFactory.createItem(TreeItem.ItemType.SYSTEM), _graph, 0, 0, 0);
        assertNotNull(item.getBinary());
    }

    @Test
    void testReadSystemItem() throws IOException {
        List<XMLElement> xml = new ArrayList<>();
        xml.add(new XMLElement("NickName", "updateFormula"));
        xml.add(new XMLElement("Name", "system"));
        xml.add(new XMLElement("Identifier", "1"));
        xml.add(new XMLElement("X-coordinate", "41"));
        xml.add(new XMLElement("Y-coordinate", "22"));
        xml.add(new XMLElement("Binary", "echo \"Bonjour\""));
        xml.add(new XMLElement("InputFiles", "2"));
        xml.add(new XMLElement("OutputFiles", "1"));
        xml.add(new XMLElement("Parameters", "3"));

        SystemItem item = new SystemItem(_graph, xml);
        assertTrue("updateFormula".equals(item.getNickName())
                && "system".equals(item.getName())
                && 1 == item.getInternalId()
                && item.getLocation().equals(new Point(41, 22))
                && "echo \"Bonjour\"".equals(item.getBinary()));
    }

    @Test
    void testSettingValue() throws IOException {
        SystemItem item = new SystemItem(TreeItemFactory.createItem(TreeItem.ItemType.SYSTEM), _graph, 0, 0, 0);

        final String binary = "cat $1";
        item.setBinary(binary);
        assertEquals(binary, item.getBinary());
    }

    @Test
    void testExecution() throws IOException, ArianeRuntimeException {
        SystemItem item = new SystemItem(TreeItemFactory.createItem(TreeItem.ItemType.SYSTEM), _graph, 0, 0, 0);

        item.setBinary("echo \"Bonjour\"");
        item.primitiveExecute();
        assertEquals("SUCCESS", item.getOutputResultSlot().getValue());
    }

    @Test
    public void testExecutionWithOneUnsetParameter() throws IOException, ArianeRuntimeException {
        SystemItem item = new SystemItem(TreeItemFactory.createItem(TreeItem.ItemType.SYSTEM), _graph, 0, 1, 0);

        item.setBinary("echo $p0");
        assertThrows(ArianeRuntimeException.class, item::primitiveExecute);
    }

    @Test
    public void testExecutionWithOneParameter() throws IOException, ArianeRuntimeException {
        SystemItem item = new SystemItem(TreeItemFactory.createItem(TreeItem.ItemType.SYSTEM), _graph, 0, 1, 0);

        item.setBinary("echo $p0");

        List<ParameterSlot> parameters = item.getInputParameterSlots();
        parameters.get(0).setValue("Bonjour");
        item.primitiveExecute();
        assertEquals("SUCCESS", item.getOutputResultSlot().getValue());
    }
}
