/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.model.graph.element.etc;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.util.Locale;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import fr.greyc.pantheon.ariane.ArianeSettings;
import fr.greyc.pantheon.ariane.model.exception.ArianeArithmeticException;

public final class ArithmeticExpressionTest {

    @BeforeAll
    public static void setUpBeforeClass() {
        ArianeSettings.setCurrentLocale(new Locale("fr", "FR"));
    }

    @Test
    void checkEvaluationSequence() throws ArianeArithmeticException {
        ArithmeticExpression expression1 = new ArithmeticExpression();
        assertThrows(NullPointerException.class, () -> expression1.evaluate(null));
    }

    @Test
    void checkValidExpression1() throws ArianeArithmeticException {
        final String formula1 = "1+2";
        ArithmeticExpression expression1 = new ArithmeticExpression();
        expression1.compile(formula1, 0);
        double result1 = expression1.evaluate(null);
        assertEquals(3.0, result1, 0.0);
    }

    @Test
    void checkFloattingExpression() throws ArianeArithmeticException {
        final String formula1 = "1.5e2+2.0e-2";
        ArithmeticExpression expression1 = new ArithmeticExpression();
        expression1.compile(formula1, 0);
        double result1 = expression1.evaluate(null);
        assertEquals(150.02, result1, 0.0);
    }

    @Test
    void checkValidExpression2() throws ArianeArithmeticException {
        final String formula1 = "exp($0)";
        ArithmeticExpression expression1 = new ArithmeticExpression();
        expression1.compile(formula1, 1);
        final double[] args = {1.0};
        double result1 = expression1.evaluate(args);
        assertEquals(2.71, result1, 0.01);
    }

    @Test
    void checkValidExpression3() throws ArianeArithmeticException {
        final String formula1 = "max($0,$1)";
        ArithmeticExpression expression1 = new ArithmeticExpression();
        expression1.compile(formula1, 2);
        final double[] args = {1.0, 2.0};
        double result1 = expression1.evaluate(args);
        assertEquals(2.00, result1, 0.01);
    }

    @Test
    void checkValidExpression4() throws ArianeArithmeticException {
        final String formula1 = "($0*5)+exp(1)";
        ArithmeticExpression expression1 = new ArithmeticExpression();
        expression1.compile(formula1, 1);
        final double[] args = {1.0};
        double result1 = expression1.evaluate(args);
        assertEquals(7.71, result1, 0.01);
    }

    @Test
    void checkInValidExpressions1() {
        final String formula1 = "1+";
        ArithmeticExpression expression1 = new ArithmeticExpression();
        assertThrows(ArianeArithmeticException.class, () -> expression1.compile(formula1, 0));
//        assertThrows(ArianeArithmeticException.class, () -> expression1.evaluate(null));
    }

    @Test
    void checkInValidExpressions2() throws ArianeArithmeticException {
        final String formula1 = "(1+$0";
        ArithmeticExpression expression1 = new ArithmeticExpression();
        assertThrows(ArianeArithmeticException.class, () ->  expression1.compile(formula1, 1));
    }

    @Test
    void checkInValidArithemticExpression() throws ArianeArithmeticException {
        final String formula1 = "ln(0)";
        ArithmeticExpression expression1 = new ArithmeticExpression();
        expression1.compile(formula1, 1);
        final double[] args = {1.0};
        assertThrows(ArianeArithmeticException.class, () -> expression1.evaluate(args));
    }

    @Test
    void checkInSqrtInvalidArithmeticExpression() throws ArianeArithmeticException {
        final String formula1 = "sqrt(-1)";
        ArithmeticExpression expression1 = new ArithmeticExpression();
        expression1.compile(formula1, 1);
        final double[] args = {1.0};
        assertThrows(ArianeArithmeticException.class, () -> expression1.evaluate(args));
    }

    @Test
    void checkInSqrtValidArithmeticExpression() throws ArianeArithmeticException {
        final String formula1 = "sqrt(4)";
        ArithmeticExpression expression1 = new ArithmeticExpression();
        expression1.compile(formula1, 1);
        final double[] args = {1.0};
        double result1 = expression1.evaluate(args);
        assertEquals(2, result1, 0.0);
    }

    @Test
    void checkOperandOrder() throws ArianeArithmeticException {
        final String formula1 = "3-2";
        ArithmeticExpression expression1 = new ArithmeticExpression();
        expression1.compile(formula1, 1);
        final double[] args = {1.0};
        double result1 = expression1.evaluate(args);
        assertEquals(1.0, result1, 0.0);
    }

    @Test
    void testFloats() throws ArianeArithmeticException {
        final String formula1 = "-1e3-2e-3";
        ArithmeticExpression expression1 = new ArithmeticExpression();
        expression1.compile(formula1, 1);
        final double[] args = {1.0};
        double result1 = expression1.evaluate(args);
        assertEquals(-1000.002, result1, 0.0);
    }
}
