/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 * https://clouard.users.greyc.fr
 *
 * This software is governed by the CeCILL-B  license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.explorer;

import fr.greyc.pantheon.ariane.Ariane;
import fr.greyc.pantheon.ariane.presenter.GraphPresenter;
import fr.greyc.pantheon.ariane.presenter.Presenter;
import fr.greyc.pantheon.ariane.presenter.RootGraphPresenter;
import fr.greyc.pantheon.ariane.view.Workspace;
import fr.greyc.pantheon.ariane.view.graph.GraphEditor;
import fr.greyc.pantheon.swingext.multivieweditor.DnDTabbedPane;
import fr.greyc.pantheon.swingext.multivieweditor.TabTitle;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.awt.*;
import java.io.File;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

final class GraphEditorTest {
    private static Presenter _presenter;
    private GraphEditor _graphEditor;

    @BeforeEach
    public void setUp() {
        Ariane.initForTest();
        _presenter = Ariane.getPresenter();
        _graphEditor = _presenter.getWorkspace().getGraphEditor();
        _graphEditor.setName("MonGraphEditor");
    }

    @Test
    void whenCreateGraphExpectDivisionSizeZero() {
        GraphEditor editor = new GraphEditor(new Workspace(null, null));
        assertEquals(0, editor.getDividerSize());
    }

    @Test
    void whenAddingNewGraphExpectGraphInserted() {
        GraphPresenter graph = new RootGraphPresenter("test graph");
        _graphEditor.add(graph);
        assertEquals(1, _graphEditor.getAllComponents().size());
    }

    @Test
    void whenAddingTwiceSameGraphExpectOneTab() {
        GraphPresenter graph1 = new RootGraphPresenter("graph1");
        GraphPresenter graph2 = new RootGraphPresenter("graph2");
        _graphEditor.add(graph1);
        _graphEditor.add(graph2);
        _graphEditor.add(graph1);
        assertEquals(2, _graphEditor.getAllComponents().size());
    }

    @Test
    void whenAddingGraphFromFilenameExpectGraphWithSpecifiedNameExists() {
        String filename = "test/graph1test.aws";
        File file = new File(filename);
        GraphPresenter graph1 = new RootGraphPresenter(file.getAbsoluteFile());
        _graphEditor.add(graph1);
        assertTrue(_graphEditor.contains(filename));
    }

    @Test
    void whenRepaintAllExpectNoException() {
        GraphPresenter graph1 = new RootGraphPresenter("graph1");
        GraphPresenter graph2 = new RootGraphPresenter("graph2");
        _presenter.getWorkspace().addGraph(graph1);
        _presenter.getWorkspace().addGraph(graph2);
        _graphEditor.repaintAllGraphs();
        assertEquals(2, _graphEditor.getAllComponents().size());
    }

    @Test
    void whenAddingAndRemovingGraphExpectZeroTab() {
        GraphPresenter graph1 = new RootGraphPresenter("graph1");
        _graphEditor.add(graph1);
        _graphEditor.remove(graph1);
        assertEquals(0, _graphEditor.getAllComponents().size());
    }

    @Test
    void whenCloseAllGraphExpectNoTab() {
        GraphPresenter graph1 = new RootGraphPresenter("graph1");
        GraphPresenter graph2 = new RootGraphPresenter("graph2");
        _presenter.getWorkspace().addGraph(graph1);
        _presenter.getWorkspace().addGraph(graph2);
        _graphEditor.closeAllGraphs();
        assertEquals(0, _graphEditor.getAllComponents().size());
    }

    @Test
    void whenChangeGraphNameExpectNameChanged() {
        GraphPresenter graph = new RootGraphPresenter("graph");
        _presenter.getWorkspace().addGraph(graph);
        final String newName = "updateFormula";
        _graphEditor.setTitle(graph, newName);
        DnDTabbedPane rightTabbedPane = (DnDTabbedPane) _graphEditor.getRightComponent();
        assertTrue(((TabTitle) rightTabbedPane.getTabComponentAt(0)).getTitle().contains(newName));
    }

    @Test
    void whenChangeGraphNameExpectGraphNotDeleted() {
        String filename = "test/graph1test.aws";
        File file = new File(filename);
        GraphPresenter graph = new RootGraphPresenter(file.getAbsoluteFile());
        _graphEditor.add(graph);
        _graphEditor.setTitle(graph, "updateFormula");
        assertTrue(_graphEditor.contains(filename));
    }

    @Test
    void whenCloseGraphExpectNoTab() {
        GraphPresenter graph = new RootGraphPresenter("graph");
        _presenter.getWorkspace().addGraph(graph);
        Component c = _graphEditor.getSelectedComponent();
        _graphEditor.close(c);
        assertEquals(0, _graphEditor.getAllComponents().size());
    }
}
