/*
 * Copyright or (C) or Copr. Normandie Univ, ENSICAEN, GREYC, Caen, France
 *
 *  https://clouard.users.greyc.fr
 *
 *  This software is governed by the CeCILL-B  license under French law and
 *  abiding by the rules of distribution of free software.  You can  use,
 *  modify and/ or redistribute the software under the terms of the CeCILL-B
 *  license as circulated by CEA, CNRS and INRIA at the following URL
 *  "http://www.cecill.info".
 *
 *  In this respect, the user's attention is drawn to the risks associated
 *  with loading,  using,  modifying and/or developing or reproducing the
 *  software by the user in light of its specific status of free software,
 *  that may mean  that it is complicated to manipulate,  and  that  also
 *  therefore means  that it is reserved for developers  and  experienced
 *  professionals having in-depth computer knowledge. Users are therefore
 *  encouraged to load and test the software's suitability as regards their
 *  requirements in conditions enabling the security of their systems and/or
 *  data to be ensured and,  more generally, to use and operate it in the
 *  same conditions as regards security.
 *
 *  The fact that you are presently reading this means that you have had
 *  knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.greyc.pantheon.ariane.view.explorer;

import fr.greyc.pantheon.ariane.model.tree.TreeItem;
import fr.greyc.pantheon.ariane.model.xml.XMLElement;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import javax.swing.tree.TreePath;
import java.util.ArrayList;
import java.util.List;

import static org.junit.jupiter.api.Assertions.*;

public final class ItemTreeModelTest {
    private static final List<XMLElement> USER_OPERATOR_XML_TREE = new ArrayList<>();
    private static final String OPERATOR_NAME = "setstatus";

    @BeforeAll
    public static void setUpBeforeClass() {
        XMLElement element;
        List<XMLElement> xmlTree = addOperator(OPERATOR_NAME);
        xmlTree.addAll(addOperator(OPERATOR_NAME + "1"));

        element = new XMLElement("OperatorLibrary");

        element.addChildren(xmlTree);

        xmlTree = new ArrayList<>();
        xmlTree.add(element);

        element = new XMLElement("Ariane");
        element.addChildren(xmlTree);
        USER_OPERATOR_XML_TREE.add(element);
    }

    @Test
    void testWhenLoadingTreefromXMLThenReturnFirstItem() {
        ItemTreeModel treeItemModel = new ItemTreeModel();
        treeItemModel.setUserOperatorTree(USER_OPERATOR_XML_TREE);

        List<TreeItem> tree = treeItemModel.getUserOperatorTreeAsList();
        assertEquals(OPERATOR_NAME, tree.get(0).getName());
    }

    @Test
    void testWhenSearchExactNameOfExistingItemThenReturnTrue() {
        ItemTreeModel treeItemModel = new ItemTreeModel();
        treeItemModel.setUserOperatorTree(USER_OPERATOR_XML_TREE);

        List<TreePath> treePath = treeItemModel.getItemsWithTheSpecifiedNameInTree(OPERATOR_NAME);
        assertFalse(treePath.isEmpty());
    }

    @Test
    void testWhenSearchBeginningOfNameOfExistingItemThenReturnTrue() {
        ItemTreeModel treeItemModel = new ItemTreeModel();
        treeItemModel.setUserOperatorTree(USER_OPERATOR_XML_TREE);

        List<String> treePath = treeItemModel.getItemsWithNameStartingWith("set");
        assertFalse(treePath.isEmpty());
    }

    @Test
    void testWhenSearchNameOfExistingItemThenreturnTrue() {
        ItemTreeModel treeItemModel = new ItemTreeModel();
        treeItemModel.setUserOperatorTree(USER_OPERATOR_XML_TREE);
        TreeItem item = treeItemModel.getItemEqualsTo(OPERATOR_NAME);
        assertNotNull(item);
    }

    @Test
    void testGetItemEqualsToWhenSearchNameOfNonExistingItemThenReturnNull() {
        ItemTreeModel treeItemModel = new ItemTreeModel();
        treeItemModel.setUserOperatorTree(USER_OPERATOR_XML_TREE);

        TreeItem item = treeItemModel.getItemEqualsTo("xxx");
        assertNull(item);
    }

    @Test
    void testGetItemEqualsToPartialNameThenReturnNull() {
        ItemTreeModel treeItemModel = new ItemTreeModel();
        treeItemModel.setUserOperatorTree(USER_OPERATOR_XML_TREE);

        TreeItem treeItem = treeItemModel.getItemEqualsTo("set");
        assertNull(treeItem);
    }

    @Test
    void whenChangeCategoryInTreeThenItemHasCategoryChanged() {
        ItemTreeModel treeItemModel = new ItemTreeModel();
        // Make a copy because the list modified
        List<XMLElement> newTree = new ArrayList<>(USER_OPERATOR_XML_TREE);
        treeItemModel.setUserOperatorTree(newTree);

        MutableTreeItem tree = treeItemModel.getUserOperatorTree();

        // 1. Change category name
        final String suffix = "updateFormula";
        MutableTreeItem firstCategory = (MutableTreeItem) tree.getFirstChild();
        final String newName = suffix + firstCategory.getUserObject();
        firstCategory.setUserObject(newName);
        // 2. Apply the change to all tree items
        treeItemModel.updateItemCategory(tree);
        // 3. check
        MutableTreeItem subTree = (MutableTreeItem) firstCategory.getChildAt(0);
        TreeItem treeItem = (TreeItem) subTree.getUserObject();
        assertEquals(newName, treeItem.getCategory());
    }

    private static List<XMLElement> addOperator( String operatorName ) {
        List<XMLElement> xmlTree = new ArrayList<>();

        XMLElement element;
        element = new XMLElement("Category");
        element.setValue("category");
        xmlTree.add(element);
        element = new XMLElement("Name");
        element.setValue(operatorName);
        xmlTree.add(element);
        element = new XMLElement("Documentation");
        element.setValue("file:///home/regis/pantheon/software/ariane/" + operatorName + ".html");
        xmlTree.add(element);
        element = new XMLElement("Binary");
        element.setValue("/home/regis/pantheon/software/" + operatorName + ".sh&amp;[-m $m] [$p] [$i] [$o]");
        xmlTree.add(element);
        element = new XMLElement("Parameters");
        element.setValue("0");
        xmlTree.add(element);
        element = new XMLElement("Signature");
        xmlTree.add(element);
        element = new XMLElement("Inputs");
        element.setValue("0");
        xmlTree.add(element);
        element = new XMLElement("Outputs");
        element.setValue("0");
        xmlTree.add(element);

        element = new XMLElement("Operator");
        element.addChildren(xmlTree);
        xmlTree = new ArrayList<>();
        xmlTree.add(element);
        return xmlTree;
    }
}
